#include <app/log/Logger.h>
#include <app/config/Config.h>
#include <common/components/worker/DummyRetryableWork.h>
#include <common/components/worker/DummyWork.h>
#include <common/components/worker/RetryWorker.h>
#include <common/components/worker/Worker.h>
#include <common/components/worker/WorkQueue.h>
#include <common/nodes/TargetMapper.h>
#include <common/system/System.h>
#include <common/toolkit/ackstore/AcknowledgmentStore.h>
#include <common/toolkit/NetFilter.h>
#include <components/AckManager.h>
#include <components/DatagramListener.h>
#include <components/Flusher.h>
#include <components/HeartbeatManager.h>
#include <components/InternodeSyncer.h>
#include <net/filesystem/FhgfsOpsCommKit.h>
#include <nodes/NodeStoreEx.h>
#include <toolkit/InodeRefStore.h>
#include <app/log/Logger.h>
#include <app/App.h>
#include <net/filesystem/FhgfsOpsRemoting.h>
#include <toolkit/NoAllocBufferStore.h>
#include <common/toolkit/TimeAbs.h>


#ifndef FHGFS_VERSION
   #error FHGFS_VERSION undefined
#endif

#if !defined(FHGFS_VERSION_CODE) || (FHGFS_VERSION_CODE == 0)
   #error FHGFS_VERSION_CODE undefined
#endif


/**
 * @param mountConfig belongs to the app afterwards (and will automatically be destructed
 * by the app)
 */
void App_init(App* this, MountConfig* mountConfig)
{
   this->mountConfig = mountConfig;

   this->appResult = APPCODE_NO_ERROR;

   this->lockAckAtomicCounter = NULL;
   this->connRetriesEnabled = fhgfs_true;
   this->netBenchModeEnabled = fhgfs_false;

   this->cfg = NULL;
   this->netFilter = NULL;
   this->logger = NULL;
   this->helperd = NULL;
   this->allowedInterfaces = NULL;
   this->preferredMetaNodes = NULL;
   this->preferredStorageTargets = NULL;
   this->pagedIOBufStore = NULL;
   this->cacheBufStore = NULL;
   this->rwStatesStore = NULL;
   this->pathBufStore = NULL;
   this->msgBufStore = NULL;
   this->ackStore = NULL;
   this->inodeRefStore = NULL;
   this->localNode = NULL;
   this->mgmtNodes = NULL;
   this->metaNodes = NULL;
   this->storageNodes = NULL;
   this->targetMapper = NULL;
   this->workQueue = NULL;
   this->retryWorkQueue = NULL;

   this->dgramListener = NULL;
   this->heartbeatMgr = NULL;
   this->workerList = NULL;
   this->retryWorkerList = NULL;
   this->internodeSyncer = NULL;
   this->ackManager = NULL;
   this->flusher = NULL;

#ifdef FHGFS_DEBUG
   this->debugCounterMutex = NULL;

   this->numRPCs = 0;
   this->numRemoteReads = 0;
   this->numRemoteWrites = 0;
#endif // FHGFS_DEBUG

}

/**
 * @param mountConfig belongs to the app afterwards (and will automatically be destructed
 * by the app)
 */
App* App_construct(MountConfig* mountConfig)
{
   App* this = (App*)os_kmalloc(sizeof(*this) );

   App_init(this, mountConfig);

   return this;
}

void App_uninit(App* this)
{
   __App_workersDelete(this);

   SAFE_DESTRUCT(this->flusher, Flusher_destruct);
   SAFE_DESTRUCT(this->ackManager, AckManager_destruct);
   SAFE_DESTRUCT(this->internodeSyncer, InternodeSyncer_destruct);
   SAFE_DESTRUCT(this->retryWorkerList, PointerList_destruct);
   SAFE_DESTRUCT(this->workerList, PointerList_destruct);
   SAFE_DESTRUCT(this->heartbeatMgr, HeartbeatManager_destruct);
   SAFE_DESTRUCT(this->dgramListener, DatagramListener_destruct);

   SAFE_DESTRUCT(this->retryWorkQueue, WorkQueue_destruct);
   SAFE_DESTRUCT(this->workQueue, WorkQueue_destruct);
   SAFE_DESTRUCT(this->storageNodes, NodeStoreEx_destruct);
   SAFE_DESTRUCT(this->metaNodes, NodeStoreEx_destruct);
   SAFE_DESTRUCT(this->mgmtNodes, NodeStoreEx_destruct);

   if(this->logger)
      Logger_setAllLogLevels(this->logger, -1); // disable logging

   SAFE_DESTRUCT(this->localNode, Node_destruct);
   SAFE_DESTRUCT(this->targetMapper, TargetMapper_destruct);
   SAFE_DESTRUCT(this->inodeRefStore, InodeRefStore_destruct);
   SAFE_DESTRUCT(this->ackStore, AcknowledgmentStore_destruct);
   SAFE_DESTRUCT(this->msgBufStore, NoAllocBufferStore_destruct);
   SAFE_DESTRUCT(this->pathBufStore, NoAllocBufferStore_destruct);
   SAFE_DESTRUCT(this->rwStatesStore, NoAllocBufferStore_destruct);
   SAFE_DESTRUCT(this->cacheBufStore, NoAllocBufferStore_destruct);
   SAFE_DESTRUCT(this->pagedIOBufStore, NoAllocBufferStore_destruct);
   SAFE_DESTRUCT(this->preferredStorageTargets, UInt16List_destruct);
   SAFE_DESTRUCT(this->preferredMetaNodes, UInt16List_destruct);
   SAFE_DESTRUCT(this->allowedInterfaces, StrCpyList_destruct);
   SAFE_DESTRUCT(this->helperd, ExternalHelperd_destruct);
   SAFE_DESTRUCT(this->logger, Logger_destruct);
   SAFE_DESTRUCT(this->netFilter, NetFilter_destruct);
   SAFE_DESTRUCT(this->cfg, Config_destruct);

   SAFE_DESTRUCT(this->lockAckAtomicCounter, AtomicInt_destruct);

   SAFE_DESTRUCT(this->mountConfig, MountConfig_destruct);

#ifdef FHGFS_DEBUG
   SAFE_DESTRUCT(this->debugCounterMutex, Mutex_destruct);
#endif // FHGFS_DEBUG
}

void App_destruct(App* this)
{
   App_uninit(this);

   os_kfree(this);
}


int App_run(App* this)
{
   // init data objects & storage

   if(!__App_initDataObjects(this, this->mountConfig) )
   {
      printk_fhgfs(KERN_WARNING,
         "Configuration error: Initialization of common objects failed. "
         "(Log file may provide additional information.)\n");
      this->appResult = APPCODE_INVALID_CONFIG;
      return this->appResult;
   }

   if(!__App_initStorage(this) )
   {
      printk_fhgfs(KERN_WARNING, "Configuration error: Initialization of storage failed\n");
      this->appResult = APPCODE_INVALID_CONFIG;
      return this->appResult;
   }

   if(Config_getDebugRunStartupTests(this->cfg) )
   { // startup tests
      // not impl'ed
   }


   // init components

   if(!__App_initComponents(this) )
   {
      printk_fhgfs(KERN_WARNING, "Component initialization error. "
         "(Log file may provide additional information.)\n");
      this->appResult = APPCODE_INITIALIZATION_ERROR;
      return this->appResult;
   }

   if(Config_getDebugRunComponentTests(this->cfg) )
   { // component tests
      // not impl'ed
   }


   __App_logInfos(this);


   // start components

   if(Config_getDebugRunComponentThreads(this->cfg) )
   {
      __App_startComponents(this);

      // Note: We wait some ms for the node downloads here because the kernel would like to
      //    check the properties of the root directory directly after mount.

      HeartbeatManager_waitForMgmtInit(this->heartbeatMgr, 1000);

      if(!__App_mountServerCheck(this) )
      { // mount check failed => cancel mount
         printk_fhgfs(KERN_WARNING, "Mount sanity check failed. Canceling mount. "
            "(Log file may provide additional information. Check can be disabled with "
            "sysMountSanityCheckMS=0 in the config file.)\n");
         this->appResult = APPCODE_INITIALIZATION_ERROR;

         return this->appResult;
      }

      if(Config_getDebugRunIntegrationTests(this->cfg) )
      { // integration tests
         // not impl'ed
      }

      if(Config_getDebugRunThroughputTests(this->cfg) )
      { // throughput tests
         // not impl'ed
      }


      // mark: mount succeeded if we got here!
   }


   return this->appResult;
}

void App_stop(App* this)
{
   const char* logContext = "App (stop)";

   // note: be careful, because we don't know what has been succesfully initialized!

   // note: this can also be called from App_run() to cancel a mount!

   if(this->cfg && Config_getDebugRunComponentThreads(this->cfg) )
   {
      __App_stopComponents(this);

      if(!Config_getConnUnmountRetries(this->cfg) )
         this->connRetriesEnabled = fhgfs_false;

      // wait for termination
      __App_joinComponents(this);

      if(this->logger)
         Logger_log(this->logger, 1, logContext, "All components stopped.");
   }
}

fhgfs_bool __App_initDataObjects(App* this, MountConfig* mountConfig)
{
   const char* logContext = "App (init data objects)";

   char* interfacesFilename;
   char* preferredMetaFile;
   char* preferredStorageFile;
   size_t numPagedIOBufs;
   size_t pagedIOBufSize;
   size_t numCacheBufs;
   size_t cacheBufSize;
   size_t numRWStateBufs;
   size_t numRWStateBufNodes;
   size_t rwStateBufsSize;
   size_t numPathBufs;
   size_t pathBufsSize;
   size_t numMsgBufs;
   size_t msgBufsSize;


#ifdef FHGFS_DEBUG
   this->debugCounterMutex = Mutex_construct(); /* required for RPCs (=> logging) etc */
#endif // FHGFS_DEBUG

   this->lockAckAtomicCounter = AtomicInt_construct(0);

   this->cfg = Config_construct(mountConfig);

   if(!Config_getValid(this->cfg) )
      return fhgfs_false;


   if(!StringTk_hasLength(Config_getSysMgmtdHost(this->cfg) ) )
   {
      printk_fhgfs(KERN_WARNING, "Management host undefined\n");
      return fhgfs_false;
   }


   // load net filter (required before any connection can be made, incl. local conns)
   this->netFilter = NetFilter_construct(this->cfg);
   if(!NetFilter_getValid(this->netFilter) )
      return fhgfs_false;

   // logger (+ hostname as initial clientID, real ID will be set below)
   this->logger = Logger_construct(this, this->cfg);
   if(Config_getLogClientID(this->cfg) )
   {
      const char* hostnameCopy = System_getHostnameCopy();
      Logger_setClientID(this->logger, hostnameCopy);
      os_kfree(hostnameCopy);
   }

   this->helperd = ExternalHelperd_construct(this, this->cfg);

   // load allowed interface file
   this->allowedInterfaces = StrCpyList_construct();
   interfacesFilename = Config_getConnInterfacesFile(this->cfg);
   if(os_strlen(interfacesFilename) &&
      !Config_loadStringListFile(this->cfg, interfacesFilename, this->allowedInterfaces) )
   {
      Logger_logErrFormatted(this->logger, logContext,
         "Unable to load configured interfaces file: %s", interfacesFilename);
      return fhgfs_false;
   }

   // load preferred nodes files

   this->preferredMetaNodes = UInt16List_construct();
   preferredMetaFile = Config_getTunePreferredMetaFile(this->cfg);
   if(os_strlen(preferredMetaFile) &&
      !Config_loadUInt16ListFile(this->cfg, preferredMetaFile, this->preferredMetaNodes) )
   {
      Logger_logErrFormatted(this->logger, logContext,
         "Unable to load configured preferred meta nodes file: %s", preferredMetaFile);
      return fhgfs_false;
   }

   this->preferredStorageTargets = UInt16List_construct();
   preferredStorageFile = Config_getTunePreferredStorageFile(this->cfg);
   if(os_strlen(preferredStorageFile) &&
      !Config_loadUInt16ListFile(this->cfg, preferredStorageFile, this->preferredStorageTargets) )
   {
      Logger_logErrFormatted(this->logger, logContext,
         "Unable to load configured preferred storage nodes file: %s", preferredStorageFile);
      return fhgfs_false;
   }

   // init stores, queues etc.

   this->targetMapper = TargetMapper_construct();

   this->mgmtNodes = NodeStoreEx_construct(this, NODETYPE_Mgmt);
   if(!NodeStoreEx_getValid(this->mgmtNodes) )
      return fhgfs_false;

   this->metaNodes = NodeStoreEx_construct(this, NODETYPE_Meta);
   if(!NodeStoreEx_getValid(this->metaNodes) )
      return fhgfs_false;

   this->storageNodes = NodeStoreEx_construct(this, NODETYPE_Storage);
   if(!NodeStoreEx_getValid(this->storageNodes) )
      return fhgfs_false;

   this->workQueue = WorkQueue_construct();
   this->retryWorkQueue = WorkQueue_construct();

   if(!__App_initLocalNodeInfo(this) )
      return fhgfs_false;

   if(Config_getLogClientID(this->cfg) )
   { // set real clientID
      const char* clientID = Node_getID(this->localNode);
      Logger_setClientID(this->logger, clientID);
   }

   // prealloc buffers

   numPagedIOBufs = (Config_getTuneFileCacheTypeNum(this->cfg) == FILECACHETYPE_Paged) ?
      Config_getTunePagedIOBufNum(this->cfg) : 0;
   pagedIOBufSize = Config_getTunePagedIOBufSize(this->cfg);

   this->pagedIOBufStore = NoAllocBufferStore_construct(numPagedIOBufs, pagedIOBufSize);
   if(!NoAllocBufferStore_getValid(this->pagedIOBufStore) )
      return fhgfs_false;

   numCacheBufs = (Config_getTuneFileCacheTypeNum(this->cfg) == FILECACHETYPE_Buffered) ?
      Config_getTuneFileCacheBufNum(this->cfg) : 0;
   cacheBufSize = Config_getTuneFileCacheBufSize(this->cfg);

   this->cacheBufStore = NoAllocBufferStore_construct(numCacheBufs, cacheBufSize);
   if(!NoAllocBufferStore_getValid(this->cacheBufStore) )
      return fhgfs_false;

   numRWStateBufs = Config_getTuneMaxReadWriteNum(this->cfg);
   numRWStateBufNodes = Config_getTuneMaxReadWriteNodesNum(this->cfg);
   rwStateBufsSize = numRWStateBufNodes * (MAX(sizeof(WritefileState), sizeof(ReadfileState) ) +
      sizeof(PollSocketEx) + FHGFS_COMMKIT_MSGBUF_SIZE);

   this->rwStatesStore = NoAllocBufferStore_construct(numRWStateBufs, rwStateBufsSize);

   numPathBufs = Config_getTunePathBufNum(this->cfg);
   pathBufsSize = Config_getTunePathBufSize(this->cfg);

   this->pathBufStore = NoAllocBufferStore_construct(numPathBufs, pathBufsSize);

   numMsgBufs = Config_getTuneMsgBufNum(this->cfg);
   msgBufsSize = Config_getTuneMsgBufSize(this->cfg);

   this->msgBufStore = NoAllocBufferStore_construct(numMsgBufs, msgBufsSize);

   this->ackStore = AcknowledgmentStore_construct();

   this->inodeRefStore = InodeRefStore_construct();


   return fhgfs_true;
}

fhgfs_bool __App_initLocalNodeInfo(App* this)
{
   const char* logContext = "App (init local node info)";

   NicAddressList nicList;
   NicAddressList* sortedNicList;
   char* hostname;
   TimeAbs nowT;
   pid_t currentPID;
   char* nodeID;
   fhgfs_bool useSDP = Config_getConnUseSDP(this->cfg);
   fhgfs_bool useRDMA = Config_getConnUseRDMA(this->cfg);

   NicAddressList_init(&nicList);

   NIC_findAll(this->allowedInterfaces, useSDP, useRDMA, &nicList);

   if(!NicAddressList_length(&nicList) )
   {
      Logger_logErr(this->logger, logContext, "Couldn't find any usable NIC");
      NicAddressList_uninit(&nicList);

      return fhgfs_false;
   }

   // created sorted nicList clone
   sortedNicList = ListTk_cloneSortNicAddressList(&nicList);


   // prepare clientID and create localNode
   TimeAbs_init(&nowT);
   hostname = System_getHostnameCopy();
   currentPID = os_getCurrentPID();

   nodeID = StringTk_kasprintf("%llX-%llX-%s",
      (uint64_t)currentPID, (uint64_t)TimeAbs_getTimeval(&nowT)->tv_sec, hostname);


   this->localNode = Node_construct(this, nodeID, 0, Config_getConnClientPortUDP(this->cfg), 0,
      sortedNicList);

   Node_setFhgfsVersion(this->localNode, FHGFS_VERSION_CODE);


   // clean up
   os_kfree(nodeID);
   os_kfree(hostname);
   TimeAbs_uninit(&nowT);

   // delete nicList elems
   ListTk_kfreeNicAddressListElems(sortedNicList);
   ListTk_kfreeNicAddressListElems(&nicList);

   NicAddressList_destruct(sortedNicList);
   NicAddressList_uninit(&nicList);

   return fhgfs_true;
}

fhgfs_bool __App_initStorage(App* this)
{
   // nothing to be done here (client does not use any harddisk storage)

   return fhgfs_true;
}

fhgfs_bool __App_initComponents(App* this)
{
   const char* logContext = "App (init components)";

   Logger_log(this->logger, 4, logContext, "Initializing components...");

   this->dgramListener =
      DatagramListener_construct(this, this->localNode, Config_getConnClientPortUDP(this->cfg) );
   if(!DatagramListener_getValid(this->dgramListener) )
   {
      Logger_logErr(this->logger, logContext,
         "Initialization of DatagramListener component failed");
      return fhgfs_false;
   }

   this->heartbeatMgr = HeartbeatManager_construct(this, this->dgramListener);
   if(!HeartbeatManager_getValid(this->heartbeatMgr) )
   {
      Logger_logErr(this->logger, logContext,
         "Initialization of HeartbeatManager component failed");
      return fhgfs_false;
   }

   __App_workersInit(this);

   this->internodeSyncer = InternodeSyncer_construct(this);
   if(!InternodeSyncer_getValid(this->internodeSyncer) )
   {
      Logger_logErr(this->logger, logContext,
         "Initialization of InternodeSyncer component failed");
      return fhgfs_false;
   }

   this->ackManager = AckManager_construct(this);
   if(!AckManager_getValid(this->ackManager) )
   {
      Logger_logErr(this->logger, logContext,
         "Initialization of AckManager component failed");
      return fhgfs_false;
   }

   this->flusher = Flusher_construct(this);
   if(!Flusher_getValid(this->flusher) )
   {
      Logger_logErr(this->logger, logContext,
         "Initialization of Flusher component failed");
      return fhgfs_false;
   }

   Logger_log(this->logger, 4, logContext, "Components initialized.");

   return fhgfs_true;
}

void __App_startComponents(App* this)
{
   const char* logContext = "App (start components)";

   Logger_log(this->logger, 4, logContext, "Starting up components...");

   Thread_start( (Thread*)this->dgramListener);
   Thread_start( (Thread*)this->heartbeatMgr);

   __App_workersStart(this);

   Thread_start( (Thread*)this->internodeSyncer);
   Thread_start( (Thread*)this->ackManager);

   if(Config_getTuneFileCacheTypeNum(this->cfg) == FILECACHETYPE_Buffered)
      Thread_start( (Thread*)this->flusher);

   Logger_log(this->logger, 4, logContext, "Components running.");
}

void __App_stopComponents(App* this)
{
   const char* logContext = "App (stop components)";

   if(this->logger)
      Logger_log(this->logger, 2, logContext, "Stopping components...");

   if(this->flusher &&
      (Config_getTuneFileCacheTypeNum(this->cfg) == FILECACHETYPE_Buffered) )
      Thread_selfTerminate( (Thread*)this->flusher);
   if(this->ackManager)
      Thread_selfTerminate( (Thread*)this->ackManager);
   if(this->internodeSyncer)
      Thread_selfTerminate( (Thread*)this->internodeSyncer);

   __App_workersStop(this);

   if(this->heartbeatMgr)
      Thread_selfTerminate( (Thread*)this->heartbeatMgr);
   if(this->dgramListener)
      Thread_selfTerminate( (Thread*)this->dgramListener);
}

void __App_joinComponents(App* this)
{
   const char* logContext = "App (join components)";

   if(this->logger)
      Logger_log(this->logger, 4, logContext, "Waiting for components to self-terminate...");


   if(Config_getTuneFileCacheTypeNum(this->cfg) == FILECACHETYPE_Buffered)
      __App_waitForComponentTermination(this, (Thread*)this->flusher);

   __App_waitForComponentTermination(this, (Thread*)this->ackManager);
   __App_waitForComponentTermination(this, (Thread*)this->internodeSyncer);

   __App_workersJoin(this);

   __App_waitForComponentTermination(this, (Thread*)this->heartbeatMgr);
   __App_waitForComponentTermination(this, (Thread*)this->dgramListener);
}

/**
 * @param component the thread that we're waiting for via join(); may be NULL (in which case this
 * method returns immediately)
 */
void __App_waitForComponentTermination(App* this, Thread* component)
{
   const char* logContext = "App (wait for component termination)";

   const int timeoutMS = 2000;

   fhgfs_bool isTerminated;

   if(!component)
      return;

   isTerminated = Thread_timedjoin(component, timeoutMS);
   if(!isTerminated)
   { // print message to show user which thread is blocking
      if(this->logger)
         Logger_logFormatted(this->logger, 2, logContext,
            "Still waiting for this component to stop: %s", Thread_getName(component) );

      Thread_join(component);

      if(this->logger)
         Logger_logFormatted(this->logger, 2, logContext,
            "Component stopped: %s", Thread_getName(component) );
   }
}

void __App_logInfos(App* this)
{
   const char* logContext = "App_logInfos";

   char* localNodeID = Node_getID(this->localNode);

   size_t nicListStrLen = 1024;
   char* nicListStr = os_kmalloc(nicListStrLen);
   char* extendedNicListStr = os_kmalloc(nicListStrLen);
   NicAddressList* nicList = Node_getNicList(this->localNode);
   NicAddressListIter nicIter;


   // print software version (FHGFS_VERSION)
   Logger_logFormatted(this->logger, 1, logContext, "FhGFS Client Version: %s", FHGFS_VERSION);

   // print debug version info
   LOG_DEBUG(this->logger, 1, logContext, "--DEBUG VERSION--");

   // print local clientID
   Logger_logFormatted(this->logger, 2, logContext, "ClientID: %s", localNodeID);

   // list usable network interfaces
   NicAddressListIter_init(&nicIter, nicList);
   nicListStr[0] = 0;
   extendedNicListStr[0] = 0;

   for( ; !NicAddressListIter_end(&nicIter); NicAddressListIter_next(&nicIter) )
   {
      const char* nicTypeStr;
      char* nicAddrStr;
      NicAddress* nicAddr = NicAddressListIter_value(&nicIter);
      char* tmpStr = os_kmalloc(nicListStrLen);

      if(nicAddr->nicType == NICADDRTYPE_RDMA)
         nicTypeStr = "RDMA";
      else
      if(nicAddr->nicType == NICADDRTYPE_SDP)
         nicTypeStr = "SDP";
      else
      if(nicAddr->nicType == NICADDRTYPE_STANDARD)
         nicTypeStr = "TCP";
      else
         nicTypeStr = "Unknown";

      // short NIC info
      os_snprintf(tmpStr, nicListStrLen, "%s%s(%s) ", nicListStr, nicAddr->name, nicTypeStr);
      os_strcpy(nicListStr, tmpStr); // tmpStr to avoid writing to a buffer that we're reading
         // from at the same time

      // extended NIC info
      nicAddrStr = NIC_nicAddrToString(nicAddr);
      os_snprintf(tmpStr, nicListStrLen, "%s\n+ %s", extendedNicListStr, nicAddrStr);
      os_strcpy(extendedNicListStr, tmpStr); // tmpStr to avoid writing to a buffer that we're
         // reading from at the same time
      os_kfree(nicAddrStr);

      SAFE_KFREE(tmpStr);
   }

   NicAddressListIter_uninit(&nicIter);

   Logger_logFormatted(this->logger, 2, logContext, "Usable NICs: %s", nicListStr);
   Logger_logFormatted(this->logger, 4, logContext, "Extended list of usable NICs: %s",
      extendedNicListStr);

   // print net filters
   if(NetFilter_getNumFilterEntries(this->netFilter) )
   {
      Logger_logFormatted(this->logger, 2, logContext, "Net filters: %d",
         (int)NetFilter_getNumFilterEntries(this->netFilter) );
   }

   // print preferred nodes and targets

   if(UInt16List_length(this->preferredMetaNodes) )
   {
      Logger_logFormatted(this->logger, Log_WARNING, logContext, "Preferred nodes: %d",
         (int)UInt16List_length(this->preferredMetaNodes) );
   }

   if(UInt16List_length(this->preferredStorageTargets) )
   {
      Logger_logFormatted(this->logger, Log_WARNING, logContext, "Preferred targets: %d",
         (int)UInt16List_length(this->preferredStorageTargets) );
   }

   // clean up
   SAFE_KFREE(nicListStr);
   SAFE_KFREE(extendedNicListStr);

}

fhgfs_bool __App_workersInit(App* this)
{
   const size_t workerIDLen = 16;

   unsigned numWorkers = Config_getTuneNumWorkers(this->cfg);
   unsigned numRetryWorkers = Config_getTuneNumRetryWorkers(this->cfg);
   unsigned i;

   // standard workers

   // construct the list
   this->workerList = PointerList_construct();

   // construct the workers and add them to the list
   for(i=0; i < numWorkers; i++)
   {
      Worker* worker;
      char* workerID;

      workerID = (char*)os_kmalloc(workerIDLen);
      os_snprintf(workerID, workerIDLen, FHGFS_THREAD_NAME_PREFIX_STR "Worker/%u", i+1);

      worker = Worker_construct(this, workerID, this->workQueue);
      PointerList_append(this->workerList, worker);

      os_kfree(workerID);
   }

   // retry workers

   // construct the list
   this->retryWorkerList = PointerList_construct();

   // construct the workers and add them to the list
   for(i=0; i < numRetryWorkers; i++)
   {
      RetryWorker* worker;
      char* workerID;

      workerID = (char*)os_kmalloc(workerIDLen);
      os_snprintf(workerID, workerIDLen, FHGFS_THREAD_NAME_PREFIX_STR "RtrWrk/%u", i+1);

      worker = RetryWorker_construct(this, workerID, this->retryWorkQueue);
      PointerList_append(this->retryWorkerList, worker);

      os_kfree(workerID);
   }

   return fhgfs_true;
}

void __App_workersStart(App* this)
{
   PointerListIter iter;

   // standard workers

   PointerListIter_init(&iter, this->workerList);

   for(; !PointerListIter_end(&iter); PointerListIter_next(&iter) )
   {
      Worker* worker = (Worker*)PointerListIter_value(&iter);
      Thread_start( (Thread*)worker);
   }

   PointerListIter_uninit(&iter);

   // retry workers

   PointerListIter_init(&iter, this->retryWorkerList);

   for(; !PointerListIter_end(&iter); PointerListIter_next(&iter) )
   {
      RetryWorker* worker = (RetryWorker*)PointerListIter_value(&iter);
      Thread_start( (Thread*)worker);
   }

   PointerListIter_uninit(&iter);
}

void __App_workersStop(App* this)
{
   PointerListIter iter;

   // standard workers

   if(this->workerList) // check whether the list has been initialized
   {
      // note: need two loops here because we don't know if the worker that handles the work will be
      //    the same one that received the self-terminate-request

      PointerListIter_init(&iter, this->workerList);

      for( ; !PointerListIter_end(&iter); PointerListIter_next(&iter) )
      {
         Worker* worker = (Worker*)PointerListIter_value(&iter);

         Thread_selfTerminate( (Thread*)worker);
      }

      PointerListIter_uninit(&iter);


      PointerListIter_init(&iter, this->workerList);

      for( ; !PointerListIter_end(&iter); PointerListIter_next(&iter) )
      {
         WorkQueue_addWork(this->workQueue, (Work*)DummyWork_construct() );
      }

      PointerListIter_uninit(&iter);
   }


   // retry workers

   if(this->retryWorkerList) // check whether the list has been initialized
   {
      // note: need two loops here because we don't know if the worker that handles the work will be
      //    the same one that received the self-terminate-request

      PointerListIter_init(&iter, this->retryWorkerList);

      for( ; !PointerListIter_end(&iter); PointerListIter_next(&iter) )
      {
         RetryWorker* worker = (RetryWorker*)PointerListIter_value(&iter);

         Thread_selfTerminate( (Thread*)worker);
      }

      PointerListIter_uninit(&iter);


      PointerListIter_init(&iter, this->retryWorkerList);

      for( ; !PointerListIter_end(&iter); PointerListIter_next(&iter) )
      {
         WorkQueue_addWork(this->retryWorkQueue, (Work*)DummyRetryableWork_construct() );
      }

      PointerListIter_uninit(&iter);
   }

}

void __App_workersDelete(App* this)
{
   PointerListIter iter;

   // standard workers

   if(this->workerList) // check whether the list has been initialized
   {
      PointerListIter_init(&iter, this->workerList);

      for(; !PointerListIter_end(&iter); PointerListIter_next(&iter) )
      {
         Worker* worker = (Worker*)PointerListIter_value(&iter);

         Worker_destruct(worker);
      }

      PointerListIter_uninit(&iter);

      PointerList_clear(this->workerList);
   }


   // retry workers

   if(this->retryWorkerList) // check whether the list has been initialized
   {
      PointerListIter_init(&iter, this->retryWorkerList);

      for(; !PointerListIter_end(&iter); PointerListIter_next(&iter) )
      {
         RetryWorker* worker = (RetryWorker*)PointerListIter_value(&iter);

         RetryWorker_destruct(worker);
      }

      PointerListIter_uninit(&iter);

      PointerList_clear(this->retryWorkerList);
   }
}

void __App_workersJoin(App* this)
{
   PointerListIter iter;

   // standard workers

   if(this->workerList) // check whether the list has been initialized
   {
      PointerListIter_init(&iter, this->workerList);

      for( ; !PointerListIter_end(&iter); PointerListIter_next(&iter) )
      {
         Worker* worker = (Worker*)PointerListIter_value(&iter);

         __App_waitForComponentTermination(this, (Thread*)worker);
      }

      PointerListIter_uninit(&iter);
   }


   // retry workers

   if(this->retryWorkerList) // check whether the list has been initialized
   {
      PointerListIter_init(&iter, this->retryWorkerList);

      for( ; !PointerListIter_end(&iter); PointerListIter_next(&iter) )
      {
         RetryWorker* worker = (RetryWorker*)PointerListIter_value(&iter);

         __App_waitForComponentTermination(this, (Thread*)worker);
      }

      PointerListIter_uninit(&iter);
   }
}

/**
 * Perform some basic sanity checks to deny mount in case of an error.
 */
fhgfs_bool __App_mountServerCheck(App* this)
{
   const char* logContext = "Mount sanity check";
   Config* cfg = this->cfg;

   unsigned waitMS = Config_getSysMountSanityCheckMS(cfg);

   fhgfs_bool retVal = fhgfs_false;
   fhgfs_bool retriesEnabledOrig = this->connRetriesEnabled;
   char* helperdCheckStr;
   fhgfs_bool mgmtInitDone;
   FhgfsOpsErr statRootErr;
   fhgfs_stat statRootDetails;
   FhgfsOpsErr statStorageErr;
   int64_t storageSpaceTotal;
   int64_t storageSpaceFree;


   if(!waitMS)
      return fhgfs_true; // mount check disabled


   this->connRetriesEnabled = fhgfs_false; // NO _ R E T R I E S

   // check helper daemon

   helperdCheckStr = ExternalHelperd_getHostByName(this->helperd, "localhost");
   if(!helperdCheckStr)
   {
      Logger_logErr(this->logger, logContext, "Communication with user-space helper daemon failed. "
         "Is fhgfs-helperd running?");

      goto error_resetRetries;
   }

   os_kfree(helperdCheckStr);


   // wait for management init

   mgmtInitDone = HeartbeatManager_waitForMgmtInit(this->heartbeatMgr, waitMS);
   if(!mgmtInitDone)
   {
      Logger_logErr(this->logger, logContext, "Waiting for management server initialization "
         "timed out. Are the server settings correct? Is the management daemon running?");

      goto error_resetRetries;
   }

   // check root metadata server

   statRootErr = FhgfsOpsRemoting_statRoot(this, &statRootDetails);
   if(statRootErr != FhgfsOpsErr_SUCCESS)
   {
      Logger_logErrFormatted(this->logger, logContext, "Retrieval of FhGFS root directory entry "
         "failed. Is the metadata server running and registered at the management daemon? "
         "(Error: %s)", FhgfsOpsErr_toErrString(statRootErr) );

      goto error_resetRetries;
   }

   // check storage servers

   statStorageErr = FhgfsOpsRemoting_statStoragePath(this, &storageSpaceTotal, &storageSpaceFree);
   if(statStorageErr != FhgfsOpsErr_SUCCESS)
   {
      Logger_logErrFormatted(this->logger, logContext, "Retrieval of storage server free space "
         "info failed. Are the storage servers running and registered at the management daemon? "
         "Did you remove a storage target directory on a server? (Error: %s)",
         FhgfsOpsErr_toErrString(statStorageErr) );

      goto error_resetRetries;
   }

   retVal = fhgfs_true;


error_resetRetries:
   this->connRetriesEnabled = retriesEnabledOrig; // R E S E T _ R E T R I E S

   return retVal;
}

/**
 * Note: This is actually a Program version, not an App version, but we have it here now because
 * app provides a stable closed source part and we want this version to be fixed at compile time.
 */
const char* App_getVersionStr(void)
{
   return FHGFS_VERSION;
}
