#ifndef COMMON_H_
#define COMMON_H_

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/kthread.h>
#include <linux/sched.h> /* for TASK_COMM_LEN */
#include <linux/string.h>
#include <linux/time.h>
#include <linux/types.h>
#include <linux/version.h>
#include <linux/vmalloc.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <asm/uaccess.h>
#include <stdarg.h>
#include <linux/types.h>
#include <linux/stddef.h>
//#include <linux/compiler.h>
#include <asm/div64.h>

#include <opentk/OpenTk_Common.h>
#include <common/FhgfsTypes.h>
#include <os/OsDeps.h>


/**
 * connection (response) timeouts in ms
 * note: be careful here, because servers not responding for >30secs under high load is nothing
 * unusual, so never use CONN_SHORT_TIMEOUT for IO-related operations.
 */
#define CONN_LONG_TIMEOUT     600000
#define CONN_MEDIUM_TIMEOUT    90000
#define CONN_SHORT_TIMEOUT     30000

#define MIN(a, b) \
   ( ( (a) < (b) ) ? (a) : (b) )
#define MAX(a, b) \
   ( ( (a) < (b) ) ? (b) : (a) )

#define SAFE_VFREE(p) \
   do{ if(p) {os_vfree(p); (p)=NULL;} } while(0)

#define SAFE_VFREE_NOSET(p) \
   do{ if(p) os_vfree(p); } while(0)

#define SAFE_KFREE(p) \
   do{ if(p) {os_kfree(p); (p)=NULL;} } while(0)

#define SAFE_KFREE_NOSET(p) \
   do{ if(p) os_kfree(p); } while(0)

#define SAFE_DESTRUCT(p, destructor) \
   do{if(p) {destructor(p); (p)=NULL;} } while(0)

#define SAFE_DESTRUCT_NOSET(p, destructor) \
   do{if(p) destructor(p); } while(0)

// simply way to convert an enum to a string
#define ENUMSTR(ENUM) #ENUM


// typically used for optional out-args
#define SAFE_ASSIGN(destPointer, sourceValue) \
   do{ if(destPointer) {*(destPointer) = (sourceValue);} } while(0)


// module name (used for logging)
#define FHGFS_MODULE_NAME_STR          "fhgfs"
#define FHGFS_THREAD_NAME_PREFIX_STR   "fhgfs_"

// a printk-version that adds the module name and comm name
#define printk_fhgfs(levelStr, fmtStr, ...) \
   os_printk(levelStr FHGFS_MODULE_NAME_STR ": " "%s(%u): " fmtStr, os_getCurrentCommNameStr(), \
      (unsigned)os_getCurrentPID(), ## __VA_ARGS__)

// dumps stack in case of a buggy condition
#define FHGFS_BUG_ON(condition, msgStr) \
   do { \
      if(unlikely(condition) ) { \
         printk_fhgfs(KERN_ERR, "%s:%d: BUG: %s (dumping stack...)\n", \
            __func__, __LINE__, msgStr); \
         os_dumpStack(); \
      } \
   } while(0)



#ifdef LOG_DEBUG_MESSAGES

#define printk_fhgfs_debug(levelStr, fmtStr, ...) \
   os_printk(levelStr FHGFS_MODULE_NAME_STR ": " "%s(%u): " fmtStr, os_getCurrentCommNameStr(), \
      (unsigned)os_getCurrentPID(), ## __VA_ARGS__)

#define FHGFS_BUG_ON_DEBUG(condition, msgStr)   FHGFS_BUG_ON(condition, msgStr)

#else // !LOG_DEBUG_MESSAGES

#define printk_fhgfs_debug(levelStr, fmtStr, ...) \
   do { /* nothing */ } while(0)

#define FHGFS_BUG_ON_DEBUG(condition, msgStr) \
   do { /* nothing */ } while(0)

#endif // LOG_DEBUG_MESSAGES


// this macro mutes warnings about unused variables
#define IGNORE_UNUSED_VARIABLE(a)   do{ if( ((long)a)==1) {} } while(0)

// __copy_to/from_user() are declared with warnings for ignored return values. this macro mutes the
// warnings (because we check up front whether everything is correct).
#define IGNORE_RETURN_VALUE(a)      IGNORE_UNUSED_VARIABLE(a)

/* wrappers for get_fs()/set_fs() */
#define ACQUIRE_PROCESS_CONTEXT(fs_varname) \
   do { fs_varname = get_fs(); set_fs(KERNEL_DS); } while(0)
#define RELEASE_PROCESS_CONTEXT(fs_varname) \
   set_fs(fs_varname)


// not defined for old kernel versions
#ifndef TASK_COMM_LEN
#define TASK_COMM_LEN   16
#endif

// get effective ID of current FS user/group access
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
   #define currentFsUserID       (current->fsuid)
   #define currentFsGroupID      (current->fsgid)
#else
   #define currentFsUserID       current_fsuid()
   #define currentFsGroupID      current_fsgid()
#endif



/*
 * Debug definitions:
 * - LOG_DEBUG_MESSAGES: Enables logging of some extra debug messages that will not be
 *    available otherwise.
 * - DEBUG_REFCOUNT: Enables debugging of ObjectReferencer::refCount. Error messages will
 *    be logged if refCount is less than zero.
 */


#endif /*COMMON_H_*/
