#ifndef NETMESSAGE_H_
#define NETMESSAGE_H_

#include <common/net/sock/NetworkInterfaceCard.h>
#include <common/net/sock/Socket.h>
#include <common/toolkit/Serialization.h>
#include <common/Common.h>
#include "NetMessageTypes.h"

/**
 * Note: This "class" is not meant to be instantiated directly (consider it to be abstract).
 * It contains some "virtual" methods, as you can see in the struct NetMessage. Only the virtual
 * Method processIncoming(..) has a default implementation.
 * Derived classes have a destructor with a NetMessage-Pointer (instead of the real type)
 * because of the generic virtual destructor signature.
 * Derived classes normally have two constructors: One has no arguments and is used for
 * deserialization. The other one is the standard constructor.
 */

// common message constants
// ========================
#define NETMSG_PREFIX_STR        "fhgfs01" // must be exactly(!!) 8 bytes long
#define NETMSG_PREFIX_STR_LEN    8
#define NETMSG_MIN_LENGTH        NETMSG_HEADER_LENGTH
#define NETMSG_HEADER_LENGTH     24 /* length of the header (see struct NetMessageHeader) */
#define NETMSG_MAX_MSG_SIZE      65536    // 64kB
#define NETMSG_MAX_PAYLOAD_SIZE  ((unsigned)(NETMSG_MAX_MSG_SIZE - NETMSG_HEADER_LENGTH))

#define NETMSG_DEFAULT_USERID    (~0) // non-zero to avoid mixing up with root userID


// forward declaration
struct App;


struct NetMessageHeader;
typedef struct NetMessageHeader NetMessageHeader;
struct NetMessage;
typedef struct NetMessage NetMessage;

static inline void NetMessage_init(NetMessage* this, unsigned short msgType);
static inline NetMessage* NetMessage_construct(unsigned short msgType);
static inline void NetMessage_uninit(NetMessage* this);
static inline void NetMessage_destruct(NetMessage* this);

extern void __NetMessage_deserializeHeader(char* buf, size_t bufLen,
   struct NetMessageHeader* outHeader);
extern void __NetMessage_serializeHeader(NetMessage* this, char* buf);

extern void _NetMessage_serializeDummy(NetMessage* this, char* buf);
extern fhgfs_bool _NetMessage_deserializeDummy(NetMessage* this, const char* buf, size_t bufLen);
extern unsigned _NetMessage_calcMessageLengthDummy(NetMessage* this);


static inline unsigned NetMessage_extractMsgLengthFromBuf(char* recvBuf);
static inline fhgfs_bool NetMessage_serialize(NetMessage* this, char* buf, size_t bufLen);
static inline fhgfs_bool NetMessage_checkHeaderFeatureFlagsCompat(NetMessage* this);

static inline void NetMessage_virtualDestruct(NetMessage* this);

// virtual functions
extern fhgfs_bool NetMessage_processIncoming(NetMessage* this, struct App* app,
   fhgfs_sockaddr_in* fromAddr, struct Socket* sock, char* respBuf, size_t bufLen);
extern unsigned NetMessage_getSupportedHeaderFeatureFlagsMask(NetMessage* this);

// getters & setters
static inline unsigned short NetMessage_getMsgType(NetMessage* this);
static inline unsigned NetMessage_getMsgHeaderFeatureFlags(NetMessage* this);
static inline void NetMessage_setMsgHeaderFeatureFlags(NetMessage* this,
   unsigned msgFeatureFlags);
static inline fhgfs_bool NetMessage_isMsgHeaderFeatureFlagSet(NetMessage* this, unsigned flag);
static inline void NetMessage_addMsgHeaderFeatureFlag(NetMessage* this, unsigned flag);
static inline unsigned NetMessage_getMsgLength(NetMessage* this);
static inline void NetMessage_setMsgHeaderUserID(NetMessage* this, unsigned userID);

static inline void _NetMessage_setMsgType(NetMessage* this, unsigned short msgType);


struct NetMessageHeader
{
   // note: everything in this struct is in host byte order!
   // note: this struct (resp. the header) has a length of 24 bytes (see NETMSG_HEADER_LENGTH)

   unsigned       msgLength; // in bytes
   unsigned       msgFeatureFlags; // feature flags for derived messages (depend on msgType)
   char*          msgPrefix; // NETMSG_PREFIX_STR
   unsigned short msgType; // the type of payload, defined as NETMSGTYPE_x
   unsigned short msgPaddingShort1; // padding (available for future use)
   unsigned       msgUserID; // system user ID for per-user msg queues, stats etc.
};


struct NetMessage
{
   unsigned       msgLength; // in bytes
   unsigned       msgFeatureFlags; // feature flags for derived messages (depend on msgType)
   char*          msgPrefix; // NETMSG_PREFIX_STR
   unsigned short msgType; // the type of payload, defined as NETMSGTYPE_x
   unsigned       msgUserID; // system user ID for per-user msg queues, stats etc.


   // virtual functions
   void (*uninit) (NetMessage* this);

   void (*serializePayload) (NetMessage* this, char* buf);
   fhgfs_bool (*deserializePayload) (NetMessage* this, const char* buf, size_t bufLen);
   unsigned (*calcMessageLength) (NetMessage* this);

   fhgfs_bool (*processIncoming) (NetMessage* this, struct App* app, fhgfs_sockaddr_in* fromAddr,
      struct Socket* sock, char* respBuf, size_t bufLen);
   unsigned (*getSupportedHeaderFeatureFlagsMask) (NetMessage* this);

};


void NetMessage_init(NetMessage* this, unsigned short msgType)
{
   os_memset(this, 0, sizeof(NetMessage) ); // clear function pointers etc.

   this->msgType = msgType;
   this->msgUserID = currentFsUserID; // needs to be set to actual ID by some async flushers etc
   // this->msgFeatureFlags = 0; // zero'ed by memset
   // this->msgLength = 0; // zero'ed by memset
   // this->msgPrefix = NULL; // zero'ed by memset

   this->processIncoming = NetMessage_processIncoming;
   this->getSupportedHeaderFeatureFlagsMask = NetMessage_getSupportedHeaderFeatureFlagsMask;
}

NetMessage* NetMessage_construct(unsigned short msgType)
{
   struct NetMessage* this = os_kmalloc(sizeof(struct NetMessage) );

   NetMessage_init(this, msgType);

   return this;
}

void NetMessage_uninit(NetMessage* this)
{
}

void NetMessage_destruct(NetMessage* this)
{
   NetMessage_uninit(this);

   os_kfree(this);
}


/**
 * recvBuf must be at least NETMSG_MIN_LENGTH long
 */
unsigned NetMessage_extractMsgLengthFromBuf(char* recvBuf)
{
   unsigned msgLengthBufLen;
   unsigned msgLength;
   Serialization_deserializeUInt(recvBuf, sizeof(msgLength), &msgLength, &msgLengthBufLen);

   return msgLength;
}

fhgfs_bool NetMessage_serialize(NetMessage* this, char* buf, size_t bufLen)
{
   if(bufLen < NetMessage_getMsgLength(this) )
      return fhgfs_false;

   __NetMessage_serializeHeader(this, buf);
   this->serializePayload(this, &buf[NETMSG_HEADER_LENGTH]);

   return fhgfs_true;
}

/**
 * Check if the msg sender has set an incompatible feature flag.
 *
 * @return false if an incompatible feature flag was set
 */
fhgfs_bool NetMessage_checkHeaderFeatureFlagsCompat(NetMessage* this)
{
   unsigned unsupportedFlags = ~(this->getSupportedHeaderFeatureFlagsMask(this) );
   if(unlikely(this->msgFeatureFlags & unsupportedFlags) )
      return fhgfs_false; // an unsupported flag was set

   return fhgfs_true;
}

/**
 * Calls the virtual uninit method and kfrees the object.
 */
void NetMessage_virtualDestruct(NetMessage* this)
{
   this->uninit(this);
   os_kfree(this);
}


unsigned short NetMessage_getMsgType(NetMessage* this)
{
   return this->msgType;
}

unsigned NetMessage_getMsgHeaderFeatureFlags(NetMessage* this)
{
   return this->msgFeatureFlags;
}

void NetMessage_setMsgHeaderFeatureFlags(NetMessage* this, unsigned msgFeatureFlags)
{
   this->msgFeatureFlags = msgFeatureFlags;
}

/**
 * Test flag. (For convenience and readability.)
 *
 * @return true if given flag is set.
 */
fhgfs_bool NetMessage_isMsgHeaderFeatureFlagSet(NetMessage* this, unsigned flag)
{
   return (this->msgFeatureFlags & flag) != 0;
}

/**
 * Add another flag without clearing the previously set flags.
 */
void NetMessage_addMsgHeaderFeatureFlag(NetMessage* this, unsigned flag)
{
   this->msgFeatureFlags |= flag;
}

unsigned NetMessage_getMsgLength(NetMessage* this)
{
   if(!this->msgLength)
      this->msgLength = this->calcMessageLength(this);

   return this->msgLength;
}

void NetMessage_setMsgHeaderUserID(NetMessage* this, unsigned userID)
{
   this->msgUserID = userID;
}

void _NetMessage_setMsgType(NetMessage* this, unsigned short msgType)
{
   this->msgType = msgType;
}


#endif /*NETMESSAGE_H_*/
