#include <opentk/net/sock/OpenTk_NetworkInterfaceCard.h>
#include <common/net/sock/RDMASocket.h>
#include <common/net/sock/StandardSocket.h>
#include <common/net/sock/NetworkInterfaceCard.h>
#include <common/toolkit/ListTk.h>

#include <linux/if_arp.h>
#include <linux/in.h>
#include <linux/inetdevice.h>


#define NIC_STRING_LEN  1024

#define inaddrr(x)   (*(struct in_addr*) &ifr.x[sizeof(short) /*=sizeof sa.sin_port*/ ] )


static fhgfs_bool __NIC_fillNicAddress(struct net_device* dev, NicAddrType_t nicType,
   NicAddress* outAddr);



void NIC_findAll(StrCpyList* allowedInterfaces, fhgfs_bool useSDP, fhgfs_bool useRDMA, NicAddressList* outList)
{
   struct net_device *dev;

   // find standard TCP/IP interfaces
   __NIC_findAllTCP(allowedInterfaces, outList);


   // find SDP interfaces
   if(useSDP && __NIC_checkSDPAvailable() )
   {
      // foreach network device
      for(dev = fhgfs_NIC_getFirstNetDevice(); dev; dev = fhgfs_NIC_getNextNetDevice(dev) )
      {
         NicAddress* nicAddr = (NicAddress*)os_kmalloc(sizeof(NicAddress) );
         ssize_t metricByListPos = 0;

         if(__NIC_fillNicAddress(dev, NICADDRTYPE_SDP, nicAddr) &&
            (!StrCpyList_length(allowedInterfaces) ||
             ListTk_listContains(nicAddr->name, allowedInterfaces, &metricByListPos) ) )
         {
            nicAddr->metric = (int)metricByListPos;
            NicAddressList_append(outList, nicAddr);
         }
         else
         { // netdevice rejected => clean up
            kfree(nicAddr);
         }
      }
   }

   // find RDMA interfaces (based on TCP/IP interfaces query results)
   if(useRDMA && RDMASocket_rdmaDevicesExist() )
   {
      NicAddressList tcpInterfaces;
      NicAddressList_init(&tcpInterfaces);

      __NIC_findAllTCP(allowedInterfaces, &tcpInterfaces);

      __NIC_filterInterfacesForRDMA(&tcpInterfaces, outList);

      ListTk_kfreeNicAddressListElems(&tcpInterfaces);
      NicAddressList_uninit(&tcpInterfaces);
   }
}

void __NIC_findAllTCP(StrCpyList* allowedInterfaces, NicAddressList* outList)
{
   struct net_device *dev;

   // find standard TCP/IP interfaces

   // foreach network device
   for(dev = fhgfs_NIC_getFirstNetDevice(); dev; dev = fhgfs_NIC_getNextNetDevice(dev) )
   {
      NicAddress* nicAddr = (NicAddress*)os_kmalloc(sizeof(NicAddress) );
      ssize_t metricByListPos = 0;

      if(__NIC_fillNicAddress(dev, NICADDRTYPE_STANDARD, nicAddr) &&
         (!StrCpyList_length(allowedInterfaces) ||
          ListTk_listContains(nicAddr->name, allowedInterfaces, &metricByListPos) ) )
      {
         nicAddr->metric = (int)metricByListPos;
         NicAddressList_append(outList, nicAddr);
      }
      else
      { // netdevice rejected => clean up
         kfree(nicAddr);
      }
   }
}

fhgfs_bool __NIC_fillNicAddress(struct net_device* dev, NicAddrType_t nicType, NicAddress* outAddr)
{
   struct ifreq ifr;
   struct in_device* in_dev;
   struct in_ifaddr *ifa;

   // name
   strcpy(outAddr->name, dev->name);


   // SIOCGIFFLAGS:
   // Get interface flags
   ifr.ifr_flags = dev_get_flags(dev);

   if(ifr.ifr_flags & IFF_LOOPBACK)
      return fhgfs_false; // loopback interface => skip


   // SIOCGIFHWADDR:
   if(!dev->addr_len)
      memset(ifr.ifr_hwaddr.sa_data, 0, sizeof(ifr.ifr_hwaddr.sa_data) );
   else
      memcpy(ifr.ifr_hwaddr.sa_data, dev->dev_addr,
             min(sizeof(ifr.ifr_hwaddr.sa_data), (size_t) dev->addr_len) );
   ifr.ifr_hwaddr.sa_family = dev->type;


   // select which hardware types to process
   // (on Linux see /usr/include/linux/if_arp.h for the whole list)
   switch(ifr.ifr_hwaddr.sa_family)
   {
      case ARPHRD_LOOPBACK: return fhgfs_false;

      default:
      {
         // make sure we allow SDP for IB only (because an SDP socket domain is valid for other
         // NIC types as well, but cannot connect then, of course
         if( (nicType == NICADDRTYPE_SDP) && (ifr.ifr_hwaddr.sa_family != ARPHRD_INFINIBAND) )
            return fhgfs_false;
      } break;
   }


   // copy nicType
   outAddr->nicType = nicType;

   // copy hardware address
   memcpy(&outAddr->hwAddr, &ifr.ifr_addr.sa_data, IFHWADDRLEN);


   // ip
   // note: code based on /net/ipv4/devinet.c function inet_gifconf(..)

   in_dev = (struct in_device*)dev->ip_ptr;
   if (!in_dev || (ifa = in_dev->ifa_list) == NULL)
      return fhgfs_false;

   outAddr->ipAddr = ifa->ifa_local; // ip address
   outAddr->broadcastAddr = ifa->ifa_broadcast; // broadcast address

   // code to read multiple addresses
   /*
   for (; ifa; ifa = ifa->ifa_next)
   {
      (*(struct sockaddr_in *)&ifr.ifr_addr).sin_family = AF_INET;
      (*(struct sockaddr_in *)&ifr.ifr_addr).sin_addr.s_addr =
                        ifa->ifa_local;
   }
   */


   // SIOCGIFMETRIC:
   // Get the metric of the interface (currently not supported by the kernel)
   ifr.ifr_metric = 0;
   outAddr->metric = ifr.ifr_metric;


   return fhgfs_true;
}

/**
 * @return string will be kalloced and must be kfreed later
 */
char* NIC_nicAddrToString(NicAddress* nicAddr)
{
   char* nicAddrStr;
   char* resultStr;
   char ipStr[NICADDRESS_IP_STR_LEN];
   char* hwStr;
   const char* typeStr;

   nicAddrStr = (char*)os_kmalloc(NIC_STRING_LEN);

   NicAddress_ipToStr(nicAddr->ipAddr, ipStr);

   hwStr = NIC_hwAddrToString(nicAddr->hwAddr);

   if(nicAddr->nicType == NICADDRTYPE_RDMA)
      typeStr = "RDMA";
   else
   if(nicAddr->nicType == NICADDRTYPE_SDP)
      typeStr = "SDP";
   else
   if(nicAddr->nicType == NICADDRTYPE_STANDARD)
      typeStr = "TCP";
   else
      typeStr = "Unknown";

   snprintf(nicAddrStr, NIC_STRING_LEN,
      "%s[ip addr: %s; hw addr: %s; metric: %d; type: %s]",
      nicAddr->name, ipStr, hwStr, nicAddr->metric, typeStr);

   kfree(hwStr);


   resultStr = StringTk_strDup(nicAddrStr); // create string with min required size
   kfree(nicAddrStr);

   return resultStr;
}

/**
 * @return string will be kalloced and must be kfreed later
 */
char* NIC_nicAddrToStringLight(NicAddress* nicAddr)
{
   char* nicAddrStr;
   char* resultStr;
   char ipStr[NICADDRESS_IP_STR_LEN];
   char broadcastStr[NICADDRESS_IP_STR_LEN];
   char* hwStr;
   const char* typeStr;

   nicAddrStr = (char*)os_kmalloc(NIC_STRING_LEN);

   NicAddress_ipToStr(nicAddr->ipAddr, ipStr);

   NicAddress_ipToStr(nicAddr->broadcastAddr, broadcastStr);

   hwStr = NIC_hwAddrToString(nicAddr->hwAddr);

   if(nicAddr->nicType == NICADDRTYPE_RDMA)
      typeStr = "RDMA";
   else
   if(nicAddr->nicType == NICADDRTYPE_SDP)
      typeStr = "SDP";
   else
   if(nicAddr->nicType == NICADDRTYPE_STANDARD)
      typeStr = "TCP";
   else
      typeStr = "Unknown";

   snprintf(nicAddrStr, NIC_STRING_LEN,
      "%s[ip addr: %s; type: %s]", nicAddr->name, ipStr, typeStr);

   kfree(hwStr);


   resultStr = StringTk_strDup(nicAddrStr); // create string with min required size
   kfree(nicAddrStr);

   return resultStr;
}

/**
 * @return string will be kalloced and must be kfreed later
 */
char* NIC_hwAddrToString(char* hwAddr)
{
   char* hwAddrStr;
   int i;

   hwAddrStr = (char*)os_kmalloc(IFHWADDRLEN*3);
   // *3 for two letters per byte plus dots plus trailing zero


   { // first loop walk (without the dot)
      sprintf(hwAddrStr, "%2.2x", (unsigned char)hwAddr[0]);
   }

   for(i=1; i < IFHWADDRLEN; i++)
   {
      sprintf(&hwAddrStr[(i*3)-1], ".%2.2x", (unsigned char)hwAddr[i]);
   }

   return(hwAddrStr);
}

fhgfs_bool NIC_supportsSDP(NicAddressList* nicList)
{
   fhgfs_bool sdpSupported = fhgfs_false;

   NicAddressListIter iter;
   NicAddressListIter_init(&iter, nicList);

   for( ; !NicAddressListIter_end(&iter); NicAddressListIter_next(&iter) )
   {
      if(NicAddressListIter_value(&iter)->nicType == NICADDRTYPE_SDP)
      {
         sdpSupported = fhgfs_true;
         break;
      }
   }

   NicAddressListIter_uninit(&iter);

   return sdpSupported;
}

fhgfs_bool NIC_supportsRDMA(NicAddressList* nicList)
{
   fhgfs_bool rdmaSupported = fhgfs_false;

   NicAddressListIter iter;
   NicAddressListIter_init(&iter, nicList);

   for( ; !NicAddressListIter_end(&iter); NicAddressListIter_next(&iter) )
   {
      if(NicAddressListIter_value(&iter)->nicType == NICADDRTYPE_RDMA)
      {
         rdmaSupported = fhgfs_true;
         break;
      }
   }

   NicAddressListIter_uninit(&iter);

   return rdmaSupported;
}

void NIC_supportedCapabilities(NicAddressList* nicList, NicListCapabilities* outCapabilities)
{
   outCapabilities->supportsSDP = NIC_supportsSDP(nicList);
   outCapabilities->supportsRDMA = NIC_supportsRDMA(nicList);
}


fhgfs_bool __NIC_checkSDPAvailable(void)
{
   fhgfs_bool retVal = fhgfs_false;

   Socket* sock = (Socket*)StandardSocket_construct(PF_SDP, SOCK_STREAM, 0);
   if(Socket_getSockValid(sock) )
      retVal = fhgfs_true;

   Socket_virtualDestruct(sock);

   return retVal;
}

/**
 * Checks a list of TCP/IP interfaces for RDMA-capable interfaces.
 */
void __NIC_filterInterfacesForRDMA(NicAddressList* nicList, NicAddressList* outList)
{
   // Note: This works by binding an RDMASocket to each IP of the passed list.

   NicAddressListIter iter;
   NicAddressListIter_init(&iter, nicList);

   for( ; !NicAddressListIter_end(&iter); NicAddressListIter_next(&iter) )
   {
      RDMASocket rdmaSock;
      Socket* sock = (Socket*)&rdmaSock;
      NicAddress* nicAddr = NicAddressListIter_value(&iter);
      fhgfs_bool bindRes;

      RDMASocket_init(&rdmaSock);

      bindRes = sock->bindToAddr(sock, &nicAddr->ipAddr, 0);

      if(bindRes)
      { // we've got an RDMA-capable interface => append it to outList
         NicAddress* nicAddrCopy = os_kmalloc(sizeof(NicAddress) );

         *nicAddrCopy = *nicAddr;

         nicAddrCopy->nicType = NICADDRTYPE_RDMA;

         NicAddressList_append(outList, nicAddrCopy);
      }

      sock->uninit(sock);
   }

   NicAddressListIter_uninit(&iter);
}
