#include "NodeConnPool.h"

#include <common/net/message/control/SetChannelDirectMsg.h>
#include <common/toolkit/ListTk.h>
#include <common/toolkit/NetFilter.h>
#include <app/App.h>
#include <app/config/Config.h>
#include <common/threading/Thread.h>


#define NODECONNPOOL_SHUTDOWN_WAITTIMEMS              100


/**
 * Note: localNicCaps should to be initialized before acquiring a stream.
 *
 * @param nicList an internal copy will be created
 */
void NodeConnPool_init(NodeConnPool* this, struct App* app, unsigned short streamPort,
   NicAddressList* nicList)
{
   Config* cfg = App_getConfig(app);

   this->app = app;

   this->mutex = Mutex_construct();
   this->changeCond = Condition_construct();

   ConnectionList_init(&this->connList);

   this->nicList = ListTk_cloneNicAddressList(nicList);

   this->establishedConns = 0;
   this->availableConns = 0;

   this->maxConns = Config_getConnMaxInternodeNum(cfg);
   this->nonPrimaryExpirationCount = Config_getConnNonPrimaryExpiration(cfg);

   this->streamPort = streamPort;
   os_memset(&this->localNicCaps, 0, sizeof(this->localNicCaps) );
   os_memset(&this->stats, 0, sizeof(this->stats) );

   this->logConnErrors = fhgfs_true;
}

NodeConnPool* NodeConnPool_construct(struct App* app, unsigned short streamPort,
   NicAddressList* nicList)
{
   NodeConnPool* this = (NodeConnPool*)os_kmalloc(sizeof(NodeConnPool) );

   NodeConnPool_init(this, app, streamPort, nicList);

   return this;
}

void NodeConnPool_uninit(NodeConnPool* this)
{
   Logger* log = App_getLogger(this->app);
   const char* logContext = "NodeConn (uninit)";

   // close open connections
   if(ConnectionList_length(&this->connList) > 0)
   {
      Logger_logFormatted(log, 4, logContext, "Closing %d connections...",
         ConnectionList_length(&this->connList) );
   }

   // Note: invalidateStreamSocket changes the original connList
   //    (so we must invalidate the iterator each time)
   while(ConnectionList_length(&this->connList) )
   {
      ConnectionListIter connIter;
      PooledSocket* socket;

      ConnectionListIter_init(&connIter, &this->connList);

      socket = ConnectionListIter_value(&connIter);

      __NodeConnPool_invalidateSpecificStreamSocket(this, (Socket*)socket);

      ConnectionListIter_uninit(&connIter);
   }

   // delete nicList elems
   ListTk_kfreeNicAddressListElems(this->nicList);

   // normal clean-up
   NicAddressList_destruct(this->nicList);

   Condition_destruct(this->changeCond);
   Mutex_destruct(this->mutex);
}

void NodeConnPool_destruct(NodeConnPool* this)
{
   NodeConnPool_uninit(this);

   os_kfree(this);
}

/**
 * Note: Will block if no stream socket is immediately available.
 *
 * @return connected socket; NULL on error or pending signal
 */
Socket* NodeConnPool_acquireStreamSocket(NodeConnPool* this)
{
   return NodeConnPool_acquireStreamSocketEx(this, fhgfs_true);
}

/**
 * @param fhgfs_true to allow waiting if all avaiable conns are currently in use; if waiting is not
 * allowed, there is no difference between all conns in use and connection error (so it is expected
 * that the caller will sooner or later allow waiting to detect conn errors).
 * @return connected socket; NULL on error or pending signal
 */
Socket* NodeConnPool_acquireStreamSocketEx(NodeConnPool* this, fhgfs_bool allowWaiting)
{
   const char* logContext = "NodeConn (acquire stream)";

   Logger* log = App_getLogger(this->app);
   NetFilter* netFilter = App_getNetFilter(this->app);

   Socket* sock = NULL;
   unsigned short port;
   NicAddressListIter nicIter;
   fhgfs_bool isPrimaryInterface = fhgfs_true; // used to set expiration for non-primary interfaces;
      // "primary" means: first interface in the list that is supported by client and server

   if(unlikely(Thread_isSignalPending() ) )
      return NULL; // no need to try if a signal is pending

   Mutex_lock(this->mutex); // L O C K

   if(!this->availableConns && (this->establishedConns == this->maxConns) )
   { // wait for a conn to become available (or disconnected)

      if(!allowWaiting)
      { // all conns in use and waiting not allowed => exit
         Mutex_unlock(this->mutex); // U N L O C K
         return NULL;
      }

      while(!this->availableConns && (this->establishedConns == this->maxConns) )
      { // sleep interruptbile
         cond_wait_res_t waitRes = Condition_waitInterruptible(this->changeCond, this->mutex);
         if(unlikely(waitRes == COND_WAIT_SIGNAL) )
         { // interrupted by signal
            Mutex_unlock(this->mutex); // U N L O C K
            return NULL;
         }
      }
   }


   if(likely(this->availableConns) )
   { // established connection available => grab it
      PooledSocket* pooledSock;

      ConnectionListIter connIter;
      ConnectionListIter_init(&connIter, &this->connList);

      while(!PooledSocket_isAvailable(ConnectionListIter_value(&connIter) ) )
         ConnectionListIter_next(&connIter);

      pooledSock = ConnectionListIter_value(&connIter);
      PooledSocket_setAvailable(pooledSock, fhgfs_false);
      PooledSocket_setHasActivity(pooledSock);

      this->availableConns--;

      ConnectionListIter_uninit(&connIter);

      Mutex_unlock(this->mutex); // U N L O C K

      return (Socket*)pooledSock;
   }


   // no conn available, but maxConns not reached yet => establish a new conn

   port = this->streamPort;

   this->establishedConns++;

   Mutex_unlock(this->mutex); // U N L O C K

   // walk over all available NICs, create the corresponding socket and try to connect

   NicAddressListIter_init(&nicIter, this->nicList);

   for( ; !NicAddressListIter_end(&nicIter); NicAddressListIter_next(&nicIter) )
   {
      NicAddress* nicAddr = NicAddressListIter_value(&nicIter);
      fhgfs_bool connectRes;

      char* endpointStr = SocketTk_endpointAddrToString(&nicAddr->ipAddr, port);

      if(!NetFilter_isAllowed(netFilter, nicAddr->ipAddr) )
         goto continue_clean_endpointStr;

      switch(nicAddr->nicType)
      {
         case NICADDRTYPE_RDMA:
         { // RDMA
            if(!this->localNicCaps.supportsRDMA)
               goto continue_clean_endpointStr;

            Logger_logTopFormatted(log, LogTopic_CONN, Log_NOTICE, logContext,
               "Establishing new RDMA connection to: %s", endpointStr);
            sock = (Socket*)RDMASocket_construct();
         } break;
         case NICADDRTYPE_SDP:
         { // SDP
            if(!this->localNicCaps.supportsSDP)
               goto continue_clean_endpointStr;

            Logger_logTopFormatted(log, LogTopic_CONN, Log_NOTICE, logContext,
               "Establishing new SDP connection to: %s", endpointStr);
            sock = (Socket*)StandardSocket_constructSDP();
         } break;
         case NICADDRTYPE_STANDARD:
         { // TCP
            Logger_logTopFormatted(log, LogTopic_CONN, Log_NOTICE, logContext,
               "Establishing new TCP connection to: %s", endpointStr);
            sock = (Socket*)StandardSocket_constructTCP();
         } break;

         default:
         { // unknown
            if(this->logConnErrors)
               Logger_logTopFormatted(log, LogTopic_CONN, Log_WARNING, logContext,
                  "Skipping unknown connection type to: %s", endpointStr);
            goto continue_clean_endpointStr;
         }
      } // end of switch

      // check whether the socket was successfully initialized
      if(!Socket_getSockValid( (Socket*)sock) )
      { // socket initialization failed
         if(this->logConnErrors)
            Logger_logFormatted(log, Log_WARNING, logContext, "Socket initialization failed: %s",
               endpointStr);

         Socket_virtualDestruct(sock);

         goto continue_clean_endpointStr;
      }

      // the actual connection attempt
      __NodeConnPool_applySocketOptionsPreConnect(this, (Socket*)sock);

      connectRes = sock->connectByIP(sock, &nicAddr->ipAddr, port);

      if(connectRes)
      { // connected
         Logger_logTopFormatted(log, LogTopic_CONN, 3, logContext, "Connected: %s", endpointStr);

         __NodeConnPool_applySocketOptionsConnected(this, (Socket*)sock);

         if(!isPrimaryInterface) // non-primary => set expiration counter
         {
            PooledSocket_setExpirationCounter(
               (PooledSocket*)sock, this->nonPrimaryExpirationCount);
         }

         os_kfree(endpointStr);

         break;
      }
      else
      { // not connected
         if(this->logConnErrors)
            Logger_logTopFormatted(log, LogTopic_CONN, 3, logContext, "Connect failed: %s",
               endpointStr);

         // the next interface is definitely non-primary
         isPrimaryInterface = fhgfs_false;

         Socket_virtualDestruct(sock);
      }


   continue_clean_endpointStr:
      os_kfree(endpointStr);
   }

   Mutex_lock(this->mutex); // L O C K

   if(!NicAddressListIter_end(&nicIter) )
   { // success => add to list (as unavailable) and update stats
      PooledSocket* pooledSock = (PooledSocket*)sock;

      ConnectionList_append(&this->connList, pooledSock);
      __NodeConnPool_statsAddNic(this, PooledSocket_getNicType(pooledSock) );
   }
   else
   { // absolutely unable to connect
      NicAddressListIter nicIter;
      NicAddress* nicAddr;
      char* endpointStr;

      sock = NULL;
      this->establishedConns--;

      /* If we had been suspended (Condition_wait() above) and woken up again, we now need to wake
       *  up another thread, as we are not going to make use of this connection. */
      Condition_signal(this->changeCond);

      NicAddressListIter_init(&nicIter, this->nicList);
      nicAddr = NicAddressListIter_value(&nicIter);
      endpointStr = SocketTk_endpointAddrToString(&nicAddr->ipAddr, port);

      if(this->logConnErrors)
         Logger_logTopFormatted(log, LogTopic_CONN, Log_CRITICAL, logContext,
            "Connect failed on all available routes: %s", endpointStr);

      NicAddressListIter_uninit(&nicIter);
      os_kfree(endpointStr);
   }

   NicAddressListIter_uninit(&nicIter);

   Mutex_unlock(this->mutex); // U N L O C K

   return sock;
}

void NodeConnPool_releaseStreamSocket(NodeConnPool* this, Socket* sock)
{
   PooledSocket* pooledSock = (PooledSocket*)sock;

   // test whether this socket has expired

   if(unlikely(PooledSocket_testAndDecExpirationCounter(pooledSock) ) )
   { // this socket just expired => invalidate it
      __NodeConnPool_invalidateSpecificStreamSocket(this, sock);
      return;
   }

   // mark the socket as available

   Mutex_lock(this->mutex);

   this->availableConns++;

   PooledSocket_setAvailable(pooledSock, fhgfs_true);

   Condition_signal(this->changeCond);

   Mutex_unlock(this->mutex);
}

void NodeConnPool_invalidateStreamSocket(NodeConnPool* this, Socket* sock)
{
   Logger* log = App_getLogger(this->app);
   const char* logContext = "NodeConn (invalidate stream)";

   unsigned numInvalidated;

   /* note: we use invalidateAvailableStreams first here, because we want to keep other threads
      from trying to acquire an available socket immediately - and the parameter socket isn't
      availabe anyways. (we don't want to acquire a mutex here that would block the whole pool
      while we're just waiting for the sock-close response messages) */


   numInvalidated = __NodeConnPool_invalidateAvailableStreams(this, fhgfs_false);

   Logger_logTopFormatted(log, LogTopic_CONN, 4, logContext,
      "Invalidated %u pooled connections. (1 more invalidation pending...)", numInvalidated);

   __NodeConnPool_invalidateSpecificStreamSocket(this, sock);
}

void __NodeConnPool_invalidateSpecificStreamSocket(NodeConnPool* this, Socket* sock)
{
   const char* logContext = "NodeConn (invalidate stream)";
   Logger* log = App_getLogger(this->app);

   fhgfs_bool sockValid = fhgfs_true;
   ConnectionListIter connIter;
   PooledSocket* pooledSock = (PooledSocket*)sock;
   fhgfs_bool shutdownRes;


   Mutex_lock(this->mutex); // L O C K

   ConnectionListIter_init(&connIter, &this->connList);

   for( ; !ConnectionListIter_end(&connIter); ConnectionListIter_next(&connIter) )
   {
      Socket* listSock = (Socket*)ConnectionListIter_value(&connIter);
      if(listSock == sock)
         break;
   }

   if(ConnectionListIter_end(&connIter) )
   { // socket not in the list
      Logger_logErrFormatted(log, logContext,
         "Tried to remove a socket that was not found in the pool: %s", Socket_getPeername(sock) );
      sockValid = fhgfs_false;
   }
   else
   {
      this->establishedConns--;

      ConnectionListIter_remove(&connIter);
      __NodeConnPool_statsRemoveNic(this, PooledSocket_getNicType(pooledSock) );

      Condition_signal(this->changeCond);
   }

   Mutex_unlock(this->mutex); // U N L O C K

   if(!sockValid)
      return;


   shutdownRes = sock->shutdownAndRecvDisconnect(sock, NODECONNPOOL_SHUTDOWN_WAITTIMEMS);
   if(shutdownRes)
   {
      Logger_logTopFormatted(log, LogTopic_CONN, 3, logContext, "Disconnected: %s",
         Socket_getPeername(sock) );
   }
   else
   {
      Logger_logTopFormatted(log, LogTopic_CONN, 3, logContext, "Hard disconnect: %s",
         Socket_getPeername(sock) );
   }

   Socket_virtualDestruct(sock);
}

/**
 * Invalidate (disconnect) connections that are currently not acquired.
 *
 * @param idleStreamsOnly invalidate only conns that are marked as idle (ie don't have the activity
 * flag set).
 * @return number of invalidated streams
 */
unsigned __NodeConnPool_invalidateAvailableStreams(NodeConnPool* this, fhgfs_bool idleStreamsOnly)
{
   /* note: we have TWO STAGES here, because we don't want to hold the mutex and block everything
      while we're waiting for the conns to be dropped. */

   unsigned numInvalidated = 0; // retVal
   ConnectionListIter connIter;
   ConnectionListIter availableIter;

   ConnectionList availableConnsList;
   ConnectionList_init(&availableConnsList);

   Mutex_lock(this->mutex); // L O C K

   // STAGE 1: grab all sockets that should be disconnected

   ConnectionListIter_init(&connIter, &this->connList);

   for( ; !ConnectionListIter_end(&connIter); ConnectionListIter_next(&connIter) )
   {
      PooledSocket* sock = ConnectionListIter_value(&connIter);

      if(!PooledSocket_isAvailable(sock) )
         continue; // this one is currently in use

      if(idleStreamsOnly && PooledSocket_getHasActivity(sock) )
         continue; // idle-only requested and this one was not idle

      PooledSocket_setAvailable(sock, fhgfs_false);
      this->availableConns--;
      ConnectionList_append(&availableConnsList, sock);

      numInvalidated++;
   }

   ConnectionListIter_uninit(&connIter);

   Mutex_unlock(this->mutex); // U N L O C K


   // STAGE 2: invalidate all grabbed sockets

   ConnectionListIter_init(&availableIter, &availableConnsList);

   for( ; !ConnectionListIter_end(&availableIter); ConnectionListIter_next(&availableIter) )
   {
      PooledSocket* sock = ConnectionListIter_value(&availableIter);
      __NodeConnPool_invalidateSpecificStreamSocket(this, (Socket*)sock);
   }

   ConnectionListIter_uninit(&availableIter);
   ConnectionList_uninit(&availableConnsList);

   return numInvalidated;
}

/**
 * Disconnect all established (available) connections.
 *
 * @return number of disconnected streams
 */
unsigned NodeConnPool_disconnectAvailableStreams(NodeConnPool* this)
{
   unsigned numInvalidated;

   numInvalidated = __NodeConnPool_invalidateAvailableStreams(this, fhgfs_false);

   return numInvalidated;
}

/**
 * Note: There is no locking around dropping and resetting afterwards, so there should only be one
 * thread which calls this function (=> InternodeSyncer).
 *
 * @return number of disconnected streams
 */
unsigned NodeConnPool_disconnectAndResetIdleStreams(NodeConnPool* this)
{
   unsigned numInvalidated;

   numInvalidated = __NodeConnPool_invalidateAvailableStreams(this, fhgfs_true);

   __NodeConnPool_resetStreamsIdleFlag(this);

   return numInvalidated;
}

/**
 * Resets the activity flag of all available connections to mark them as idle.
 */
void __NodeConnPool_resetStreamsIdleFlag(NodeConnPool* this)
{
   ConnectionListIter connIter;

   Mutex_lock(this->mutex); // L O C K

   ConnectionListIter_init(&connIter, &this->connList);

   for( ; !ConnectionListIter_end(&connIter); ConnectionListIter_next(&connIter) )
   {
      PooledSocket* sock = ConnectionListIter_value(&connIter);

      if(!PooledSocket_isAvailable(sock) )
         continue; // this one is currently in use

      PooledSocket_resetHasActivity(sock);
   }

   ConnectionListIter_uninit(&connIter);

   Mutex_unlock(this->mutex); // U N L O C K
}

fhgfs_bool __NodeConnPool_applySocketOptionsPreConnect(NodeConnPool* this, Socket* sock)
{
   Config* cfg = App_getConfig(this->app);

   NicAddrType_t sockType = Socket_getSockType(sock);

   if(sockType == NICADDRTYPE_STANDARD)
   {
      StandardSocket* stdSock = (StandardSocket*)sock;
      int bufSize = Config_getConnRDMABufNum(cfg) * Config_getConnRDMABufSize(cfg); /* we're just
         re-using the rdma buffer settings here. should later be changed to separate settings */

      StandardSocket_setSoRcvBuf(stdSock, bufSize);
   }
   else
   if(sockType == NICADDRTYPE_RDMA)
   {
      RDMASocket* rdmaSock = (RDMASocket*)sock;
      RDMASocket_setBuffers(rdmaSock, Config_getConnRDMABufNum(cfg),
         Config_getConnRDMABufSize(cfg) );
      RDMASocket_setTypeOfService(rdmaSock, Config_getConnRDMATypeOfService(cfg));
   }

   return fhgfs_true;
}


fhgfs_bool __NodeConnPool_applySocketOptionsConnected(NodeConnPool* this, Socket* sock)
{
   Logger* log = App_getLogger(this->app);
   const char* logContext = "NodeConn (apply socket options)";

   fhgfs_bool allSuccessful = fhgfs_true;

   char* sendBuf;
   size_t sendBufLen;
   SetChannelDirectMsg directMsg;
   size_t sendRes;
   NicAddrType_t sockType = Socket_getSockType(sock);

   // apply general socket options
   if( (sockType == NICADDRTYPE_STANDARD) ||
       (sockType == NICADDRTYPE_SDP) )
   {
      StandardSocket* standardSock = (StandardSocket*)sock;
      fhgfs_bool noDelayRes;
      fhgfs_bool keepAliveRes;

      noDelayRes = StandardSocket_setTcpNoDelay(standardSock, fhgfs_true);
      if(!noDelayRes)
      {
         Logger_log(log, 3, logContext, "Failed to enable TcpNoDelay");
         allSuccessful = fhgfs_false;
      }

      keepAliveRes = StandardSocket_setSoKeepAlive(standardSock, fhgfs_true);
      if(!keepAliveRes)
      {
         Logger_log(log, 3, logContext, "Failed to enable SoKeepAlive");
         allSuccessful = fhgfs_false;
      }
   }

   // send control messages

   SetChannelDirectMsg_initFromValue(&directMsg, 0);
   sendBufLen = NetMessage_getMsgLength( (NetMessage*)&directMsg);
   sendBuf = (char*)os_kmalloc(sendBufLen);

   NetMessage_serialize( (NetMessage*)&directMsg, sendBuf, sendBufLen);
   sendRes = sock->send(sock, sendBuf, sendBufLen, 0);
   if(sendRes <= 0)
   {
      Logger_log(log, 2, logContext, "Failed to set channel to indirect mode");
      allSuccessful = fhgfs_false;
   }

   os_kfree(sendBuf);
   SetChannelDirectMsg_uninit( (NetMessage*)&directMsg);

   return allSuccessful;
}

void NodeConnPool_getStats(NodeConnPool* this, NodeConnPoolStats* outStats)
{
   Mutex_lock(this->mutex);

   *outStats = this->stats;

   Mutex_unlock(this->mutex);
}

/**
 * Gets the first socket peer name for a connection to the specified nicType.
 * Only works on currently available connections.
 *
 * @outBuf buffer for the peer name string (contains "n/a" if no such connection exists or is
 * currently not available).
 * @param outIsNonPrimary set to true if conn in outBuf has an expiration counter, false otherwise
 * @return number of chars written to buf
 */
unsigned NodeConnPool_getFirstPeerName(NodeConnPool* this, NicAddrType_t nicType, ssize_t outBufLen,
   char* outBuf, fhgfs_bool* outIsNonPrimary)
{
   unsigned numWritten = 0; // return value
   ConnectionListIter connIter;
   fhgfs_bool foundMatch = fhgfs_false;

   Mutex_lock(this->mutex);

   ConnectionListIter_init(&connIter, &this->connList);

   for( ; !ConnectionListIter_end(&connIter); ConnectionListIter_next(&connIter) )
   {
      PooledSocket* pooledSock = ConnectionListIter_value(&connIter);
      Socket* sock = (Socket*)pooledSock;

      if(PooledSocket_isAvailable(pooledSock) &&
         (Socket_getSockType(sock) == nicType) )
      { // found a match => print to buf and stop
         numWritten += os_scnprintf(outBuf, outBufLen, "%s", Socket_getPeername(sock) );

         *outIsNonPrimary = (PooledSocket_getExpirationCounter(pooledSock) > 0);

         foundMatch = fhgfs_true;
         break;
      }
   }

   ConnectionListIter_uninit(&connIter);

   if(!foundMatch)
   { // print "n/a"
      numWritten += os_scnprintf(outBuf, outBufLen, "busy");

      *outIsNonPrimary = fhgfs_false;
   }

   Mutex_unlock(this->mutex);

   return numWritten;
}


/**
 * Increase stats counter for number of established conns (by NIC type).
 */
void __NodeConnPool_statsAddNic(NodeConnPool* this, NicAddrType_t nicType)
{
   switch(nicType)
   {
      case NICADDRTYPE_RDMA:
      {
         (this->stats.numEstablishedRDMA)++;
      } break;

      case NICADDRTYPE_SDP:
      {
         (this->stats.numEstablishedSDP)++;
      } break;

      default:
      {
         (this->stats.numEstablishedStd)++;
      } break;
   }
}

/**
 * Decrease stats counter for number of established conns (by NIC type).
 */
void __NodeConnPool_statsRemoveNic(NodeConnPool* this, NicAddrType_t nicType)
{
   switch(nicType)
   {
      case NICADDRTYPE_RDMA:
      {
         (this->stats.numEstablishedRDMA)--;
      } break;

      case NICADDRTYPE_SDP:
      {
         (this->stats.numEstablishedSDP)--;
      } break;

      default:
      {
         (this->stats.numEstablishedStd)--;
      } break;
   }
}

/**
 * @param streamPort value 0 will be ignored
 * @return fhgfs_true if port has changed
 */
fhgfs_bool NodeConnPool_updateStreamPort(NodeConnPool* this, unsigned short streamPort)
{
   Logger* log = App_getLogger(this->app);
   const char* logContext = "NodeConn (update stream port)";

   fhgfs_bool portHasChanged = fhgfs_false; // retVal

   Mutex_lock(this->mutex); // L O C K

   if(streamPort && (streamPort != this->streamPort) )
   {
      this->streamPort = streamPort;
      portHasChanged = fhgfs_true;
   }

   Mutex_unlock(this->mutex); // U N L O C K

   if(portHasChanged)
   {
      unsigned numInvalidated = __NodeConnPool_invalidateAvailableStreams(this, fhgfs_false);
      if(numInvalidated)
      {
         Logger_logFormatted(log, 4, logContext,
            "Invalidated %u pooled connections (due to port change)", numInvalidated);
      }
   }

   return portHasChanged;
}
