#ifndef NODECONNPOOL_H_
#define NODECONNPOOL_H_

#include <app/log/Logger.h>
#include <app/App.h>
#include <common/net/sock/NetworkInterfaceCard.h>
#include <common/net/sock/RDMASocket.h>
#include <common/net/sock/StandardSocket.h>
#include <common/threading/Mutex.h>
#include <common/threading/Condition.h>
#include <common/toolkit/ListTk.h>
#include <common/toolkit/SocketTk.h>
#include <common/Common.h>
#include "ConnectionList.h"
#include "ConnectionListIter.h"


// forward declaration
struct App;

struct NodeConnPoolStats;
typedef struct NodeConnPoolStats NodeConnPoolStats;

struct NodeConnPool;
typedef struct NodeConnPool NodeConnPool;

extern void NodeConnPool_init(NodeConnPool* this, struct App* app, unsigned short streamPort,
   NicAddressList* nicList);
extern NodeConnPool* NodeConnPool_construct(struct App* app, unsigned short streamPort,
   NicAddressList* nicList);
extern void NodeConnPool_uninit(NodeConnPool* this);
extern void NodeConnPool_destruct(NodeConnPool* this);

extern Socket* NodeConnPool_acquireStreamSocket(NodeConnPool* this);
extern Socket* NodeConnPool_acquireStreamSocketEx(NodeConnPool* this, fhgfs_bool allowWaiting);
extern void NodeConnPool_releaseStreamSocket(NodeConnPool* this, Socket* sock);
extern void NodeConnPool_invalidateStreamSocket(NodeConnPool* this, Socket* sock);
extern unsigned NodeConnPool_disconnectAvailableStreams(NodeConnPool* this);
extern unsigned NodeConnPool_disconnectAndResetIdleStreams(NodeConnPool* this);
extern fhgfs_bool NodeConnPool_updateStreamPort(NodeConnPool* this, unsigned short streamPort);

extern void __NodeConnPool_invalidateSpecificStreamSocket(NodeConnPool* this, Socket* sock);
extern unsigned __NodeConnPool_invalidateAvailableStreams(NodeConnPool* this,
   fhgfs_bool idleStreamsOnly);
extern void __NodeConnPool_resetStreamsIdleFlag(NodeConnPool* this);
extern fhgfs_bool __NodeConnPool_applySocketOptionsPreConnect(NodeConnPool* this, Socket* sock);
extern fhgfs_bool __NodeConnPool_applySocketOptionsConnected(NodeConnPool* this, Socket* sock);

extern void __NodeConnPool_statsAddNic(NodeConnPool* this, NicAddrType_t nicType);
extern void __NodeConnPool_statsRemoveNic(NodeConnPool* this, NicAddrType_t nicType);

extern void NodeConnPool_getStats(NodeConnPool* this, NodeConnPoolStats* outStats);
extern unsigned NodeConnPool_getFirstPeerName(NodeConnPool* this, NicAddrType_t nicType,
   ssize_t outBufLen, char* outBuf, fhgfs_bool* outIsNonPrimary);

// getters & setters
static inline NicAddressList* NodeConnPool_getNicList(NodeConnPool* this);
static inline void NodeConnPool_setLocalNicCaps(NodeConnPool* this,
   NicListCapabilities* localNicCaps);
static inline unsigned short NodeConnPool_getStreamPort(NodeConnPool* this);
static inline void NodeConnPool_setLogConnErrors(NodeConnPool* this, fhgfs_bool logConnErrors);


struct NodeConnPoolStats
{
   unsigned numEstablishedStd;
   unsigned numEstablishedSDP;
   unsigned numEstablishedRDMA;
};

struct NodeConnPool
{
   struct App* app;

   NicAddressList* nicList; // must not be changed (rwlock would be an option otherwise)
   ConnectionList connList;

   unsigned short streamPort;
   NicListCapabilities localNicCaps;

   unsigned availableConns; // available established conns
   unsigned establishedConns; // not equal to connList.size!!
   unsigned maxConns;
   int nonPrimaryExpirationCount; // expiration counter for conns over non-primary interfaces
   
   NodeConnPoolStats stats;
   
   fhgfs_bool logConnErrors; // false to disable logging during acquireStream() (e.g. for helperd)

   Mutex* mutex;
   Condition* changeCond;
};


NicAddressList* NodeConnPool_getNicList(NodeConnPool* this)
{
   return this->nicList;
}

/**
 * @param localNicCaps will be copied
 */
void NodeConnPool_setLocalNicCaps(NodeConnPool* this, NicListCapabilities* localNicCaps)
{
   this->localNicCaps = *localNicCaps;
}

unsigned short NodeConnPool_getStreamPort(NodeConnPool* this)
{
   unsigned short retVal;
   
   Mutex_lock(this->mutex); // L O C K
   
   retVal = this->streamPort;
   
   Mutex_unlock(this->mutex); // U N L O C K

   return retVal;
}

void NodeConnPool_setLogConnErrors(NodeConnPool* this, fhgfs_bool logConnErrors)
{
   this->logConnErrors = logConnErrors;
}


#endif /*NODECONNPOOL_H_*/
