#include <app/App.h>
#include <common/nodes/Node.h>
#include <common/nodes/NodeConnPool.h>
#include <common/nodes/Node.h>
#include <nodes/NodeStoreEx.h>
#include <toolkit/NoAllocBufferStore.h>
#include "MessagingTk.h"


/**
 * Note: This version will kmalloc a 4KB buf.
 *
 * @param outBuf contains the received message; will be kmalloced and has to be
 * kfreed by the caller (only if the return value is greater than 0)
 * @return 0 on logic error (e.g. message to big), negative socket error code on socket error (might
 * also be 0 on timeout) and message length otherwise
 */
ssize_t __MessagingTk_recvMsgBufKMalloc(App* app, Socket* sock, char** outBuf)
{
   size_t bufSize = 4*1024;

   ssize_t recvRes;

   *outBuf = os_kmalloc(bufSize);

   recvRes = MessagingTk_recvMsgBuf(app, sock, *outBuf, bufSize);

   // clean-up
   if(unlikely(recvRes <= 0) )
   {
      os_kfree(*outBuf);
      *outBuf = NULL;
   }

   return recvRes;
}


/**
 * Note: This version will acquire a buf from the msgBufStore.
 *
 * @param outBuf contains the received message; will be acquired and has to be
 * released to the msgBufStore by the caller (only if the return value is greater than 0)
 * @return 0 on logic error (e.g. message to big), negative socket error code on socket error (might
 * also be 0 on timeout) and message length otherwise
 */
ssize_t __MessagingTk_recvMsgBufStore(App* app, Socket* sock, char** outBuf)
{
   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);
   size_t bufSize = NoAllocBufferStore_getBufSize(bufStore);

   ssize_t recvRes;

   *outBuf = NoAllocBufferStore_waitForBuf(bufStore);

   recvRes = MessagingTk_recvMsgBuf(app, sock, *outBuf, bufSize);

   // clean-up
   if(unlikely(recvRes <= 0) )
   {
      NoAllocBufferStore_addBuf(bufStore, *outBuf);
      *outBuf = NULL;
   }

   return recvRes;
}


/**
 * Sends a request message to a node and receives the response.
 *
 * @param rrArgs:
 * .node receiver
 * .requestMsg the message that should be sent to the receiver
 * .respMsgType expected response message type
 * .outRespBuf response buffer if successful (must be returned to store by the caller)
 * .outRespMsg response message if successful (must be deleted by the caller)
 * @return fhgfs_true on success, fhgfs_false otherwise
 */
fhgfs_bool __MessagingTk_requestResponseWithRRArgsComm(App* app,
   MessagingTk_RequestResponseArgs* rrArgs)
{
   /* note: keep in mind that there are multiple alternative response buf alloc types avilable,
      e.g. "kmalloc" or "get from store". */

   Logger* log = App_getLogger(app);
   const char* logContext = "Messaging (RPC)";

   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);
   NodeConnPool* connPool = Node_getConnPool(rrArgs->node);

   size_t sendBufLen;
   ssize_t respRes = 0;
   ssize_t sendRes;

   // cleanup init
   Socket* sock = NULL;
   char* sendBuf = NULL;
   rrArgs->outRespBuf = NULL;
   rrArgs->outRespMsg = NULL;

   // connect

   sock = NodeConnPool_acquireStreamSocket(connPool);
   if(unlikely(!sock) )
   { // not connected
      if(!rrArgs->connEstablishFailedLogged && !Thread_isSignalPending() )
      { // only log once and only if user didn't manually interrupt with signal (to avoid log spam)
         Logger_logFormatted(log, Log_WARNING, logContext,
            "Unable to connect to: %s", Node_getNodeIDWithTypeStr(rrArgs->node) );
         Logger_logFormatted(log, Log_DEBUG, logContext,
            "Message type: %hu", NetMessage_getMsgType(rrArgs->requestMsg) );

         rrArgs->connEstablishFailedLogged = fhgfs_true;
      }
      
      return fhgfs_false;
   }

   // prepare sendBuf

   sendBufLen = NetMessage_getMsgLength(rrArgs->requestMsg);
   sendBuf = (char*)os_kmalloc(sendBufLen);
   NetMessage_serialize(rrArgs->requestMsg, sendBuf, sendBufLen);

   // send request

   sendRes = sock->send(sock, sendBuf, sendBufLen, 0);

   SAFE_KFREE(sendBuf);

   if(unlikely(sendRes != (ssize_t)sendBufLen) )
      goto socket_exception;

   // receive response (with alternative buf alloc styles)

   if(rrArgs->respBufType == MessagingTkBufType_BufStore)
      respRes = __MessagingTk_recvMsgBufStore(app, sock, &rrArgs->outRespBuf);
   else
      respRes = __MessagingTk_recvMsgBufKMalloc(app, sock, &rrArgs->outRespBuf);

   if(unlikely(respRes <= 0) )
   { // error
      if(!rrArgs->connErrLogged)
      {
         Logger_logFormatted(log, Log_WARNING, logContext,
            "Failed to receive response from: %s (%s)",
            Node_getNodeIDWithTypeStr(rrArgs->node), Socket_getPeername(sock) );
         Logger_logFormatted(log, Log_DEBUG, logContext,
            "Expected response type: %u", rrArgs->respMsgType);
      }

      goto socket_invalidate;
   }

   // got response => deserialize it
   rrArgs->outRespMsg = NetMessageFactory_createFromBuf(rrArgs->outRespBuf, respRes);

   if(unlikely(NetMessage_getMsgType(rrArgs->outRespMsg) != rrArgs->respMsgType) )
   { // response invalid (wrong msgType)
      if(!rrArgs->connInvalidRespLogged)
      {
         Logger_logErrFormatted(log, logContext,
            "Received invalid response type: %hu; expected: %d. Disconnecting: %s (%s)",
            NetMessage_getMsgType(rrArgs->outRespMsg), rrArgs->respMsgType,
            Node_getNodeIDWithTypeStr(rrArgs->node), Socket_getPeername(sock) );

         rrArgs->connInvalidRespLogged = fhgfs_true;
      }

      goto socket_invalidate;
   }

   // correct response => return it (through rrArgs)

   NodeConnPool_releaseStreamSocket(connPool, sock);

   return fhgfs_true;


   // error handling (something went wrong)...

   socket_exception:
   {
      if(!rrArgs->connErrLogged)
      {
         Logger_logErrFormatted(log, logContext,
            "Communication error: Node: %s (ErrCode: %lld; Message type: %hu)",
            Node_getNodeIDWithTypeStr(rrArgs->node),
            (long long)( (sendRes <= 0) ? sendRes : respRes),
            NetMessage_getMsgType(rrArgs->requestMsg) );
         
         rrArgs->connErrLogged = fhgfs_true;
      }
   }

   socket_invalidate:
   {
      NodeConnPool_invalidateStreamSocket(connPool, sock);
   }

   // clean up

   if(rrArgs->outRespMsg)
   {
      NetMessage_virtualDestruct(rrArgs->outRespMsg);
      rrArgs->outRespMsg = NULL;
   }

   if(rrArgs->outRespBuf)
   {
      if(rrArgs->respBufType == MessagingTkBufType_BufStore)
         NoAllocBufferStore_addBuf(bufStore, rrArgs->outRespBuf);
      else
         os_kfree(rrArgs->outRespBuf);

      rrArgs->outRespBuf = NULL;
   }

   SAFE_KFREE(sendBuf);

   return fhgfs_false;
}

