#include <app/App.h>
#include <app/log/Logger.h>
#include <common/net/message/NetMessage.h>
#include <common/net/message/storage/lookup/LookupIntentMsg.h>
#include <common/net/message/storage/lookup/LookupIntentRespMsg.h>
#include <common/storage/Metadata.h>
#include <common/storage/StorageDefinitions.h>
#include <common/storage/StorageErrors.h>
#include <common/toolkit/MessagingTk.h>
#include <common/toolkit/VersionTk.h>
#include <nodes/NodeStoreEx.h>
#include <toolkit/NoAllocBufferStore.h>

#include "MetadataTk.h"


/**
 * Used to initialize struct CreateInfo. This function always should be used, to make sure,
 * values are not forgotten.
 */
void CreateInfo_init(App* app, struct inode* parentDirInode, const char* entryName,
   int mode, CreateInfo* outCreateInfo)
{
   outCreateInfo->userID = currentFsUserID;

   // groupID and mode logic taken from inode_init_owner()
   if (parentDirInode && parentDirInode->i_mode & S_ISGID)
   {
      outCreateInfo->groupID = parentDirInode->i_gid;
      if (S_ISDIR(mode))
         mode |= S_ISGID;
   } else
      outCreateInfo->groupID = currentFsGroupID;

   outCreateInfo->entryName = entryName;
   outCreateInfo->mode      = mode;
   outCreateInfo->preferredStorageTargets = App_getPreferredStorageTargets(app);
   outCreateInfo->preferredMetaTargets    = App_getPreferredMetaNodes(app);
}


/**
 * Check whether parent entry has a subentry with the specified name/ID/owner and stat the entry
 * (if it is owned by the same metadata server).
 *
 * @param outStatRes result of stat operation, only set if general return value indicates success
 * @param outFhgfsStat only set if outStatRes indicates success
 * @return FhgfsOpsErr_SUCCESS if assumed values match actual values, FhgfsOpsErr_PATHNOTEXISTS
 * if they do not match (note that stat result comes as separate out arg).
 */
FhgfsOpsErr MetadataTk_validateAndStatSubentry(App* app, const EntryInfo* parentEntryInfo,
   const char* entryName, const EntryInfo* entryInfo, FhgfsOpsErr* outStatRes,
   fhgfs_stat* outFhgfsStat)
{
   Logger* log = App_getLogger(app);
   Config* cfg = App_getConfig(app);
   const char* logContext = "Validate & stat subentry";

   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);
   NodeStoreEx* nodes = App_getMetaNodes(app);

   FhgfsOpsErr retVal;
   Node* parentNode;

   LookupIntentMsg requestMsg;
   char* respBuf;
   NetMessage* respMsg;
   FhgfsOpsErr requestRes;
   LookupIntentRespMsg* lookupResp;
   int responseFlags;
   FhgfsOpsErr lookupRes;
   FhgfsOpsErr revalidateRes;

   EntryInfo updatedEntryInfo; // entryInfo + entryName

   fhgfs_bool supportsQuota;

   parentNode = NodeStoreEx_referenceNode(nodes, parentEntryInfo->ownerNodeID);
   if(unlikely(!parentNode) )
   {
      Logger_logErrFormatted(log, logContext, "Unknown metadata node: %hu",
         parentEntryInfo->ownerNodeID);
      return FhgfsOpsErr_UNKNOWNNODE;
   }

   EntryInfo_dup(entryInfo, &updatedEntryInfo);
   EntryInfo_updateFileName(&updatedEntryInfo, entryName);

   supportsQuota = (Config_getQuotaEnabled(cfg) && VersionTk_checkRequiredRelease(
      VERSIONTK_REQUIRED_VERSION_QUOTA_SUPPORT, Node_getFhgfsVersion(parentNode) ) );

   // prepare request msg
   LookupIntentMsg_initFromEntryInfo(&requestMsg, parentEntryInfo, &updatedEntryInfo);

   LookupIntentMsg_addIntentStat(&requestMsg);

   if (supportsQuota)
      NetMessage_addMsgHeaderFeatureFlag((NetMessage*)&requestMsg, LOOKUPINTENTMSG_FLAG_USE_QUOTA);

   // communicate
   requestRes = MessagingTk_requestResponseRetry(app,
      parentNode, (NetMessage*)&requestMsg, NETMSGTYPE_LookupIntentResp, &respBuf, &respMsg);

   NodeStoreEx_releaseNode(nodes, &parentNode);

   if(unlikely(requestRes != FhgfsOpsErr_SUCCESS) )
   { // communication error
      retVal = requestRes;
      goto cleanup_request;
   }

   // handle result

   lookupResp = (LookupIntentRespMsg*)respMsg;
   responseFlags = LookupIntentRespMsg_getResponseFlags(lookupResp);

   lookupRes = LookupIntentRespMsg_getLookupResult(lookupResp);
   if(lookupRes != FhgfsOpsErr_SUCCESS)
   {
      retVal = lookupRes;
      goto cleanup_response;
   }

   if(unlikely(!(responseFlags & LOOKUPINTENTRESPMSG_FLAG_REVALIDATE) ) )
   {
      Logger_logErrFormatted(log, logContext, "Unexpected revalidate info missing: %s",
         entryInfo->fileName);
      retVal = FhgfsOpsErr_INTERNAL;
      goto cleanup_response;
   }

   revalidateRes = LookupIntentRespMsg_getRevalidateResult(lookupResp);
   if(revalidateRes != FhgfsOpsErr_SUCCESS)
   {
      retVal = revalidateRes;
      goto cleanup_response;
   }

   if(unlikely(!(responseFlags & LOOKUPINTENTRESPMSG_FLAG_STAT) ) )
   {
      Logger_logErrFormatted(log, logContext, "Unexpected stat info missing: %s",
         entryInfo->fileName);
      retVal = FhgfsOpsErr_INTERNAL;
      goto cleanup_response;
   }

   // if we got here, the entry info was successfully revalidated
   retVal = FhgfsOpsErr_SUCCESS;

   *outStatRes = LookupIntentRespMsg_getStatResult(lookupResp);
   if(*outStatRes != FhgfsOpsErr_SUCCESS)
      goto cleanup_response; // (no retVal setting here, this is a separate result)
   else
   { // apply stat data to out-arg
      StatData* statData = LookupIntentRespMsg_getStatData(lookupResp);
      StatData_getOsStat(statData, outFhgfsStat);
   }


   // clean-up
cleanup_response:
   NetMessage_virtualDestruct(respMsg);
   NoAllocBufferStore_addBuf(bufStore, respBuf);

cleanup_request:
   LookupIntentMsg_uninit( (NetMessage*)&requestMsg);
   EntryInfo_uninit(&updatedEntryInfo);


   return retVal;
}

/**
 * Find owner and ID of a subentry and stat it (if it is owned by the same metadata server).
 *
 * @param outEntryInfo attribs set only in case of success (and must then be kfreed by the
 * caller)
 * @param outStatRes result of stat operation, only set if general return value indicates success
 * @param outFhgfsStat only set if outStatRes indicates success
 * @return lookup result (note that stat result is a separate out-arg)
 */
FhgfsOpsErr MetadataTk_lookupAndStatSubentry(App* app, const EntryInfo* parentEntryInfo,
   const char* entryName, EntryInfo* outEntryInfo, FhgfsOpsErr* outStatRes,
   fhgfs_stat* outFhgfsStat)
{
   Logger* log = App_getLogger(app);
   Config* cfg = App_getConfig(app);
   const char* logContext = "Lookup & stat subentry";

   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);
   NodeStoreEx* nodes = App_getMetaNodes(app);

   FhgfsOpsErr retVal;
   uint16_t parentNodeID = parentEntryInfo->ownerNodeID;
   Node* parentNode;

   LookupIntentMsg requestMsg;
   char* respBuf;
   NetMessage* respMsg;
   FhgfsOpsErr requestRes;
   LookupIntentRespMsg* lookupResp;
   int responseFlags;
   FhgfsOpsErr lookupRes;

   fhgfs_bool supportsQuota;

   parentNode = NodeStoreEx_referenceNode(nodes, parentNodeID);
   if(unlikely(!parentNode) )
   {
      Logger_logErrFormatted(log, logContext, "Unknown metadata node: %hu", parentNodeID);
      return FhgfsOpsErr_UNKNOWNNODE;
   }

   supportsQuota = (Config_getQuotaEnabled(cfg) && VersionTk_checkRequiredRelease(
      VERSIONTK_REQUIRED_VERSION_QUOTA_SUPPORT, Node_getFhgfsVersion(parentNode) ) );

   // prepare request msg
   LookupIntentMsg_initFromName(&requestMsg, parentEntryInfo, entryName);

   LookupIntentMsg_addIntentStat(&requestMsg);

   if (supportsQuota)
      NetMessage_addMsgHeaderFeatureFlag((NetMessage*)&requestMsg, LOOKUPINTENTMSG_FLAG_USE_QUOTA);

   // communicate
   requestRes = MessagingTk_requestResponseRetry(app,
      parentNode, (NetMessage*)&requestMsg, NETMSGTYPE_LookupIntentResp, &respBuf, &respMsg);

   NodeStoreEx_releaseNode(nodes, &parentNode);

   if(unlikely(requestRes != FhgfsOpsErr_SUCCESS) )
   { // communication error
      retVal = requestRes;
      goto cleanup_request;
   }

   // handle result

   lookupResp = (LookupIntentRespMsg*)respMsg;
   responseFlags = LookupIntentRespMsg_getResponseFlags(lookupResp);

   lookupRes = LookupIntentRespMsg_getLookupResult(lookupResp);
   if(lookupRes != FhgfsOpsErr_SUCCESS)
   {
      retVal = lookupRes;
      goto cleanup_response;
   }

   if(unlikely(!(responseFlags & LOOKUPINTENTRESPMSG_FLAG_STAT) ) )
   {
      Logger_logErrFormatted(log, logContext, "Unexpected stat info missing: %s", entryName);
      retVal = FhgfsOpsErr_INTERNAL;
      goto cleanup_response;
   }

   // if we got here, the entry lookup was successful
   retVal = FhgfsOpsErr_SUCCESS;

   LookupIntentRespMsg_toEntryInfo(lookupResp, outEntryInfo);

   *outStatRes = LookupIntentRespMsg_getStatResult(lookupResp);
   if(*outStatRes != FhgfsOpsErr_SUCCESS)
      goto cleanup_response; // (no retVal setting here, this is a separate result)
   else
   { // apply stat data to out-arg
      StatData* statData = LookupIntentRespMsg_getStatData(lookupResp);

      StatData_getOsStat(statData, outFhgfsStat);
   }


   // clean-up
cleanup_response:
   NetMessage_virtualDestruct(respMsg);
   NoAllocBufferStore_addBuf(bufStore, respBuf);

cleanup_request:
   LookupIntentMsg_uninit( (NetMessage*)&requestMsg);

   return retVal;
}

/**
 * Lookup or create a file and stat it in a single remote request.
 *
 * @param outInfo outInfo::outEntryInfo attribs set only in case of success (and must then be
 * kfreed by the caller).
 * @return success if we got the basic outInfo::outEntryInfo; this means either file existed
 * and O_EXCL wasn't specified or file didn't exist and was created.
 */
FhgfsOpsErr MetadataTk_lookupCreateStatFile(App* app, const LookupCreateStatInfoIn* inInfo,
   LookupCreateStatInfoOut* outInfo)
{
   Logger* log = App_getLogger(app);
   Config* cfg = App_getConfig(app);
   const char* logContext = "Lookup/create & stat file";

   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);
   NodeStoreEx* nodes = App_getMetaNodes(app);

   FhgfsOpsErr retVal;
   uint16_t parentNodeID = inInfo->parentEntryInfo->ownerNodeID;

   const CreateInfo* createInfo = inInfo->createInfo;

   Node* parentNode;

   LookupIntentMsg requestMsg;
   char* respBuf;
   NetMessage* respMsg;
   FhgfsOpsErr requestRes;
   LookupIntentRespMsg* lookupResp;
   int responseFlags;
   FhgfsOpsErr createRes;

   fhgfs_bool supportsQuota;

   parentNode = NodeStoreEx_referenceNode(nodes, parentNodeID);
   if(unlikely(!parentNode) )
   {
      Logger_logErrFormatted(log, logContext, "Unknown metadata node: %hu", parentNodeID);
      return FhgfsOpsErr_UNKNOWNNODE;
   }

   supportsQuota = (Config_getQuotaEnabled(cfg) && VersionTk_checkRequiredRelease(
      VERSIONTK_REQUIRED_VERSION_QUOTA_SUPPORT, Node_getFhgfsVersion(parentNode) ) );

   // prepare request msg
   LookupIntentMsg_initFromName(&requestMsg, inInfo->parentEntryInfo, createInfo->entryName);

   LookupIntentMsg_addIntentCreate(&requestMsg, createInfo->userID, createInfo->groupID,
      createInfo->mode, createInfo->preferredStorageTargets);

   if(inInfo->isExclusiveCreate)
      LookupIntentMsg_addIntentCreateExclusive(&requestMsg);

   LookupIntentMsg_addIntentStat(&requestMsg);

   if (supportsQuota)
      NetMessage_addMsgHeaderFeatureFlag((NetMessage*)&requestMsg, LOOKUPINTENTMSG_FLAG_USE_QUOTA);

   // communicate
   requestRes = MessagingTk_requestResponseRetry(app,
      parentNode, (NetMessage*)&requestMsg, NETMSGTYPE_LookupIntentResp, &respBuf, &respMsg);

   NodeStoreEx_releaseNode(nodes, &parentNode);

   if(unlikely(requestRes != FhgfsOpsErr_SUCCESS) )
   { // communication error
      retVal = requestRes;
      goto cleanup_request;
   }

   // handle result

   lookupResp = (LookupIntentRespMsg*)respMsg;
   responseFlags = LookupIntentRespMsg_getResponseFlags(lookupResp);


   if(unlikely(!(responseFlags & LOOKUPINTENTRESPMSG_FLAG_CREATE) ) )
   {
      Logger_logErrFormatted(log, logContext, "Unexpected create info missing: %s",
         createInfo->entryName);
      retVal = FhgfsOpsErr_INTERNAL;
      goto cleanup_response;
   }

   createRes = LookupIntentRespMsg_getCreateResult(lookupResp);
   retVal = createRes;

   if(createRes != FhgfsOpsErr_SUCCESS)
   { // creation or lookup (if no O_EXCL given) didn't succeed

      /* note: due to server not doing lookups again if file was created after first lookup check,
         we cannot be sure here to have entryInfo, so we must fall back to the non-intent
         methods. */

      goto cleanup_response;
   }

   if(unlikely(!(responseFlags & LOOKUPINTENTRESPMSG_FLAG_STAT) ) )
   {
      Logger_logErrFormatted(log, logContext, "Unexpected stat info missing: %s",
         createInfo->entryName);
      retVal = FhgfsOpsErr_INTERNAL;
      goto cleanup_response;
   }

   // if we got here, the entry lookup or creation was successful
   retVal = FhgfsOpsErr_SUCCESS;

   LookupIntentRespMsg_toEntryInfo(lookupResp, outInfo->entryInfo);

   outInfo->statRes = LookupIntentRespMsg_getStatResult(lookupResp);
   if(outInfo->statRes != FhgfsOpsErr_SUCCESS)
      goto cleanup_response; // (no retVal setting here, this is a separate result)
   else
   { // apply stat data to out-arg
      StatData* statData = LookupIntentRespMsg_getStatData(lookupResp);
      fhgfs_stat* outFhgfsStat = outInfo->fhgfsStat;

      StatData_getOsStat(statData, outFhgfsStat);

   }


   // clean-up
cleanup_response:
   NetMessage_virtualDestruct(respMsg);
   NoAllocBufferStore_addBuf(bufStore, respBuf);

cleanup_request:
   LookupIntentMsg_uninit( (NetMessage*)&requestMsg);

   return retVal;
}


/**
 * @param outEntryInfo contained values will be kalloced and need to be kfreed with
 * FhgfsInode_freeEntryMinInfoVals() later.
 */
void MetadataTk_getRootEntryInfoCopy(App* app, EntryInfo* outEntryInfo)
{
   NodeStoreEx* nodes = App_getMetaNodes(app);

   uint16_t outOwnerNodeID = NodeStoreEx_getRootNodeNumID(nodes);
   const char* outParentEntryID = StringTk_strDup("");
   const char* outEntryID = StringTk_strDup(META_ROOTDIR_ID_STR);
   const char* outDirName = StringTk_strDup("");
   DirEntryType outEntryType = (DirEntryType) DirEntryType_DIRECTORY;
   int outStatFlags = 0;

   EntryInfo_init(outEntryInfo, outOwnerNodeID, outParentEntryID, outEntryID, outDirName,
      outEntryType, outStatFlags);

}
