#include <toolkit/NoAllocBufferStore.h>
#include <common/net/message/nodes/GetNodesMsg.h>
#include <common/net/message/nodes/GetNodesRespMsg.h>
#include <common/net/message/nodes/GetTargetMappingsMsg.h>
#include <common/net/message/nodes/GetTargetMappingsRespMsg.h>
#include <common/threading/Thread.h>
#include <common/toolkit/MessagingTk.h>
#include <common/toolkit/Random.h>
#include "NodesTk.h"


/**
 * Download node list from given source node.
 * 
 * @param sourceNode the node from which node you want to download
 * @param nodeType which type of node list you want to download
 * @param outNodeList caller is responsible for the deletion of the received nodes
 * @param outRootID may be NULL if caller is not interested
 * @return true if download successful
 */
fhgfs_bool NodesTk_downloadNodes(App* app, Node* sourceNode, NodeType nodeType,
   NodeList* outNodeList, uint16_t* outRootNodeID)
{
   fhgfs_bool retVal = fhgfs_false;
   
   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);
   
   GetNodesMsg msg;
   
   FhgfsOpsErr commRes;
   char* respBuf = NULL;
   NetMessage* respMsg = NULL;
   GetNodesRespMsg* respMsgCast;

   // prepare request
   GetNodesMsg_initFromValue(&msg, nodeType);
   
   
   // connect & communicate
   commRes = MessagingTk_requestResponse(app,
      sourceNode, (NetMessage*)&msg, NETMSGTYPE_GetNodesResp, &respBuf, &respMsg);
   
   if(unlikely(commRes != FhgfsOpsErr_SUCCESS) )
      goto cleanup_request;
   
   // handle result
   respMsgCast = (GetNodesRespMsg*)respMsg;

   GetNodesRespMsg_parseNodeList(app, respMsgCast, outNodeList);
   
   if(outRootNodeID)
      *outRootNodeID = GetNodesRespMsg_getRootNumID(respMsgCast);
   
   retVal = fhgfs_true;

   // cleanup
   NetMessage_virtualDestruct(respMsg);
   NoAllocBufferStore_addBuf(bufStore, respBuf);
   
cleanup_request:
   GetNodesMsg_uninit( (NetMessage*)&msg);
   

   return retVal;
}

/**
 * Downloads target mappings from given source node.
 *
 * @return fhgfs_true if download successful
 */
fhgfs_bool NodesTk_downloadTargetMappings(App* app, Node* sourceNode, UInt16List* outTargetIDs,
   UInt16List* outNodeIDs)
{
   fhgfs_bool retVal = fhgfs_false;

   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);

   GetTargetMappingsMsg msg;

   FhgfsOpsErr commRes;
   char* respBuf = NULL;
   NetMessage* respMsg = NULL;
   GetTargetMappingsRespMsg* respMsgCast;

   // prepare request
   GetTargetMappingsMsg_init(&msg);

   // communicate

   commRes = MessagingTk_requestResponse(app,
      sourceNode, (NetMessage*)&msg, NETMSGTYPE_GetTargetMappingsResp, &respBuf, &respMsg);

   if(unlikely(commRes != FhgfsOpsErr_SUCCESS) )
      goto cleanup_request;

   // handle result

   respMsgCast = (GetTargetMappingsRespMsg*)respMsg;

   GetTargetMappingsRespMsg_parseTargetIDs(respMsgCast, outTargetIDs);
   GetTargetMappingsRespMsg_parseNodeIDs(respMsgCast, outNodeIDs);

   retVal = fhgfs_true;

   // cleanup

   NetMessage_virtualDestruct(respMsg);
   NoAllocBufferStore_addBuf(bufStore, respBuf);

cleanup_request:
   GetTargetMappingsMsg_uninit( (NetMessage*)&msg);


   return retVal;
}

/**
 * Walk over all nodes in the given store and drop established (available) connections.
 *
 * @return number of dropped connections
 */
unsigned NodesTk_dropAllConnsByStore(NodeStoreEx* nodes)
{
   unsigned numDroppedConns = 0;

   Node* node = NodeStoreEx_referenceFirstNode(nodes);
   while(node)
   {
      NodeConnPool* connPool = Node_getConnPool(node);

      numDroppedConns += NodeConnPool_disconnectAvailableStreams(connPool);

      node = NodeStoreEx_referenceNextNodeAndReleaseOld(nodes, node); // iterate to next node
   }

   return numDroppedConns;
}
