#include "Serialization.h"
#include <os/OsDeps.h>

/**
 * @return 0 on error (e.g. strLen is greater than bufLen), used buffer size otherwise
 */
unsigned Serialization_serializeStr(char* buf, unsigned strLen, const char* strStart)
{
   size_t bufPos = 0;

   // write length field
   bufPos += Serialization_serializeUInt(&buf[bufPos], strLen);

   // write raw string
   os_memcpy(&buf[bufPos], strStart, strLen);
   bufPos += strLen;

   // write termination char
   buf[bufPos] = 0;

   return Serialization_serialLenStr(strLen);
}

/**
 * @return fhgfs_false on error (e.g. strLen is greater than bufLen)
 */
fhgfs_bool Serialization_deserializeStr(const char* buf, size_t bufLen,
   unsigned* outStrLen, const char** outStrStart, unsigned* outLen)
{
   size_t bufPos = 0;
   unsigned strLenFieldLen = 0; // "=0" to mute false compiler warning

   // check min length
   if(unlikely(bufLen < Serialization_serialLenStr(0) ) )
      return fhgfs_false;


   // length field
   Serialization_deserializeUInt(&buf[bufPos], bufLen-bufPos, outStrLen, &strLenFieldLen);
   bufPos += strLenFieldLen;

   // string start
   *outStrStart = &buf[bufPos];
   bufPos += *outStrLen;

   // required outLen
   *outLen = Serialization_serialLenStr(*outStrLen);

   // check length and terminating zero
   if(unlikely( (bufLen < (*outLen) ) || ( (*outStrStart)[*outStrLen] != 0) ) )
      return fhgfs_false;

   return fhgfs_true;
}

unsigned Serialization_serialLenStr(unsigned strLen)
{
   // strLenField + str + terminating zero
   return Serialization_serialLenUInt() + strLen + 1;
}

/**
 * Note: adds padding to achieve 4 byte alignment
 *
 * @return 0 on error (e.g. strLen is greater than bufLen), used buffer size otherwise
 */
unsigned Serialization_serializeStrAlign4(char* buf, unsigned strLen, const char* strStart)
{
   size_t bufPos = 0;

   // write length field
   bufPos += Serialization_serializeUInt(&buf[bufPos], strLen);

   // write raw string
   os_memcpy(&buf[bufPos], strStart, strLen);
   bufPos += strLen;

   // write termination char
   buf[bufPos] = 0;

   return Serialization_serialLenStrAlign4(strLen);
}

/**
 * @return fhgfs_false on error (e.g. strLen is greater than bufLen)
 */
fhgfs_bool Serialization_deserializeStrAlign4(const char* buf, size_t bufLen,
   unsigned* outStrLen, const char** outStrStart, unsigned* outLen)
{
   size_t bufPos = 0;
   unsigned strLenFieldLen = 0; // "=0" to mute false compiler warning

   // check min length
   if(unlikely(bufLen < Serialization_serialLenStrAlign4(0) ) )
      return fhgfs_false;


   // length field
   Serialization_deserializeUInt(&buf[bufPos], bufLen-bufPos, outStrLen, &strLenFieldLen);
   bufPos += strLenFieldLen;

   // string start
   *outStrStart = &buf[bufPos];
   bufPos += *outStrLen;

   // required outLen (incl alignment padding)
   *outLen = Serialization_serialLenStrAlign4(*outStrLen);

   // check length and terminating zero
   if(unlikely( (bufLen < (*outLen) ) || ( (*outStrStart)[*outStrLen] != 0) ) )
      return fhgfs_false;

   return fhgfs_true;
}

unsigned Serialization_serialLenStrAlign4(unsigned strLen)
{
   // strLenField + str + terminating zero
   unsigned rawLen = Serialization_serialLenUInt() + strLen + 1;
   unsigned remainder = rawLen % 4;
   unsigned alignedLen = remainder ? (rawLen + 4 - remainder) : rawLen;

   return alignedLen;
}

/**
 * Serialization of a NicList.
 *
 * @return 0 on error, used buffer size otherwise
 */
unsigned Serialization_serializeNicList(char* buf, NicAddressList* nicList)
{
   unsigned nicListSize = NicAddressList_length(nicList);
   NicAddressListIter iter;
   unsigned i;

   size_t bufPos = 0;

   // elem count info field

   bufPos += Serialization_serializeUInt(&buf[bufPos], nicListSize);


   // serialize each element of the nicList

   NicAddressListIter_init(&iter, nicList);

   for(i=0; i < nicListSize; i++, NicAddressListIter_next(&iter) )
   {
      NicAddress* nicAddr = NicAddressListIter_value(&iter);
      const size_t minNameSize = MIN(sizeof(nicAddr->name), SERIALIZATION_NICLISTELEM_NAME_SIZE);

      // ipAddress
      *(unsigned*)&buf[bufPos] = nicAddr->ipAddr;

      bufPos += 4;

      // nicType
      buf[bufPos] = (char)nicAddr->nicType;

      bufPos += 1;

      // name
      os_memcpy(&buf[bufPos], &(nicAddr->name), minNameSize);
      
      bufPos += SERIALIZATION_NICLISTELEM_NAME_SIZE;
   }

   NicAddressListIter_uninit(&iter);

   return Serialization_serialLenNicList(nicList);
}

/**
 * Pre-processes a serialized NicList for deserialization via deserializeNicList().
 *
 * @return fhgfs_false on error or inconsistency
 */
fhgfs_bool Serialization_deserializeNicListPreprocess(const char* buf, size_t bufLen,
   unsigned* outNicListElemNum, const char** outNicListStart, unsigned* outLen)
{
   size_t bufPos = 0;

   unsigned elemNumFieldLen;
   if(!Serialization_deserializeUInt(&buf[bufPos], bufLen-bufPos, outNicListElemNum,
      &elemNumFieldLen) )
   {
      return fhgfs_false;
   }

   bufPos += elemNumFieldLen;

   *outLen = bufPos + *outNicListElemNum * SERIALIZATION_NICLISTELEM_SIZE;

   if(unlikely(bufLen < *outLen) )
      return fhgfs_false;

   *outNicListStart = &buf[bufPos];

   return fhgfs_true;
}

/**
 * Deserializes a NicList.
 * (requires pre-processing via nicListBufToMsg() )
 */
void Serialization_deserializeNicList(unsigned nicListElemNum, const char* nicListStart,
   NicAddressList* outNicList)
{
   const char* currentNicListPos = nicListStart;
   unsigned i;


   for(i=0; i < nicListElemNum; i++)
   {
      NicAddress* nicAddr = os_kmalloc(sizeof(NicAddress) );
      const size_t minNameSize = MIN(sizeof(nicAddr->name), SERIALIZATION_NICLISTELEM_NAME_SIZE);
      os_memset(nicAddr, 0, sizeof(NicAddress) ); // clear unused fields

      // ipAddress
      nicAddr->ipAddr = *(unsigned*)currentNicListPos;

      currentNicListPos += 4;

      // nicType
      nicAddr->nicType = (NicAddrType_t)*currentNicListPos;

      currentNicListPos += 1;
      
      // name
      os_memcpy(&nicAddr->name, currentNicListPos, minNameSize);
      (nicAddr->name)[minNameSize-1] = 0; // make sure the string is zero-terminated

      currentNicListPos += SERIALIZATION_NICLISTELEM_NAME_SIZE;
      

      NicAddressList_append(outNicList, nicAddr);
   }

}

unsigned Serialization_serialLenNicList(NicAddressList* nicList)
{
   // elem count info field + elements
   return Serialization_serialLenUInt() +
      NicAddressList_length(nicList) * SERIALIZATION_NICLISTELEM_SIZE;
}

/**
 * Serialization of a NodeList.
 *
 * @return 0 on error, used buffer size otherwise
 */
unsigned Serialization_serializeNodeList(char* buf, NodeList* nodeList)
{
   unsigned nodeListSize = NodeList_length(nodeList);
   NodeListIter iter;
   unsigned i;
   
   size_t bufPos = 0;

   // elem count info field

   bufPos += Serialization_serializeUInt(&buf[bufPos], nodeListSize);


   // serialize each element of the nodeList

   NodeListIter_init(&iter, nodeList);

   for(i=0; i < nodeListSize; i++, NodeListIter_next(&iter) )
   {
      Node* node = NodeListIter_value(&iter);
      NodeConnPool* connPool = Node_getConnPool(node);

      const char* nodeID = Node_getID(node);
      uint16_t nodeNumID = Node_getNumID(node);
      uint16_t portUDP = Node_getPortUDP(node);
      uint16_t portTCP = Node_getPortTCP(node);
      NicAddressList* nicList = NodeConnPool_getNicList(connPool);
      char nodeType = (char)Node_getNodeType(node);
      unsigned fhgfsVersion = Node_getFhgfsVersion(node);

      // nodeID
      bufPos += Serialization_serializeStr(&buf[bufPos], os_strlen(nodeID), nodeID);

      // nodeNumID
      bufPos += Serialization_serializeUShort(&buf[bufPos], nodeNumID);

      // portUDP
      bufPos += Serialization_serializeUShort(&buf[bufPos], portUDP);

      // portTCP
      bufPos += Serialization_serializeUShort(&buf[bufPos], portTCP);

      // nicList
      bufPos += Serialization_serializeNicList(&buf[bufPos], nicList);

      // nodeType
      bufPos += Serialization_serializeChar(&buf[bufPos], nodeType);

      // fhgfsVersion
      bufPos += Serialization_serializeUInt(&buf[bufPos], fhgfsVersion);
   }

   NodeListIter_uninit(&iter);
   
   return bufPos;
}

/**
 * Pre-processes a serialized NodeList for deserialization via deserializeNodeList().
 *
 * @return fhgfs_false on error or inconsistency
 */
fhgfs_bool Serialization_deserializeNodeListPreprocess(const char* buf, size_t bufLen,
   unsigned* outNodeListElemNum, const char** outNodeListStart, unsigned* outLen)
{
   // Note: This is a fairly inefficient implementation (requiring redundant pre-processing),
   //    but efficiency doesn't matter for the typical use-cases of this method

   unsigned i;   
   
   size_t bufPos = 0;

   // elem count info field
   unsigned elemNumFieldLen;

   if(!Serialization_deserializeUInt(&buf[bufPos], bufLen-bufPos, outNodeListElemNum,
      &elemNumFieldLen) )
      return fhgfs_false;

   bufPos += elemNumFieldLen;


   *outNodeListStart = &buf[bufPos];

   for(i=0; i < *outNodeListElemNum; i++)
   {
      unsigned nodeIDLen = 0;
      const char* nodeID = NULL;
      unsigned idBufLen = 0;

      uint16_t nodeNumID = 0;
      unsigned nodeNumIDBufLen = 0;

      uint16_t portUDP = 0;
      unsigned udpBufLen = 0;

      uint16_t portTCP = 0;
      unsigned tcpBufLen = 0;

      unsigned nicListElemNum = 0;
      const char* nicListStart = NULL;
      unsigned nicListBufLen = 0;

      char nodeType = 0;
      unsigned nodeTypeBufLen = 0;
      
      unsigned fhgfsVersion = 0;
      unsigned versionBufLen = 0;


      // nodeID
      if(!Serialization_deserializeStr(&buf[bufPos], bufLen-bufPos,
         &nodeIDLen, &nodeID, &idBufLen) )
         return fhgfs_false;

      bufPos += idBufLen;

      // nodeNumID
      if(unlikely(!Serialization_deserializeUShort(&buf[bufPos], bufLen-bufPos, &nodeNumID,
         &nodeNumIDBufLen) ) )
         return fhgfs_false;

      bufPos += nodeNumIDBufLen;

      // portUDP
      if(unlikely(!Serialization_deserializeUShort(&buf[bufPos], bufLen-bufPos, &portUDP,
         &udpBufLen) ) )
         return fhgfs_false;

      bufPos += udpBufLen;

      // portTCP
      if(unlikely(!Serialization_deserializeUShort(&buf[bufPos], bufLen-bufPos, &portTCP,
         &tcpBufLen) ) )
         return fhgfs_false;

      bufPos += tcpBufLen;

      // nicList
      if(unlikely(!Serialization_deserializeNicListPreprocess(&buf[bufPos], bufLen-bufPos,
         &nicListElemNum, &nicListStart, &nicListBufLen) ) )
         return fhgfs_false;

      bufPos += nicListBufLen;

      // nodeType
      if(unlikely(!Serialization_deserializeChar(&buf[bufPos], bufLen-bufPos, &nodeType,
         &nodeTypeBufLen) ) )
         return fhgfs_false;

      bufPos += nodeTypeBufLen;

      // fhgfsVersion
      if(unlikely(!Serialization_deserializeUInt(&buf[bufPos], bufLen-bufPos, &fhgfsVersion,
         &versionBufLen) ) )
         return fhgfs_false;

      bufPos += versionBufLen;
   }

   *outLen = bufPos;

   return fhgfs_true;
}

/**
 * Deserializes a NodeList.
 * (requires pre-processing)
 *
 * Note: Nodes will be constructed and need to be deleted later
 */
void Serialization_deserializeNodeList(App* app, unsigned nodeListElemNum,
   const char* nodeListStart, NodeList* outNodeList)
{
   // Note: This is a fairly inefficient implementation (containing redundant pre-processing),
   //    but efficiency doesn't matter for the typical use-cases of this method

   unsigned i;
   
   size_t bufPos = 0;
   size_t bufLen = ~0; // fake bufLen to max value (has already been verified during pre-processing)
   const char* buf = nodeListStart;


   for(i=0; i < nodeListElemNum; i++)
   {
      unsigned nodeIDLen = 0;
      const char* nodeID = NULL;
      unsigned idBufLen = 0;

      uint16_t nodeNumID = 0;
      unsigned nodeNumIDBufLen = 0;

      uint16_t portUDP = 0;
      unsigned udpBufLen = 0;

      uint16_t portTCP = 0;
      unsigned tcpBufLen = 0;

      unsigned nicListElemNum = 0;
      const char* nicListStart = NULL;
      unsigned nicListBufLen = 0;

      char nodeType = 0;
      unsigned nodeTypeBufLen = 0;

      unsigned fhgfsVersion = 0;
      unsigned versionBufLen = 0;

      NicAddressList nicList;
      Node* node = NULL;
      
      NicAddressList_init(&nicList);

      // nodeID
      Serialization_deserializeStr(&buf[bufPos], bufLen-bufPos, &nodeIDLen, &nodeID, &idBufLen);

      bufPos += idBufLen;

      // nodeNumID
      Serialization_deserializeUShort(&buf[bufPos], bufLen-bufPos, &nodeNumID, &nodeNumIDBufLen);

      bufPos += nodeNumIDBufLen;

      // portUDP
      Serialization_deserializeUShort(&buf[bufPos], bufLen-bufPos, &portUDP, &udpBufLen);

      bufPos += udpBufLen;

      // portTCP
      Serialization_deserializeUShort(&buf[bufPos], bufLen-bufPos, &portTCP, &tcpBufLen);

      bufPos += tcpBufLen;

      // nicList
      Serialization_deserializeNicListPreprocess(&buf[bufPos], bufLen-bufPos,
         &nicListElemNum, &nicListStart, &nicListBufLen);

      bufPos += nicListBufLen;

      Serialization_deserializeNicList(nicListElemNum, nicListStart, &nicList);

      // nodeType
      Serialization_deserializeChar(&buf[bufPos], bufLen-bufPos, &nodeType, &nodeTypeBufLen);

      bufPos += nodeTypeBufLen;

      // fhgfsVersion
      Serialization_deserializeUInt(&buf[bufPos], bufLen-bufPos, &fhgfsVersion, &versionBufLen);

      bufPos += versionBufLen;


      // construct node
      node = Node_construct(app, nodeID, nodeNumID, portUDP, portTCP, &nicList);
      
      Node_setNodeType(node, (NodeType)nodeType);
      Node_setFhgfsVersion(node, fhgfsVersion);

      // append node to outList
      NodeList_append(outNodeList, node);

      // cleanup
      ListTk_kfreeNicAddressListElems(&nicList);
      NicAddressList_uninit(&nicList);
   }

}

unsigned Serialization_serialLenNodeList(NodeList* nodeList)
{
   // note: this is basically serializeNodeList() (without the serialization ;)

   unsigned nodeListSize = NodeList_length(nodeList);
   NodeListIter iter;
   unsigned i;
   
   size_t bufPos = 0;

   // elem count info field

   bufPos += Serialization_serialLenUInt();


   // serialize each element of the nodeList

   NodeListIter_init(&iter, nodeList);

   for(i=0; i < nodeListSize; i++, NodeListIter_next(&iter) )
   {
      Node* node = NodeListIter_value(&iter);
      NodeConnPool* connPool = Node_getConnPool(node);

      const char* nodeID = Node_getID(node);
      NicAddressList* nicList = NodeConnPool_getNicList(connPool);

      // nodeID
      bufPos += Serialization_serialLenStr(os_strlen(nodeID) );

      // nodeNumID
      bufPos += Serialization_serialLenUShort();

      // portUDP
      bufPos += Serialization_serialLenUShort();

      // portTCP
      bufPos += Serialization_serialLenUShort();

      // nicList
      bufPos += Serialization_serialLenNicList(nicList);

      // nodeType
      bufPos += Serialization_serialLenChar();

      // fhgfsVersion
      bufPos += Serialization_serialLenUInt();
   }
   
   NodeListIter_uninit(&iter);

   return bufPos;
}



/**
 * Serialization of a StrCpyList.
 *
 * Note: We keep the serialiazation format compatible with string vec serialization
 */
unsigned Serialization_serializeStrCpyList(char* buf, StrCpyList* list)
{
   StrCpyListIter iter;

   size_t requiredLen = Serialization_serialLenStrCpyList(list);
   size_t listSize = StrCpyList_length(list);
   size_t i;

   size_t bufPos = 0;

   // totalBufLen info field

   bufPos += Serialization_serializeUInt(&buf[bufPos], (unsigned)requiredLen);

   // elem count info field

   bufPos += Serialization_serializeUInt(&buf[bufPos], listSize);


   // store each element of the list as a raw zero-terminated string

   StrCpyListIter_init(&iter, list);

   for(i=0; i < listSize; i++, StrCpyListIter_next(&iter) )
   {
      char* currentElem = StrCpyListIter_value(&iter);

      size_t serialElemLen = os_strlen(currentElem) + 1; // +1 for the terminating zero

      os_memcpy(&buf[bufPos], currentElem, serialElemLen);

      bufPos += serialElemLen;
   }

   StrCpyListIter_uninit(&iter);

   return requiredLen;
}

/**
 * Pre-processes a serialized StrCpyList.
 *
 * @return fhgfs_false on error or inconsistency
 */
fhgfs_bool Serialization_deserializeStrCpyListPreprocess(const char* buf, size_t bufLen,
   unsigned* outElemNum, const char** outListStart, unsigned* outLen)
{
   size_t bufPos = 0;

   const char* outListEnd;

   unsigned bufLenFieldLen = 0; // "=0" to mute false compiler warning
   unsigned elemNumFieldLen = 0; // "=0" to mute false compiler warning

   // totalBufLen info field
   if(unlikely(!Serialization_deserializeUInt(&buf[bufPos], bufLen-bufPos, outLen,
      &bufLenFieldLen) ) )
      return fhgfs_false;
   bufPos += bufLenFieldLen;

   // elem count field
   if(unlikely(!Serialization_deserializeUInt(&buf[bufPos], bufLen-bufPos, outElemNum,
      &elemNumFieldLen) ) )
      return fhgfs_false;
   bufPos += elemNumFieldLen;

   *outListStart = &buf[bufPos];

   outListEnd = &buf[*outLen] - 1;

   if(unlikely(
       (*outLen > bufLen) ||
       (*outLen < (2 * Serialization_serialLenUInt() ) ) || // bufLenField + numElemsField
       (*outElemNum && (*outListEnd != 0) ) ) )
      return fhgfs_false;

   return fhgfs_true;
}

/**
 * Deserializes a StrCpyList.
 * (requires pre-processing)
 *
 * @return fhgfs_false on error or inconsistency
 */
fhgfs_bool Serialization_deserializeStrCpyList(unsigned listBufLen, unsigned elemNum,
   const char* listStart, StrCpyList* outList)
{
   const char* currentPos = listStart;
   unsigned elemsBufLen;
   const char* listEndPos;
   unsigned lastElemLen;
   unsigned i;

   elemsBufLen =
      listBufLen - Serialization_serialLenUInt() -
      Serialization_serialLenUInt(); // bufLenField & numElemsField

   listEndPos = listStart + elemsBufLen - 1;


   // read each list element as a raw zero-terminated string
   // and make sure that we do not read beyond the specified end position

   for(i=0; (i < elemNum) && (currentPos <= listEndPos); i++, currentPos += lastElemLen)
   {
      const char* currentElem = currentPos;

      StrCpyList_append(outList, currentElem);

      lastElemLen = os_strlen(currentElem) + 1; // +1 for the terminating zero
   }

   // check whether all of the elements were read (=consistency)
   return ( (i==elemNum) && (currentPos > listEndPos) );
}



unsigned Serialization_serialLenStrCpyList(StrCpyList* list)
{
   StrCpyListIter iter;

   // bufLen-field + numElems-field
   unsigned requiredLen = Serialization_serialLenUInt() + Serialization_serialLenUInt();

   StrCpyListIter_init(&iter, list);

   for( ; !StrCpyListIter_end(&iter); StrCpyListIter_next(&iter) )
   {
      char* currentElem = StrCpyListIter_value(&iter);

      requiredLen += os_strlen(currentElem) + 1; // +1 for the terminating zero
   }

   StrCpyListIter_uninit(&iter);

   return requiredLen;
}

/**
 * Serialization of a StrCpyVec.
 *
 * Note: We keep the serialiazation format compatible with string list serialization
 */
unsigned Serialization_serializeStrCpyVec(char* buf, StrCpyVec* vec)
{
   size_t requiredLen = Serialization_serialLenStrCpyVec(vec);
   size_t listSize = StrCpyVec_length(vec);
   size_t i;

   size_t bufPos = 0;

   // totalBufLen info field

   bufPos += Serialization_serializeUInt(&buf[bufPos], (unsigned)requiredLen);

   // elem count info field

   bufPos += Serialization_serializeUInt(&buf[bufPos], listSize);


   // store each element of the list as a raw zero-terminated string

   for(i=0; i < listSize; i++)
   {
      char* currentElem = StrCpyVec_at(vec, i);

      size_t serialElemLen = os_strlen(currentElem) + 1; // +1 for the terminating zero

      os_memcpy(&buf[bufPos], currentElem, serialElemLen);

      bufPos += serialElemLen;
   }

   return requiredLen;
}

/**
 * Pre-processes a serialized StrCpyVec.
 *
 * Note: We keep the serialization format compatible to the string list format.
 *
 * @return fhgfs_false on error or inconsistency
 */
fhgfs_bool Serialization_deserializeStrCpyVecPreprocess(const char* buf,
   size_t bufLen, unsigned* outElemNum, const char** outListStart, unsigned* outLen)
{
   // serialization format is 100% compatible to the list version, so we can just use the list
      // version code here...

   return Serialization_deserializeStrCpyListPreprocess(buf, bufLen,
      outElemNum, outListStart, outLen);
}

/**
 * Deserializes a StrCpyVec.
 * (requires pre-processing)
 *
 * @return fhgfs_false on error or inconsistency
 */
fhgfs_bool Serialization_deserializeStrCpyVec(unsigned listBufLen, unsigned elemNum,
   const char* listStart, StrCpyVec* outVec)
{
   const char* currentPos = listStart;
   unsigned elemsBufLen;
   const char* listEndPos;
   unsigned lastElemLen;
   unsigned i;

   elemsBufLen =
      listBufLen - Serialization_serialLenUInt() -
      Serialization_serialLenUInt(); // bufLenField & numElemsField

   listEndPos = listStart + elemsBufLen - 1;


   // read each list element as a raw zero-terminated string
   // and make sure that we do not read beyond the specified end position

   for(i=0; (i < elemNum) && (currentPos <= listEndPos); i++, currentPos += lastElemLen)
   {
      const char* currentElem = currentPos;

      StrCpyVec_append(outVec, currentElem);

      lastElemLen = os_strlen(currentElem) + 1; // +1 for the terminating zero
   }

   // check whether all of the elements were read (=consistency)
   return ( (i==elemNum) && (currentPos > listEndPos) );
}



unsigned Serialization_serialLenStrCpyVec(StrCpyVec* vec)
{
   size_t i = 0;
   size_t vecLen = StrCpyVec_length(vec);

   // bufLen-field + numElems-field
   unsigned requiredLen = Serialization_serialLenUInt() + Serialization_serialLenUInt();

   for( ; i < vecLen; i++)
   {
      char* currentElem = StrCpyVec_at(vec, i);

      requiredLen += os_strlen(currentElem) + 1; // +1 for the terminating zero
   }

   return requiredLen;
}


/**
 * Serialization of a IntList.
 *
 * Note: We keep the serialization format compatible with int vec serialization
 */
unsigned Serialization_serializeIntCpyList(char* buf, IntCpyList* list)
{
   IntCpyListIter iter;
   unsigned i;

   unsigned requiredLen = Serialization_serialLenIntCpyList(list);
   unsigned listSize = IntCpyList_length(list);

   size_t bufPos = 0;

   // totalBufLen info field

   bufPos += Serialization_serializeUInt(&buf[bufPos], requiredLen);

   // elem count info field

   bufPos += Serialization_serializeUInt(&buf[bufPos], listSize);


   // store each element of the list

   IntCpyListIter_init(&iter, list);

   for(i=0; i < listSize; i++, IntCpyListIter_next(&iter) )
   {
      int currentValue = IntCpyListIter_value(&iter);

      bufPos += Serialization_serializeInt(&buf[bufPos], currentValue);
   }

   IntCpyListIter_uninit(&iter);

   return requiredLen;
}

/**
 * Pre-processes a serialized IntList().
 *
 * @return fhgfs_false on error or inconsistency
 */
fhgfs_bool Serialization_deserializeIntCpyListPreprocess(const char* buf, size_t bufLen,
   unsigned* outElemNum, const char** outListStart, unsigned* outLen)
{
   size_t bufPos = 0;

   unsigned bufLenFieldLen = 0; // "=0" to mute false compiler warning
   unsigned elemNumFieldLen = 0; // "=0" to mute false compiler warning

   // totalBufLen info field
   if(unlikely(!Serialization_deserializeUInt(&buf[bufPos], bufLen-bufPos, outLen,
      &bufLenFieldLen) ) )
      return fhgfs_false;
   bufPos += bufLenFieldLen;

   // elem count field
   if(unlikely(!Serialization_deserializeUInt(&buf[bufPos], bufLen-bufPos, outElemNum,
      &elemNumFieldLen) ) )
      return fhgfs_false;
   bufPos += elemNumFieldLen;

   *outListStart = &buf[bufPos];

   if( (*outLen > bufLen) ||
      ( (*outLen - bufPos) != (*outElemNum * Serialization_serialLenInt() ) ) )
      return fhgfs_false;

   return fhgfs_true;
}

/**
 * Deserializes a IntList.
 * (requires pre-processing)
 *
 * @return fhgfs_false on error or inconsistency
 */
fhgfs_bool Serialization_deserializeIntCpyList(unsigned listBufLen, unsigned elemNum,
   const char* listStart, IntCpyList* outList)
{
   unsigned i;

   unsigned elemsBufLen =
      listBufLen - Serialization_serialLenUInt() -
      Serialization_serialLenUInt(); // bufLenField & numElemsField

   size_t bufPos = 0;

   // read each list element

   for(i=0; i < elemNum; i++)
   {
      int value;
      unsigned valueLen;

      if(!Serialization_deserializeInt(&listStart[bufPos], elemsBufLen-bufPos, &value, &valueLen) )
         return fhgfs_false;

      bufPos += valueLen;

      IntCpyList_append(outList, value);
   }

   return fhgfs_true;
}



unsigned Serialization_serialLenIntCpyList(IntCpyList* list)
{
   // bufLen-field + numElems-field + numElems*elemSize
   unsigned requiredLen = Serialization_serialLenUInt() + Serialization_serialLenUInt() +
         IntCpyList_length(list)*Serialization_serialLenInt();

   return requiredLen;
}

/**
 * Note: We keep the serialiazation format compatible with string list serialization
 */
unsigned Serialization_serializeIntCpyVec(char* buf, IntCpyVec* vec)
{
   // IntCpyVecs are derived from ListCpyLists, so we can just do a cast here and use the list
      // version code...

   return Serialization_serializeIntCpyList(buf, (IntCpyList*)vec);
}

/**
 * Pre-processes a serialized IntVec().
 *
 * @return fhgfs_false on error or inconsistency
 */
fhgfs_bool Serialization_deserializeIntCpyVecPreprocess(const char* buf, size_t bufLen,
   unsigned* outElemNum, const char** outListStart, unsigned* outLen)
{
   // serialization format is 100% compatible to the list version, so we can just use the list
      // version code here...

   return Serialization_deserializeIntCpyListPreprocess(buf, bufLen,
      outElemNum, outListStart, outLen);
}

/**
 * Deserializes a IntVec.
 * (requires pre-processing)
 *
 * @return fhgfs_false on error or inconsistency
 */
fhgfs_bool Serialization_deserializeIntCpyVec(unsigned listBufLen, unsigned elemNum,
   const char* listStart, IntCpyVec* outVec)
{
   unsigned i;

   unsigned elemsBufLen =
      listBufLen - Serialization_serialLenUInt() -
      Serialization_serialLenUInt(); // bufLenField & numElemsField

   size_t bufPos = 0;

   // read each list element

   for(i=0; i < elemNum; i++)
   {
      int value;
      unsigned valueLen;

      if(!Serialization_deserializeInt(&listStart[bufPos], elemsBufLen-bufPos, &value, &valueLen) )
         return fhgfs_false;

      bufPos += valueLen;

      IntCpyVec_append(outVec, value);
   }

   return fhgfs_true;
}

unsigned Serialization_serialLenIntCpyVec(IntCpyVec* vec)
{
   // IntCpyVecs are derived from IntCpyLists and serialization format for both is 100% compatible,
      // so we can just do a cast here and use the list version code...

   return Serialization_serialLenIntCpyList( (IntCpyList*)vec);
}

/**
 * Serialization of a UInt16List.
 *
 * Note: We keep the serialization format compatible with int vec serialization
 */
unsigned Serialization_serializeUInt16List(char* buf, UInt16List* list)
{
   UInt16ListIter iter;
   unsigned i;

   unsigned requiredLen = Serialization_serialLenUInt16List(list);
   unsigned listSize = UInt16List_length(list);

   size_t bufPos = 0;

   // totalBufLen info field

   bufPos += Serialization_serializeUInt(&buf[bufPos], requiredLen);

   // elem count info field

   bufPos += Serialization_serializeUInt(&buf[bufPos], listSize);


   // store each element of the list

   UInt16ListIter_init(&iter, list);

   for(i=0; i < listSize; i++, UInt16ListIter_next(&iter) )
   {
      uint16_t currentValue = UInt16ListIter_value(&iter);

      bufPos += Serialization_serializeUShort(&buf[bufPos], currentValue);
   }

   UInt16ListIter_uninit(&iter);

   return requiredLen;
}

/**
 * Pre-processes a serialized UInt16List().
 *
 * @return fhgfs_false on error or inconsistency
 */
fhgfs_bool Serialization_deserializeUInt16ListPreprocess(const char* buf, size_t bufLen,
   unsigned* outElemNum, const char** outListStart, unsigned* outLen)
{
   size_t bufPos = 0;

   unsigned bufLenFieldLen = 0; // "=0" to mute false compiler warning
   unsigned elemNumFieldLen = 0; // "=0" to mute false compiler warning

   // totalBufLen info field
   if(unlikely(!Serialization_deserializeUInt(&buf[bufPos], bufLen-bufPos, outLen,
      &bufLenFieldLen) ) )
      return fhgfs_false;
   bufPos += bufLenFieldLen;

   // elem count field
   if(unlikely(!Serialization_deserializeUInt(&buf[bufPos], bufLen-bufPos, outElemNum,
      &elemNumFieldLen) ) )
      return fhgfs_false;
   bufPos += elemNumFieldLen;

   *outListStart = &buf[bufPos];

   if( (*outLen > bufLen) ||
      ( (*outLen - bufPos) != (*outElemNum * Serialization_serialLenUShort() ) ) )
      return fhgfs_false;

   return fhgfs_true;
}

/**
 * Deserializes a UInt16List.
 * (requires pre-processing)
 *
 * @return fhgfs_false on error or inconsistency
 */
fhgfs_bool Serialization_deserializeUInt16List(unsigned listBufLen, unsigned elemNum,
   const char* listStart, UInt16List* outList)
{
   unsigned i;

   unsigned elemsBufLen =
      listBufLen - Serialization_serialLenUInt() -
      Serialization_serialLenUInt(); // bufLenField & numElemsField

   size_t bufPos = 0;

   // read each list element

   for(i=0; i < elemNum; i++)
   {
      uint16_t value;
      unsigned valueLen;

      if(!Serialization_deserializeUShort(&listStart[bufPos], elemsBufLen-bufPos,
         &value, &valueLen) )
         return fhgfs_false;

      bufPos += valueLen;

      UInt16List_append(outList, value);
   }

   return fhgfs_true;
}



unsigned Serialization_serialLenUInt16List(UInt16List* list)
{
   // bufLen-field + numElems-field + numElems*elemSize
   unsigned requiredLen = Serialization_serialLenUInt() + Serialization_serialLenUInt() +
         UInt16List_length(list)*Serialization_serialLenUShort();

   return requiredLen;
}

/**
 * Note: We keep the serialiazation format compatible with string list serialization
 */
unsigned Serialization_serializeUInt16Vec(char* buf, UInt16Vec* vec)
{
   // UInt16Vecs are derived from UInt16Lists, so we can just do a cast here and use the list
      // version code...

   return Serialization_serializeUInt16List(buf, (UInt16List*)vec);
}

/**
 * Pre-processes a serialized UInt16Vec().
 *
 * @return fhgfs_false on error or inconsistency
 */
fhgfs_bool Serialization_deserializeUInt16VecPreprocess(const char* buf, size_t bufLen,
   unsigned* outElemNum, const char** outListStart, unsigned* outLen)
{
   // serialization format is 100% compatible to the list version, so we can just use the list
      // version code here...

   return Serialization_deserializeUInt16ListPreprocess(buf, bufLen,
      outElemNum, outListStart, outLen);
}

/**
 * Deserializes a UInt16Vec.
 * (requires pre-processing)
 *
 * @return fhgfs_false on error or inconsistency
 */
fhgfs_bool Serialization_deserializeUInt16Vec(unsigned listBufLen, unsigned elemNum,
   const char* listStart, UInt16Vec* outVec)
{
   unsigned i;

   unsigned elemsBufLen =
      listBufLen - Serialization_serialLenUInt() -
      Serialization_serialLenUInt(); // bufLenField & numElemsField

   size_t bufPos = 0;

   // read each list element

   for(i=0; i < elemNum; i++)
   {
      uint16_t value;
      unsigned valueLen;

      if(!Serialization_deserializeUShort(&listStart[bufPos], elemsBufLen-bufPos,
         &value, &valueLen) )
         return fhgfs_false;

      bufPos += valueLen;

      UInt16Vec_append(outVec, value);
   }

   return fhgfs_true;
}

unsigned Serialization_serialLenUInt16Vec(UInt16Vec* vec)
{
   // UInt16Vecs are derived from UInt16Lists and serialization format for both is 100% compatible,
      // so we can just do a cast here and use the list version code...

   return Serialization_serialLenUInt16List( (UInt16List*)vec);
}

unsigned Serialization_serializePath(char* buf, Path* path)
{
   size_t bufPos = 0;

   // StrCpyList
   bufPos += Serialization_serializeStrCpyList(&buf[bufPos], Path_getPathElems(path) );

   // isAbsolute
   bufPos += Serialization_serializeBool(&buf[bufPos], Path_isAbsolute(path) );

   return bufPos;
}

fhgfs_bool Serialization_deserializePathPreprocess(const char* buf, size_t bufLen,
   struct PathDeserializationInfo* outInfo, unsigned* outLen)
{
   size_t bufPos = 0;

   unsigned absFieldLen = 0; // "=0" to mute false compiler warning
   unsigned elemsBufLen = 0; // "=0" to mute false compiler warning

   // StrCpyList
   if(unlikely(!Serialization_deserializeStrCpyListPreprocess(&buf[bufPos], bufLen-bufPos,
      &outInfo->elemNum, &outInfo->elemListStart, &elemsBufLen) ) )
      return fhgfs_false;

   outInfo->elemsBufLen = elemsBufLen;

   bufPos += elemsBufLen;

   // isAbsolute
   if(unlikely(!Serialization_deserializeBool(&buf[bufPos], bufLen-bufPos, &outInfo->isAbsolute,
      &absFieldLen) ) )
      return fhgfs_false;

   bufPos += absFieldLen;

   *outLen = bufPos;

   return fhgfs_true;
}

fhgfs_bool Serialization_deserializePath(struct PathDeserializationInfo* info, Path* outPath)
{
   // StrCpyList
   if(unlikely(!Serialization_deserializeStrCpyList(
      info->elemsBufLen, info->elemNum, info->elemListStart, Path_getPathElems(outPath) ) ) )
      return fhgfs_false;

   // isAbsolute
   Path_setAbsolute(outPath, info->isAbsolute);

   return fhgfs_true;
}

unsigned Serialization_serialLenPath(Path* path)
{
   // elemsList + isAbsolute
   return Serialization_serialLenStrCpyList(Path_getPathElems(path) ) +
      Serialization_serialLenBool();
}

uint16_t Serialization_cpu_to_be16(uint16_t value)
{
   return __cpu_to_be16(value);
}

uint16_t Serialization_be16_to_cpu(uint16_t value)
{
   return __be16_to_cpu(value);
}

uint32_t Serialization_cpu_to_be32(uint32_t value)
{
   return __cpu_to_be32(value);
}

uint32_t Serialization_be32_to_cpu(uint32_t value)
{
   return __be32_to_cpu(value);
}

uint64_t Serialization_cpu_to_be64(uint64_t value)
{
   return __cpu_to_be64(value);
}

uint64_t Serialization_be64_to_cpu(uint64_t value)
{
   return __be64_to_cpu(value);
}

uint32_t Serialization_cpu_to_le16(uint16_t value)
{
   return __cpu_to_le16(value);
}

uint32_t Serialization_le16_to_cpu(uint16_t value)
{
   return __le16_to_cpu(value);
}

uint32_t Serialization_cpu_to_le32(uint32_t value)
{
   return __cpu_to_le32(value);
}

uint32_t Serialization_le32_to_cpu(uint32_t value)
{
   return __le32_to_cpu(value);
}

uint64_t Serialization_cpu_to_le64(uint64_t value)
{
   return __cpu_to_le64(value);
}

uint64_t Serialization_le64_to_cpu(uint64_t value)
{
   return __le64_to_cpu(value);
}


