#include <app/App.h>
#include <toolkit/NoAllocBufferStore.h>
#include <common/net/message/control/AckMsgEx.h>
#include <common/net/message/nodes/HeartbeatMsgEx.h>
#include <common/net/message/nodes/HeartbeatRequestMsgEx.h>
#include <common/net/message/nodes/RemoveNodeMsgEx.h>
#include <common/net/message/nodes/RemoveNodeRespMsg.h>
#include <common/toolkit/MessagingTk.h>
#include <common/toolkit/NodesTk.h>
#include <common/toolkit/Random.h>
#include <common/toolkit/Time.h>
#include "HeartbeatManager.h"


void HeartbeatManager_init(HeartbeatManager* this, App* app, DatagramListener* dgramLis)
{
   // call super constructor
   Thread_init( (Thread*)this, FHGFS_THREAD_NAME_PREFIX_STR "HBeatMgr", __HeartbeatManager_run);

   this->componentValid = fhgfs_true;

   this->mgmtInitDoneMutex = Mutex_construct();
   this->mgmtInitDoneCond = Condition_construct();

   this->app = app;
   this->cfg = App_getConfig(app);
   this->dgramLis = dgramLis;
   this->helperd = App_getHelperd(app);

   this->mgmtNodes = App_getMgmtNodes(app);
   this->metaNodes = App_getMetaNodes(app);
   this->storageNodes = App_getStorageNodes(app);

   this->nodeRegistered = fhgfs_false;
   this->mgmtInitDone = fhgfs_false;
}

struct HeartbeatManager* HeartbeatManager_construct(App* app, DatagramListener* dgramLis)
{
   struct HeartbeatManager* this = (HeartbeatManager*)os_kmalloc(sizeof(*this) );

   HeartbeatManager_init(this, app, dgramLis);

   return this;
}

void HeartbeatManager_uninit(HeartbeatManager* this)
{
   Condition_destruct(this->mgmtInitDoneCond);
   Mutex_destruct(this->mgmtInitDoneMutex);

   Thread_uninit( (Thread*)this);
}

void HeartbeatManager_destruct(HeartbeatManager* this)
{
   HeartbeatManager_uninit(this);

   os_kfree(this);
}

void __HeartbeatManager_run(Thread* this)
{
   HeartbeatManager* thisCast = (HeartbeatManager*)this;

   const char* logContext = "HeartbeatManager (run)";
   Logger* log = App_getLogger(thisCast->app);

   Logger_logFormatted(log, 4, logContext, "Searching for nodes...");

   __HeartbeatManager_mgmtInit(thisCast);

   __HeartbeatManager_requestLoop(thisCast);

   __HeartbeatManager_signalMgmtInitDone(thisCast);

   if(thisCast->nodeRegistered)
      __HeartbeatManager_unregisterNode(thisCast);

   Logger_log(log, 4, logContext, "Component stopped.");
}

void __HeartbeatManager_requestLoop(HeartbeatManager* this)
{
   int sleepTimeMS = 5000;

   const int mgmtInitIntervalMS = 5000;
   const int reregisterIntervalMS = 300000;
   const int downloadNodesIntervalMS = 300000;

   Time* lastMgmtInitT;
   Time* lastReregisterT;
   Time* lastDownloadNodesT;


   Thread* thisThread = (Thread*)this;

   lastMgmtInitT = Time_construct();
   lastReregisterT = Time_construct();
   lastDownloadNodesT = Time_construct();

   while(!_Thread_waitForSelfTerminateOrder(thisThread, sleepTimeMS) )
   {
      // mgmt init
      if(!this->mgmtInitDone)
      {
         if(Time_elapsedMS(lastMgmtInitT) > mgmtInitIntervalMS)
         {
            __HeartbeatManager_mgmtInit(this);

            Time_setToNow(lastMgmtInitT);
         }

         continue;
      }

      // the following things only happen after successful management init...

      // re-register
      if(Time_elapsedMS(lastReregisterT) > reregisterIntervalMS)
      {
         __HeartbeatManager_reregisterNode(this);

         Time_setToNow(lastReregisterT);
      }

      // download & sync nodes
      if(Time_elapsedMS(lastDownloadNodesT) > downloadNodesIntervalMS)
      {
         __HeartbeatManager_downloadAndSyncNodes(this);
         __HeartbeatManager_downloadAndSyncTargetMappings(this);

         Time_setToNow(lastDownloadNodesT);
      }

   }  // end of while loop

   Time_destruct(lastMgmtInitT);
   Time_destruct(lastReregisterT);
   Time_destruct(lastDownloadNodesT);
}


void __HeartbeatManager_mgmtInit(HeartbeatManager* this)
{
   static fhgfs_bool waitForMgmtLogged = fhgfs_false; // to avoid log spamming

   const char* logContext = "Init";
   Logger* log = App_getLogger(this->app);

   if(!waitForMgmtLogged)
   {
      const char* hostname = Config_getSysMgmtdHost(this->cfg);
      unsigned short port = Config_getConnMgmtdPortUDP(this->cfg);

      Logger_logFormatted(log, Log_WARNING, logContext,
         "Waiting for fhgfs-mgmtd@%s:%hu...", hostname, port);
      waitForMgmtLogged = fhgfs_true;
   }

   if(!__HeartbeatManager_waitForMgmtHeartbeat(this) )
      return;

   Logger_log(log, 3, logContext, "Management node found. Downloading node groups...");

   __HeartbeatManager_downloadAndSyncNodes(this);
   __HeartbeatManager_downloadAndSyncTargetMappings(this);

   Logger_log(log, 3, logContext, "Node registration...");

   if(__HeartbeatManager_registerNode(this) )
   { // download nodes again now that we will receive notifications about add/remove (avoids race)
      __HeartbeatManager_downloadAndSyncNodes(this);
      __HeartbeatManager_downloadAndSyncTargetMappings(this);
   }

   __HeartbeatManager_signalMgmtInitDone(this);

   Logger_log(log, 3, logContext, "Init complete.");
}

fhgfs_bool __HeartbeatManager_waitForMgmtHeartbeat(HeartbeatManager* this)
{
   const int waitTimeoutMS = 400;
   const int numRetries = 1;

   char heartbeatReqBuf[NETMSG_MIN_LENGTH];
   Thread* thisThread = (Thread*)this;

   const char* hostname = Config_getSysMgmtdHost(this->cfg);
   unsigned short port = Config_getConnMgmtdPortUDP(this->cfg);
   fhgfs_in_addr ipAddr;
   int i;

   HeartbeatRequestMsgEx msg;

   if(NodeStoreEx_getSize(this->mgmtNodes) )
      return fhgfs_true;


   // prepare request message

   HeartbeatRequestMsgEx_init(&msg);
   NetMessage_serialize( (NetMessage*)&msg, heartbeatReqBuf, NETMSG_MIN_LENGTH);
   HeartbeatRequestMsgEx_uninit( (NetMessage*)&msg);


   // resolve name, send message, wait for incoming heartbeat

   for(i=0; (i <= numRetries) && !Thread_getSelfTerminate(thisThread); i++)
   {
      fhgfs_bool resolveRes = SocketTk_getHostByName(this->helperd, hostname, &ipAddr);
      if(resolveRes)
      { // we got an ip => send request and wait
         int tryTimeoutWarpMS = Random_getNextInRange(-(waitTimeoutMS/4), waitTimeoutMS/4);

         DatagramListener_sendtoIP(this->dgramLis, heartbeatReqBuf, NETMSG_MIN_LENGTH, 0,
            ipAddr, port);

         if(NodeStoreEx_waitForFirstNode(this->mgmtNodes, waitTimeoutMS + tryTimeoutWarpMS) )
         {
            return fhgfs_true; // heartbeat received => we're done
         }
      }
   }

   return fhgfs_false;
}

void __HeartbeatManager_downloadAndSyncNodes(HeartbeatManager* this)
{
   Node* mgmtNode;
   Node* localNode;

   NodeList metaNodesList;
   UInt16List addedMetaNodes;
   UInt16List removedMetaNodes;
   uint16_t rootNodeID;

   NodeList storageNodesList;
   UInt16List addedStorageNodes;
   UInt16List removedStorageNodes;


   mgmtNode = NodeStoreEx_referenceFirstNode(this->mgmtNodes);
   if(!mgmtNode)
      return;

   localNode = App_getLocalNode(this->app);

   // metadata nodes

   NodeList_init(&metaNodesList);
   UInt16List_init(&addedMetaNodes);
   UInt16List_init(&removedMetaNodes);

   if(NodesTk_downloadNodes(this->app, mgmtNode, NODETYPE_Meta, &metaNodesList, &rootNodeID) )
   {
      NodeStoreEx_syncNodes(this->metaNodes,
         &metaNodesList, &addedMetaNodes, &removedMetaNodes, localNode);
      NodeStoreEx_setRootNodeNumID(this->metaNodes, rootNodeID, fhgfs_false);
      __HeartbeatManager_printSyncResults(this, NODETYPE_Meta, &addedMetaNodes, &removedMetaNodes);
   }

   NodeList_uninit(&metaNodesList);
   UInt16List_uninit(&addedMetaNodes);
   UInt16List_uninit(&removedMetaNodes);

   // storage nodes

   NodeList_init(&storageNodesList);
   UInt16List_init(&addedStorageNodes);
   UInt16List_init(&removedStorageNodes);

   if(NodesTk_downloadNodes(this->app, mgmtNode, NODETYPE_Storage, &storageNodesList, NULL) )
   {
      NodeStoreEx_syncNodes(this->storageNodes,
         &storageNodesList, &addedStorageNodes, &removedStorageNodes, localNode);
      __HeartbeatManager_printSyncResults(this,
         NODETYPE_Storage, &addedStorageNodes, &removedStorageNodes);
   }

   NodeList_uninit(&storageNodesList);
   UInt16List_uninit(&addedStorageNodes);
   UInt16List_uninit(&removedStorageNodes);


   NodeStoreEx_releaseNode(this->mgmtNodes, &mgmtNode);
}

void __HeartbeatManager_printSyncResults(HeartbeatManager* this, NodeType nodeType,
   UInt16List* addedNodes, UInt16List* removedNodes)
{
   const char* logContext = "Sync";
   Logger* log = App_getLogger(this->app);

   if(UInt16List_length(addedNodes) )
      Logger_logFormatted(log, 2, logContext, "Nodes added (sync results): %d (Type: %s)",
         (int)UInt16List_length(addedNodes), Node_nodeTypeToStr(nodeType) );

   if(UInt16List_length(removedNodes) )
      Logger_logFormatted(log, 2, logContext, "Nodes removed (sync results): %d (Type: %s)",
         (int)UInt16List_length(removedNodes), Node_nodeTypeToStr(nodeType) );
}

void __HeartbeatManager_downloadAndSyncTargetMappings(HeartbeatManager* this)
{
   TargetMapper* targetMapper = App_getTargetMapper(this->app);
   Node* mgmtNode;
   fhgfs_bool downloadRes;

   UInt16List targetIDs;
   UInt16List nodeIDs;

   mgmtNode = NodeStoreEx_referenceFirstNode(this->mgmtNodes);
   if(!mgmtNode)
      return;

   UInt16List_init(&targetIDs);
   UInt16List_init(&nodeIDs);

   downloadRes = NodesTk_downloadTargetMappings(this->app, mgmtNode, &targetIDs, &nodeIDs);
   if(downloadRes)
      TargetMapper_syncTargetsFromLists(targetMapper, &targetIDs, &nodeIDs);

   UInt16List_uninit(&targetIDs);
   UInt16List_uninit(&nodeIDs);
   NodeStoreEx_releaseNode(this->mgmtNodes, &mgmtNode);
}


/**
 * @return fhgfs_true if an ack was received for the heartbeat, fhgfs_false otherwise
 */
fhgfs_bool __HeartbeatManager_registerNode(HeartbeatManager* this)
{
   static fhgfs_bool registrationFailureLogged = fhgfs_false; // to avoid log spamming

   const char* logContext = "Registration";
   Logger* log = App_getLogger(this->app);

   Node* mgmtNode;

   NoAllocBufferStore* bufStore = App_getMsgBufStore(this->app);
   Node* localNode = App_getLocalNode(this->app);
   const char* localNodeID = Node_getID(localNode);
   NicAddressList* nicList = Node_getNicList(localNode);
   const char* ackID = "registration"; // just a dummy for direct reply, not used in ackStore

   HeartbeatMsgEx msg;
   char* respBuf;
   NetMessage* respMsg;
   FhgfsOpsErr requestRes;
   //AckMsgEx* hbResp; // response value not needed currently

   mgmtNode = NodeStoreEx_referenceFirstNode(this->mgmtNodes);
   if(!mgmtNode)
      return fhgfs_false;

   HeartbeatMsgEx_initFromNodeData(&msg, localNodeID, 0, NODETYPE_Client, 0, nicList);
   HeartbeatMsgEx_setPorts(&msg, Config_getConnClientPortUDP(this->cfg), 0);
   HeartbeatMsgEx_setFhgfsVersion(&msg, FHGFS_VERSION_CODE);
   HeartbeatMsgEx_setAckID(&msg, ackID);

   // connect & communicate
   requestRes = MessagingTk_requestResponse(this->app, mgmtNode,
      (NetMessage*)&msg, NETMSGTYPE_Ack, &respBuf, &respMsg);

   if(unlikely(requestRes != FhgfsOpsErr_SUCCESS) )
   { // request/response failed
      goto cleanup_request;
   }

   // handle result
   // hbResp = (AckMsgEx*)respMsg; // response value not needed currently

   this->nodeRegistered = fhgfs_true;


   // clean-up
   NetMessage_virtualDestruct(respMsg);
   NoAllocBufferStore_addBuf(bufStore, respBuf);


cleanup_request:
   HeartbeatMsgEx_uninit( (NetMessage*)&msg);

   NodeStoreEx_releaseNode(this->mgmtNodes, &mgmtNode);


   // log registration result
   if(this->nodeRegistered)
      Logger_log(log, 2, logContext, "Node registration successful.");
   else
   if(!registrationFailureLogged)
   {
      Logger_log(log, 1, logContext,
         "Node registration failed. Management node offline? Will keep on trying...");
      registrationFailureLogged = fhgfs_true;
   }

   return this->nodeRegistered;
}

/**
 * Note: This just sends a heartbeat to the mgmt node to re-new our existence information
 * (in case the mgmt node assumed our death for some reason like network errors etc.)
 */
void __HeartbeatManager_reregisterNode(HeartbeatManager* this)
{
   Node* mgmtNode;

   Node* localNode = App_getLocalNode(this->app);
   const char* localNodeID = Node_getID(localNode);
   NicAddressList* nicList = Node_getNicList(localNode);

   HeartbeatMsgEx msg;

   mgmtNode = NodeStoreEx_referenceFirstNode(this->mgmtNodes);
   if(!mgmtNode)
      return;

   HeartbeatMsgEx_initFromNodeData(&msg, localNodeID, 0, NODETYPE_Client, 0, nicList);
   HeartbeatMsgEx_setPorts(&msg, Config_getConnClientPortUDP(this->cfg), 0);
   HeartbeatMsgEx_setFhgfsVersion(&msg, FHGFS_VERSION_CODE);

   DatagramListener_sendMsgToNode(this->dgramLis, mgmtNode, (NetMessage*)&msg);


   HeartbeatMsgEx_uninit( (NetMessage*)&msg);

   NodeStoreEx_releaseNode(this->mgmtNodes, &mgmtNode);
}

/**
 * @return fhgfs_true if an ack was received for the heartbeat, fhgfs_false otherwise
 */
fhgfs_bool __HeartbeatManager_unregisterNode(HeartbeatManager* this)
{
   const char* logContext = "Deregistration";
   Logger* log = App_getLogger(this->app);

   // Note: Be careful not to use Datagrams here, because this is called during App stop and hence
   //    the DGramLis is probably not even listening for responses anymore

   Node* mgmtNode;

   NoAllocBufferStore* bufStore = App_getMsgBufStore(this->app);
   Node* localNode = App_getLocalNode(this->app);
   const char* localNodeID = Node_getID(localNode);
   uint16_t localNodeNumID = Node_getNumID(localNode);

   RemoveNodeMsgEx msg;
   char* respBuf;
   NetMessage* respMsg;
   FhgfsOpsErr requestRes;
   //RemoveNodeRespMsg* rmResp; // response value not needed currently
   fhgfs_bool nodeUnregistered = fhgfs_false;

   mgmtNode = NodeStoreEx_referenceFirstNode(this->mgmtNodes);
   if(!mgmtNode)
      return fhgfs_false;

   RemoveNodeMsgEx_initFromNodeData(&msg, localNodeID, localNodeNumID, NODETYPE_Client);

   // connect & communicate
   requestRes = MessagingTk_requestResponse(this->app, mgmtNode,
      (NetMessage*)&msg, NETMSGTYPE_RemoveNodeResp, &respBuf, &respMsg);

   if(unlikely(requestRes != FhgfsOpsErr_SUCCESS) )
   { // request/response failed
      goto cleanup_request;
   }

   // handle result
   // rmResp = (RemoveNodeRespMsg*)respMsg; // response value not needed currently

   nodeUnregistered = fhgfs_true;


   // clean-up
   NetMessage_virtualDestruct(respMsg);
   NoAllocBufferStore_addBuf(bufStore, respBuf);


cleanup_request:
   HeartbeatMsgEx_uninit( (NetMessage*)&msg);

   NodeStoreEx_releaseNode(this->mgmtNodes, &mgmtNode);


   // log deregistration result
   if(nodeUnregistered)
      Logger_log(log, 2, logContext, "Node deregistration successful.");
   else
   {
      Logger_log(log, 1, logContext,
         "Node deregistration failed. Management node offline?");
      Logger_logFormatted(log, 1, logContext,
         "In case you didn't enable automatic removal, you will have to remove this ClientID "
         "manually from the system: %s", localNodeID);
   }

   return nodeUnregistered;
}

/**
 * @param timeoutMS 0 to wait infinitely (which is probably never what you want)
 */
fhgfs_bool HeartbeatManager_waitForMgmtInit(HeartbeatManager* this, int timeoutMS)
{
   fhgfs_bool retVal;

   Mutex_lock(this->mgmtInitDoneMutex);

   if(!timeoutMS)
   {
      while(!this->mgmtInitDone)
            Condition_wait(this->mgmtInitDoneCond, this->mgmtInitDoneMutex);
   }
   else
   if(!this->mgmtInitDone)
      Condition_timedwait(this->mgmtInitDoneCond, this->mgmtInitDoneMutex, timeoutMS);

   retVal = this->mgmtInitDone;

   Mutex_unlock(this->mgmtInitDoneMutex);

   return retVal;
}

void __HeartbeatManager_signalMgmtInitDone(HeartbeatManager* this)
{
   Mutex_lock(this->mgmtInitDoneMutex);

   this->mgmtInitDone = fhgfs_true;

   Condition_broadcast(this->mgmtInitDoneCond);

   Mutex_unlock(this->mgmtInitDoneMutex);
}
