#ifndef INTERNODESYNCER_H_
#define INTERNODESYNCER_H_

#include <app/config/Config.h>
#include <app/log/Logger.h>
#include <common/toolkit/list/PointerListIter.h>
#include <common/toolkit/MetadataTk.h>
#include <common/threading/AtomicInt.h>
#include <common/threading/Thread.h>
#include <common/threading/Mutex.h>
#include <common/threading/Condition.h>
#include <common/net/message/NetMessage.h>
#include <net/filesystem/RemotingIOInfo.h>
#include <nodes/NodeStoreEx.h>
#include <toolkit/ExternalHelperd.h>


/*
 * note: some part of this is declared in the include_closed tree.
 */

/*
 * note: Delayed...Entry-queues management is integrated into InternodeSyncer, because it needs
 * direct access to the queues via iterators and relies on special access patterns, e.g.
 * it must be the only one removing entries from the queue (to keep iterators valid).
 */


// forward declarations...

struct InternodeSyncer;
typedef struct InternodeSyncer InternodeSyncer;

struct DelayedCloseEntry;
typedef struct DelayedCloseEntry DelayedCloseEntry;
struct DelayedEntryUnlockEntry;
typedef struct DelayedEntryUnlockEntry DelayedEntryUnlockEntry;
struct DelayedRangeUnlockEntry;
typedef struct DelayedRangeUnlockEntry DelayedRangeUnlockEntry;


extern void InternodeSyncer_init(InternodeSyncer* this, App* app);
extern InternodeSyncer* InternodeSyncer_construct(App* app);
extern void InternodeSyncer_uninit(InternodeSyncer* this);
extern void InternodeSyncer_destruct(InternodeSyncer* this);

extern void InternodeSyncer_delayedCloseAdd(InternodeSyncer* this, const EntryInfo* entryInfo,
   const RemotingIOInfo* ioInfo);
extern void InternodeSyncer_delayedEntryUnlockAdd(InternodeSyncer* this,
   const EntryInfo* entryInfo, const RemotingIOInfo* ioInfo, int64_t clientFD);
extern void InternodeSyncer_delayedRangeUnlockAdd(InternodeSyncer* this,
   const EntryInfo* entryInfo, const RemotingIOInfo* ioInfo, int ownerPID);

extern void _InternodeSyncer_requestLoop(InternodeSyncer* this);

extern void __InternodeSyncer_run(Thread* this);
extern void __InternodeSyncer_delayedCloseComm(InternodeSyncer* this);
extern void __InternodeSyncer_delayedEntryUnlockComm(InternodeSyncer* this);
extern void __InternodeSyncer_delayedRangeUnlockComm(InternodeSyncer* this);
extern void __InternodeSyncer_delayedClosePrepareRemoting(InternodeSyncer* this,
   DelayedCloseEntry* closeEntry, EntryInfo** outEntryInfo, RemotingIOInfo* outIOInfo);
extern void __InternodeSyncer_delayedEntryUnlockPrepareRemoting(InternodeSyncer* this,
   DelayedEntryUnlockEntry* closeEntry, EntryInfo** outEntryInfo, RemotingIOInfo* outIOInfo);
extern void __InternodeSyncer_delayedRangeUnlockPrepareRemoting(InternodeSyncer* this,
   DelayedRangeUnlockEntry* closeEntry, EntryInfo** outEntryInfo, RemotingIOInfo* outIOInfo);
extern void __InternodeSyncer_delayedCloseFreeEntry(InternodeSyncer* this,
   DelayedCloseEntry* closeEntry);
extern void __InternodeSyncer_delayedEntryUnlockFreeEntry(InternodeSyncer* this,
   DelayedEntryUnlockEntry* closeEntry);
extern void __InternodeSyncer_delayedRangeUnlockFreeEntry(InternodeSyncer* this,
   DelayedRangeUnlockEntry* closeEntry);
extern void __InternodeSyncer_dropIdleConns(InternodeSyncer* this);
extern unsigned __InternodeSyncer_dropIdleConnsByStore(InternodeSyncer* this, NodeStoreEx* nodes);

// getters & setters
static inline fhgfs_bool InternodeSyncer_getValid(InternodeSyncer* this);
extern size_t InternodeSyncer_getDelayedCloseQueueSize(InternodeSyncer* this);
extern size_t InternodeSyncer_getDelayedEntryUnlockQueueSize(InternodeSyncer* this);
extern size_t InternodeSyncer_getDelayedRangeUnlockQueueSize(InternodeSyncer* this);


struct DelayedCloseEntry
{
   Time ageT; // time when this entry was created (to compute entry age)

   EntryInfo entryInfo;

   // fields for RemotingIOInfo
   const char* fileHandleID;
   unsigned accessFlags; // OPENFILE_ACCESS_... flags

   /* value as in struct FhgfsInodeFileHandle{}. However in contrary to struct RemotingIOInfo{}
    * it is NOT used as reference pointer, but the value will be os_kmalloced. Reason is that we are
    * purely closed source here and do not know the size or anything about struct AtomicInt{}.
    * Can be updated once the client code is under GPL.
    */
   AtomicInt* maxUsedTargetIndex; // (used as a reference in ioInfo)
};

struct DelayedEntryUnlockEntry
{
   Time ageT; // time when this entry was created (to compute entry age)

   EntryInfo entryInfo;

   // fields for RemotingIOInfo
   const char* fileHandleID;

   int64_t clientFD;
};

struct DelayedRangeUnlockEntry
{
   Time ageT; // time when this entry was created (to compute entry age)

   EntryInfo entryInfo;

   // fields for RemotingIOInfo
   const char* fileHandleID;

   int ownerPID;
};


struct InternodeSyncer
{
   Thread thread; // base class
   
   fhgfs_bool componentValid; // whether component init was successful
   
   App* app;   
   Config* cfg;

   Mutex* delayedCloseMutex; // for delayedCloseQueue
   PointerList* delayedCloseQueue; /* remove from head, add to tail (important, because we rely on
      validity of an iterator even when a new entry was added) */
   Mutex* delayedEntryUnlockMutex; // for delayedEntryUnlockQueue
   PointerList* delayedEntryUnlockQueue; /* remove from head, add to tail (important, because we
      rely on validity of an iterator even when a new entry was added) */
   Mutex* delayedRangeUnlockMutex; // for delayedRangeUnlockQueue
   PointerList* delayedRangeUnlockQueue; /* remove from head, add to tail (important, because we
      rely on validity of an iterator even when a new entry was added) */
};



fhgfs_bool InternodeSyncer_getValid(InternodeSyncer* this)
{
   return this->componentValid;
}


#endif /*INTERNODESYNCER_H_*/
