#include <app/log/Logger.h>
#include <app/App.h>
#include <toolkit/NoAllocBufferStore.h>
#include <common/net/message/session/FSyncLocalFileMsg.h>
#include <common/net/message/session/FSyncLocalFileRespMsg.h>
#include <common/net/message/NetMessage.h>
#include <common/nodes/TargetMapper.h>
#include <common/threading/Thread.h>
#include <common/toolkit/MessagingTk.h>
#include <common/toolkit/VersionTk.h>
#include <common/components/worker/WorkQueue.h>
#include <nodes/NodeStoreEx.h>
#include "FSyncLocalFileWork.h"


void FSyncLocalFileWork_process(Work* this, char* bufIn, unsigned bufInLen,
   char* bufOut, unsigned bufOutLen)
{
   FSyncLocalFileWork* thisCast = (FSyncLocalFileWork*)this;

   TargetMapper* targetMapper = App_getTargetMapper(thisCast->app);
   NodeStoreEx* nodes = App_getStorageNodes(thisCast->app);
   uint16_t targetID = UInt16Vec_at(thisCast->targetIDs, thisCast->nodeIndex);

   FhgfsOpsErr resolveErr;
   fhgfs_bool supportsSessionCheckAndSyncOnClose;

   Node* node = NodeStoreEx_referenceNodeByTargetID(nodes, targetID, targetMapper, &resolveErr);

   if(unlikely(!node) )
   { // unable to resolve targetID
      (thisCast->nodeResults)[thisCast->nodeIndex] = resolveErr;
      goto node_not_found;
   }

   // check the version of the server and disable the unsupported features
   // disable the unsupported features at this method then is a second version check in the
   // __FSyncLocalFileWork_communicate not required
   supportsSessionCheckAndSyncOnClose = VersionTk_checkRequiredRelease(
      VERSIONTK_REQUIRED_VERSION_2012_10_R2, Node_getFhgfsVersion(node));
   if(!supportsSessionCheckAndSyncOnClose)
   { // disables syncOnClose and sessionCheck because the server doesn't support it
      thisCast->checkSession = fhgfs_false;
      thisCast->doSyncOnClose = fhgfs_false;
   }

   if(thisCast->forceRemoteFlush || thisCast->checkSession || thisCast->doSyncOnClose)
   { // check if communication is required, got node reference => begin communication
      FhgfsOpsErr commRes = __FSyncLocalFileWork_communicate(
         thisCast, node, bufIn, bufInLen, bufOut, bufOutLen);

      NodeStoreEx_releaseNode(nodes, &node);

      // handle retry
      if(unlikely(commRes == FhgfsOpsErr_COMMUNICATION) )
      { // comm error occurred => prepare retry (if allowed)
         if(__FSyncLocalFileWork_prepareRetry(thisCast) )
         {
            // retry prepared => don't inc the done-counter and don't touch any member variables
            return;
         }
      }
   }
   else
   { // old server and no force remote flush set
      (thisCast->nodeResults)[thisCast->nodeIndex] = FhgfsOpsErr_SUCCESS;
      NodeStoreEx_releaseNode(nodes, &node);
   }

node_not_found:
   SynchronizedCounter_incCount(thisCast->counter);
}

FhgfsOpsErr __FSyncLocalFileWork_communicate(FSyncLocalFileWork* this, Node* node,
   char* bufIn, unsigned bufInLen, char* bufOut, unsigned bufOutLen)
{
   Node* localNode = App_getLocalNode(this->app);
   char* localNodeID = Node_getID(localNode);
   NoAllocBufferStore* bufStore = App_getMsgBufStore(this->app);

   FSyncLocalFileMsg requestMsg;
   char* respBuf;
   NetMessage* respMsg;
   FSyncLocalFileRespMsg* fsyncResp;

   FhgfsOpsErr* nodeResult = &(this->nodeResults)[this->nodeIndex];
   uint16_t targetID = UInt16Vec_at(this->targetIDs, this->nodeIndex);

   // prepare request message
   FSyncLocalFileMsg_initFromSession(&requestMsg, localNodeID, this->fileHandleID, targetID);
   NetMessage_setMsgHeaderUserID( (NetMessage*)&requestMsg, this->msgUserID);

   if(this->isMirror)
      NetMessage_addMsgHeaderFeatureFlag( (NetMessage*)&requestMsg,
         FSYNCLOCALFILEMSG_FLAG_ISMIRROR);

   // required version is checked in FSyncLocalFileWork_process and if session check is not
   // supported by the server, the value of this->checkSession was set to false
   if(this->checkSession && this->firstWriteDoneForTarget)
      NetMessage_addMsgHeaderFeatureFlag( (NetMessage*)&requestMsg,
         FSYNCLOCALFILEMSG_FLAG_SESSION_CHECK);

   // required version is checked in FSyncLocalFileWork_process and if syncOnClose is not
   // supported by the server, the value of this->doSyncOnClose was set to false
   if(!this->forceRemoteFlush && !this->doSyncOnClose)
      NetMessage_addMsgHeaderFeatureFlag( (NetMessage*)&requestMsg,
         FSYNCLOCALFILEMSG_FLAG_NO_SYNC);

   // connect & communicate

   *nodeResult = MessagingTk_requestResponse(this->app,
      node, (NetMessage*)&requestMsg, NETMSGTYPE_FSyncLocalFileResp, &respBuf, &respMsg);

   if(likely(*nodeResult == FhgfsOpsErr_SUCCESS) )
   {
      // handle result
      fsyncResp = (FSyncLocalFileRespMsg*)respMsg;
      *nodeResult = FSyncLocalFileRespMsg_getValue(fsyncResp);

      // clean up
      NetMessage_virtualDestruct(respMsg);
      NoAllocBufferStore_addBuf(bufStore, respBuf);
   }

   // clean up
   FSyncLocalFileMsg_uninit( (NetMessage*)&requestMsg);

   return *nodeResult;
}

/**
 * @return fhgfs_true if a retry was prepared (e.g. caller shouldn't notify waiters), fhgfs_false
 * if no retry allowed and we're done
 */
fhgfs_bool __FSyncLocalFileWork_prepareRetry(FSyncLocalFileWork* this)
{
   RetryableWork* thisRetryWork = (RetryableWork*)this;
   Config* cfg = App_getConfig(this->app);
   unsigned maxNumRetries = Config_getConnNumCommRetries(cfg);

   if(RetryableWork_getIsInterrupted(thisRetryWork) )
   { // interrupted by initiator => no retry
      (this->nodeResults)[this->nodeIndex] = FhgfsOpsErr_INTERRUPTED;
   }
   else
   if( App_getConnRetriesEnabled(this->app) &&
       ( (!maxNumRetries) ||
         (thisRetryWork->currentRetryNum < maxNumRetries) ) )
   { // we have retries left => prepare and enqueue new (retry-)work

      // note: we have to create new work here because the worker will destroy this instance
      // note: we prefer to construct a new work instance and init it via the constructors
      //    (instead of copying from this instance) to be saver wrt side-effects

      WorkQueue* retryQ = App_getRetryWorkQueue(this->app);

      FSyncLocalFileWork* newWork;
      RetryableWork* newRetryWork;

      newWork = FSyncLocalFileWork_construct(this->app, this->nodeIndex,
         this->fileHandleID, this->targetIDs, this->nodeResults, this->counter,
         this->forceRemoteFlush, this->firstWriteDoneForTarget, this->checkSession,
         this->doSyncOnClose);
      FSyncLocalFileWork_setMsgUserID(newWork, this->msgUserID);
      FSyncLocalFileWork_setIsMirror(newWork, this->isMirror);

      newRetryWork = (RetryableWork*)newWork;
      RetryableWork_setInterruptor(newRetryWork, thisRetryWork->isInterrupted);

      newRetryWork->retryMS = MessagingTk_getRetryWaitMS(thisRetryWork->currentRetryNum);
      if(newRetryWork->retryMS)
      { // "retryWaitMS != 0" means we have to init the time object
         fhgfs_Time_init(&newRetryWork->lastRetryT);
      }

      newRetryWork->currentRetryNum = thisRetryWork->currentRetryNum + 1;

      WorkQueue_addWork(retryQ, (Work*)newWork);

      return fhgfs_true;
   }

   return fhgfs_false;
}
