#ifndef FSYNCLOCALFILEWORK_H_
#define FSYNCLOCALFILEWORK_H_

#include <app/App.h>
#include <common/Common.h>
#include <common/net/message/NetMessage.h>
#include <common/net/sock/Socket.h>
#include <common/components/worker/RetryableWork.h>
#include <common/toolkit/SynchronizedCounter.h>
#include <common/toolkit/vector/StrCpyVec.h>
#include <common/toolkit/vector/IntCpyVec.h>
#include <common/toolkit/vector/Int64CpyVec.h>
#include <toolkit/BitStore.h>


struct FSyncLocalFileWork;
typedef struct FSyncLocalFileWork FSyncLocalFileWork;

static inline void FSyncLocalFileWork_init(FSyncLocalFileWork* this, App* app, unsigned nodeIndex,
   const char* fileHandleID, UInt16Vec* targetIDs, FhgfsOpsErr* nodeResults,
   SynchronizedCounter* counter, fhgfs_bool forceRemoteFlush, fhgfs_bool firstWriteDoneForTarget,
   fhgfs_bool checkSession, fhgfs_bool doSyncOnClose);
static inline FSyncLocalFileWork* FSyncLocalFileWork_construct(App* app, unsigned nodeIndex,
   const char* fileHandleID, UInt16Vec* targetIDs, FhgfsOpsErr* nodeResults,
   SynchronizedCounter* counter, fhgfs_bool forceRemoteFlush, fhgfs_bool firstWriteDoneForTarget,
   fhgfs_bool checkSession, fhgfs_bool doSyncOnClose);
static inline void FSyncLocalFileWork_uninit(Work* this);
static inline void FSyncLocalFileWork_destruct(Work* this);

extern FhgfsOpsErr __FSyncLocalFileWork_communicate(FSyncLocalFileWork* this, Node* node,
   char* bufIn, unsigned bufInLen, char* bufOut, unsigned bufOutLen);
extern fhgfs_bool __FSyncLocalFileWork_prepareRetry(FSyncLocalFileWork* this);

// virtual functions
extern void FSyncLocalFileWork_process(Work* this, char* bufIn, unsigned bufInLen,
   char* bufOut, unsigned bufOutLen);

// getters & setters
static inline fhgfs_bool FSyncLocalFileWork_getIsMirror(FSyncLocalFileWork* this);
static inline void FSyncLocalFileWork_setIsMirror(FSyncLocalFileWork* this, fhgfs_bool isMirror);
static inline void FSyncLocalFileWork_setMsgUserID(FSyncLocalFileWork* this, unsigned msgUserID);


struct FSyncLocalFileWork
{
   RetryableWork retryableWork;

   App* app;

   unsigned nodeIndex;
   const char* fileHandleID;
   UInt16Vec* targetIDs;

   unsigned msgUserID; // user ID to be set in net msg header
   fhgfs_bool isMirror; // fhgfs_true if this is a mirror chunk file

   FhgfsOpsErr* nodeResults;
   SynchronizedCounter* counter;

   fhgfs_bool forceRemoteFlush; // true if the server must a sync the file (if a flush was called)

   fhgfs_bool checkSession; // true if the server must check the session

   fhgfs_bool doSyncOnClose; // true if the server must a sync the file (if a file was closed)

   fhgfs_bool firstWriteDoneForTarget; /* true if the first chunk was written to the storage target,
                                          it's needed for the session check*/
};


void FSyncLocalFileWork_init(FSyncLocalFileWork* this, App* app, unsigned nodeIndex,
   const char* fileHandleID, UInt16Vec* targetIDs, FhgfsOpsErr* nodeResults,
   SynchronizedCounter* counter, fhgfs_bool forceRemoteFlush, fhgfs_bool firstWriteDoneForTarget,
   fhgfs_bool checkSession, fhgfs_bool doSyncOnClose)
{
   RetryableWork_init( (RetryableWork*)this);

   this->app = app;

   this->nodeIndex = nodeIndex;
   this->fileHandleID = fileHandleID;
   this->targetIDs = targetIDs;

   this->msgUserID = NETMSG_DEFAULT_USERID;
   this->isMirror = fhgfs_false;

   this->nodeResults = nodeResults;
   this->counter = counter;

   this->forceRemoteFlush = forceRemoteFlush;

   this->firstWriteDoneForTarget = firstWriteDoneForTarget;

   this->checkSession = checkSession;

   this->doSyncOnClose = doSyncOnClose;

   // assign virtual functions
   ( (Work*)this)->uninit = FSyncLocalFileWork_uninit;

   ( (Work*)this)->process = FSyncLocalFileWork_process;
}

struct FSyncLocalFileWork* FSyncLocalFileWork_construct(App* app, unsigned nodeIndex,
   const char* fileHandleID, UInt16Vec* targetIDs, FhgfsOpsErr* nodeResults,
   SynchronizedCounter* counter, fhgfs_bool forceRemoteFlush, fhgfs_bool firstWriteDoneForTarget,
   fhgfs_bool checkSession, fhgfs_bool doSyncOnClose)
{
   struct FSyncLocalFileWork* this = (FSyncLocalFileWork*)os_kmalloc(sizeof(*this) );

   FSyncLocalFileWork_init(this, app, nodeIndex, fileHandleID, targetIDs, nodeResults, counter,
      forceRemoteFlush, firstWriteDoneForTarget, checkSession, doSyncOnClose);

   return this;
}

void FSyncLocalFileWork_uninit(Work* this)
{
   RetryableWork_uninit(this);
}

void FSyncLocalFileWork_destruct(Work* this)
{
   FSyncLocalFileWork_uninit(this);

   os_kfree(this);
}

fhgfs_bool FSyncLocalFileWork_getIsMirror(FSyncLocalFileWork* this)
{
   return this->isMirror;
}

void FSyncLocalFileWork_setIsMirror(FSyncLocalFileWork* this, fhgfs_bool isMirror)
{
   this->isMirror = isMirror;
}

void FSyncLocalFileWork_setMsgUserID(FSyncLocalFileWork* this, unsigned msgUserID)
{
   this->msgUserID = msgUserID;
}

#endif /*FSYNCLOCALFILEWORK_H_*/
