#ifndef FHGFSINODE_H_
#define FHGFSINODE_H_

#include <app/App.h>
#include <app/config/Config.h>
#include <common/Common.h>
#include <common/threading/AtomicInt64.h>
#include <common/threading/Mutex.h>
#include <common/threading/RWLock.h>
#include <common/storage/EntryInfo.h>
#include <common/storage/StorageDefinitions.h>
#include <common/storage/StorageErrors.h>
#include <common/storage/Metadata.h>
#include <common/toolkit/Time.h>
#include <common/toolkit/HashTk.h>
#include <common/toolkit/MetadataTk.h>
#include <common/toolkit/tree/IntMapIter.h>
#include <filesystem/FhgfsInode.h>
#include <net/filesystem/RemotingIOInfo.h>
#include <toolkit/BitStore.h>
#include "FhgfsOpsSuper.h"

#include <linux/fs.h>
#include <linux/vfs.h>


/*
 * The rest of this is declared in the include_open tree.
 */


#define FHGFS_INODE_ROOT_INO                 2 /* traditional root inode number */
#define FHGFS_INODE_MAXRESERVED_INO          FHGFS_INODE_ROOT_INO

#define FHGFS_INODE_FILEHANDLES_NUM          3 /* r, rw, w (see "enum FileHandleType") */

#define FHGFS_INODE(inodePointer)            ( (FhgfsInode*)inodePointer)
#define FHGFS_VFSINODE(fhgfsInodePointer)    ( (struct inode*)fhgfsInodePointer)

struct StripePattern; // forward declaration

struct FhgfsInodeFileHandle;
typedef struct FhgfsInodeFileHandle FhgfsInodeFileHandle;

enum FileHandleType;
typedef enum FileHandleType FileHandleType;

enum FileBufferType;
typedef enum FileBufferType FileBufferType;

struct CacheBuffer;
typedef struct CacheBuffer CacheBuffer;

struct FhgfsInode;
typedef struct FhgfsInode FhgfsInode;

enum FileHandleType // (note: used as array index for FhgfsInode::fileHandles)
   {FileHandleType_READ=0, FileHandleType_RW=1, FileHandleType_WRITE=2};

enum FileBufferType
   {FileBufferType_NONE=0, FileBufferType_READ=1, FileBufferType_WRITE=2};

struct CacheBuffer
{
   char* buf; // the file contents cache buffer
   //size_t bufLen; // length of buf (not needed, because we have bufUsageMaxLen)
   size_t bufUsageLen; // how much of the buffer is already used up
   size_t bufUsageMaxLen; // how much of the buffer may be used (is min(bufLen, chunk_end) )

   loff_t fileOffset; // the file offset where this buffer starts

   enum FileBufferType bufType;
};

// mem cache helpers
extern void FhgfsInode_initOnce(FhgfsInode* this);
extern void FhgfsInode_allocInit(FhgfsInode* this);
extern void FhgfsInode_destroyUninit(FhgfsInode* this);


extern fhgfs_bool FhgfsInode_removeRangeLockPID(FhgfsInode* this, int pid);
extern void FhgfsInode_addRangeLockPID(FhgfsInode* this, int pid);

extern FileHandleType __FhgfsInode_openFlagsToHandleType(int openFlags);
extern FhgfsOpsErr __FhgfsInode_referenceTrunc(FhgfsInode* this, App* app, int openFlags);

extern FhgfsOpsErr FhgfsInode_referenceHandle(FhgfsInode* this, int openFlags,
   fhgfs_bool allowRWHandle, FileHandleType* outHandleType);

extern FhgfsOpsErr FhgfsInode_releaseHandle(FhgfsInode* this, FileHandleType handleType);

extern void FhgfsInode_getRefIOInfo(FhgfsInode* this, FileHandleType handleType,
   unsigned accessFlags, RemotingIOInfo* outIOInfo);

extern void FhgfsInode_fileCacheExclusiveLock(FhgfsInode* this);
extern void FhgfsInode_fileCacheExclusiveUnlock(FhgfsInode* this);
extern int FhgfsInode_fileCacheExclusiveTryLock(FhgfsInode* this);

extern struct CacheBuffer* Fhgfsinode_getFileCacheBuffer(FhgfsInode* this);

extern uint64_t FhgfsInode_generateInodeID(struct super_block* sb, const char* entryID,
   int entryIDLen);

// getters & setters

static inline EntryInfo* FhgfsInode_getEntryInfo(FhgfsInode* this);

static inline DirEntryType FhgfsInode_getDirEntryType(FhgfsInode* this);

static inline int FhgfsInode_getNumRangeLockPIDs(FhgfsInode* this);
static inline fhgfs_bool FhgfsInode_getIsFileOpen(FhgfsInode* this);
static inline fhgfs_bool FhgfsInode_getIsFileOpenByMultipleReaders(FhgfsInode* this);

// inliners
static inline void Fhgfsinode_appendLock(FhgfsInode* this);
static inline void Fhgfsinode_appendUnlock(FhgfsInode* this);

static inline void __FhgfsInode_initIOInfo(FhgfsInode* this, FhgfsInodeFileHandle* fileHandle,
   unsigned accessFlags, RemotingIOInfo* outIOInfo);

static inline void _FhgfsInode_initRootEntryInfo(FhgfsInode* this);

static inline void FhgfsInode_entryInfoReadLock(FhgfsInode* this);
static inline void FhgfsInode_entryInfoReadUnlock(FhgfsInode* this);
static inline void FhgfsInode_entryInfoWriteLock(FhgfsInode* this);
static inline void FhgfsInode_entryInfoWriteUnlock(FhgfsInode* this);

static inline fhgfs_bool FhgfsInode_compareEntryID(FhgfsInode* this, const char* otherEntryID);
static inline void FhgfsInode_UpdateEntryInfoOnRenameUnlocked(FhgfsInode* this, const EntryInfo* newParentInfo,
   const char* newEntryName);
static inline void FhgfsInode_UpdateEntryInfoUnlocked(FhgfsInode* this, const EntryInfo* newEntryInfo);


static inline int FhgfsInode_getStripeCount(FhgfsInode* this);

static inline int FhgfsInode_handleTypeToOpenFlags(FileHandleType handleType);

static inline void FhgfsInode_invalidateCache(FhgfsInode* this);

static inline void FhgfsInode_incNumDirtyPages(FhgfsInode* this);
static inline void FhgfsInode_decNumDirtyPages(FhgfsInode* this);
static inline uint64_t FhgfsInode_getNumDirtyPages(FhgfsInode* this);
static inline void FhgfsInode_setNumDirtyPages(FhgfsInode* this, uint64_t value);
static inline fhgfs_bool FhgfsInode_getHasDirtyPages(FhgfsInode* this);


struct FhgfsInodeFileHandle
{
   const char* fileHandleID;
   unsigned refCount;

   /*
    * zero-based target index in stripe pattern (-1 means "none").
    * Used for fsync() and as hint to the meta server about the highest storage target accessed by
    * a file read/write. So if a file has several stripes, but writes so far only went to the first
    * stripe target, other storage targets/servers don't need to be notified about file close.
    */
   AtomicInt maxUsedTargetIndex; // zero-based target index in stripe pattern (-1 means "none")

   BitStore firstWriteDone;   /* bit is set if the first chunk was written to a storage
                              target successful */
};

struct FhgfsInode
{
   struct inode vfs_inode;

   EntryInfo entryInfo;
   RWLock entryInfoLock;

   Time dataCacheTime; /* last time we updated file contents (monotonic clock) or dir attribs;
                          protected by i_lock */

   Mutex fileHandlesMutex;
   FhgfsInodeFileHandle fileHandles[FHGFS_INODE_FILEHANDLES_NUM]; // use FileHandleType as index
   struct StripePattern* pattern; // initialized when file is opened; multiple readers allowed

   Mutex rangeLockPIDsMutex;
   AtomicInt numRangeLockPIDs; // modified only with mutex held, but readable without it
   IntMap rangeLockPIDs; /* all PIDs (TGIDs) that used range locking and have this file open
                            (values are dummies) */

   Mutex appendMutex; // for atomic writes with append-flag

   struct CacheBuffer fileCacheBuffer; // current buffer cache entry (protected by fileCacheLock)
   RWLock fileCacheLock; /* for sync'ed access to cacheEntry and page cache (required to ensure
      order, e.g. when a new r/w-call comes in while we're about to disable caching) */

   AtomicInt64 numPageCacheDirtyPages; // number of written (=> dirty) bytes since last flush

   loff_t asyncPageError; // position of an async page write error, -1 if all is ok
};


int FhgfsInode_getNumRangeLockPIDs(FhgfsInode* this)
{
   return AtomicInt_read(&this->numRangeLockPIDs);
}

/**
 * Test whether a given file inode is currently open.
 *
 * @return fhgfs_true if the file is currently open.
 */
fhgfs_bool FhgfsInode_getIsFileOpen(FhgfsInode* this)
{
   fhgfs_bool retVal = fhgfs_false;
   int i;

   // Mutex_lock(&this->fileHandlesMutex); // L O C K

   for(i=0; i < FHGFS_INODE_FILEHANDLES_NUM; i++)
   {
      if(this->fileHandles[i].refCount)
      { // we found an open file handle
         retVal = fhgfs_true;
         break;
      }
   }

   // Mutex_unlock(&this->fileHandlesMutex); // U N L O C K

   return retVal;
}

/**
 * Test whether a given file inode is currently open for reading by multiple readers.
 *
 * Note: This is specifically intended to give a hint for read-ahead disabling (because the current
 * buffered mode logic could do read-ahead trashing with multiple concurrent readers on the same
 * file). Since this is just a hint, we don't care about mutex locking here.
 *
 * @return fhgfs_true if the file is currently open by more than one reader (including read+write).
 */
fhgfs_bool FhgfsInode_getIsFileOpenByMultipleReaders(FhgfsInode* this)
{
   unsigned numReaders = this->fileHandles[FileHandleType_READ].refCount +
      this->fileHandles[FileHandleType_RW].refCount;

   return (numReaders > 1);
}

/**
 * Acquire the internal append mutex
 *
 * Note: remember to call _appendUnlock()!!
 */
void Fhgfsinode_appendLock(FhgfsInode* this)
{
   Mutex_lock(&this->appendMutex);
}

void Fhgfsinode_appendUnlock(FhgfsInode* this)
{
   Mutex_unlock(&this->appendMutex);
}


/**
 * Get the EntryInfo of this inode
 */
EntryInfo* FhgfsInode_getEntryInfo(FhgfsInode* this)
{
      return &this->entryInfo;
}


/**
 * Initialize the root inode. We can not do that at mount time as we allow mounts without a
 * connection to the management server.
 */
void _FhgfsInode_initRootEntryInfo(FhgfsInode* this)
{
   struct inode* vfs_inode = &this->vfs_inode;
   struct super_block* sb  = vfs_inode->i_sb;
   fhgfs_bool hasRootInfo = FhgfsOps_getHasRootEntryInfo(sb);

   // unlikely as it will happen only once
   if (unlikely (unlikely(!hasRootInfo) && (vfs_inode->i_ino == FHGFS_INODE_ROOT_INO) ) )
   {
      /* root inode, EntryInfo not yet set during mount, as the meta server might be unknown.
       * Set it now. */

      App* app = FhgfsOps_getApp(sb);

      RWLock_writeLock(&this->entryInfoLock); // Write-LOCK

      hasRootInfo = FhgfsOps_getHasRootEntryInfo(sb); // verify under a lock
      if (unlikely(hasRootInfo) )
      {
         // another thread updated it already, nothing to do
      }
      else
      {
         EntryInfo_uninit(&this->entryInfo);

         MetadataTk_getRootEntryInfoCopy(app, &this->entryInfo);

         FhgfsOps_setHasRootEntryInfo(sb, fhgfs_true);
      }

      RWLock_writeUnlock(&this->entryInfoLock); // Write-UNLOCK

   }

}

/*
 * Get an EntryInfo read-lock.
 *
 * Note: If FhgfsInode is the root inode and not yet initialized, we are going to that using
 *       a writelock.
 */
void FhgfsInode_entryInfoReadLock(FhgfsInode* this)
{
   _FhgfsInode_initRootEntryInfo(this); // NOTE: might temporarily writelock entryInfoLock

   RWLock_readLock(&this->entryInfoLock); // Read-LOCK
}

void FhgfsInode_entryInfoReadUnlock(FhgfsInode* this)
{
   RWLock_readUnlock(&this->entryInfoLock);
}

/**
 * Note: Also might initialize the root-inode.
 */
void FhgfsInode_entryInfoWriteLock(FhgfsInode* this)
{
   _FhgfsInode_initRootEntryInfo(this); // NOTE: might temporarily writelock entryInfoLock

   RWLock_writeLock(&this->entryInfoLock); // Write-LOCK
}

void FhgfsInode_entryInfoWriteUnlock(FhgfsInode* this)
{
   RWLock_writeUnlock(&this->entryInfoLock);
}



/**
 * Compares given entryID with the entryID of this inode.
 *
 * Note: This is intended to be called from methods like __FhgfsOps_compareInodeID, which is
 * called e.g. by ilookup5 with the inode_hash_lock held, so this method may not sleep
 *
 * @return fhgfs_true if given ID matches, fhgfs_false otherwise
 */
fhgfs_bool FhgfsInode_compareEntryID(FhgfsInode* this, const char* otherEntryID)
{
   if(this->vfs_inode.i_ino == FHGFS_INODE_ROOT_INO)
   { // root node => owner info is not stored in the inode
      return !strcmp(otherEntryID, META_ROOTDIR_ID_STR);
   }
   else
      return !strcmp(otherEntryID, this->entryInfo.entryID);
}

/**
 * Return the DirEntryType of the given inode
 */
DirEntryType FhgfsInode_getDirEntryType(FhgfsInode* this)
{
   return this->entryInfo.entryType;
}

/**
 * Note: For internal use only (assigns filehandle->maxUsedTargetIndex pointer to outIOInfo).
 */
void __FhgfsInode_initIOInfo(FhgfsInode* this, FhgfsInodeFileHandle* fileHandle,
   unsigned accessFlags, RemotingIOInfo* outIOInfo)
{
   outIOInfo->app = FhgfsOps_getApp(this->vfs_inode.i_sb);

   outIOInfo->fileHandleID = fileHandle->fileHandleID;
   outIOInfo->pattern = this->pattern;
   outIOInfo->accessFlags = accessFlags;

   outIOInfo->maxUsedTargetIndex = &fileHandle->maxUsedTargetIndex;

   outIOInfo->userID = this->vfs_inode.i_uid;
   outIOInfo->groupID = this->vfs_inode.i_gid;
}


/**
 * Update EntryInfo on rename
 *
 * Note: newParentInfo and newName are not owned by this object
 *
 * Note: FhgfsInode->entryInfoLock already needs to be write-locked by the caller.
 */
void FhgfsInode_UpdateEntryInfoOnRenameUnlocked(FhgfsInode* this, const EntryInfo* newParentInfo,
   const char* newName)
{
   EntryInfo_updateParentEntryID(&this->entryInfo, newParentInfo->entryID);
   EntryInfo_updateFileName(&this->entryInfo, newName);

   if (!DirEntryType_ISDIR(this->entryInfo.entryType) )
   {  // only update the ownerNodeID if it is not a directory
      EntryInfo_setOwnerNodeID(&this->entryInfo, newParentInfo->ownerNodeID);
   }
}

/**
 * Update an existing EntryInfo with a new one
 *
 * Note: this->entryInfoLock already needs to be write-locked by the caller.
 */
void FhgfsInode_UpdateEntryInfoUnlocked(FhgfsInode* this, const EntryInfo* newEntryInfo)
{
   EntryInfo_uninit(&this->entryInfo);
   EntryInfo_copy(newEntryInfo, &this->entryInfo);
}


int FhgfsInode_getStripeCount(FhgfsInode* this)
{
   return UInt16Vec_length(this->pattern->getStripeTargetIDs(this->pattern));
}

/**
 * Converts FileHandleType to OPENFILE_ACCESS_... fhgfs flags.
 *
 * Note: This is especially required when a handle was referenced with allowRWHandle, because then
 * you don't know the OPENFILE_ACCESS_... flags.
 */
int FhgfsInode_handleTypeToOpenFlags(FileHandleType handleType)
{
   switch(handleType)
   {
      case FileHandleType_WRITE:
      {
         return OPENFILE_ACCESS_WRITE;
      } break;

      case FileHandleType_RW:
      {
         return OPENFILE_ACCESS_READWRITE;
      } break;

      default:
      {
         return OPENFILE_ACCESS_READ;
      } break;
   }
}

/**
 * Invalidate the cache of an inode
 */
void FhgfsInode_invalidateCache(FhgfsInode* this)
{
   Time_setZero(&this->dataCacheTime);
}

void FhgfsInode_incNumDirtyPages(FhgfsInode* this)
{
   AtomicInt64_inc(&this->numPageCacheDirtyPages);
}

void FhgfsInode_decNumDirtyPages(FhgfsInode* this)
{
   AtomicInt64_dec(&this->numPageCacheDirtyPages);
}

uint64_t FhgfsInode_getNumDirtyPages(FhgfsInode* this)
{
   return AtomicInt64_read(&this->numPageCacheDirtyPages);
}

void FhgfsInode_setNumDirtyPages(FhgfsInode* this, uint64_t value)
{
   return AtomicInt64_set(&this->numPageCacheDirtyPages, value);
}

/**
 * Test whether a given file inode is currently write-opened.
 *
 * @return fhgfs_true if the file is currently open.
 */
fhgfs_bool FhgfsInode_getHasDirtyPages(FhgfsInode* this)
{
   return !!FhgfsInode_getNumDirtyPages(this);
}


#endif /* FHGFSINODE_H_ */
