#include <common/Common.h> // (placed up here for LINUX_VERSION_CODE definition)


/**
 * NFS export is probably not worth the backport efforts for kernels before 2.6.29
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)

#include <common/storage/Metadata.h>
#include <filesystem/FhgfsInode.h>
#include <filesystem/FhgfsOpsInode.h>
#include <net/filesystem/FhgfsOpsRemoting.h>
#include <os/OsTypeConversion.h>
#include "FhgfsOpsExport.h"


/**
 * Operations for NFS export (and open_by_handle).
 */
const struct export_operations fhgfs_export_ops =
{
   .encode_fh       = FhgfsOpsExport_encodeNfsFileHandle,
   .fh_to_dentry    = FhgfsOpsExport_nfsFileHandleToDentry,
};



/**
 * Single-byte handle type identifier that will be returned by _encodeNfsFileHandle and later be
 * used by _nfsFileHandleToDentry to decode the handle.
 *
 * note: in-tree file systems use "enum fid_type" in linux/exportfs.h; this is our own version of
 * it to let _nfsFileHandleToDentry know how to decode the FhgfsNfsFileHandle structure.
 * note: according to linux/exportfs.h, "the filesystem must not use the value '0' or '0xff'" as
 * valid types; 0xff means given handle buffer size is too small.
 */
enum FhgfsNfsHandleType
{
   FhgfsNfsHandle_STANDARD    = 0xf1,     /* some arbitrary number that doesn't conflict with others
                                             in linux/exportfs.h (though that wouldn't be a real
                                             conflict) to identify our standard valid handle. */
   FhgfsNfsHandle_INVALID     = 0xfe,     /* special meaning: error occured, invalid handle
                                             (this is only a hint for _nfsFileHandleToDentry,
                                             because _encodeNfsFileHandle has no way to return an
                                             error to the calling kernel code). */
   FhgfsNfsHandle_BUFTOOSMALL = 0xff,     /* special meaning for callers: given handle buffer
                                             was too small */
};

typedef enum FhgfsNfsHandleType FhgfsNfsHandleType;


/**
 * Encode a file handle (typically for NFS) that can later be used to lookup an inode via
 * _nfsFileHanleToDentry().
 *
 * @param max_len file_handle_buf array length (=> length in 4-byte words), will be set to actually
 * used or desired length.
 * @param parent_inode/connectable if set, this means we should try to create a connectable file
 * handle, so that we can later do a parent dir lookup from it.
 *
 * @return FhgfsNfsHandleType_...
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,5,0)

int FhgfsOpsExport_encodeNfsFileHandle(struct dentry* dentry, __u32* file_handle_buf, int* max_len,
   int connectable)
{
   struct inode* inode = dentry->d_inode;
   struct inode* parent_inode = NULL;

#else // LINUX_VERSION_CODE >= KERNEL_VERSION(3,5,0)

int FhgfsOpsExport_encodeNfsFileHandle(struct inode* inode, __u32* file_handle_buf, int* max_len,
   struct inode* parent_inode)
{

#endif // LINUX_VERSION_CODE

   FhgfsNfsHandleType retVal = FhgfsNfsHandle_STANDARD;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   FhgfsNfsFileHandle* fhgfsNfsHandle = (void*)file_handle_buf;

   size_t fhgfsNfsHandleLen = sizeof(FhgfsNfsFileHandle);
   size_t givenHandleByteLength = (*max_len) * sizeof(__u32);

   EntryInfo* entryInfo;
   fhgfs_bool parseParentIDRes;
   fhgfs_bool parseEntryIDRes;

   IGNORE_UNUSED_VARIABLE(parent_inode); // (for parent_inode declaration in old kernels)

   *max_len = (fhgfsNfsHandleLen + sizeof(__u32)-1) / sizeof(__u32); /* set desired/used max_len for
      caller (4-byte words; "+sizeof(u32)-1" rounds up) */

   // check whether given buf length is enough for our handle
   if(givenHandleByteLength < fhgfsNfsHandleLen)
      return FhgfsNfsHandle_BUFTOOSMALL;

   // get entryInfo and serialize it in a special small format
   // (normal string-based serialization would use too much buffer space)

   FhgfsInode_entryInfoReadLock(fhgfsInode); // LOCK EntryInfo
   entryInfo = FhgfsInode_getEntryInfo(fhgfsInode);

   parseParentIDRes = __FhgfsOpsExport_parseEntryIDForNfsHandle(entryInfo->parentEntryID,
      &fhgfsNfsHandle->parentEntryIDCounter, &fhgfsNfsHandle->parentEntryIDTimestamp,
      &fhgfsNfsHandle->parentEntryIDNodeID);

   if(unlikely(!parseParentIDRes) )
   { // parsing failed (but we have no real way to return an error)
      retVal = FhgfsNfsHandle_INVALID;
      goto cleanup;
   }

   parseEntryIDRes = __FhgfsOpsExport_parseEntryIDForNfsHandle(entryInfo->entryID,
      &fhgfsNfsHandle->entryIDCounter, &fhgfsNfsHandle->entryIDTimestamp,
      &fhgfsNfsHandle->entryIDNodeID);

   if(unlikely(!parseEntryIDRes) )
   { // parsing failed (but we have no real way to return an error)
      retVal = FhgfsNfsHandle_INVALID;
      goto cleanup;
   }

   fhgfsNfsHandle->ownerNodeID = entryInfo->ownerNodeID;
   fhgfsNfsHandle->entryType = entryInfo->entryType;


cleanup:
   FhgfsInode_entryInfoReadUnlock(fhgfsInode); // UNLOCK EntryInfo

   return retVal;
}

/**
 * Lookup an inode based on a file handle that was previously created via _encodeNfsFileHandle().
 *
 * @param fid file handle buffer
 * @param fh_len fid buffer length in 4-byte words
 * @param fileid_type FhgfsNfsHandleType_... (as returned by _encodeNfsFileHandle).
 */
struct dentry* FhgfsOpsExport_nfsFileHandleToDentry(struct super_block *sb, struct fid *fid,
   int fh_len, int fileid_type)
{
   FhgfsNfsFileHandle* fhgfsNfsHandle = (FhgfsNfsFileHandle*)fid->raw;
   FhgfsNfsHandleType handleType = (FhgfsNfsHandleType)fileid_type;

   size_t fhgfsNfsHandleLen = sizeof(FhgfsNfsFileHandle);
   size_t givenHandleByteLength = fh_len * sizeof(__u32); // fh_len is in 4-byte words


   // check if this handle is valid
   if(unlikely(handleType != FhgfsNfsHandle_STANDARD) )
   {
      printk_fhgfs_debug(KERN_INFO, "%s: Called with invalid handle type: 0x%x\n",
         __func__, fileid_type);

      return NULL;
   }

   // check if given handle length is valid
   if(unlikely(givenHandleByteLength < fhgfsNfsHandleLen) )
   {
      printk_fhgfs_debug(KERN_INFO, "%s: Called with too small handle length: "
         "%d bytes; (handle type: 0x%x)\n",
         __func__, (int)givenHandleByteLength, fileid_type);

      return NULL;
   }


   return __FhgfsOpsExport_lookupDentryFromNfsHandle(sb, fhgfsNfsHandle);
}

/**
 * Check whether a dentry/inode for the nfs handle exists in the local cache and try server lookup
 * otherwise.
 */
struct dentry* __FhgfsOpsExport_lookupDentryFromNfsHandle(struct super_block *sb,
   FhgfsNfsFileHandle* fhgfsNfsHandle)
{
   fhgfs_bool entryRes;
   char* entryID = NULL;
   char* parentEntryID = NULL;

   ino_t inodeHash; // (simply the inode number for us)
   struct inode* inode;
   FhgfsInodeComparisonInfo comparisonInfo;

   EntryInfo entryInfo;

   // generate entryID string

   entryRes = __FhgfsOpsExport_entryIDFromNfsHandle(fhgfsNfsHandle->entryIDCounter,
      fhgfsNfsHandle->entryIDTimestamp, fhgfsNfsHandle->entryIDNodeID, &entryID);

   if(unlikely(!entryRes) )
      goto err_cleanup_entryids;

   inodeHash = FhgfsInode_generateInodeID(sb, entryID, strlen(entryID) );

   comparisonInfo.inodeHash = inodeHash;
   comparisonInfo.entryID = entryID;

   inode = ilookup5(sb, inodeHash, __FhgfsOps_compareInodeID,
      &comparisonInfo); // (ilookup5 calls iget() on match)

   if(!inode)
   { // not found in cache => try to get it from mds

      App* app = FhgfsOps_getApp(sb);
      fhgfs_bool parentRes;
      fhgfs_stat fhgfsStat;
      struct kstat kstat;
      FhgfsOpsErr statRes = FhgfsOpsErr_SUCCESS;

      // generate parentEntryID string

      parentRes = __FhgfsOpsExport_entryIDFromNfsHandle(fhgfsNfsHandle->parentEntryIDCounter,
         fhgfsNfsHandle->parentEntryIDTimestamp, fhgfsNfsHandle->parentEntryIDNodeID,
         &parentEntryID);

      if(unlikely(!parentRes) )
         goto err_cleanup_entryids;

      // init entry info

      EntryInfo_init(&entryInfo, fhgfsNfsHandle->ownerNodeID, parentEntryID, entryID,
         StringTk_strDup("<nfs_fh>"), fhgfsNfsHandle->entryType, 0);

      // communicate

      statRes = FhgfsOpsRemoting_statDirect(app, &entryInfo, &fhgfsStat);

      if(statRes != FhgfsOpsErr_SUCCESS)
         goto err_cleanup_entryinfo;

      // entry found => create inode

      OsTypeConv_kstatFhgfsToOs(&fhgfsStat, &kstat);

      kstat.ino = inodeHash;

      inode = __FhgfsOps_newInode(sb, &kstat, 0, &entryInfo);
   }


   return d_obtain_alias(inode); // (d_obtain_alias can also handle pointer error codes and NULL)


err_cleanup_entryids:
   SAFE_KFREE(parentEntryID);
   SAFE_KFREE(entryID);

   return ERR_PTR(-ESTALE);

err_cleanup_entryinfo:
   EntryInfo_uninit(&entryInfo);

   return ERR_PTR(-ESTALE);
}

/**
 * Parse an entryID string to get its three components.
 *
 * Note: META_ROOTDIR_ID_STR is a special case (all three components are set to 0).
 *
 * @return fhgfs_false on error
 */
fhgfs_bool __FhgfsOpsExport_parseEntryIDForNfsHandle(const char* entryID, uint32_t* outCounter,
   uint32_t* outTimestamp, uint16_t* outNodeID)
{
   const int numEntryIDComponents = 3; // sscanf must find 3 components in a valid entryID string

   uint32_t nodeID32; // just a tmp variable, because 16-bit outNodeID cannot be used with sscanf %X
   int scanRes;

   if(!strcmp(META_ROOTDIR_ID_STR, entryID) )
   { // special case: this is the root ID, which doesn't have the usual three components
      *outCounter = 0;
      *outTimestamp = 0;
      *outNodeID = 0;

      return fhgfs_true;
   }

   scanRes = os_sscanf(entryID, "%X-%X-%X", outCounter, outTimestamp, &nodeID32);

   if(unlikely(scanRes != numEntryIDComponents) )
   { // parsing failed
      printk_fhgfs_debug(KERN_INFO, "%s: Parsing of entryID failed. entryID: %s\n",
         __func__, entryID);

      return fhgfs_false;
   }

   *outNodeID = nodeID32;

   return fhgfs_true;
}

/**
 * Generate an entryID string from the NFS handle components.
 *
 * @param outEntryID will be kmalloced on success and needs to be kfree'd by the caller
 * @return fhgfs_false on error
 */
fhgfs_bool __FhgfsOpsExport_entryIDFromNfsHandle(uint32_t counter, uint32_t timestamp,
   uint16_t nodeID, char** outEntryID)
{
   if(!counter && !timestamp && !nodeID)
   { // special case: root ID
      *outEntryID = StringTk_strDup(META_ROOTDIR_ID_STR);
   }
   else
      *outEntryID = StringTk_kasprintf("%X-%X-%X", counter, timestamp, (uint32_t)nodeID);

   return (*outEntryID != NULL); // (just in case kmalloc failed)
}

#endif // LINUX_VERSION_CODE
