#include <app/log/Logger.h>
#include <app/App.h>
#include <common/threading/AtomicInt.h>
#include <common/toolkit/StringTk.h>
#include <common/toolkit/vector/StrCpyVec.h>
#include <common/storage/striping/StripePattern.h>
#include <common/storage/Path.h>
#include <common/storage/StorageDefinitions.h>
#include <filesystem/FhgfsInode.h>
#include <filesystem/FhgfsOpsHelper.h>
#include <filesystem/FsDirInfo.h>
#include <toolkit/CacheStore.h>
#include <toolkit/InodeRefStore.h>
#include <toolkit/NoAllocBufferStore.h>
#include "FhgfsOpsDir.h"
#include "FhgfsOpsHelper.h"


/**
 * Log file system operations and optionally additional messages.
 * Used to trace function calls and to print the path the function is to operate on.
 *
 * @param level         The log level
 * @param app           Fhgfs client data structure
 * @param dentry        Common vfs directory entry
 * @param logContext    Usually the name of the calling function
 * @param msgStr        Optional message string, may be NULL
 * @param ...           Optional arguments according to format given in msgStr
 */
void FhgfsOpsHelper_logOpMsg(int level, App* app, struct dentry* dentry, const char *logContext,
   const char *msgStr, ...)
{
   Logger* log = App_getLogger(app);
   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf = NULL; // NoAllocBufferStore_addBuf() can detect a wrong NULL...
   int pathError = 0;
   char* path;
   const char* noPath = "n/a (no dentry)";

   if(level > Logger_getLogLevel(log) )
      return;

   // Dentry might be NULL, for example if an open file was deleted. We cannot figure out
   // the path then.
   if(dentry)
   {
      path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);
      if (IS_ERR(path))
      {
         path = (char*)noPath;
         pathError = 1;
      }
   }
   else
      path = (char*)noPath;

   if(msgStr)
   { // All of that is to add the path (filename) before msgStr
      va_list ap;
      char* newMsg;

      newMsg = kmalloc(LOGGER_LOGBUF_SIZE, GFP_NOFS);
      if(newMsg)
         snprintf(newMsg, LOGGER_LOGBUF_SIZE, "called. Path: %s. %s", path, msgStr);
      else // vmalloc failed. Likely an out memory situation, we still try to print msgStr
         newMsg = (char*)msgStr;

      va_start(ap, msgStr);
      Logger_logFormattedVA(log, level, logContext, newMsg, ap);
      va_end(ap);

      if(newMsg != msgStr)
         kfree(newMsg);
   }
   else // msgStr is NULL, no need to concatenate the path with NULL
      Logger_logFormatted(log, level, logContext, "called. Path: %s", path);

   if(dentry && !pathError)
      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
}

/**
 * Refreshes the dirInfo by retrieving a new version from the nodes (but only if required or
 * forced).
 *
 * Note: This method guarantees either valid local names range for currentServerOffset or empty
 * dirContents list on completion.
 *
 * @param dirInfo holds the relevant serverOffset for the refresh
 * @param forceUpdate use this to force an update (e.g. when the user seeked)
 * @return negative linux error code on error, 0 otherwise
 */
int FhgfsOpsHelper_refreshDirInfoIncremental(App* app, const EntryInfo* entryInfo,
   FsDirInfo* dirInfo, fhgfs_bool forceUpdate)
{
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOpsHelper (refresh dir info incremental)";

   const unsigned maxNames = 100; // max number of retrieved names

   int retVal = 0;
   FhgfsOpsErr listRes = FhgfsOpsErr_SUCCESS;
   StrCpyVec* dirContents = FsDirInfo_getDirContents(dirInfo);
   size_t dirContentsLen = StrCpyVec_length(dirContents);
   size_t currentContentsPos = FsDirInfo_getCurrentContentsPos(dirInfo); // pos inside contents vec

   LOG_DEBUG_FORMATTED(log, Log_SPAM, logContext, "called. EntryID: %s, Owner: %hu",
      entryInfo->entryID, entryInfo->ownerNodeID);
   IGNORE_UNUSED_VARIABLE(log);
   IGNORE_UNUSED_VARIABLE(logContext);


   if(forceUpdate)
   { // user seeked backwards (or something else makes an update necessary)
      LOG_DEBUG(log, Log_SPAM, logContext, "forced update of dir contents");

      listRes = FhgfsOpsRemoting_listdirFromOffset(entryInfo, dirInfo, maxNames);
   }
   else
   if(FsDirInfo_getEndOfDir(dirInfo) && (currentContentsPos >= dirContentsLen) )
   { // we reached the end of the dir contents => do nothing
      LOG_DEBUG(log, Log_SPAM, logContext, "reached end of dir contents by offset");

      StrCpyVec_clear(FsDirInfo_getDirContents(dirInfo) );
   }
   else
   if(currentContentsPos < dirContentsLen)
   { // inside the local contents region => do nothing
      LOG_DEBUG_FORMATTED(log, Log_SPAM, logContext,
         "offset inside local contents region: %lld <= %lld",
         (long long)currentContentsPos, (long long)dirContentsLen);
   }
   else
   { // initial retrieval or offset outside current local contents region
      LOG_DEBUG(log, Log_SPAM, logContext, "retrieving by offset");

      listRes = FhgfsOpsRemoting_listdirFromOffset(entryInfo, dirInfo, maxNames);
   }


   if(unlikely(listRes != FhgfsOpsErr_SUCCESS) )
   { // error
      StrCpyVec_clear(FsDirInfo_getDirContents(dirInfo) );
      FsDirInfo_setCurrentContentsPos(dirInfo, 0);

      retVal = FhgfsOpsErr_toSysErr(listRes);
   }

   return retVal;
}


/**
 * Flush cache buffer and return it to the store.
 *
 * @param discardCacheOnError fhgfs_true to discard a write cache if the remote write
 * was not successful; fhgfs_false to just release it to the store in this case
 */
FhgfsOpsErr FhgfsOpsHelper_flushCache(App* app, FhgfsInode* fhgfsInode,
   fhgfs_bool discardCacheOnError)
{
   FhgfsOpsErr retVal;

   FhgfsInode_fileCacheExclusiveLock(fhgfsInode); // L O C K

   retVal = __FhgfsOpsHelper_flushCacheUnlocked(app, fhgfsInode, discardCacheOnError);

   FhgfsInode_fileCacheExclusiveUnlock(fhgfsInode); // U N L O C K

   return retVal;
}

/**
 * Flush cache buffer and return it to the store, but return immediately if the cache lock cannot
 * be acquired immediately.
 *
 * @param discardCacheOnError fhgfs_true to discard a write cache if the remote write
 * was not successful; fhgfs_false to just release it to the store in this case
 * @return FhgfsOpsErr_INUSE if cache lock is contended and nothing has been flushed.
 */
FhgfsOpsErr FhgfsOpsHelper_flushCacheNoWait(App* app, FhgfsInode* fhgfsInode,
   fhgfs_bool discardCacheOnError)
{
   int gotLock; // 1 if we got the lock, 0 otherwise
   FhgfsOpsErr retVal;

   gotLock = FhgfsInode_fileCacheExclusiveTryLock(fhgfsInode); // (T R Y) L O C K
   if(gotLock != 1)
      return FhgfsOpsErr_INUSE;

   retVal = __FhgfsOpsHelper_flushCacheUnlocked(app, fhgfsInode, discardCacheOnError);

   FhgfsInode_fileCacheExclusiveUnlock(fhgfsInode); // U N L O C K

   return retVal;
}

/**
 * Flush cache buffer and return it to the store.
 *
 * Note: Unlocked, so caller must hold inode cache lock.
 *
 * @param discardCacheOnError fhgfs_true to discard a write cache if the remote write
 * was not successful; fhgfs_false to just release it to the store in this case
 */
FhgfsOpsErr __FhgfsOpsHelper_flushCacheUnlocked(App* app, FhgfsInode* fhgfsInode,
   fhgfs_bool discardCacheOnError)
{
   /* note: we don't take a handle to an open file here, because we wouldn't know whether the
      given handle is a read-only handle that just needs to flush the write cache to establish a new
      read cache. that's why we use the _writeStatlessInode() method here. */

   CacheBuffer* cacheBuffer = Fhgfsinode_getFileCacheBuffer(fhgfsInode);
   enum FileBufferType cacheType = cacheBuffer->bufType;
   ssize_t writeRes;

   if(cacheType == FileBufferType_NONE)
      return FhgfsOpsErr_SUCCESS;

   if(cacheType == FileBufferType_READ)
   { // file has a read cache => just discard the cached data
      __FhgfsOpsHelper_discardCache(app, fhgfsInode);
      return FhgfsOpsErr_SUCCESS;
   }

   // file has a write cache => send the cached data to the storage nodes

   writeRes = FhgfsOpsHelper_writeStatelessInode(fhgfsInode, cacheBuffer->buf,
      cacheBuffer->bufUsageLen, cacheBuffer->fileOffset);
   if(unlikely(writeRes < (ssize_t)cacheBuffer->bufUsageLen) )
   { // write did not succeed
      if(discardCacheOnError)
         __FhgfsOpsHelper_discardCache(app, fhgfsInode);

      if(writeRes > 0)
         return FhgfsOpsErr_NOSPACE;

      return -writeRes;
   }

   // write succeeded

   __FhgfsOpsHelper_discardCache(app, fhgfsInode);

   return FhgfsOpsErr_SUCCESS;
}

/**
 * Buffered writing.
 * Tries to add written data to an existing cache or flushes the cache (if any) and delegates to
 * writeCacheFlushed().
 *
 * Note: This method also updates FsFileInfo cache hits counter.
 *
 * @return number of bytes written or negative fhgfs error code
 */
ssize_t FhgfsOpsHelper_writeCached(const char __user *buf, size_t size,
   loff_t offset, FhgfsInode* fhgfsInode, FsFileInfo* fileInfo, RemotingIOInfo* ioInfo)
{
   App* app = ioInfo->app;
   NoAllocBufferStore* cacheStore = App_getCacheBufStore(app);

   ssize_t retVal;
   CacheBuffer* cacheBuffer;
   enum FileBufferType cacheType;
   size_t bufLenLeft; // remaining buffer length
   loff_t currentCacheEndOffset;
   int userBufCopyRes;

   FhgfsInode_fileCacheExclusiveLock(fhgfsInode); // L O C K

   cacheBuffer = Fhgfsinode_getFileCacheBuffer(fhgfsInode);
   cacheType = cacheBuffer->bufType;

   if(cacheType == FileBufferType_NONE)
   { // file doesn't have any cache buffer currently

      // update cache hits by guessing whether this would have been a cache hit if we had a buffer
      loff_t virtualCacheStart = FsFileInfo_getLastWriteOffset(fileInfo);
      loff_t virtualCacheEnd = virtualCacheStart + NoAllocBufferStore_getBufSize(cacheStore) - 1;

      if( (offset >= virtualCacheStart) && (offset <= virtualCacheEnd) )
         FsFileInfo_incCacheHits(fileInfo); // could have been a cache hit
      else
         FsFileInfo_decCacheHits(fileInfo); // would probably not have been a cache hit

      retVal = __FhgfsOpsHelper_writeCacheFlushed(buf, size, offset, fhgfsInode, fileInfo, ioInfo);
      goto unlock_and_exit;
   }

   if(cacheType == FileBufferType_READ)
   { // file has a read cache => just discard the cached data
      FsFileInfo_decCacheHits(fileInfo);

      __FhgfsOpsHelper_discardCache(app, fhgfsInode);

      retVal = __FhgfsOpsHelper_writeCacheFlushed(buf, size, offset, fhgfsInode, fileInfo, ioInfo);
      goto unlock_and_exit;
   }

   // file has a write cache => can we extend it or do we have to flush it?

   currentCacheEndOffset = cacheBuffer->fileOffset + cacheBuffer->bufUsageLen;
   bufLenLeft = cacheBuffer->bufUsageMaxLen - cacheBuffer->bufUsageLen;
   if( (offset != currentCacheEndOffset) ||
       (size > bufLenLeft) )
   { // offset doesn't fit or not enough cache room left => flush cache
      FhgfsOpsErr flushRes;

      FsFileInfo_decCacheHits(fileInfo);

      flushRes = __FhgfsOpsHelper_flushCacheUnlocked(app, fhgfsInode, fhgfs_false);
      if(unlikely(flushRes != FhgfsOpsErr_SUCCESS) )
      { // flush failed
         retVal = -flushRes;
         goto unlock_and_exit;
      }

      retVal = __FhgfsOpsHelper_writeCacheFlushed(buf, size, offset, fhgfsInode, fileInfo, ioInfo);
      goto unlock_and_exit;
   }

   // offset fits and we have enough room left for the write => copy data to cache buf (and update)

   FsFileInfo_incCacheHits(fileInfo);

   userBufCopyRes = __os_copy_from_user(&(cacheBuffer->buf)[cacheBuffer->bufUsageLen], buf, size);
   if(unlikely(userBufCopyRes) )
   { // copy failed
      Logger* log = App_getLogger(app);
      Logger_log(log, Log_DEBUG, __func__, "Buffer copy from userspace failed (invalid buffer)");

      retVal = -FhgfsOpsErr_ADDRESSFAULT;
      goto unlock_and_exit;
   }

   cacheBuffer->bufUsageLen += size;

   bufLenLeft = cacheBuffer->bufUsageMaxLen - cacheBuffer->bufUsageLen; // re-calc remaining bufLen

   // write has been completely cached => check whether the cache is full now

   if(!bufLenLeft)
   { // cache buf used up => flush it
      FhgfsOpsErr flushRes = __FhgfsOpsHelper_flushCacheUnlocked(app, fhgfsInode, fhgfs_false);

      if(unlikely(flushRes != FhgfsOpsErr_SUCCESS) )
      { // flush failed
         retVal = -flushRes;
         goto unlock_and_exit;
      }
   }

   // write was successful if we got here

   retVal = size;


unlock_and_exit:
   FhgfsInode_fileCacheExclusiveUnlock(fhgfsInode); // U N L O C K

   return retVal;
}

/**
 * Buffered reading.
 * Tries to read data from an existing cache or flushes the cache and delegates to
 * readCacheFlushed().
 *
 * @return number of bytes read or negative fhgfs error code
 */
ssize_t FhgfsOpsHelper_readCached(char __user *buf, size_t size, loff_t offset,
   FhgfsInode* fhgfsInode, FsFileInfo* fileInfo, RemotingIOInfo* ioInfo)
{
   App* app = ioInfo->app;
   NoAllocBufferStore* cacheStore = App_getCacheBufStore(app);

   ssize_t retVal;
   CacheBuffer* cacheBuffer;
   enum FileBufferType cacheType;
   loff_t readEndOffset = offset + size - 1;
   loff_t cacheEndOffset;
   ssize_t remoteReadSize;
   loff_t cacheCopyOffset;
   size_t cacheCopySize;
   int userBufCopyRes;

   FhgfsInode_fileCacheExclusiveLock(fhgfsInode); // L O C K

   cacheBuffer = Fhgfsinode_getFileCacheBuffer(fhgfsInode);
   cacheType = cacheBuffer->bufType;

   if(cacheType == FileBufferType_NONE)
   { // file has no buffer attached

      // update cache hits by guessing whether this would have been a cache hit if we had a buffer
      loff_t virtualCacheStart = FsFileInfo_getLastReadOffset(fileInfo);
      loff_t virtualCacheEnd = virtualCacheStart + NoAllocBufferStore_getBufSize(cacheStore) - 1;

      if( (virtualCacheEnd >= offset) && (virtualCacheStart <= readEndOffset) )
         FsFileInfo_incCacheHits(fileInfo); // range overlap => could have been a cache hit
      else
         FsFileInfo_decCacheHits(fileInfo); // would probably not have been a cache hit

      retVal = __FhgfsOpsHelper_readCacheFlushed(buf, size, offset, fhgfsInode, fileInfo, ioInfo);
      goto unlock_and_exit;
   }

   if(cacheType == FileBufferType_WRITE)
   { // file has a write cache => flush it
      FhgfsOpsErr flushRes;

      FsFileInfo_decCacheHits(fileInfo);

      flushRes = __FhgfsOpsHelper_flushCacheUnlocked(app, fhgfsInode, fhgfs_false);

      if(unlikely(flushRes != FhgfsOpsErr_SUCCESS) )
      { // flush failed
         retVal = -flushRes;
         goto unlock_and_exit;
      }

      retVal = __FhgfsOpsHelper_readCacheFlushed(buf, size, offset, fhgfsInode, fileInfo, ioInfo);
      goto unlock_and_exit;
   }

   // file has a read cache => does it overlap with the read range?

   cacheEndOffset = cacheBuffer->fileOffset + cacheBuffer->bufUsageLen - 1;

   if( (cacheEndOffset < offset) ||
       (cacheBuffer->fileOffset > readEndOffset) )
   { // cache range and read range do not overlap => flush
      FhgfsOpsErr flushRes;

      FsFileInfo_decCacheHits(fileInfo);

      flushRes = __FhgfsOpsHelper_flushCacheUnlocked(app, fhgfsInode, fhgfs_false);

      if(unlikely(flushRes != FhgfsOpsErr_SUCCESS) )
      { // flush failed
         retVal = -flushRes;
         goto unlock_and_exit;
      }

      retVal = __FhgfsOpsHelper_readCacheFlushed(buf, size, offset, fhgfsInode, fileInfo, ioInfo);
      goto unlock_and_exit;
   }

   // cache range and read range are overlapping. three possible cases to be handled:
   // read range before/inside/behind the cache range

   FsFileInfo_incCacheHits(fileInfo);

   remoteReadSize = 0;

   if(offset < cacheBuffer->fileOffset)
   { // read begins before cache => remote-read and copy the rest
      ssize_t readRes;

      remoteReadSize = cacheBuffer->fileOffset - offset;

      readRes = FhgfsOpsRemoting_readfile(buf, remoteReadSize, offset, ioInfo);
      if(readRes < remoteReadSize)
      { // error or end-of-file => invalidate cache
         __FhgfsOpsHelper_discardCache(app, fhgfsInode);

         retVal = readRes;
         goto unlock_and_exit;
      }
   }

   // remote read (if any) succeeded => copy part from cache
   cacheCopyOffset = (offset <= cacheBuffer->fileOffset) ? 0 : (offset - cacheBuffer->fileOffset);
   cacheCopySize = MIN(cacheBuffer->fileOffset + cacheBuffer->bufUsageLen, offset + size) -
      (cacheBuffer->fileOffset + cacheCopyOffset);

   userBufCopyRes = __os_copy_to_user(
      &buf[remoteReadSize], &(cacheBuffer->buf)[cacheCopyOffset], cacheCopySize);
   if(unlikely(userBufCopyRes) )
   { // copy failed
      Logger* log = App_getLogger(app);
      Logger_log(log, Log_DEBUG, __func__, "Buffer copy to userspace failed (invalid buffer)");

      retVal = -FhgfsOpsErr_ADDRESSFAULT;
      goto unlock_and_exit;
   }


   // has everything been read or is there a remainder behind the cacheBuf?

   if(readEndOffset > cacheEndOffset)
   { // there is a remainder behind the cacheBuf => discard cache (to allow new caching)
      loff_t remainderBufOffset = remoteReadSize + cacheCopySize;
      loff_t remainderFileOffset = cacheEndOffset + 1;
      size_t remainderSize = readEndOffset - cacheEndOffset;
      ssize_t remainderRes;

      __FhgfsOpsHelper_discardCache(app, fhgfsInode);

      remainderRes = __FhgfsOpsHelper_readCacheFlushed(&buf[remainderBufOffset],
         remainderSize, remainderFileOffset, fhgfsInode, fileInfo, ioInfo);
      if(unlikely(remainderRes < 0) )
      { // reading failed
         retVal = remainderRes;
         goto unlock_and_exit;
      }

      retVal = remoteReadSize + cacheCopySize + remainderRes;
      goto unlock_and_exit;
   }

   // read was successful if we got here

   retVal = size;


unlock_and_exit:
   FhgfsInode_fileCacheExclusiveUnlock(fhgfsInode); // U N L O C K

   return retVal;
}



/**
 * Try to cache the write data or delegate to the normal remoting writefile().
 *
 * Note: Use this only when definitely no file cache entry exists.
 * Note: Unlocked, so caller must hold inode cache lock.
 *
 * @return number of bytes written or negative fhgfs error code
 */
ssize_t __FhgfsOpsHelper_writeCacheFlushed(const char __user *buf,
   size_t size, loff_t offset, FhgfsInode* fhgfsInode, FsFileInfo* fileInfo, RemotingIOInfo* ioInfo)
{
   App* app = ioInfo->app;
   NoAllocBufferStore* cacheStore = App_getCacheBufStore(app);
   InodeRefStore* refStore = App_getInodeRefStore(app);

   StripePattern* pattern = ioInfo->pattern;
   size_t currentChunkSize;
   size_t maxCacheLen;
   CacheBuffer* cacheBuffer;
   int userBufCopyRes;

   // caching disabled?
   if(!FsFileInfo_getAllowCaching(fileInfo) || (FsFileInfo_getCacheHits(fileInfo) <= 0) )
      return FhgfsOpsRemoting_writefile(buf, size, offset, ioInfo);

      // check whether we're (probably) writing a very small file
   if(!offset && (size < FSFILEINFO_CACHE_MINWRITE_THRESHOLD) )
   { // caching would induce more overhead than write-through (for small files) => write through

      /* commented out: because I don't see a real reason why we shouldn't cache it anyways, since
         it's not unlikely that there might come more data after this. */

      //return FhgfsOpsRemoting_writefile(buf, size, offset, ioInfo);
   }

   // check whether the write size is larger than a single cacheBuf or would span multiple chunks
   currentChunkSize = StripePattern_getChunkEnd(pattern, offset) - offset + 1;
   maxCacheLen = MIN(currentChunkSize, NoAllocBufferStore_getBufSize(cacheStore) );
   if(size >= maxCacheLen) // (Note: '=' because we don't want a completely filled up buffer)
   { // scenario not allowed => write through
      return FhgfsOpsRemoting_writefile(buf, size, offset, ioInfo);
   }

   // looks like we got something to cache here (data fits completely into a single cacheBuf)

   // create new cache (if buffer available)

   cacheBuffer = Fhgfsinode_getFileCacheBuffer(fhgfsInode);

   #ifdef FHGFS_DEBUG
      FHGFS_BUG_ON(cacheBuffer->buf, "Looks like we're about to leak a cache buffer");
   #endif // FHGFS_DEBUG

   cacheBuffer->buf = NoAllocBufferStore_instantBuf(cacheStore);
   if(!cacheBuffer->buf)
   { // no cache buffer left in the store => write through
      return FhgfsOpsRemoting_writefile(buf, size, offset, ioInfo);
   }

   cacheBuffer->bufType = FileBufferType_WRITE; // (needed here for _discardCache() below)

   // init cache entry fields and copy data to cacheBuf
   userBufCopyRes = __os_copy_from_user(cacheBuffer->buf, buf, size);
   if(unlikely(userBufCopyRes) )
   { // copy failed
      Logger* log = App_getLogger(app);
      Logger_log(log, Log_DEBUG, __func__, "Buffer copy from userspace failed (invalid buffer)");

      __FhgfsOpsHelper_discardCache(app, fhgfsInode);
      return -FhgfsOpsErr_ADDRESSFAULT;
   }

   cacheBuffer->bufUsageLen = size;
   cacheBuffer->bufUsageMaxLen = maxCacheLen;
   cacheBuffer->fileOffset = offset;

   // add to inode ref store for async flush
   InodeRefStore_addAndReferenceInode(refStore, FHGFS_VFSINODE(fhgfsInode) );

   return size;
}


/**
 * Try to read-ahead data or just do a plain normal remoting readfile().
 *
 * Note: Use this only when definitely no file cache entry exists.
 * Note: Unlocked, so caller must hold inode cache lock.
 *
 * @return number of bytes read or negative fhgfs error code
 */
ssize_t __FhgfsOpsHelper_readCacheFlushed(char __user *buf, size_t size, loff_t offset,
   FhgfsInode* fhgfsInode, FsFileInfo* fileInfo, RemotingIOInfo* ioInfo)
{
   App* app = ioInfo->app;
   NoAllocBufferStore* cacheStore = App_getCacheBufStore(app);
   InodeRefStore* refStore = App_getInodeRefStore(app);

   StripePattern* pattern = ioInfo->pattern;
   size_t currentChunkSize;
   size_t maxCacheLen;
   CacheBuffer* cacheBuffer;
   ssize_t readRes;
   int userBufCopyRes;


   // caching disabled?
   if(!FsFileInfo_getAllowCaching(fileInfo) || (FsFileInfo_getCacheHits(fileInfo) <= 0) ||
      FhgfsInode_getIsFileOpenByMultipleReaders(fhgfsInode) )
      return FhgfsOpsRemoting_readfile(buf, size, offset, ioInfo);

   // check whether the read size is larger than a single cacheBuf or would span multiple chunks
   currentChunkSize = StripePattern_getChunkEnd(pattern, offset) - offset + 1;
   maxCacheLen = MIN(currentChunkSize, NoAllocBufferStore_getBufSize(cacheStore) );
   if(size >= maxCacheLen) // (Note: '=' because we don't want a completely used up buffer)
   { // scenario not allowed => read direct
      return FhgfsOpsRemoting_readfile(buf, size, offset, ioInfo);
   }

   // looks like we got something to cache here (read size is smaller than a single cacheBuf)

   // create new cache (if buffer available)

   cacheBuffer = Fhgfsinode_getFileCacheBuffer(fhgfsInode);

   #ifdef FHGFS_DEBUG
      FHGFS_BUG_ON(cacheBuffer->buf, "Looks like we're about to leak a cache buffer");
   #endif // FHGFS_DEBUG

   cacheBuffer->buf = NoAllocBufferStore_instantBuf(cacheStore);
   if(!cacheBuffer->buf)
   { // no cache buffer left in the store => read direct
      return FhgfsOpsRemoting_readfile(buf, size, offset, ioInfo);
   }

   cacheBuffer->bufType = FileBufferType_READ; // (needed here for _discardCache() below)

   // we got a buffer => read as much as possible into the cache buffer

   if(!offset && (size < FSFILEINFO_CACHE_SLOWSTART_READLEN) )
      maxCacheLen = FSFILEINFO_CACHE_SLOWSTART_READLEN; /* reduce read-ahead at file start for small
         reads. good if e.g. a process is only looking at file starts */


   readRes = FhgfsOpsRemoting_readfile(cacheBuffer->buf, maxCacheLen, offset, ioInfo);
   if(readRes <= 0)
   { // error or immediate end of file
      __FhgfsOpsHelper_discardCache(app, fhgfsInode);
      return readRes;
   }

   // init cache entry fields and copy data from cache buffer to user buffer
   userBufCopyRes = __os_copy_to_user(buf, cacheBuffer->buf, MIN(size, (size_t)readRes) );
   if(unlikely(userBufCopyRes) )
   { // copy failed
      Logger* log = App_getLogger(app);
      Logger_log(log, Log_DEBUG, __func__, "Buffer copy to userspace failed (invalid buffer)");

      __FhgfsOpsHelper_discardCache(app, fhgfsInode);
      return -FhgfsOpsErr_ADDRESSFAULT;
   }

   cacheBuffer->bufUsageLen = readRes;
   cacheBuffer->bufUsageMaxLen = maxCacheLen;
   cacheBuffer->fileOffset = offset;

   // add to inode ref store for async flush
   InodeRefStore_addAndReferenceInode(refStore, FHGFS_VFSINODE(fhgfsInode) );

   return MIN(size, (size_t)readRes);
}

/**
 * Discard the current cache buffer and return it to the store.
 *
 * Note: Unlocked, so caller must hold the inode cache lock.
 */
void __FhgfsOpsHelper_discardCache(App* app, FhgfsInode* fhgfsInode)
{
   NoAllocBufferStore* cacheStore = App_getCacheBufStore(app);
   InodeRefStore* refStore = App_getInodeRefStore(app);

   CacheBuffer* cacheBuffer = Fhgfsinode_getFileCacheBuffer(fhgfsInode);

   #ifdef FHGFS_DEBUG
   if( (cacheBuffer->buf == NULL) || (cacheBuffer->bufType == FileBufferType_NONE) )
   {
      FHGFS_BUG_ON(1, "Attempting to discard an invalid cache buffer");
      return;
   }
   #endif // FHGFS_DEBUG

   NoAllocBufferStore_addBuf(cacheStore, cacheBuffer->buf);

   cacheBuffer->buf = NULL; // (NULL'ing required for debug sanity checks)
   cacheBuffer->bufType = FileBufferType_NONE;

   // remove inode from async flush store
   InodeRefStore_removeAndReleaseInode(refStore, FHGFS_VFSINODE(fhgfsInode) );
}

/**
 * Reads chunk by chunk from the servers and zero-fills the buffer if a server returns EOF.
 * So the caller must make sure that the file actually has at least the requested size.
 *
 * Note: Intended for sparse file reading.
 * Note: There is also a similar version for kernel buffers.
 */
FhgfsOpsErr FhgfsOpsHelper_readOrClearUser(App* app, char __user *buf, size_t size,
   loff_t offset, FsFileInfo* fileInfo, RemotingIOInfo* ioInfo)
{
   size_t toBeRead = size;
   loff_t currentOffset = offset;
   StripePattern* pattern = ioInfo->pattern;

   while(toBeRead)
   {
      size_t currentBufOffset = size - toBeRead;
      size_t currentChunkSize = StripePattern_getChunkEnd(
         pattern, currentOffset) - currentOffset + 1;
      size_t currentReadSize = MIN(currentChunkSize, toBeRead);
      ssize_t currentReadRes;

      currentReadRes = FhgfsOpsRemoting_readfile(&buf[currentBufOffset], currentReadSize,
         currentOffset, ioInfo);

      if(unlikely(currentReadRes < 0) )
         return -currentReadRes;

      if( (size_t)currentReadRes < currentReadSize)
      { // zero-fill the remainder
         __os_clear_user(&buf[currentBufOffset+currentReadRes], currentReadSize-currentReadRes);
      }

      currentOffset += currentReadSize;
      toBeRead -= currentReadSize;
   }

   return FhgfsOpsErr_SUCCESS;
}

/**
 * Reads chunk by chunk from the servers and zero-fills the buffer if a server returns EOF.
 * So the caller must make sure that the file actually has at least the requested size.
 *
 * Note: Intended for sparse file reading.
 * Note: There is also a similar version for user buffers.
 *
 * @return  cummulated read-result or negative (FhGFS) error code
 */
ssize_t FhgfsOpsHelper_readOrClearKernel(App* app, char* buf, size_t size,
   loff_t offset, FsFileInfo* fileInfo, RemotingIOInfo* ioInfo)
{
   size_t toBeRead = size;
   loff_t currentOffset = offset;
   StripePattern* pattern = ioInfo->pattern;
   loff_t sumReadRes = 0;

   while(toBeRead)
   {
      size_t currentBufOffset = size - toBeRead;
      size_t currentChunkSize = StripePattern_getChunkEnd(
         pattern, currentOffset) - currentOffset + 1;
      size_t currentReadSize = MIN(currentChunkSize, toBeRead);
      ssize_t currentReadRes;

      currentReadRes = FhgfsOpsRemoting_readfile(&buf[currentBufOffset], currentReadSize,
         currentOffset, ioInfo);

      if(unlikely(currentReadRes < 0) )
         return currentReadRes;

      if( (size_t)currentReadRes < currentReadSize)
      { // zero-fill the remainder
         os_memset(&buf[currentBufOffset+currentReadRes], 0, currentReadSize-currentReadRes);
      }

      currentOffset += currentReadSize;
      toBeRead -= currentReadSize;
      sumReadRes += currentReadRes;
   }

   return sumReadRes;
}

/**
 * Takes two paths that are relative to the mount point and creates a new path that is relative from
 * pathFromStr to pathToStr.
 *
 * @param outPathRelativeToStr will be kalloc'ed and needs to be kfree'd by the caller
 */
void FhgfsOpsHelper_getRelativeLinkStr(const char* pathFromStr, const char* pathToStr,
   char** outPathRelativeToStr)
{
   // the idea here is to prepend a "../" to pathTo for every parent dir of pathFrom

   int i;
   int pathFromLen; // (must be signed, because it might be negative for certain paths)
   Path pathFrom;
   Path pathTo; // will be modified to become the relative result path
   StrCpyList* pathFromElems;
   StrCpyList* pathToElems;

   Path_initFromString(&pathFrom, pathFromStr);
   Path_initFromString(&pathTo, pathToStr);

   pathFromElems = Path_getPathElems(&pathFrom);
   pathToElems = Path_getPathElems(&pathTo);

   pathFromLen = StrCpyList_length(pathFromElems);

   // insert a ".." for every parent dir in pathFrom

   // (note: -1 to excluse the final path element)
   for(i = 0; i < (pathFromLen-1); i++)
   {
      StrCpyList_addHead(pathToElems, "..");
   }

   Path_setAbsolute(&pathTo, fhgfs_false);

   *outPathRelativeToStr = Path_getPathAsStrCopy(&pathTo);

   Path_uninit(&pathFrom);
   Path_uninit(&pathTo);
}

/**
 * Note: creates the symlink as a normal file and writes the target to it
 *
 * @param mode access mode (permission flags)
 * @return 0 on success, negative linux error code otherwise
 */
int FhgfsOpsHelper_symlink(App* app, const EntryInfo* parentInfo, const char* to,
   struct CreateInfo* createInfo, EntryInfo* outEntryInfo)
{
   //Logger* log = App_getLogger(app);
   //const char* logContext = "FhgfsOpsHelper_symlink";

   int retVal = 0;

   size_t toStrLen = os_strlen(to);
   FhgfsOpsErr mkRes;
   AtomicInt maxUsedTargetIndex;
   RemotingIOInfo ioInfo;
   FhgfsOpsErr openRes;
   size_t numTargets;
   ssize_t writeRes;
   FhgfsOpsErr closeRes;
   BitStore firstWriteDone;

   AtomicInt_init(&maxUsedTargetIndex, -1);

   // create the file

   mkRes = FhgfsOpsRemoting_mkfile(app, parentInfo, createInfo, outEntryInfo);
   if(mkRes != FhgfsOpsErr_SUCCESS)
   { // error
      retVal = FhgfsOpsErr_toSysErr(mkRes);
      goto err_exit;
   }

   // open the file

   RemotingIOInfo_initOpen(app, OPENFILE_ACCESS_WRITE, &maxUsedTargetIndex, &ioInfo);

   openRes = FhgfsOpsRemoting_openfile(outEntryInfo, &ioInfo);
   if(openRes != FhgfsOpsErr_SUCCESS)
   { // error
      FhgfsOpsRemoting_unlinkfile(app, parentInfo, createInfo->entryName);

      retVal = FhgfsOpsErr_toSysErr(openRes);
      goto err_cleanup;
   }

   numTargets = UInt16Vec_length(ioInfo.pattern->getStripeTargetIDs(ioInfo.pattern));
   BitStore_initWithSizeAndReset(&firstWriteDone, numTargets);
   ioInfo.firstWriteDone = &firstWriteDone;
   ioInfo.userID = createInfo->userID;
   ioInfo.groupID = createInfo->groupID;

   // write link-destination to the file

   writeRes = FhgfsOpsRemoting_writefile(to, toStrLen, 0, &ioInfo);
   if(writeRes < (ssize_t)toStrLen)
   { // error
      FhgfsOpsHelper_closefileWithAsyncRetry(outEntryInfo, &ioInfo);
      FhgfsOpsRemoting_unlinkfile(app, parentInfo, createInfo->entryName);

      retVal = (writeRes < 0) ? FhgfsOpsErr_toSysErr(-writeRes) : FhgfsOpsErr_INTERNAL;
      goto err_cleanup;
   }


   // close the file

   closeRes = FhgfsOpsHelper_closefileWithAsyncRetry(outEntryInfo, &ioInfo);

   if(closeRes != FhgfsOpsErr_SUCCESS)
   { // error
      retVal = FhgfsOpsErr_toSysErr(closeRes);
      goto err_cleanup;
   }


   // clean-up

   BitStore_uninit(&firstWriteDone);
   RemotingIOInfo_freeVals(&ioInfo);

   return retVal;


err_cleanup:
   if (ioInfo.firstWriteDone)
      BitStore_uninit(ioInfo.firstWriteDone);

   RemotingIOInfo_freeVals(&ioInfo);
   EntryInfo_uninit(outEntryInfo);

err_exit:
   return retVal;
}

/**
 * Reads a symlink.
 *
 * @return number of read bytes or negative linux error code
 */
int FhgfsOpsHelper_readlink(App* app, const EntryInfo* entryInfo, char __user* buf, int size)
{
   return FhgfsOpsHelper_readStateless(app, entryInfo, buf, size, 0);
}

/**
 * Opens a file (the file must exist), reads data from it and closes it.
 *
 * Note: This is really slow currently, because of the open/close overhead.
 *
 * @return number of read bytes or negative linux error code
 */
ssize_t FhgfsOpsHelper_readStateless(App* app, const EntryInfo* entryInfo,
   char __user *buf, size_t size, loff_t offset)
{
   int retVal = -EREMOTEIO;

   AtomicInt maxUsedTargetIndex;
   RemotingIOInfo ioInfo;
   FhgfsOpsErr openRes;
   ssize_t readRes;
   size_t numTargets;
   FhgfsOpsErr closeRes;
   BitStore firstWriteDone;

   AtomicInt_init(&maxUsedTargetIndex, -1);

   // open file

   RemotingIOInfo_initOpen(app, OPENFILE_ACCESS_READ, &maxUsedTargetIndex, &ioInfo);

   openRes = FhgfsOpsRemoting_openfile(entryInfo, &ioInfo);

   if(openRes != FhgfsOpsErr_SUCCESS)
   { // error
      retVal = FhgfsOpsErr_toSysErr(openRes);
      goto clean_up;
   }

   numTargets = UInt16Vec_length(ioInfo.pattern->getStripeTargetIDs(ioInfo.pattern));
   BitStore_initWithSizeAndReset(&firstWriteDone, numTargets);
   ioInfo.firstWriteDone = &firstWriteDone;

   // read file

   readRes = FhgfsOpsRemoting_readfile(buf, size, offset, &ioInfo);
   if(readRes < 0)
   { // error
      FhgfsOpsHelper_closefileWithAsyncRetry(entryInfo, &ioInfo);

      retVal = FhgfsOpsErr_toSysErr(-readRes);
      goto clean_up;
   }

   retVal = readRes;

   // close the file

   closeRes = FhgfsOpsHelper_closefileWithAsyncRetry(entryInfo, &ioInfo);

   if(closeRes != FhgfsOpsErr_SUCCESS)
   { // error
      retVal = FhgfsOpsErr_toSysErr(closeRes);
   }


   // clean-up
clean_up:
   if (ioInfo.firstWriteDone)
      BitStore_uninit(ioInfo.firstWriteDone);

   RemotingIOInfo_freeVals(&ioInfo);

   return retVal;
}

/**
 * Opens a file (the file must exist), writes the data to it and closes it.
 *
 * Note: This is really slow, because of the open/close overhead - use _writeStatelessInode()
 * method instead, if possible.
 *
 * @return number of written bytes or negative linux error code
 */
ssize_t FhgfsOpsHelper_writeStateless(App* app, const EntryInfo* entryInfo,
   const char __user *buf, size_t size, loff_t offset, unsigned uid, unsigned gid)
{
   int retVal = -EREMOTEIO;

   AtomicInt maxUsedTargetIndex;
   RemotingIOInfo ioInfo;
   FhgfsOpsErr openRes;
   ssize_t writeRes;
   FhgfsOpsErr closeRes;
   BitStore firstWriteDone;
   size_t numTargets;

   AtomicInt_init(&maxUsedTargetIndex, -1);

   // open file

   RemotingIOInfo_initOpen(app, OPENFILE_ACCESS_WRITE, &maxUsedTargetIndex, &ioInfo);

   openRes = FhgfsOpsRemoting_openfile(entryInfo, &ioInfo);

   if(openRes != FhgfsOpsErr_SUCCESS)
   { // error
      retVal = FhgfsOpsErr_toSysErr(openRes);
      goto clean_up;
   }

   numTargets = UInt16Vec_length(ioInfo.pattern->getStripeTargetIDs(ioInfo.pattern));
   BitStore_initWithSizeAndReset(&firstWriteDone, numTargets);
   ioInfo.firstWriteDone = &firstWriteDone;
   ioInfo.userID = uid;
   ioInfo.groupID = gid;

   // write file

   writeRes = FhgfsOpsRemoting_writefile(buf, size, offset, &ioInfo);
   if(writeRes < 0)
   { // error
      FhgfsOpsHelper_closefileWithAsyncRetry(entryInfo, &ioInfo);

      retVal = FhgfsOpsErr_toSysErr(-writeRes);
      goto clean_up;
   }

   retVal = writeRes;

   // close file

   closeRes = FhgfsOpsHelper_closefileWithAsyncRetry(entryInfo, &ioInfo);

   if(closeRes != FhgfsOpsErr_SUCCESS)
   { // error
      retVal = FhgfsOpsErr_toSysErr(closeRes);
   }


   // clean-up
clean_up:
   if (ioInfo.firstWriteDone)
      BitStore_uninit(ioInfo.firstWriteDone);

   RemotingIOInfo_freeVals(&ioInfo);

   return retVal;
}

/**
 * Writes the given buffer by getting a reference handle from the inode and releasing that handle
 * immediately after the write.
 *
 * Note: This can safe the overhead for remote open/close if the file is already open for writing
 * (or reading+writing) by any other process, so it is way better than the _writeStateless()
 * method.
 *
 * @return bytes written or negative fhgfs error code
 */
ssize_t FhgfsOpsHelper_writeStatelessInode(FhgfsInode* fhgfsInode, const char __user *buf,
   size_t size, loff_t offset)
{
   FileHandleType handleType;
   RemotingIOInfo ioInfo;

   FhgfsOpsErr referenceRes;
   ssize_t writeRes;
   FhgfsOpsErr releaseRes;


   // open file

   referenceRes = FhgfsInode_referenceHandle(fhgfsInode, OPENFILE_ACCESS_WRITE, fhgfs_true,
      &handleType);
   if(unlikely(referenceRes != FhgfsOpsErr_SUCCESS) )
   { // error
      return -referenceRes;
   }

   // write file

   FhgfsInode_getRefIOInfo(fhgfsInode, handleType, FhgfsInode_handleTypeToOpenFlags(handleType),
      &ioInfo);

   writeRes = FhgfsOpsRemoting_writefile(buf, size, offset, &ioInfo);
   if(unlikely(writeRes < 0) )
   { // error
      FhgfsInode_releaseHandle(fhgfsInode, handleType);

      return writeRes;
   }

   // close file

   releaseRes = FhgfsInode_releaseHandle(fhgfsInode, handleType);
   if(unlikely(releaseRes != FhgfsOpsErr_SUCCESS) )
   { // error
      return -releaseRes;
   }

   return writeRes;
}
