#include <app/log/Logger.h>
#include <app/App.h>
#include <app/config/Config.h>
#include <common/toolkit/vector/StrCpyVec.h>
#include <common/storage/StorageErrors.h>
#include <common/toolkit/StringTk.h>
#include <common/toolkit/MetadataTk.h>
#include <os/OsCompat.h>
#include <os/OsTypeConversion.h>
#include "FhgfsOpsSuper.h"
#include "FhgfsOpsDir.h"
#include "FhgfsOpsInode.h"
#include "FhgfsOpsFile.h"
#include "FhgfsOpsHelper.h"

#include <linux/namei.h>
#include <linux/backing-dev.h>


static struct kmem_cache* FhgfsInodeCache = NULL;


struct inode_operations fhgfs_file_inode_ops =
{
   .getattr    = FhgfsOps_getattr,
   .permission = FhgfsOps_permission,
   .setattr    = FhgfsOps_setattr,
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,8,0)
   .truncate   = FhgfsOps_truncate, // (truncation is done via setattr since linux-3.8)
#endif // LINUX_VERSION_CODE
};

struct inode_operations fhgfs_file_pagecache_inode_ops =
{
   .getattr    = FhgfsOps_getattr,
   .permission = FhgfsOps_permission,
   .setattr    = FhgfsOps_setattr,
};

struct inode_operations fhgfs_symlink_inode_ops =
{
   .readlink      = generic_readlink, /* FhgfsOps_readlink, */ // default is fine for us currently
   .follow_link   = FhgfsOps_follow_link,
   .put_link      = FhgfsOps_put_link,
   .getattr       = FhgfsOps_getattr,
   .permission    = FhgfsOps_permission,
   .setattr       = FhgfsOps_setattr,
};

static struct inode_operations fhgfs_dir_inode_ops =
{
   .lookup     = FhgfsOps_lookupIntent,
   .create     = FhgfsOps_createIntent,
   .link       = FhgfsOps_link,
   .unlink     = FhgfsOps_unlink,
   .mknod      = FhgfsOps_mknod,
   .symlink    = FhgfsOps_symlink,
   .mkdir      = FhgfsOps_mkdir,
   .rmdir      = FhgfsOps_rmdir,
   .rename     = FhgfsOps_rename,
   .getattr    = FhgfsOps_getattr,
   .permission = FhgfsOps_permission,
   .setattr    = FhgfsOps_setattr,
};

static struct inode_operations fhgfs_special_inode_ops =
{
    .setattr       = FhgfsOps_setattr,
};


static void FhgfsOps_newAttrToInode(struct iattr* iAttr, struct inode* outInode);


/**
 * Find out whether a directory entry exists. Add the dentry and create/associate the
 * corresponding inode in case it does exist.
 *
 * Note: Newer version, superseding the old _lookup() with more efficient messaging based on
 * combined intent requests.
 *
 * @param dentry the thing that we are looking for.
 * @param flags LOOKUP_...
 * @return NULL if we're using the given new dentry, or pointer to another dentry if we
 * find out that it already existed (e.g. from a NFS handle), or ERR_PTR(x) with x being a negative
 * linux error code.
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,6,0)
struct dentry* FhgfsOps_lookupIntent(struct inode* parentDir, struct dentry* dentry,
   struct nameidata* nameidata)
#else
struct dentry* FhgfsOps_lookupIntent(struct inode* parentDir, struct dentry* dentry,
   unsigned flags)
#endif // LINUX_VERSION_CODE
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_lookupIntent";

   struct dentry* returnDentry = NULL; // can be NULL or existing dentry or ERR_PTR
   FhgfsOpsErr statRes;
   fhgfs_stat fhgfsStat;
   FhgfsInode* fhgfsParentInode = FHGFS_INODE(parentDir);
   EntryInfo newEntryInfo;
   const char* entryName = dentry->d_name.name;
   fhgfs_bool freeNewEntryInfo = fhgfs_false;

#if 0
   #if LINUX_VERSION_CODE < KERNEL_VERSION(3,6,0)
      unsigned flags = nameidata ? nameidata->flags : 0;
   #endif // LINUX_VERSION_CODE

   #ifdef KERNEL_HAS_LOOKUP_CONTINUE
      // old kernels need extra CONTINUE|PARENT check to ensure we're at the last element
      unsigned isLookupCreate = (flags & LOOKUP_CREATE) &&
         !(flags & (LOOKUP_CONTINE | LOOKUP_PARENT) );
   #else
      // newer kernels set CREATE only on the last element and CONTINUE got killed
      // (linux commits: remove CONTINUE check: 407938e79edcad; kill CONTINUE: 49084c3bb2055c401)
      unsigned isLookupCreate = (flags & LOOKUP_CREATE);
   #endif // KERNEL_HAS_LOOKUP_CONTINUE
#endif

   if(unlikely(Logger_getLogLevel(log) >= Log_SPAM) )
      FhgfsOpsHelper_logOp(Log_SPAM, app, dentry, logContext);

   if(unlikely(dentry->d_name.len > NAME_MAX) )
      return ERR_PTR(-ENAMETOOLONG);


   /* check if root inode attribs have been fetched already
      (because the kernel doesn't do lookup/revalidate for the root inode) */

   if(unlikely(!FhgfsOps_getIsRootInited(dentry->d_sb) &&
      parentDir->i_ino == FHGFS_INODE_ROOT_INO) )
   {
      int initRootRes;
      int permRes;

      initRootRes = __FhgfsOps_refreshInode(app, parentDir, NULL);
      if(initRootRes)
         return ERR_PTR(initRootRes);

      FhgfsOps_setIsRootInited(dentry->d_sb, fhgfs_true);

      // root permissions might have changed now => recheck permissions

      permRes = os_generic_permission(parentDir, MAY_EXEC);
      if(permRes)
         return ERR_PTR(permRes);
   }


   // retrieve remote stat info for given entry...

   if(unlikely(IS_ROOT(dentry) ) )
   { // root inode is special (though the kernel never actually does a root lookup => unlikely)
      MetadataTk_getRootEntryInfoCopy(app, &newEntryInfo);
      freeNewEntryInfo = fhgfs_true;

      statRes = FhgfsOpsRemoting_statDirect(app, &newEntryInfo, &fhgfsStat);
   }
   else
   { // just a normal subentry of some dir => lookup entryInfo and stat it
#if 0
      if(isLookupCreate)
      { // create intent flag set => don't do anything and let _createIntent() handle it
         statRes = FhgfsOpsErr_PATHNOTEXISTS;
      }
      else
#endif
      { // create flag not set, do our lookup
         FhgfsOpsErr lookupRes;
         EntryInfo* parentInfo;

         FhgfsInode_entryInfoReadLock(fhgfsParentInode); // LOCK EntryInfo

         parentInfo = FhgfsInode_getEntryInfo(fhgfsParentInode);

         lookupRes = FhgfsOpsRemoting_lookupAndStatSubentry(
            app, parentInfo, entryName, &newEntryInfo, &statRes, &fhgfsStat);

         FhgfsInode_entryInfoReadUnlock(fhgfsParentInode); // UNLOCK EntryInfo

         if(lookupRes != FhgfsOpsErr_SUCCESS)
            statRes = lookupRes; // entry not found
         else
         { // lookup successful (entry exists)
            freeNewEntryInfo = fhgfs_true;

            if(statRes == FhgfsOpsErr_NOTOWNER)
            { // metadata not owned by parent/lookup server => need separate stat remoting
               statRes = FhgfsOpsRemoting_statDirect(app, &newEntryInfo, &fhgfsStat);
            }
         }
      }

   }

   // handle result of stat retrieval (e.g. create new inode)...

   if(statRes != FhgfsOpsErr_SUCCESS)
   { // stat error (e.g. entry doesn't exist)

      if(statRes == FhgfsOpsErr_PATHNOTEXISTS)
      {
         /* note: "not exists" does not mean that we return a lookup error. the kernel will check
            after lookup() whether we attached an inode to this dentry and handle it accordingly. */

         #ifndef KERNEL_HAS_S_D_OP
            dentry->d_op = &fhgfs_dentry_ops; // (see below for KERNEL_HAS_S_D_OP comments)
         #endif // KERNEL_HAS_S_D_OP

         d_add(dentry, NULL);
      }
      else
         returnDentry = ERR_PTR(FhgfsOpsErr_toSysErr(statRes) );

   }
   else
   { // entry exists => create inode

      struct kstat kstat;
      struct inode* newInode;

      OsTypeConv_kstatFhgfsToOs(&fhgfsStat, &kstat);

      kstat.ino = FhgfsInode_generateInodeID(dentry->d_sb,
         newEntryInfo.entryID, strlen(newEntryInfo.entryID) );

      newInode = __FhgfsOps_newInode(parentDir->i_sb, &kstat, 0, &newEntryInfo);

      freeNewEntryInfo = fhgfs_false; // newEntryInfo now owned or freed by _newInode()

      if(unlikely(!newInode || IS_ERR(newInode) ) )
         returnDentry = IS_ERR(newInode) ? ERR_PTR(PTR_ERR(newInode) ) : ERR_PTR(-EACCES);
      else
      { // new inode created

         #ifndef KERNEL_HAS_S_D_OP
            /* per-dentry d_ops are deprecated as of linux 2.6.38 (commit c8aebb0c9f8c7471643d5f).
               they are now set on the superblock. (individual per-dentry d_ops could still be set
               with d_set_d_op(), but are not required for us.) */

            dentry->d_op = &fhgfs_dentry_ops;
         #endif // KERNEL_HAS_S_D_OP

         if (S_ISDIR(newInode->i_mode) )
            returnDentry = d_materialise_unique(dentry, newInode);
         else
         {
            returnDentry = d_splice_alias(newInode, dentry); /* (d_splice_alias() also replaces a
               disconnected dentry that was created from a nfs handle) */
         }
      }

   }

   // clean-up
   if(freeNewEntryInfo)
      EntryInfo_uninit(&newEntryInfo);

   return returnDentry;
}


int FhgfsOps_getattr(struct vfsmount* mnt, struct dentry* dentry, struct kstat* kstat)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Config* cfg = App_getConfig(app);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_getattr";

   int retVal = 0;
   struct inode* inode = dentry->d_inode;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   if(unlikely(Logger_getLogLevel(log) >= 5) )
      FhgfsOpsHelper_logOp(5, app, dentry, logContext);

   /* note: we assumed that we could live without refreshInode() here, because either lookup() or
      revalidate() was called before. But that assumption is wrong when a open file is fstat'ed
      (which is relevant for e.g. "tail -f"), so we only use this optimization for closed files.
      The root dir inode is also an important exception here, because it is not being revalidated
      by the kernel.

      note on dirs: we could always refresh dirs here because of caching and tools like "find"
      relying on an updated link count, but while a user is creating subdirs, the stat from "find"
      can never be up-to-date, so that would also be quite useless. */

   if(IS_ROOT(dentry) ||
      FhgfsInode_getIsFileOpen(fhgfsInode) ||
      Config_getTuneRefreshOnGetAttr(cfg) )
   {
      retVal = __FhgfsOps_refreshInode(app, inode, NULL);
   }

   if(!retVal)
   {
      generic_fillattr(inode, kstat);

      kstat->blksize = Config_getTuneInodeBlockSize(cfg);

      if(IS_ROOT(dentry) )
         kstat->ino = FHGFS_INODE_ROOT_INO; // root => assign the constant root inode number
      else
      if (is_32bit_api() && kstat->ino > UINT_MAX)
      {  // 32-bit applications cannot handle 64bit inodes and would fail with EOVERFLOW
         kstat->ino = kstat->ino >> 32;
      }
   }

   // clean-up

   return retVal;
}

/**
 * @return 0 on success, negative linux error code otherwise
 */
int FhgfsOps_setattr(struct dentry* dentry, struct iattr* iattr)
{
   // note: called for chmod(), chown(), utime(), truncate()
   // note: changed fields can be determined by the iattr->ia_valid flags (ATTR_...)

   App* app = FhgfsOps_getApp(dentry->d_sb);
   Config* cfg = App_getConfig(app);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_setattr";

   int retVal = 0;
   int inodeChangeRes;
   SettableFileAttribs fhgfsAttr;
   int validFhgfsAttribs;
   FhgfsOpsErr setAttrRes = FhgfsOpsErr_SUCCESS;
   struct inode* inode = dentry->d_inode;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   if(unlikely(Logger_getLogLevel(log) >= 5) )
   {
      FhgfsOpsHelper_logOpMsg(5, app, dentry, logContext,
         (iattr->ia_valid & ATTR_SIZE) ? "(with trunc)" : "");
   }

   inodeChangeRes = inode_change_ok(inode, iattr);
   if(inodeChangeRes < 0)
      return inodeChangeRes;

   #ifdef KERNEL_HAS_ATTR_OPEN
      /* we do trunc during open message on meta server, so we don't want this redundant trunc
         (and ctime/mtime update) from the kernel */
      if(iattr->ia_valid & ATTR_OPEN)
         return 0; //
   #endif

   if(iattr->ia_valid & ATTR_SIZE)
   { // make sure we only update size of regular files
      if(!S_ISREG(inode->i_mode) )
         iattr->ia_valid &= ~ATTR_SIZE;
   }

   if(S_ISREG(inode->i_mode) )
   { // flush all dirty file data
      filemap_write_and_wait(inode->i_mapping);

      if(Config_getTuneFileCacheTypeNum(cfg) == FILECACHETYPE_Buffered)
         FhgfsOpsHelper_flushCache(app, fhgfsInode, fhgfs_false);
   }

   OsTypeConv_iattrOsToFhgfs(iattr, &fhgfsAttr, &validFhgfsAttribs);

   if(validFhgfsAttribs)
   {
      FhgfsInode_entryInfoReadLock(fhgfsInode); // LOCK EntryInfo

      setAttrRes = FhgfsOpsRemoting_setAttr(app, FhgfsInode_getEntryInfo(fhgfsInode), &fhgfsAttr,
         validFhgfsAttribs);

      FhgfsInode_entryInfoReadUnlock(fhgfsInode); // UNLOCK EntryInfo
   }

   /* todo: truncation should be integrated into Remoting_setAttr to save the time to send an
      extra msg */

   if(setAttrRes == FhgfsOpsErr_SUCCESS)
   {
      if (validFhgfsAttribs)
         FhgfsOps_newAttrToInode(iattr, inode);

      // all right so far => handle truncation
      if(iattr->ia_valid & ATTR_SIZE)
      {
         FhgfsInode_entryInfoReadLock(fhgfsInode); // LOCK EntryInfo

         setAttrRes = FhgfsOpsRemoting_truncfile(app, FhgfsInode_getEntryInfo(fhgfsInode), iattr->ia_size);

         FhgfsInode_entryInfoReadUnlock(fhgfsInode); // UNLOCK EntryInfo

         if( (setAttrRes == FhgfsOpsErr_SUCCESS) &&
             (Config_getTuneFileCacheTypeNum(cfg) == FILECACHETYPE_Paged) )
         {
            FhgfsOps_vmtruncate(inode, iattr->ia_size);
         }
      }
   }

   if(setAttrRes != FhgfsOpsErr_SUCCESS)
      retVal = FhgfsOpsErr_toSysErr(setAttrRes);

   return retVal;
}

/**
 * Set inode attributes after a successful setattr
 */
void FhgfsOps_newAttrToInode(struct iattr* iAttr, struct inode* outInode)
{
   spin_lock(&outInode->i_lock);

   if(iAttr->ia_valid & ATTR_MODE)
      outInode->i_mode = iAttr->ia_mode;

   if(iAttr->ia_valid & ATTR_UID)
      outInode->i_uid   = iAttr->ia_uid;

   if(iAttr->ia_valid & ATTR_GID)
      outInode->i_gid    = iAttr->ia_gid;

   if(iAttr->ia_valid & ATTR_MTIME_SET)
   {
      outInode->i_mtime  = iAttr->ia_mtime;
   }
   else
   if(iAttr->ia_valid & ATTR_MTIME)
   { // set mtime to "now"
      TimeAbs now;
      TimeAbs_init(&now);

      outInode->i_mtime.tv_sec = TimeAbs_getTimeval(&now)->tv_sec;

      TimeAbs_uninit(&now);
   }

   if(iAttr->ia_valid & ATTR_ATIME_SET)
   {
      outInode->i_atime             = iAttr->ia_atime;
   }
   else
   if(iAttr->ia_valid & ATTR_ATIME)
   { // set atime to "now"
      TimeAbs now;
      TimeAbs_init(&now);

      outInode->i_atime.tv_sec = TimeAbs_getTimeval(&now)->tv_sec;

      TimeAbs_uninit(&now);
   }

   if(iAttr->ia_valid & ATTR_CTIME)
   {
      TimeAbs now;
      TimeAbs_init(&now);

      outInode->i_ctime.tv_sec = TimeAbs_getTimeval(&now)->tv_sec;

      TimeAbs_uninit(&now);
   }

   spin_unlock(&outInode->i_lock);
}



/**
 * Create directory.
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,3,0)
int FhgfsOps_mkdir(struct inode* dir, struct dentry* dentry, int mode)
#else
int FhgfsOps_mkdir(struct inode* dir, struct dentry* dentry, umode_t mode)
#endif // LINUX_VERSION_CODE
{
   struct super_block* sb = dentry->d_sb;
   App* app = FhgfsOps_getApp(sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_mkdir";

   int retVal = 0;
   FhgfsOpsErr mkRes;
   FhgfsInode* fhgfsParentInode = FHGFS_INODE(dir);
   EntryInfo newEntryInfo;
   const char* entryName = dentry->d_name.name;
   struct CreateInfo createInfo;

   if(unlikely(Logger_getLogLevel(log) >= 5) )
      FhgfsOpsHelper_logOp(5, app, dentry, logContext);

   mode |= S_IFDIR; // just make sure this is a dir

   CreateInfo_init(app, dir, entryName, mode, &createInfo);

   FhgfsInode_entryInfoReadLock(fhgfsParentInode); // LOCK EntryInfo

   mkRes = FhgfsOpsRemoting_mkdir(app, FhgfsInode_getEntryInfo(fhgfsParentInode), &createInfo,
      &newEntryInfo);

   FhgfsInode_entryInfoReadUnlock(fhgfsParentInode); // UNLOCK EntryInfo

   if(mkRes != FhgfsOpsErr_SUCCESS)
   {
      retVal = FhgfsOpsErr_toSysErr(mkRes);
   }
   else
   { // remote success => create the local inode
      retVal = __FhgfsOps_instantiateInode(dentry, &newEntryInfo, NULL);

      dir->i_ctime = dir->i_mtime = current_fs_time(sb);
   }

   return retVal;
}

int FhgfsOps_rmdir(struct inode* dir, struct dentry* dentry)
{
   struct super_block* sb = dentry->d_sb;
   App* app = FhgfsOps_getApp(sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_rmdir";

   int retVal = 0;
   FhgfsOpsErr rmRes;
   FhgfsInode* fhgfsParentInode = FHGFS_INODE(dir);
   const char* entryName = dentry->d_name.name;

   if(unlikely(Logger_getLogLevel(log) >= 5) )
      FhgfsOpsHelper_logOp(5, app, dentry, logContext);

   FhgfsInode_entryInfoReadLock(fhgfsParentInode); // LOCK EntryInfo

   rmRes = FhgfsOpsRemoting_rmdir(app, FhgfsInode_getEntryInfo(fhgfsParentInode), entryName);

   FhgfsInode_entryInfoReadUnlock(fhgfsParentInode); // UNLOCK EntryInfo

   if(rmRes != FhgfsOpsErr_SUCCESS)
   {
      retVal = FhgfsOpsErr_toSysErr(rmRes);
   }
   else
   { // remote success
      clear_nlink(dentry->d_inode);

      dir->i_ctime = dir->i_mtime = current_fs_time(sb);
   }


   return retVal;
}

/**
 * Create file based on combined intent message.
 *
 * @param isExclusiveCreate true if this is an exclusive (O_EXCL) create
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,3,0)
int FhgfsOps_createIntent(struct inode* dir, struct dentry* dentry, int mode,
   struct nameidata* nameidata)
#elif LINUX_VERSION_CODE < KERNEL_VERSION(3,6,0)
int FhgfsOps_createIntent(struct inode* dir, struct dentry* dentry, umode_t mode,
   struct nameidata* nameidata)
#else
int FhgfsOps_createIntent(struct inode* dir, struct dentry* dentry, umode_t mode,
   bool isExclusiveCreate)
#endif // LINUX_VERSION_CODE
{
   struct super_block* sb = dentry->d_sb;
   App* app = FhgfsOps_getApp(sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_createIntent";

   int retVal = 0;
   FhgfsOpsErr mkRes;
   FhgfsInode* fhgfsParentInode = FHGFS_INODE(dir);
   EntryInfo newEntryInfo;
   const char* entryName = dentry->d_name.name;
   struct CreateInfo createInfo;
   LookupCreateStatInfoIn inInfo; // input data for combo-request
   fhgfs_stat statData;
   LookupCreateStatInfoOut outInfo; // result data of combo-request

   #if LINUX_VERSION_CODE < KERNEL_VERSION(3,6,0)
      fhgfs_bool isExclusiveCreate = (nameidata && (nameidata->intent.open.flags & O_EXCL) );
   #endif // LINUX_VERSION_CODE


   if(Logger_getLogLevel(log) >= 4)
      FhgfsOpsHelper_logOp(4, app, dentry, logContext);

   FhgfsInode_entryInfoReadLock(fhgfsParentInode); // LOCK EntryInfo

   mode |= S_IFREG; // just make sure we create a file here

   CreateInfo_init(app, dir, entryName, mode, &createInfo);
   LookupCreateStatInfoIn_init(&inInfo, FhgfsInode_getEntryInfo(fhgfsParentInode), &createInfo,
      isExclusiveCreate);
   LookupCreateStatInfoOut_prepare(&outInfo, &newEntryInfo, &statData);

   mkRes = FhgfsOpsRemoting_lookupCreateStatFile(app, &inInfo, &outInfo);

   FhgfsInode_entryInfoReadUnlock(fhgfsParentInode); // UNLOCK EntryInfo

   if(mkRes != FhgfsOpsErr_SUCCESS)
   {
      d_drop(dentry); // avoid leaving a negative dentry behind
      retVal = FhgfsOpsErr_toSysErr(mkRes);
      goto outErr;
   }

   if(unlikely(outInfo.statRes != FhgfsOpsErr_SUCCESS) )
   { // something went wrong before the server could stat the file
      d_drop(dentry); // avoid leaving a negative dentry behind
      retVal = FhgfsOpsErr_toSysErr(outInfo.statRes);

      // with mkRes == FhgfsOpsErr_SUCCESS we have initialized the entryInfo, but now don't need it
      EntryInfo_uninit(outInfo.entryInfo);

      goto outErr;
   }

   // remote success => create local inode
   retVal = __FhgfsOps_instantiateInode(dentry, outInfo.entryInfo, outInfo.fhgfsStat);

   dir->i_ctime = dir->i_mtime = current_fs_time(sb);

   // clean-up
outErr:

   return retVal;
}

int FhgfsOps_unlink(struct inode* dir, struct dentry* dentry)
{
   struct super_block* sb = dentry->d_sb;
   App* app = FhgfsOps_getApp(sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_unlink";

   int retVal = 0;
   FhgfsOpsErr unlinkRes;
   FhgfsInode* fhgfsParentInode = FHGFS_INODE(dir);
   const char* entryName = dentry->d_name.name;

   if(Logger_getLogLevel(log) >= 4)
      FhgfsOpsHelper_logOp(4, app, dentry, logContext);

   FhgfsInode_entryInfoReadLock(fhgfsParentInode); // LOCK EntryInfo

   unlinkRes = FhgfsOpsRemoting_unlinkfile(app, FhgfsInode_getEntryInfo(fhgfsParentInode),
      entryName);

   FhgfsInode_entryInfoReadUnlock(fhgfsParentInode); // UNLOCK EntryInfo

   if(unlinkRes != FhgfsOpsErr_SUCCESS)
   {
      retVal = FhgfsOpsErr_toSysErr(unlinkRes);

      if(retVal == -ENOENT)
         d_drop(dentry);
   }
   else
   { // remote success
      dir->i_ctime = dir->i_mtime = current_fs_time(sb);

      if(dentry->d_inode)
         drop_nlink(dentry->d_inode);
   }

   return retVal;
}

/**
 * Create special file.
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,3,0)
int FhgfsOps_mknod(struct inode* dir, struct dentry* dentry, int mode, dev_t dev)
#else
int FhgfsOps_mknod(struct inode* dir, struct dentry* dentry, umode_t mode, dev_t dev)
#endif // LINUX_VERSION_CODE
{
   struct super_block* sb = dentry->d_sb;
   App* app = FhgfsOps_getApp(sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_mknod";

   int retVal = 0;
   FhgfsOpsErr mkRes;
   FhgfsInode* fhgfsParentInode = FHGFS_INODE(dir);
   EntryInfo newEntryInfo;
   const char* entryName = dentry->d_name.name;
   struct CreateInfo createInfo;

   if(Logger_getLogLevel(log) >= 4)
      FhgfsOpsHelper_logOp(4, app, dentry, logContext);

   CreateInfo_init(app, dir, entryName, mode, &createInfo);

   FhgfsInode_entryInfoReadLock(fhgfsParentInode); // LOCK EntryInfo

   mkRes = FhgfsOpsRemoting_mkfile(app, FhgfsInode_getEntryInfo(fhgfsParentInode),
      &createInfo, &newEntryInfo);

   FhgfsInode_entryInfoReadUnlock(fhgfsParentInode); // UNLOCK EntryInfo

   if(mkRes != FhgfsOpsErr_SUCCESS)
   {
      d_drop(dentry); // avoid leaving a negative dentry behind
      retVal = FhgfsOpsErr_toSysErr(mkRes);
      goto outErr;
   }

   retVal = __FhgfsOps_instantiateInode(dentry, &newEntryInfo, NULL);

   dir->i_ctime = dir->i_mtime = current_fs_time(sb);


outErr:
   // clean-up

   return retVal;
}


/**
 * @param dir directory for the new link
 * @param dentry dentry of the new entry that we want to create
 * @param to where the symlink points to
 */
int FhgfsOps_symlink(struct inode* dir, struct dentry* dentry, const char* to)
{
   struct super_block* sb = dentry->d_sb;
   App* app = FhgfsOps_getApp(sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_symlink";

   int retVal;
   int mode = S_IFLNK | S_IRWXU | S_IRWXG | S_IRWXO;
   FhgfsInode* fhgfsParentInode = FHGFS_INODE(dir);
   EntryInfo newEntryInfo;
   const char* entryName = dentry->d_name.name;
   struct CreateInfo createInfo;

   if(unlikely(Logger_getLogLevel(log) >= 4) )
      FhgfsOpsHelper_logOp(4, app, dentry, logContext);


   CreateInfo_init(app, dir, entryName, mode, &createInfo);

   FhgfsInode_entryInfoReadLock(fhgfsParentInode); // LOCK EntryInfo

   retVal = FhgfsOpsHelper_symlink(app, FhgfsInode_getEntryInfo(fhgfsParentInode), to, &createInfo,
      &newEntryInfo);

   FhgfsInode_entryInfoReadUnlock(fhgfsParentInode); // UNLOCK EntryInfo

   if(retVal)
   { // error occurred
      d_drop(dentry); // avoid leaving a negative dentry behind
   }
   else
   { // remote success => create local inode
      retVal = __FhgfsOps_instantiateInode(dentry, &newEntryInfo, NULL);

      dir->i_ctime = dir->i_mtime = current_fs_time(sb);
   }


   return retVal;
}

/**
 * @param fromFileDentry the already existing dentry
 * @param toDirInode parent (directory) inode of the new dentry
 * @param toFileDentry the new dentry (link) that we want to create
 */
int FhgfsOps_link(struct dentry* fromFileDentry, struct inode* toDirInode,
   struct dentry* toFileDentry)
{
   // note: we don't support hardlinks yet, but have a config option to create symlinks instead
      // of hardlinks.

   struct super_block* sb = toFileDentry->d_sb;
   App* app = FhgfsOps_getApp(sb);
   Logger* log = App_getLogger(app);
   Config* cfg = App_getConfig(app);
   const char* logContext = "FhgfsOps_link";

   struct inode* fileInode = fromFileDentry->d_inode;
   FhgfsInode* fhgfsFileInode = FHGFS_INODE(fileInode);

   FhgfsInode* fhgfsToDirInode    = FHGFS_INODE(toDirInode);

   struct inode* fromDirInode    = fromFileDentry->d_parent->d_inode;
   FhgfsInode* fhgfsFromDirInode  = FHGFS_INODE(fromDirInode);

   const char* fromFileName = fromFileDentry->d_name.name;
   unsigned    fromFileLen  = fromFileDentry->d_name.len;

   const char* toFileName   = toFileDentry->d_name.name;
   unsigned    toFileLen    = toFileDentry->d_name.len;

   EntryInfo* fromDirInfo;
   EntryInfo* toDirInfo;
   EntryInfo* fromFileInfo;

   int retVal;


   if(unlikely(Logger_getLogLevel(log) >= Log_DEBUG) )
   {
      // TODO: resolve complete path instead of name only
      Logger_logFormatted(log, Log_DEBUG, logContext, "called. From: %s; To: %s",
         fromFileName, toFileName);
   }

   FhgfsInode_entryInfoReadLock(fhgfsFromDirInode); // LOCK EntryInfo
   FhgfsInode_entryInfoReadLock(fhgfsToDirInode);   // LOCK EntryInfo
   FhgfsInode_entryInfoReadLock(fhgfsFileInode);    // LOCK EntryInfo

   fromDirInfo   = FhgfsInode_getEntryInfo(fhgfsFromDirInode);
   toDirInfo     = FhgfsInode_getEntryInfo(fhgfsToDirInode);
   fromFileInfo  = FhgfsInode_getEntryInfo(fhgfsFileInode);


#if 0 // TODO: Re-add here once we support inter directory hard-links
   ihold(fileInode);
   inc_nlink(fileInode);
#endif

   if (strcmp(fromDirInfo->entryID, toDirInfo->entryID) == 0)
   {  // same inode, so hard link within the same directory
      int linkRes;

      // TODO: Remove it here once we support inter directory hard-links
      ihold(fileInode);
      inc_nlink(fileInode);

      linkRes = FhgfsOpsRemoting_hardlink(app, fromFileName, fromFileLen, fromFileInfo,
         fromDirInfo, toFileName, toFileLen, toDirInfo);

      retVal = FhgfsOpsErr_toSysErr(linkRes);

      if (retVal) // TODO: Remove it here once we support inter directory hard-links
      {
         d_drop(toFileDentry);
         iput(fileInode);
         drop_nlink(fileInode);
      }
      else
      {
         d_instantiate(toFileDentry, fileInode);
      }
   }
   else
   {
      if(!Config_getSysCreateHardlinksAsSymlinks(cfg) )
      { // creation of hardlinks not allowed/supported
         FhgfsOpsHelper_logOpMsg(2, app, toFileDentry,logContext, "Hardlinks not supported yet.");
         retVal = -EPERM; // do not allow creation of hardlink
      }
      else
         retVal = FhgfsOps_hardlinkAsSymlink(fromFileDentry, toDirInode, toFileDentry);
   }

   FhgfsInode_entryInfoReadUnlock(fhgfsFileInode);    // LOCK fromParentEntryInfo
   FhgfsInode_entryInfoReadUnlock(fhgfsToDirInode);   // LOCK toParentEntryInfo
   FhgfsInode_entryInfoReadUnlock(fhgfsFromDirInode); // LOCK EntryInfo

#if 0 // TODO: Re-add here once we support inter directory hard-links
   if (retVal)
      d_drop(toFileDentry);
      iput(fileInode);
      drop_nlink(fileInode);

   else
   {
      d_instantiate(toFileDentry, fileInode);
   }
#endif

   return retVal;
}

/**
 * create symlink instead of hardlink
 *
 * @param oldDentry the already existing link
 * @param dir directory of the new link
 * @param newDentry the new link that we want to create
 *
 * Note:
 */
int FhgfsOps_hardlinkAsSymlink(struct dentry* oldDentry, struct inode* dir,
   struct dentry* newDentry)
{
   struct super_block* sb = newDentry->d_sb;
   App* app = FhgfsOps_getApp(sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_link";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);

   int retVal;
   int mode = S_IFLNK | S_IRWXU | S_IRWXG | S_IRWXO;
   FhgfsInode* fhgfsParentInode = FHGFS_INODE(dir);
   EntryInfo newEntryInfo;
   const char* entryName = newDentry->d_name.name;

   char *oldPathStr, *newPathStr;
   char *oldPathStoreBuf, *newPathStoreBuf;
   char *oldPathTmp, *newPathTmp;
   char* oldRelativePathStr;

   struct CreateInfo createInfo;

   // resolve oldDentry to full path
   oldPathTmp = __FhgfsOps_pathResolveToStoreBuf(bufStore, oldDentry, &oldPathStoreBuf);
   if (unlikely(IS_ERR(oldPathTmp) ) )
   {
      int error = IS_ERR(oldPathTmp);
      Logger_logFormatted(log, 2, logContext, "Error in link(): %d", error);
      retVal = error;
      goto errorOldPath;
   }

   oldPathStr = StringTk_strDup(oldPathTmp);
   NoAllocBufferStore_addBuf(bufStore, oldPathStoreBuf);

   // resolve newDentry to full path
   newPathTmp = __FhgfsOps_pathResolveToStoreBuf(bufStore, newDentry, &newPathStoreBuf);
   if (unlikely(IS_ERR(newPathTmp) ) )
   {
      int error = IS_ERR(oldPathTmp);
      Logger_logFormatted(log, 2, logContext, "Error in link(): %d", error);
      retVal = error;
      goto errorNewPath;
   }

   newPathStr = StringTk_strDup(newPathTmp);
   NoAllocBufferStore_addBuf(bufStore, newPathStoreBuf);

   // resolve relative path from new to old
   FhgfsOpsHelper_getRelativeLinkStr(newPathStr, oldPathStr, &oldRelativePathStr);

   if(unlikely(Logger_getLogLevel(log) >= 4) )
   {
      Logger_logFormatted(log, 4, logContext, "called (as symlink). Path: %s; To: %s",
         newPathStr, oldRelativePathStr);
   }

   CreateInfo_init(app, dir, entryName, mode, &createInfo);

   FhgfsInode_entryInfoReadLock(fhgfsParentInode); // LOCK EntryInfo

   retVal = FhgfsOpsHelper_symlink(app, FhgfsInode_getEntryInfo(fhgfsParentInode),
      oldRelativePathStr, &createInfo, &newEntryInfo);

   FhgfsInode_entryInfoReadUnlock(fhgfsParentInode); // UNLOCK EntryInfo

   if(retVal)
   { // error occurred
      d_drop(newDentry); // avoid leaving a negative dentry behind
   }
   else
   { // remote success => create local inode
      retVal = __FhgfsOps_instantiateInode(newDentry, &newEntryInfo, NULL);

      dir->i_ctime = dir->i_mtime = current_fs_time(sb);
   }

   // clean-up

   kfree(oldRelativePathStr);
   kfree(newPathStr);
errorNewPath:
   kfree(oldPathStr);
errorOldPath:
   return retVal;
}


void* FhgfsOps_follow_link(struct dentry* dentry, struct nameidata* nd)
{
   void* retVal = NULL;

   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_follow_link";

   int readRes;
   char* bufPage = (char*)__os_get_free_page();
   char* destination = bufPage;

   struct inode* inode = dentry->d_inode;
   FhgfsInode* fhgfsParentInode = FHGFS_INODE(inode);

   if(unlikely(Logger_getLogLevel(log) >= 5) )
      FhgfsOpsHelper_logOp(5, app, dentry, logContext);

   FhgfsInode_entryInfoReadLock(fhgfsParentInode); // LOCK EntryInfo

   readRes = FhgfsOpsHelper_readlink(app, FhgfsInode_getEntryInfo(fhgfsParentInode), bufPage,
      PAGE_SIZE-1);

   FhgfsInode_entryInfoReadUnlock(fhgfsParentInode); // UNLOCK EntryInfo

   if(readRes < 0)
   {
      destination = ERR_PTR(readRes);
   }
   else
      bufPage[readRes] = 0;

   if(readRes == (PAGE_SIZE-1) )
   { // link destination too long
      destination = ERR_PTR(-ENAMETOOLONG);
   }

   // store link destination
   nd_set_link(nd, destination);

   // clean-up

   if(IS_ERR(destination) )
      os_free_page( (unsigned long)bufPage);

   // Note: free_page() is called by the put_link method in the success case

   return retVal;
}


void FhgfsOps_put_link(struct dentry *dentry, struct nameidata* nd, void* p)
{
   char* destination = nd_get_link(nd);

   if(!IS_ERR(destination) )
      os_free_page( (unsigned long)destination);
}


int FhgfsOps_rename(struct inode* inodeDirFrom, struct dentry* dentryFrom,
   struct inode* inodeDirTo, struct dentry* dentryTo)
{
   struct super_block* sb = dentryFrom->d_sb;
   App* app = FhgfsOps_getApp(sb);
   Logger* log = App_getLogger(app);
   const char* logContext = __func__;
   FhgfsOpsErr renameRes;

   struct inode* fromEntryInode = dentryFrom->d_inode; // no need to lock it, see other in kernel fs
   FhgfsInode* fhgfsFromEntryInode = FHGFS_INODE(fromEntryInode);
   DirEntryType entryType = FhgfsInode_getDirEntryType(fhgfsFromEntryInode);

   EntryInfo* fromDirInfo; // EntryInfo about the 'from' directory
   EntryInfo* toDirInfo;   // EntryInfo about the 'to' directory

   int retVal = 0;

   FhgfsInode* fhgfsFromDirInode = FHGFS_INODE(inodeDirFrom);
   FhgfsInode* fhgfsToDirInode   = FHGFS_INODE(inodeDirTo);

   const char* oldName = dentryFrom->d_name.name;
   unsigned oldLen = dentryFrom->d_name.len;

   const char* newName = dentryTo->d_name.name;
   unsigned newLen = dentryTo->d_name.len;

   if(unlikely(Logger_getLogLevel(log) >= Log_DEBUG) )
      Logger_logFormatted(log, Log_DEBUG, logContext, "called. "
         "From: %s; To: %s", oldName, newName); // TODO: resolve complete path instead of name only

   FhgfsInode_entryInfoReadLock(fhgfsFromDirInode); // LOCK EntryInfo
   FhgfsInode_entryInfoReadLock(fhgfsToDirInode);   // LOCK EntryInfo

   // note the fileInode also needs to be locked to prevent reference/release during a rename
   FhgfsInode_entryInfoWriteLock(fhgfsFromEntryInode);   // LOCK EntryInfo (renamed file dir)

   fromDirInfo = FhgfsInode_getEntryInfo(fhgfsFromDirInode);
   toDirInfo   = FhgfsInode_getEntryInfo(fhgfsToDirInode);

   renameRes = FhgfsOpsRemoting_rename(app, oldName, oldLen, entryType, fromDirInfo,
      newName, newLen, toDirInfo);
   if(renameRes != FhgfsOpsErr_SUCCESS)
   {
      int logLevel = Log_NOTICE;
      EntryInfo* fromEntryInfo = FhgfsInode_getEntryInfo(fhgfsFromEntryInode);

      if( (renameRes == FhgfsOpsErr_PATHNOTEXISTS) || (renameRes == FhgfsOpsErr_INUSE) )
         logLevel = Log_DEBUG; // don't bother user with non-error messages

      /* We cannot rename onto an existing empty directory, yet. So Log_DEBUG only.
       * TODO: Once we support these renames we further need to restrict to empty
       * (link count = 2) dirs */
      if (renameRes == FhgfsOpsErr_EXISTS &&
          dentryTo->d_inode && S_ISDIR(dentryTo->d_inode->i_mode) )
         logLevel = Log_DEBUG;

      Logger_logFormatted(log, logLevel, logContext,
         "Rename failed: %s fromDirID: %s oldName: %s toDirID: %s newName: %s EntryID: %s",
         FhgfsOpsErr_toErrString(renameRes),
         fromDirInfo->entryID, oldName, toDirInfo->entryID, newName, fromEntryInfo->entryID);

      retVal = FhgfsOpsErr_toSysErr(renameRes);
   }
   else
   { // remote success
      fromEntryInode->i_ctime = inodeDirFrom->i_ctime = inodeDirFrom->i_mtime =
         inodeDirTo->i_ctime = inodeDirTo->i_mtime = current_fs_time(sb);

      /* TODO: Future releases might keep the inode on the original server, so the ownerNodeID
       *       is not going to be updated then and also must not be wrongly set in our inode. */
      FhgfsInode_UpdateEntryInfoOnRenameUnlocked(fhgfsFromEntryInode, toDirInfo, newName);
   }

   FhgfsInode_entryInfoWriteUnlock(fhgfsFromEntryInode);   // UNLOCK EntryInfo (renamed file dir)
   FhgfsInode_entryInfoReadUnlock(fhgfsToDirInode);   // UNLOCK ToDirEntryInfo
   FhgfsInode_entryInfoReadUnlock(fhgfsFromDirInode); // UNLOCK FromDirEntryInfo

   LOG_DEBUG_FORMATTED(log, Log_DEBUG, logContext, "remoting complete. result: %d", (int)renameRes);


   // clean-up

   return retVal;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,8,0)

/**
 * Note: i_size of the inode is set to the desired size before this is called
 *
 * Note: since linux-3.8, truncation is done via setattr
 */
void FhgfsOps_truncate(struct inode* inode)
{
   App* app = FhgfsOps_getApp(inode->i_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_truncate";

   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   loff_t size = i_size_read(inode);

   if(unlikely(Logger_getLogLevel(log) >= 5) )
   {
      struct dentry* dentry = d_find_alias(inode); // calls dget_locked (can return NULL)

      FhgfsOpsHelper_logOp(5, app, dentry, logContext);

      if(dentry)
         dput(dentry);
   }

   FhgfsInode_entryInfoReadLock(fhgfsInode); // LOCK EntryInfo

   // (see below why return val is not checked)
   FhgfsOpsRemoting_truncfile(app, FhgfsInode_getEntryInfo(fhgfsInode), size);

   FhgfsInode_entryInfoReadUnlock(fhgfsInode); // UNLOCK EntryInfo


   return; // note: this is an internal function which gives us no way to return a result.
}

#endif // LINUX_VERSION_CODE

/**
 * Note: This is almost a copy of general vmtruncate(), just with inode->i_lock around the i_size
 * updates.
 * Note: This is not called directly, but via setattr
 *
 * @param offset file offset for truncation
 */
int FhgfsOps_vmtruncate(struct inode* inode, loff_t offset)
{
   if(i_size_read(inode) < offset)
   {
      unsigned long limit;

      limit = current->signal->rlim[RLIMIT_FSIZE].rlim_cur;
      if (limit != RLIM_INFINITY && offset > (loff_t)limit)
         goto out_sig;

      if (offset > (loff_t)(inode->i_sb->s_maxbytes) )
         goto out_big;

      spin_lock(&inode->i_lock);
      i_size_write(inode, offset);
      spin_unlock(&inode->i_lock);
   }
   else
   {
      struct address_space *mapping = inode->i_mapping;

      /*
       * truncation of in-use swapfiles is disallowed - it would
       * cause subsequent swapout to scribble on the now-freed
       * blocks.
       */
      if(IS_SWAPFILE(inode) )
         return -ETXTBSY;

      spin_lock(&inode->i_lock);
      i_size_write(inode, offset);
      spin_unlock(&inode->i_lock);

      /*
       * unmap_mapping_range is called twice, first simply for
       * efficiency so that truncate_inode_pages does fewer
       * single-page unmaps.  However after this first call, and
       * before truncate_inode_pages finishes, it is possible for
       * private pages to be COWed, which remain after
       * truncate_inode_pages finishes, hence the second
       * unmap_mapping_range call must be made for correctness.
       */
      unmap_mapping_range(mapping, offset + PAGE_SIZE - 1, 0, 1);
      truncate_inode_pages(mapping, offset);
      unmap_mapping_range(mapping, offset + PAGE_SIZE - 1, 0, 1);
   }

   return 0;

out_sig:
   send_sig(SIGXFSZ, current, 0);
out_big:
   return -EFBIG;
}

/**
 * Note: Call this once during module init (and remember to call _destroyInodeCache() )
 */
fhgfs_bool FhgfsOps_initInodeCache(void)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17)
   unsigned long cacheFlags = SLAB_RECLAIM_ACCOUNT;
#else
   unsigned long cacheFlags = SLAB_RECLAIM_ACCOUNT | SLAB_MEM_SPREAD;
#endif // LINUX_VERSION_CODE

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
   FhgfsInodeCache = kmem_cache_create("fhgfs_inode_cache", sizeof(FhgfsInode), 0,
      cacheFlags, FhgfsOps_initInodeOnce, NULL);
#else
   FhgfsInodeCache = kmem_cache_create("fhgfs_inode_cache", sizeof(FhgfsInode), 0,
      cacheFlags, FhgfsOps_initInodeOnce);
#endif // LINUX_VERSION_CODE


   if(FhgfsInodeCache == NULL)
      return fhgfs_false;

   return fhgfs_true;
}

void FhgfsOps_destroyInodeCache(void)
{
   if(FhgfsInodeCache)
      kmem_cache_destroy(FhgfsInodeCache);
}


struct inode* FhgfsOps_alloc_inode(struct super_block *sb)
{
   App* app = FhgfsOps_getApp(sb);
   Config* cfg = App_getConfig(app);

   FhgfsInode* fhgfsInode;

   fhgfsInode = kmem_cache_alloc(FhgfsInodeCache, GFP_KERNEL);
   if(unlikely(!fhgfsInode) )
      return NULL;

   FhgfsInode_allocInit(fhgfsInode);

   fhgfsInode->vfs_inode.i_blkbits = Config_getTuneInodeBlockBits(cfg);

   return (struct inode*)fhgfsInode;
}

void FhgfsOps_destroy_inode(struct inode* inode)
{
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   FhgfsInode_destroyUninit(fhgfsInode);

   kmem_cache_free(FhgfsInodeCache, inode);
}

/**
 * Creates a new inode, inits it from the kstat, inits the ops (depending on the mode)
 * and hashes it.
 *
 * Note: Make sure everything is set in the kstat _before_ you call this, because we hash
 * the inode in here (so it can be found and accessed by others when this method returns).
 * Note: Consider using the _instantiateInode()-wrapper instead of calling this directly for new
 * files/dirs.
 *
 * @param kstat must have a valid .ino (inode number)
 * @param dev set to 0 if not required (only used for special files)
 * @param entryInfo contained strings will just be moved to the new inode or free'd in case of an
 * error (or cached inode), so don't access the given entryInfo anymore after calling this.
 * @return NULL if not successful
 */
struct inode* __FhgfsOps_newInode(struct super_block* sb, struct kstat* kstat, dev_t dev,
   EntryInfo* entryInfo)
{
   App* app = FhgfsOps_getApp(sb);
   Config* cfg = App_getConfig(app);

   FhgfsInode* fhgfsInode;

   FhgfsInodeComparisonInfo comparisonInfo =
   {
      .inodeHash = kstat->ino, // pre-set by caller
      .entryID = entryInfo->entryID,
   };


   // check inode cache for an existing inode with this ID (and get it) or allocate a new one

   struct inode* inode = iget5_locked(sb, kstat->ino,
      __FhgfsOps_compareInodeID, __FhgfsOps_initNewInodeDummy, &comparisonInfo);

   if(unlikely(!inode || IS_ERR(inode) ) )
      goto cleanup_entryInfo; // allocation of new inode failed

   fhgfsInode = FHGFS_INODE(inode);

   if( !(inode->i_state & I_NEW) )
   {  // Found an existing inode, which is possibly actively used. We still need to update it.
      FhgfsInode_entryInfoWriteLock(fhgfsInode); // LOCK EntryInfo
      FhgfsInode_UpdateEntryInfoUnlocked(fhgfsInode, entryInfo);
      FhgfsInode_entryInfoWriteUnlock(fhgfsInode); // UNLOCK EntryInfo

      spin_lock(&inode->i_lock);
      __FhgfsOps_applyStatDataToInodeUnlocked(kstat, inode); // already locked
      Time_setToNow(&fhgfsInode->dataCacheTime);
      spin_unlock(&inode->i_lock);

      goto outNoCleanUp; // we found a matching existing inode => no init needed
   }

   /* note: new inodes are protected by the I_NEW flag from access by other threads until we
    *       call unlock_new_inode(). */

   // init this fresh new inode...

   __FhgfsOps_applyStatDataToInodeUnlocked(kstat, inode); // no one can access inode yet => unlocked

   inode->i_ino = kstat->ino; // pre-set by caller

   inode->i_flags |= S_NOATIME | S_NOCMTIME; // timestamps updated by server

   // move values (no actual string copy)
   EntryInfo_copy(entryInfo, FhgfsInode_getEntryInfo(fhgfsInode) );

   switch (kstat->mode & S_IFMT)
   {
      case S_IFREG: // regular file
      {
         if(Config_getTuneFileCacheTypeNum(cfg) == FILECACHETYPE_Paged)
         { // with pagecache
            inode->i_op = &fhgfs_file_pagecache_inode_ops;
            inode->i_fop = &fhgfs_file_pagecache_ops;
            inode->i_data.a_ops = &fhgfs_address_pagecache_ops;
         }
         else
         { // no pagecache (=> either none or simple cache)
            inode->i_op = &fhgfs_file_inode_ops;
            inode->i_fop = &fhgfs_file_ops;
            inode->i_data.a_ops = &fhgfs_address_ops;
         }

         #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,32)
            inode->i_data.backing_dev_info = FhgfsOps_getBdi(sb);
         #endif

      } break;

      case S_IFDIR: // directory
      {
         inode->i_op = &fhgfs_dir_inode_ops;
         inode->i_fop = &fhgfs_dir_ops;
      } break;

      case S_IFLNK: // symlink
      {
         //inode->i_op = &page_symlink_inode_operations;
         inode->i_op = &fhgfs_symlink_inode_ops;
      } break;

      default: // pipes and other special files
      {
         inode->i_op = &fhgfs_special_inode_ops;
         init_special_inode(inode, kstat->mode, dev);
      } break;
   }


   unlock_new_inode(inode); // remove I_NEW flag, so the inode can be accessed by others

   return inode;


   // error occured
cleanup_entryInfo:
   EntryInfo_uninit(entryInfo);

   // found an existing inode
outNoCleanUp:
   return inode;
}

/**
 * Retrieves the attribs from the metadata node, generates an inode number (ID) and instantiates
 * a local version of the inode for the dentry.
 *
 * Note: This is a wrapper for _newInode() that also retrieves the the entry attributes (so the
 * entry must exist already on the server).
 *
 * @param dentry the new entry
 * @param entryInfo contained values (not the struct itself!) will be owned by the inode
 *    (or kfreed on error). So those values need to be allocated by the caller,
 *    but MUST NOT be free'ed by the caller.
 * @param fhgfsStat may be given to avoid extra remoting, or may be NULL (in which case a remote
 * stat will be done).
 * @return 0 on success, negative linux error code otherwise
 */
int __FhgfsOps_instantiateInode(struct dentry* dentry, EntryInfo* entryInfo, fhgfs_stat* fhgfsStat)
{
   const char* logContext = "__FhgfsOps_instantiateInode";
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   int retVal = 0;
   fhgfs_stat fhgfsStatInternal;
   fhgfs_stat* actualStatInfo; // points either external given or internal stat data
   FhgfsOpsErr statRes = FhgfsOpsErr_SUCCESS;

   LOG_DEBUG_FORMATTED(log, 5, logContext,
      "called (%s).", fhgfsStat ? "with stat info" : "without stat info");
   IGNORE_UNUSED_VARIABLE(log);
   IGNORE_UNUSED_VARIABLE(logContext);


   if(fhgfsStat)
      actualStatInfo = fhgfsStat;
   else
   { // no stat data given by caller => request from server
      actualStatInfo = &fhgfsStatInternal;

      statRes = FhgfsOpsRemoting_statDirect(app, entryInfo, &fhgfsStatInternal);
   }


   if(statRes != FhgfsOpsErr_SUCCESS)
   { // error
      EntryInfo_uninit(entryInfo);

      retVal = FhgfsOpsErr_toSysErr(statRes);
   }
   else
   { // success (entry exists on server or was already given by caller)
      struct kstat kstat;
      struct inode* newInode;

      OsTypeConv_kstatFhgfsToOs(actualStatInfo, &kstat);

      kstat.ino = FhgfsInode_generateInodeID(dentry->d_sb, entryInfo->entryID,
         strlen(entryInfo->entryID) );

      newInode = __FhgfsOps_newInode(dentry->d_sb, &kstat, 0, entryInfo);
      if(unlikely(!newInode || IS_ERR(newInode) ) )
         retVal = IS_ERR(newInode) ? PTR_ERR(newInode) : -EACCES;
      else
      { // new inode created
         d_instantiate(dentry, newInode);
      }
   }

   return retVal;
}

/**
 * Compare ID of given cachedInode with ID from comparison info arg.
 * This is called by iget5_locked() to make sure that we don't have an instance of a given inode
 * already (e.g. due to a hardlink or from an NFS handle) before we allocate a new one.
 *
 * Note: This is called with a spin_lock (inode_hash_lock) held, so we may not sleep.
 *
 * @param voidComparisonInfo opaque data pointer as passed to iget5_locked.
 * @return 0 if IDs don't match, !=0 on match.
 */
int __FhgfsOps_compareInodeID(struct inode* cachedInode, void* voidComparisonInfo)
{
   FhgfsInodeComparisonInfo* comparisonInfo = voidComparisonInfo;

   if(cachedInode->i_ino != comparisonInfo->inodeHash)
   { // entryID string hashes don't match

      return 0;
   }
   else
   { // inode hashes match => compare entryID strings

      FhgfsInode* fhgfsInode = FHGFS_INODE(cachedInode);
      const char* searchEntryID = comparisonInfo->entryID;

      return FhgfsInode_compareEntryID(fhgfsInode, searchEntryID);
   }

}

/**
 * A dummy (to be passed to iget5_locked() ), which actually does nothing.
 *
 * We prefer to do initialization afterwards on the new inode (which is safe, because the inode
 * is still marked as new after iget5_locked returns). This method is called with the
 * inode_hash_lock spin lock held and we don't want to keep that lock longer than necessary.
 *
 * @return 0 on success, !=0 on error (specific error code is not checked by calling code).
 */
int __FhgfsOps_initNewInodeDummy(struct inode* newInode, void* newInodeInfo)
{
   return 0;
}


/**
 * Flush file content caches of the given inode.
 *
 * @return 0 on success, negative linux error code otherwise.
 */
int __FhgfsOps_flushInodeFileCache(App* app, struct inode* inode)
{
   Config* cfg = App_getConfig(app);
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   if(inode->i_mapping)
   { // flush out file contents for correct file size
      int flushRes = write_inode_now(inode, 1);
      if(flushRes < 0)
         return flushRes;
   }

   if(S_ISREG(inode->i_mode) && (Config_getTuneFileCacheTypeNum(cfg) == FILECACHETYPE_Buffered) )
   { // regular file and buffered mode => flush write cache for correct file size
      FhgfsOpsErr flushRes = FhgfsOpsHelper_flushCache(app, fhgfsInode, fhgfs_false);
      if(unlikely(flushRes != FhgfsOpsErr_SUCCESS) )
         return FhgfsOpsErr_toSysErr(flushRes);
   }

   return 0;
}

/**
 * Called when we want to check whether the inode has changed on the server (or to update
 * the inode attribs after we've made changes to a file/dir).
 * (If it has changed, we invalidate the caches.)
 *
 * Note: This method will indirectly aqcuire the i_lock spinlock.
 *
 * @param fhgfsStat if NULL is given, this method will perform a remote stat; if not NULL, no
 * remoting is needed. (note that usually  _flushInodeFileCache() or similar should be called before
 * retrieving stat info).
 * @return 0 on success (validity), negative linux error code if no longer valid
 */
int __FhgfsOps_refreshInode(App* app, struct inode* inode, fhgfs_stat* fhgfsStat)
{
   const char* logContext = "FhgfsOps_refreshInode";
   Logger* log = App_getLogger(app);
   Config* cfg = App_getConfig(app);

   int retVal = 0;
   struct kstat kstat;
   int flushRes;
   FhgfsOpsErr statRes;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   time_t oldMTime;
   loff_t oldSize;
   unsigned cacheElapsedMS;
   fhgfs_bool mtimeSizeInvalidate;
   fhgfs_bool timeoutInvalidate;

   LOG_DEBUG_FORMATTED(log, 5, logContext,
      "called (%s).", fhgfsStat ? "with stat info" : "without stat info");
   IGNORE_UNUSED_VARIABLE(log);
   IGNORE_UNUSED_VARIABLE(logContext);


   if(fhgfsStat)
   { // stat info given by caller => no remoting needed
      OsTypeConv_kstatFhgfsToOs(fhgfsStat, &kstat);
   }
   else
   { // no stat info given by caller => get it from server
      fhgfs_stat fhgfsStatInternal;

      // flush file contents for correct stat size
      flushRes = __FhgfsOps_flushInodeFileCache(app, inode);
      if(flushRes < 0)
         return flushRes;

      FhgfsInode_entryInfoReadLock(fhgfsInode); // LOCK EntryInfo

      statRes = FhgfsOpsRemoting_statDirect(app, FhgfsInode_getEntryInfo(fhgfsInode),
         &fhgfsStatInternal);

      FhgfsInode_entryInfoReadUnlock(fhgfsInode); // UNLOCK EntryInfo

      if(statRes != FhgfsOpsErr_SUCCESS)
      { // error or entry doesn't exist anymore
         retVal = FhgfsOpsErr_toSysErr(statRes);
         goto cleanup;
      }

      OsTypeConv_kstatFhgfsToOs(&fhgfsStatInternal, &kstat);
   }

   // stat succeeded, so the entry still exists

   if(inode->i_ino == FHGFS_INODE_ROOT_INO)
   { // root node deserves special handling (less checking)
      __FhgfsOps_applyStatDataToInode(&kstat, inode);

      goto cleanup;
   }

   // check whether entry is still the same object type

   if(unlikely( (inode->i_mode & S_IFMT) != (kstat.mode & S_IFMT) ) )
   { // object type changed => limit damage by marking it as bad

      /* note: this is quite impossible since we're verifying by entryID and an ID should
         never ever be assigned to another object type. */

      umode_t savedMode = inode->i_mode; // save mode

      make_bad_inode(inode);

      inode->i_mode = savedMode; // restore mode

      printk_fhgfs(KERN_WARNING, "%s: Inode object type changed unexpectedly.\n", logContext);

      // invalidate cached pages
      if(!S_ISDIR(inode->i_mode) )
      {
         invalidate_remote_inode(inode);
      }

      retVal = -ENOENT;
      goto cleanup;
   }

   // apply new stat data

   spin_lock(&inode->i_lock); // I _ L O C K

   oldMTime = inode->i_mtime.tv_sec;
   oldSize = i_size_read(inode);

   __FhgfsOps_applyStatDataToInodeUnlocked(&kstat, inode);

   // compare previous size/mtime to detect modifications by other clients

   mtimeSizeInvalidate =
      (inode->i_mtime.tv_sec != oldMTime) || (i_size_read(inode) != oldSize);
   cacheElapsedMS = Time_elapsedMS(&fhgfsInode->dataCacheTime);
   timeoutInvalidate = cacheElapsedMS > Config_getTunePageCacheValidityMS(cfg);

   if(S_ISDIR(inode->i_mode) )
   {
      Time_setToNow(&fhgfsInode->dataCacheTime);
   }
   else
   if(mtimeSizeInvalidate || timeoutInvalidate)
   { // file contents changed => invalidate non-dirty pages
      spin_unlock(&inode->i_lock); // I _ U N L O C K
      invalidate_remote_inode(inode); // might sleep => unlocked
      spin_lock(&inode->i_lock); // I _ R E L O C K

      Time_setToNow(&fhgfsInode->dataCacheTime); // after invalidate_... to avoid race condition
   }

   spin_unlock(&inode->i_lock); // I _ U N L O C K


   // clean up
cleanup:

   return retVal;
}

/**
 * Note: Refreshes inode and mapping only if cache validity timeout expired.
 *
 * @return 0 on success (validity), negative linux error code if no longer valid
 */
int __FhgfsOps_revalidateMapping(App* app, struct inode* inode)
{
   const char* logContext = "FhgfsOps_revalidateMapping";
   Logger* log = App_getLogger(app);
   Config* cfg = App_getConfig(app);

   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   unsigned cacheElapsedMS;
   fhgfs_bool timeoutInvalidate;
   fhgfs_bool needRefresh = fhgfs_false;

   LOG_DEBUG_FORMATTED(log, 5, logContext, "called.");
   IGNORE_UNUSED_VARIABLE(log);
   IGNORE_UNUSED_VARIABLE(logContext);

   spin_lock(&inode->i_lock); // I _ L O C K

   cacheElapsedMS = Time_elapsedMS(&fhgfsInode->dataCacheTime);
   timeoutInvalidate = cacheElapsedMS > Config_getTunePageCacheValidityMS(cfg);

   if(!S_ISDIR(inode->i_mode) && timeoutInvalidate)
      needRefresh = fhgfs_true;

   spin_unlock(&inode->i_lock); // I _ U N L O C K

   if(needRefresh)
      return __FhgfsOps_refreshInode(app, inode, NULL);

   return 0;
}
