#include <app/App.h>
#include <app/config/Config.h>
#include <common/nodes/Node.h>
#include <filesystem/FhgfsOpsFile.h>
#include <os/OsCompat.h>
#include <filesystem/helper/IoctlHelper.h>
#include "FhgfsOpsSuper.h"
#include "FhgfsOpsHelper.h"
#include "FhgfsOpsIoctl.h"
#include "FhgfsOpsInode.h"

#ifdef CONFIG_COMPAT
#include <asm/compat.h>
#endif

#include <linux/mount.h>
#include <opentk/gpl_compat/GPLCompat.h>


/* old kernels do not know the TCGETS ioctl definition yet, see
   asm-generic/ioctls.h in recent kernels for more details */
#ifndef TCGETS
#define TCGETS   0x5401
#endif

static long FhGFS_ioctlGetCfgFile(struct file *file, void __user *argp);
static long FhGFS_ioctlGetRuntimeCfgFile(struct file *file, void __user *argp);
static long FhGFS_ioctlCreateFile(struct file *file, void __user *argp);


/**
 * Execute FhGFS IOCTLs.
 *
 * @param file the file the ioctl was opened for
 * @param cmd the ioctl command supposed to be done
 * @param arg in and out argument
 */
long FhgfsOps_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
   struct dentry *dentry = file->f_dentry;
   struct inode *inode = dentry->d_inode;
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_ioctl";

   Logger_logFormatted(log, 5, logContext, "Called IOCTL cmd: %u", cmd);

   switch(cmd)
   {
      case FHGFS_IOC_GET_CFG_FILE:
      { /* get the client config file */
         return FhGFS_ioctlGetCfgFile(file, (void __user *) arg);
      } break;

      case FHGFS_IOC_GET_RUNTIME_CFG_FILE:
      { /* get the runtime client config file */
         return FhGFS_ioctlGetRuntimeCfgFile(file, (void __user *) arg);
      } break;

      case FHGFS_IOC_GETVERSION:
      case FHGFS_IOC_GETVERSION_OLD:
      {
         /* useful for userspace filesystems or user space nfs server, such
          * as as unfs3 */
         return put_user(inode->i_generation, (int __user *) arg);
      } break;

      case TCGETS:
      {
         // this seems to be an ioctl called isatty(). No idea what it is fow, but it floods
         // logs, if we have debugging enabled. So for now we catch it without any logs.
         return -ENOTTY;
      } break;

      case FHGFS_IOC_CREATE_FILE:
      {
         return FhGFS_ioctlCreateFile(file, (void __user *) arg);
      } break;

      case FHGFS_IOC_TEST_IS_FHGFS:
      {
         const char returnStr[] = FHGFS_IOCTL_TEST_STRING;
         int strLen = strlen(returnStr) + 1;

         /* tell user space we are really on FhGFS */
         void __user* argp = (void __user *)arg;

         int cpRes = copy_to_user(argp, returnStr, strLen); // also calls access_ok
         if(cpRes)
         {
            LOG_DEBUG_FORMATTED(log, 2, logContext, "copy_to_user failed()");
            return -EINVAL;
         }

      } break;

      default:
      {
         LOG_DEBUG_FORMATTED(log, 1, logContext, "Unknown IOCTL: %u", cmd);
         return -EINVAL;
      } break;
   }

   return 0;
}

#ifdef CONFIG_COMPAT
/**
 * Compatibility ioctl method for 64-bit kernels called from 32-bit user space
 */
long FhgfsOps_compatIoctl(struct file *file, unsigned int cmd, unsigned long arg)
{
   struct dentry *dentry = file->f_dentry;
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = __func__;

   Logger_logFormatted(log, 5, logContext, "Called IOCTL cmd: %u", cmd);

   switch(cmd)
   {
      case FHGFS_IOC32_GETVERSION_OLD:
      case FHGFS_IOC32_GETVERSION:
      {
         cmd = FHGFS_IOC_GETVERSION;
      } break;
   }

   return FhgfsOps_ioctl(file, cmd, (unsigned long) compat_ptr(arg));
}
#endif


static long FhGFS_ioctlGetCfgFile(struct file *file, void __user *argp)
{
   struct dentry *dentry = file->f_dentry;
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = __func__;

   struct Fhgfs_ioctlGetCfgFile __user *confFile = (struct Fhgfs_ioctlGetCfgFile*)argp;
   Config* cfg = App_getConfig(app);
   char* fileName = Config_getCfgFile(cfg);
   int cpRes;

   int cfgFileStrLen = strlen(fileName) + 1; // strlen() does not count \0
   if (unlikely(cfgFileStrLen > FHGFS_IOCTL_CFG_MAX_PATH))
   {
      Logger_logFormatted(log, 2, logContext, "Config file has too long path (%d vs max %d)!",
         cfgFileStrLen, FHGFS_IOCTL_CFG_MAX_PATH);
      return -EINVAL;
   }

   if(!access_ok(VERIFY_WRITE, confFile, sizeof(*confFile) ) )
   {
      Logger_logFormatted(log, 2, logContext, "access_ok() denied to write to conf_file");
      return -EINVAL;
   }

   cpRes = copy_to_user(&confFile->path[0], fileName, cfgFileStrLen); // also calls access_ok
   if(cpRes)
   {
      LOG_DEBUG_FORMATTED(log, 2, logContext, "copy_to_user failed()");
      return -EINVAL;
   }

   return 0;
}

static long FhGFS_ioctlGetRuntimeCfgFile(struct file *file, void __user *argp)
{
   struct dentry *dentry = file->f_dentry;
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = __func__;

   struct Fhgfs_ioctlGetCfgFile __user *confFile = (struct Fhgfs_ioctlGetCfgFile*)argp;
   char* fileName = vmalloc(FHGFS_IOCTL_CFG_MAX_PATH);

   Node* localNode = App_getLocalNode(app);
   char* localNodeID = Node_getID(localNode);

   int cpRes;
   int cfgFileStrLen;

   if(!access_ok(VERIFY_WRITE, confFile, sizeof(*confFile) ) )
   {
      Logger_logFormatted(log, 2, logContext, "access_ok() denied to write to conf_file");
      return -EINVAL;
   }

   cfgFileStrLen = os_scnprintf(fileName, FHGFS_IOCTL_CFG_MAX_PATH, "/proc/fs/fhgfs/%s/config",
      localNodeID);
   if (cfgFileStrLen <= 0)
   {
      Logger_logFormatted(log, 2, logContext, "buffer too small");
      vfree(fileName);
      return -EINVAL;
   }

   cfgFileStrLen++; // os_scnprintf(...) does not count \0

   cpRes = copy_to_user(&confFile->path[0], fileName, cfgFileStrLen); // also calls access_ok
   if(cpRes)
   {
      LOG_DEBUG_FORMATTED(log, 2, logContext, "copy_to_user failed()");
      return -EINVAL;
   }

   vfree(fileName);
   return 0;
}

/**
 * handle ioctl to create a file
 *
 * @return 0 on success, negative linux error code otherwise
 */
static long FhGFS_ioctlCreateFile(struct file *file, void __user *argp)
{
   struct dentry *dentry = file->f_dentry;
   struct inode *inode = dentry->d_inode;
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = __func__;
   EntryInfo parentInfo;
   EntryInfo newEntryInfo; // only needed for FhgfsOpsRemoting_mkfile(), not really used
   struct Fhgfs_ioctlMkFile fileInfo;
   struct CreateInfo createInfo = { .preferredStorageTargets = NULL };
   int retVal = 0;
   FhgfsOpsErr mkRes;

   #if LINUX_VERSION_CODE < KERNEL_VERSION(3,9,0)
      struct vfsmount* mnt = file->f_vfsmnt;
   #else
      struct vfsmount* mnt = file->f_path.mnt;
   #endif


   Logger_logFormatted(log, Log_SPAM, logContext, "Create file from ioctl");

   retVal = os_generic_permission(inode, MAY_WRITE | MAY_EXEC);
   if (retVal)
      return retVal;

   retVal = os_mnt_want_write(mnt); // check and rw-reference counter
   if (retVal)
      return retVal;

   memset(&fileInfo, 0, sizeof(struct Fhgfs_ioctlMkFile) ); // required for SAVE_KFREEs() below

   retVal = IoctlHelper_ioctlCreateFileCopyFromUser(app, argp, &fileInfo); // user to kernel space
   if(retVal)
   {
      LOG_DEBUG_FORMATTED(log, Log_SPAM, logContext,
         "Copy from user of struct Fhgfs_ioctlMkFile failed");
      goto out;
   }

   CreateInfo_init(app, inode, fileInfo.entryName, fileInfo.mode,
      &createInfo);

   retVal = IoctlHelper_ioctlCreateFileTargetsToList(app, &fileInfo, &createInfo); // target list
   if (retVal)
      goto out;

   // only use the provided UID and GID if we are root
   /* note: this means we create the file with different uid/gid without notifying the caller. maybe
      we should change that. */
   if (currentFsUserID == 0 || currentFsGroupID == 0)
   {
      createInfo.userID  = fileInfo.uid;
      createInfo.groupID = fileInfo.gid;
   }

   EntryInfo_init(&parentInfo, fileInfo.ownerNodeID, fileInfo.parentParentEntryID,
      fileInfo.parentEntryID, fileInfo.parentName, DirEntryType_DIRECTORY, STATFLAG_HINT_INLINE);

   switch (fileInfo.fileType)
   {
      case DT_REG:
      {
         down_read(&inode->i_sb->s_umount);

         LOG_DEBUG_FORMATTED(log, Log_SPAM, logContext, "Going to create file %s",
            createInfo.entryName);

         mkRes = FhgfsOpsRemoting_mkfile(app, &parentInfo, &createInfo, &newEntryInfo);
         if(mkRes != FhgfsOpsErr_SUCCESS)
         {
            LOG_DEBUG_FORMATTED(log, Log_SPAM, logContext, "mkfile failed");
            retVal = FhgfsOpsErr_toSysErr(mkRes);
         }
         else
            EntryInfo_uninit(&newEntryInfo); // delete unused object

         up_read(&inode->i_sb->s_umount);
      } break;

      case DT_LNK:
      {
         down_read(&inode->i_sb->s_umount);

         mkRes = FhgfsOpsHelper_symlink(app, &parentInfo, fileInfo.symlinkTo,
            &createInfo, &newEntryInfo);
         if(mkRes)
            retVal = FhgfsOpsErr_toSysErr(mkRes);
         else
            EntryInfo_uninit(&newEntryInfo); // delete unused object

         up_read(&inode->i_sb->s_umount);
      } break;

      default:
      {
         // unsupported file type
         LOG_DEBUG_FORMATTED(log, Log_NOTICE, logContext, "Unsupported file type: %d",
            fileInfo.fileType);
         retVal = -EINVAL;
         goto out;
      } break;
   }


out:
   SAFE_KFREE(fileInfo.parentParentEntryID);
   SAFE_KFREE(fileInfo.parentEntryID);
   SAFE_KFREE(fileInfo.parentName);
   SAFE_KFREE(fileInfo.entryName);
   SAFE_KFREE(fileInfo.symlinkTo);
   SAFE_KFREE(fileInfo.prefTargets);
   SAFE_DESTRUCT(createInfo.preferredStorageTargets, UInt16List_destruct);

   os_mnt_drop_write(mnt); // release the rw-reference counter

   return retVal;
}

