/*
 * fhgfs page cache methods
 *
 */

#include <app/log/Logger.h>
#include <app/App.h>
#include <app/config/Config.h>
#include <common/toolkit/vector/StrCpyVec.h>
#include <common/toolkit/vector/IntCpyVec.h>
#include <common/toolkit/LockingTk.h>
#include <common/storage/StorageErrors.h>
#include <common/toolkit/StringTk.h>
#include <common/storage/striping/StripePattern.h>
#include <filesystem/ProcFs.h>
#include <os/OsTypeConversion.h>
#include "FhgfsOpsDir.h"
#include "FhgfsOpsFile.h"
#include "FhgfsOpsHelper.h"
#include "FhgfsOpsInode.h"
#include "FhgfsOpsPages.h"
#include "FhgfsOpsSuper.h"
#include "FhgfsOps_versions.h"

#include <opentk/gpl_compat/GPLCompat.h>

#include <linux/writeback.h>
#include <linux/mm.h>
#include <linux/mpage.h>
#include <linux/backing-dev.h>
#include <linux/pagemap.h>

#ifdef CONFIG_COMPAT
#include <asm/compat.h>
#endif

#define INITIAL_FIND_PAGES (256)    // search initially for this number of pages

#if ((INITIAL_FIND_PAGES) > (FHGFS_MAX_PAGE_LIST_SIZE))
   #error // trigger a compilation error as we would end up with memory corruption in real live
#endif


// forward declarations
struct fhgfsWritePageHelper;
typedef struct fhgfsWritePageHelper fhgfsWrPgHelper;


static void _FhgfsOpsPages_releasePageList(struct page** pageList, unsigned size);
static fhgfs_bool _FhgfsOpsPages_handleWBPage(fhgfsWrPgHelper* wrPage, int pgCount, fhgfs_bool*
   outDone);
static int _FhgfsOpsPages_getDirtyPages(fhgfsWrPgHelper* wrPage);
static int _FhgfsOpsPages_pageListToVector(Logger* log, struct list_head *pageList,
   unsigned numPages, struct page** pgArray);
static void _FhgfsOpsPages_handleShortRead(App* app, struct inode* inode, struct page **pageArray,
   FhgfsPageVec* pageVec, int pgIdx, RemotingIOInfo* ioInfo, struct address_space* mapping,
   struct pagevec* lru_pvec);
static fhgfs_bool _FhgfsOpsPages_addToLRU(Logger* log, struct address_space* mapping,
   struct page* page, struct pagevec* lru_pvec);



/**
 * A struct with variables to be exchanged between fhgfs writepages functions
 */
struct fhgfsWritePageHelper
{
      struct address_space* mapping;
      struct writeback_control* wbc;

      FhgfsInode* fhgfsInode;
      FileHandleType handleType;
      RemotingIOInfo ioInfo;
      fhgfs_bool have_handle;

      struct page **pageList; // assigned by find_get_pages_tag()
      pgoff_t pageIdxEnd;
      pgoff_t pageIdx; // page index where to start looking for dirty pages

      struct currentPage
      {
            struct page *page;
            size_t usedSize; // usedSize of used data of this page
            pgoff_t nextContIdx; // page->index + 1
      } curPage;

      FhgfsPageVec pageVec;

      int searchTag;

      App* app;
      Logger* log;
};


/**
 * If the meta server has told us for some reason a wrong file-size (i_size) the caller would
 * wrongly discard data beyond i_size. So we are going to correct i_size here.
 *
 * This could be removed once we are sure the meta server *always* has the correct i_size
 * (which might never be the case, due to concurrent writes and truncates).
 *
 * Note: This is the non-inlined version. Only call it from
 *    FhgfsOpsPages_incInodeFileSizeOnPagedRead()
 */
void __FhgfsOpsPages_incInodeFileSizeOnPagedRead(struct inode* inode, loff_t offset, size_t readRes)
{
   App* app = FhgfsOps_getApp(inode->i_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "Paged-read";
   loff_t i_size;

   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   EntryInfo* entryInfo = FhgfsInode_getEntryInfo(fhgfsInode);

   // refresh the inode first, hopefully that is sufficient
   __FhgfsOps_refreshInode(app, inode, NULL);

   i_size = i_size_read(inode);

   if (unlikely(readRes && (offset + readRes > i_size) ) )
   {  // _refreshInode was not sufficient, force a meta-update
      FhgfsOpsErr refreshRes = FhgfsOpsRemoting_refreshEntry(app, entryInfo);

      if (refreshRes != FhgfsOpsErr_SUCCESS)
      {
         Logger_logErr(log, logContext, "Meta Refresh failed.");

         // TODO: Return an io-error to the application to notify it about a too short read?
      }

      __FhgfsOps_refreshInode(app, inode, NULL); // again try to refresh the inode


      /* note on i_lock/i_size: make sure we only increase i_size and do not decrease it
         (e.g. in case we're racing with a concurrent writer at a higher offset) */
      spin_lock(&inode->i_lock); // L O C K

      i_size = i_size_read(inode);

      if (unlikely(readRes && (offset + readRes > i_size) ) )
      {  // All attempts to update the remote inode size to the position that we read failed.

         /* i_size_write() sugguests to lock i_mutex, but we might end up here from write_begin()
          * (FhgfsOps_write_begin), which aready has i_mutex locked.  The more common callers
          * readpages() and readpage() do not have i_mutex, though. */

         i_size_write(inode, offset + readRes);

         spin_unlock(&inode->i_lock); // U N L O C K

         /* note: this situation can also be "normal" with a concurrent trunc, so we have to be
            careful regarding user warnings and error return values. */
         Logger_log(log, Log_WARNING, logContext, "Failed to increase MDS inode size to the "
            "expected value. (Application might read less data than expected or file was truncated "
            "during read operation.)");
      }
      else
         spin_unlock(&inode->i_lock); // U N L O C K
   }

   Logger_logFormatted(log, Log_DEBUG, logContext,
      "EntryID: %s Correcting inode size from %lld to %lld",
      entryInfo->entryID, i_size, offset + readRes);
}


/**
 * release (put) all pages we got a reference on with find_get_pages_tag()
 */
static void _FhgfsOpsPages_releasePageList(struct page** pageList, unsigned size)
{
   unsigned idx;

   for(idx = 0; idx < size; idx++) {
      struct page* page = pageList[idx];
      page_cache_release(page);
   }
}

/**
 * Do all required operations before writing the current page
 */
static fhgfs_bool _FhgfsOpsPages_handleWBPage(fhgfsWrPgHelper* wrPage, int pgCount, fhgfs_bool* outDone)
{
   loff_t i_size;
   struct page* page = wrPage->curPage.page;

   if(page->index > wrPage->pageIdxEnd)
   { // behind end => get out (note: can't be range_cyclic, because then "end == -1")
      *outDone = fhgfs_true;
      return fhgfs_false;
   }

   if(pgCount == 0)
   {
      lock_page(page);
      wrPage->pageVec.offset = page_offset(page);
   }
   else
   if(!trylock_page(page) )
      return fhgfs_false; // we already got at least one page

   if(page->mapping != wrPage->mapping)
   { // page concurrently disappeared (has been truncated or invalidated)
      unlock_page(page);
      return fhgfs_false;
   }

   if(!PageDirty(page) )
   { // someone already wrote the page
      unlock_page(page);
      return fhgfs_false;
   }

   if(wrPage->curPage.nextContIdx && (page->index != wrPage->curPage.nextContIdx) )
   { // non-consecutive page (page does not follow directly after the previously proccessed page)
      unlock_page(page);
      return fhgfs_false;
   }

   if(wrPage->wbc->sync_mode != WB_SYNC_NONE)
      wait_on_page_writeback(page);

   if(PageWriteback(page) || !clear_page_dirty_for_io(page) )
   {
      unlock_page(page);
      return fhgfs_false;
   }

   set_page_writeback(page);

   i_size = i_size_read(wrPage->mapping->host);
   if(page_offset(page) >= i_size)
   {
      // page offset behind end of file
      // NOTE: (page_offset + page_length) > i_size will be handled below
      *outDone = fhgfs_true;
      unlock_page(page);
      end_page_writeback(page);
      return fhgfs_false;
   }

   // (page_offset + page_length) > i_size
   wrPage->curPage.usedSize = min(i_size - page_offset(page), (loff_t)PAGE_CACHE_SIZE);

   return fhgfs_true;
}

/**
 * Get dirty pages of the given file. We also already get the Fhgfs reference handle here, if there
 * are dirty pages.
 */
static int _FhgfsOpsPages_getDirtyPages(fhgfsWrPgHelper* wrPage)
{
   int retVal;
   int stripePages = 0; // number of pages of a complete stripe set
   int foundPages;      // number of pages we already found
   unsigned newFoundPages;   // number of pages we find in further runs of find_get_pages_tag()
   struct page **pageList = wrPage->pageList;

   // how many pages we still need to find
   int numPagesToFind = MIN(INITIAL_FIND_PAGES, wrPage->wbc->nr_to_write);

   foundPages = find_get_pages_tag(wrPage->mapping, &wrPage->pageIdx, wrPage->searchTag,
      numPagesToFind, pageList);
   if (!foundPages)
      return 0; // nothing to do for now

   pageList += foundPages;

   /* So far we only got an initial number of pages, but the file stripe might have more pages
    * and if possible we want to write a complete stripe set. In order to know the stripe set
    * we first have to get the file/inode reference handle (we need it anyway later on).
    */

   if (!wrPage->have_handle)
   {
      fhgfs_bool referenceRes = FhgfsInode_referenceHandle(wrPage->fhgfsInode,
         OPENFILE_ACCESS_WRITE, fhgfs_true, &(wrPage->handleType) );

      if(unlikely(referenceRes != FhgfsOpsErr_SUCCESS) )
      { // error getting reference
         int i;

         Logger_logErrFormatted(wrPage->log, __func__, "Unable to get file handle. "
            "ErrMsg: %s", FhgfsOpsErr_toErrString(referenceRes) );

         retVal = FhgfsOpsErr_toSysErr(referenceRes);
         mapping_set_error(wrPage->mapping, retVal);

         // find_get_pages_tag() increased the page reference counter, so we need to deacrease it
         for (i = 0; i < foundPages; i++)
            put_page(wrPage->pageList[i]);

         return retVal;
      }

      wrPage->have_handle = fhgfs_true;
   }

   FhgfsInode_getRefIOInfo(wrPage->fhgfsInode, wrPage->handleType,
      FhgfsInode_handleTypeToOpenFlags(wrPage->handleType), &wrPage->ioInfo);

   stripePages = RemotingIOInfo_getNumPagesPerStripe(&wrPage->ioInfo);
   numPagesToFind = MIN(stripePages - foundPages, FHGFS_MAX_PAGE_LIST_SIZE - foundPages) ;

   /* Is it really a good idea to limit us to the number of pages the kernel requested or
    * would it be better to use stripePages as limit only? */
   numPagesToFind = MIN(numPagesToFind, wrPage->wbc->nr_to_write - foundPages);
   if (numPagesToFind < 0)
   {
      // no need to search for even more
      retVal = foundPages;
      goto out_alloc;
   }

   do
   {
      newFoundPages = find_get_pages_tag(wrPage->mapping, &wrPage->pageIdx,
                  PAGECACHE_TAG_DIRTY,
                  numPagesToFind, pageList);
      foundPages += newFoundPages;
      numPagesToFind -= newFoundPages;
      pageList += newFoundPages;

      // find_get_pages_tag() increases wrPage->pageIdx itself
   } while (newFoundPages && numPagesToFind  && wrPage->pageIdx <= wrPage->pageIdxEnd);

#if 0
   // check if we have a complete stripe and we are not in sync-mode and we also already
   // did something (users mostly do not write complete stripes, but also might only write a few
   // bytes...).
   // !wrPage->pageVec.pages => we already allocated it below, so we already have been here before
   if(foundPages < stripePages && wrPage->wbc->sync_mode == WB_SYNC_NONE && wrPage->pageVec.pages)
   {
      _FhgfsOpsPages_releasePageList(wrPage->pageList, foundPages);
      return 0;
   }
#endif

   retVal = foundPages;

out_alloc:
   if (!wrPage->pageVec.pages)
   {
      // first run, not initialized yet
      int allocPages = MAX(INITIAL_FIND_PAGES, stripePages);

      int allocSize = sizeof(FhgfsPage) * allocPages;
      wrPage->pageVec.pages = vmalloc(allocSize);

      if (!wrPage->pageVec.pages)
         retVal = -ENOMEM;
   }

   return retVal;
}

/**
 * @return 0 on success, negative linux error code otherwise
 */
int FhgfsOps_writepagesVec(struct address_space* mapping, struct writeback_control* wbc)
{
   App* app = FhgfsOps_getApp(mapping->host->i_sb);

   struct inode* inode = mapping->host;

   int retVal = 0;

   // do we need to wrap around (to the beginning) once we already scanned all pages to the end?
   fhgfs_bool cycled = fhgfs_false;

   /* loop-terminator (used by inner loop to terminate outer loop)
    * will be set to true, if pgIdx > end (non-cycling mode) or if the file was truncated */
   fhgfs_bool done = fhgfs_false;

   fhgfs_bool range_whole = fhgfs_false; // Shall we write the whole file range?

   fhgfsWrPgHelper wrPage =
   {
         .mapping        = mapping,
         .wbc            = wbc,
         .pageIdxEnd     = -1,
         .have_handle    = fhgfs_false,
         .app            = app,
         .log            = App_getLogger(app),

         .pageVec        = {
                              .pages = NULL,
                           },

         .fhgfsInode = FHGFS_INODE(inode),

   };

   wrPage.pageList = vmalloc(sizeof (*wrPage.pageList) * FHGFS_MAX_PAGE_LIST_SIZE);
   if (wrPage.pageList == NULL)
      return -ENOMEM;

   #ifdef LOG_DEBUG_MESSAGES
   {
      struct dentry* dentry = d_find_alias(inode); // calls dget_locked

      FhgfsOpsHelper_logOpDebug(app, dentry, __func__, "(nr_to_write: %ld = %lluKB)",
         wbc->nr_to_write, (long long) (wbc->nr_to_write << PAGE_CACHE_SHIFT) / 1024);

      if(dentry)
         dput(dentry);
   }
   #endif // LOG_DEBUG_MESSAGES

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,17)
   if(wbc->sync_mode == WB_SYNC_NONE)
      wrPage.pageIdx = mapping->writeback_index; // start from prev offset
   else
   { // called for sync => whole file sweep
      wrPage.pageIdx = 0;
      cycled = fhgfs_true; // already at beginning => no wrap-around needed
      range_whole = fhgfs_true;
   }

   if(wbc->start || wbc->end)
   {
      wrPage.pageIdx    = wbc->start >> PAGE_CACHE_SHIFT;
      wrPage.pageIdxEnd = wbc->end >> PAGE_CACHE_SHIFT;
      cycled = fhgfs_true;
   }
#else
   if(wbc->range_cyclic)
   {
      wrPage.pageIdx    = mapping->writeback_index; // start from prev offset
      wrPage.pageIdxEnd = -1; // (note: end is unsigned)

      if(wrPage.pageIdx == 0)
         cycled = fhgfs_true; // already at beginning => no wrap-around needed
   }
   else
   {
      wrPage.pageIdx    = wbc->range_start >> PAGE_CACHE_SHIFT;
      wrPage.pageIdxEnd = wbc->range_end >> PAGE_CACHE_SHIFT;
      cycled = fhgfs_true;

      if( (wbc->range_start == 0) && (wbc->range_end == FHGFS_LLONG_MAX) )
         range_whole = fhgfs_true; // whole file sweep

      #ifndef KERNEL_HAS_TAG_PAGES_FOR_WRITEBACK
         // deadlock prevention before tag_pages_for_writeback() was introduced, but seems
         //  to prevent complete writeout for kernels that have this function
         if( (wbc->sync_mode == WB_SYNC_ALL) && (wbc->range_end == LLONG_MAX) )
            wrPage.pageIdxEnd = i_size_read(inode) >> PAGE_CACHE_SHIFT;
      #endif
   }
#endif // LINUX_VERSION_CODE

#ifdef KERNEL_HAS_WBC_TAGGED_WRITEPAGES
   if (wbc->sync_mode == WB_SYNC_ALL)
      wrPage.searchTag = PAGECACHE_TAG_TOWRITE;
   else
      wrPage.searchTag = PAGECACHE_TAG_DIRTY;
#else
   wrPage.searchTag = PAGECACHE_TAG_DIRTY;
#endif

   // printk_fhgfs_debug(KERN_INFO, "%s: sync: %d pageIdx: %d pageIdxEnd:%d cyc: %d seachTag %d\n",
   //   __func__, wbc->sync_mode, wrPage.pageIdx, wrPage.pageIdxEnd, !cycled, wrPage.searchTag);
   
   // Loop used to wrap around the index
   while (1)
   {
      ssize_t writeRes = 0;

#ifdef KERNEL_HAS_TAG_PAGES_FOR_WRITEBACK
   #ifdef KERNEL_HAS_WBC_TAGGED_WRITEPAGES
      if (wbc->sync_mode == WB_SYNC_ALL || wbc->tagged_writepages)
   #else
      if (wbc->sync_mode == WB_SYNC_ALL)
   #endif
         tag_pages_for_writeback(mapping, wrPage.pageIdx, wrPage.pageIdxEnd);
#endif

      /* TODO: hmm, actually writeRes == 0 is weird number, as writeres refers to the number of
       *       written bytes, so 0 should impossible and should be handled as error... */
      while(!done && (wrPage.pageIdx <= wrPage.pageIdxEnd) && writeRes >= 0)
      {
         int numDirty; // number of dirty pages we found in this round
         int numCont = 0; // number of dirty pages presently processed with continuous offsets

         int idx; // walks over number of pages

         // re-initialize before each numDirty round
         wrPage.pageVec.size = 0;
         wrPage.curPage.nextContIdx = 0;

         numDirty = _FhgfsOpsPages_getDirtyPages(&wrPage);
         if (numDirty < 0)
         {
            retVal = numDirty; // numDirty has the error code in that case
            goto out;
         }

         //printk_fhgfs_debug(KERN_INFO, "numDirty: %d done: %d pageIdx: %ld pageIdxEnd: %ld "
         //   "writeRes: %ld \n", numDirty, (int) done, wrPage.pageIdx,
         //    wrPage.pageIdxEnd, writeRes);

         for(idx = 0; idx < numDirty; idx++)
         {
            fhgfs_bool handlePageRes;
            FhgfsPage* fhgfsPage;

            wrPage.curPage.page = wrPage.pageList[idx];

            // do all kind of page magic here and check if we need to do more work
            handlePageRes = _FhgfsOpsPages_handleWBPage(&wrPage, numCont, &done);
            if (handlePageRes == fhgfs_false)
               break;

            // now add this current page to our page vector
            fhgfsPage         = &wrPage.pageVec.pages[wrPage.pageVec.size];

            fhgfsPage->data   = kmap(wrPage.curPage.page);
            fhgfsPage->length = wrPage.curPage.usedSize;

            wrPage.pageVec.size++;
            wrPage.curPage.nextContIdx = wrPage.curPage.page->index + 1;
            numCont++;
         }

         // printk_fhgfs_debug(KERN_INFO, "numCont=%d\n", numCont);
         if (numCont > 0)
         {
            loff_t offset = wrPage.pageVec.offset; // offset of the first page of the vector

            writeRes = FhgfsOpsRemoting_RWfileVec(&wrPage.pageVec, offset, &wrPage.ioInfo,
               FHGFS_RWTYPE_WRITE);

            if (unlikely(writeRes < 0) ) // TODO: Handle writeRes == 0
            {
               LOG_DEBUG_FORMATTED(wrPage.log, 1, __func__, "error: %s",
                  FhgfsOpsErr_toErrString(-writeRes) );

               retVal = FhgfsOpsErr_toSysErr(-writeRes);
               mapping_set_error(mapping, retVal);

               // store the offset of the error, so that future writes from user space, which
               // go through FhgfsOps_aio_write() immediatly fail
               if (retVal != -EAGAIN && !wrPage.fhgfsInode->asyncPageError)
                  wrPage.fhgfsInode->asyncPageError = fhgfs_true;
            }
         }

         // handle all pages we wrote
         for(idx = 0; idx < numCont; idx++)
         {
            struct page* page = wrPage.pageList[idx];

            kunmap(page); // page data access done, so unmap it

            // error cases
            if (retVal < 0)
            {
               if (retVal == -EAGAIN)
                  redirty_page_for_writepage(wbc, page);
               else
               {
                  SetPageError(wrPage.pageList[idx]);
                  FhgfsInode_decNumDirtyPages(wrPage.fhgfsInode);
               }
            }
            else
               FhgfsInode_decNumDirtyPages(wrPage.fhgfsInode);

            end_page_writeback(page);
            unlock_page(page);
         }

         // reset the page index to the next page and try to continue there
         if (numCont == 0 && numDirty > 0)
            wrPage.pageIdx = wrPage.pageList[0]->index + 1;

         _FhgfsOpsPages_releasePageList(wrPage.pageList, numDirty); // release (put) pages

         if (retVal >= 0)
            wbc->nr_to_write -= numDirty; /* decrease by number of successfully written pages
                                           * TODO: Right now we do not know if there were any
                                           *       pages of this set successfully written, so
                                           *       FhgfsOpsRemoting_RWfileVec() should get a
                                           *       parameter for that */

         // NOTE: wrPage.pageIdx already increased by find_get_pages_tag(), unless numDirty == 0

         if (numDirty == 0 || retVal < 0)
            break; // pageIdx won't be increased automatically then and need to break out ourselves

         // do not write more than requested
         if (wbc->nr_to_write <= 0 && wbc->sync_mode == WB_SYNC_NONE)
            done = 1;
      }

      if (retVal < 0)
         break;

      if(cycled || done)
         break;

      // range_cyclic: we hit the last page and there is more work to be done
      cycled = fhgfs_true;         // prevent another cycling
      wrPage.pageIdx = 0; //  => wrap back to the start of the file
      wrPage.pageIdxEnd = -1;

   }

out:

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,17)

   if(range_whole)
      mapping->writeback_index = wrPage.pageIdx;

#else // LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)

   #ifdef KERNEL_HAS_WBC_NONRWRITEINDEXUPDATE

      if(!wbc->no_nrwrite_index_update)

   #endif // KERNEL_HAS_WBC_NONRWRITEINDEXUPDATE
      {
         if (wbc->range_cyclic || (range_whole && (wbc->nr_to_write > 0) ) )
            mapping->writeback_index = wrPage.pageIdx;
      }

#endif // LINUX_VERSION_CODE

   if (wrPage.have_handle)
   {
      FhgfsOpsErr releaseRes = FhgfsInode_releaseHandle(wrPage.fhgfsInode, wrPage.handleType);

      if(unlikely(releaseRes != FhgfsOpsErr_SUCCESS) )
      {
         Logger_logErrFormatted(wrPage.log, __func__, "Unable to release file handle. ErrMsg: %s",
            FhgfsOpsErr_toErrString(releaseRes) );

         if(!retVal)
         { // no other error yet
            retVal = FhgfsOpsErr_toSysErr(releaseRes);
            mapping_set_error(mapping, retVal);
         }
      }

   }

   SAFE_VFREE(wrPage.pageVec.pages);
   vfree(wrPage.pageList);

   // printk_fhgfs_debug(KERN_INFO, "%s: retVal: %d\n", __func__, retVal);

   return retVal;
}

/**
 * Build an array of pages from the given page-list, only continous pages count
 *
 * @outNewHead The new list head, if we did not process the entire list
 * @return The page array (vector)
 *
 */
static int _FhgfsOpsPages_pageListToVector(Logger* log, struct list_head *pageList,
   unsigned numPages, struct page** pgArray)
{
      unsigned nextContIdx; // next expected (continous) page index
      unsigned numContPages = 0;
      struct page *page;

      if (unlikely(list_empty(pageList) ) )
      {
         Logger_logFormatted(log, 2, __func__, "Bug! Got an empty List!");
#ifdef FHGFS_DEBUG
         os_dumpStack();
#endif
         return -EIO;
      }

      // walk over the page list here and add continous pages into our page-array
      // TODO: Why are we walking backwards?
      // Note: first expected next index is the current page index, so it will definitely be a hit
     //        and numContPages will be inc'ed to 1
      nextContIdx = list_entry(pageList->prev, struct page, lru)->index;
      list_for_each_entry_reverse(page, pageList, lru)
      {
         if (page->index == nextContIdx && numContPages < numPages)
         {
            pgArray[numContPages] = page;
            numContPages++;
            nextContIdx++;
         }
         else
         {
            /* Non-continous page, as the the kernel will issue another read-ahead itself
             * we do not need to handle it. Handling it ourselves will even cause a slow down, as
             * there somehow come up requests for a single page then */
            break;
         }
      }

      // printk_fhgfs_debug(KERN_INFO, "%s: cont-pages: %d\n", __func__, numContPages);
      return numContPages;
}

/**
 * Handle a short read, which might have 2 two reasons
 * 1) Another client truncated the file, as this client has another inode-size
 *    the kernel read-ahead would try forever to read missing data
 * 2) Sparse files, so one or more chunks have less data than following chunks
 *
 * As we do not know which applies, we always update the inode size first
 *
 * TODO: Once we are open source, this is supposed to go into FhgfsOpsRemoting_RWfileVec(), which
 *       knows which stripes it still has data for (important for efficient sparse handling, as
 *       we right now trow away already read data from other storage targets
 * */
static void _FhgfsOpsPages_handleShortRead(App* app, struct inode* inode, struct page **pageArray,
   FhgfsPageVec* pageVec, int pgIdx, RemotingIOInfo* ioInfo, struct address_space* mapping,
   struct pagevec* lru_pvec)
{
   loff_t iSize;
   loff_t newIsize = -1;
   struct page* page;
   loff_t pageFileOffset;

   page = pageArray[pgIdx]; // use the idx of the first failed page
   pageFileOffset = page->index << PAGE_CACHE_SHIFT; // file-offset of that page

   iSize = i_size_read(inode);

   if (iSize > pageFileOffset)
   {
      // So we read less data than the size of the file should be, try to update the file size

      __FhgfsOps_refreshInode(app, inode, NULL);
      newIsize = i_size_read(inode);
   }

   // check if we need to handle a sparse file
   if (newIsize > pageFileOffset && pageVec->size >= pgIdx)
   {
      /* this will handle sparse files, but only till the end of the storage chunk. We do not
       * handle further chunks, as the kernel read-ahead code will retry unknown pages
       * (so the next chunk) automatically */

      struct StripePattern* pattern = ioInfo->pattern;
      int idx;

      // + 1 here as offsets start with zero
      size_t chunkBytesLeft =
         1 + StripePattern_getChunkEnd(pattern, pageFileOffset) - pageFileOffset;
      unsigned chunkPagesLeft = chunkBytesLeft / PAGE_SIZE;
      unsigned numSparsePages;

      numSparsePages = min(chunkPagesLeft, (unsigned) pageVec->size - pgIdx);

      for (idx = 0; idx < numSparsePages; idx++)
      {
         FhgfsPage* fhgfsPage;
         int sparseIdx = idx + pgIdx;
         struct page* page = pageArray[sparseIdx];
         Logger* log = App_getLogger(app);

         fhgfsPage = &(pageVec->pages[sparseIdx]);

         memset(fhgfsPage->data, 0, PAGE_SIZE);

         list_del(&page->lru);
         _FhgfsOpsPages_addToLRU(log, mapping, page, lru_pvec);
      }
   }
}


/**
 * Add the page to the LRU page cache
 */
static fhgfs_bool _FhgfsOpsPages_addToLRU(Logger* log, struct address_space* mapping,
   struct page* page, struct pagevec* lru_pvec)
{
   int addRes;

#if ((LINUX_VERSION_CODE < KERNEL_VERSION(2,6,30) ) )
   addRes = add_to_page_cache(page, mapping, page->index, GFP_KERNEL);
   if (addRes)
   {
      LOG_DEBUG_FORMATTED(log, 5, __func__, "add_to_page_cache failed");
      page_cache_release(page);
   }
   else
   {  // success
      if (!pagevec_add(lru_pvec, page))
         __pagevec_lru_add_file(lru_pvec); // no free space in lru_pvec left
   }
#else
   addRes = os_addToPageCacheLRU(page, mapping, page->index, GFP_NOFS);

   /* NOTE: When the page is allocated by mm/vfs read-ahead it also already has a reference.
    *       Those references are later on dropped, if we do not remove the page
    *       from list_head* pageList. However, we do remove pages from this list, if we
    *       successfully read data for it. Once we add the page to the page-cache and add it
    *       to the LRU list, this reference is also required, as the page otherwise might be
    *       used for other data, which would lead to data corruption.
    *       Now add_to_page_cache_lru() (actually the subsequent lru_cache_add_file() ) gets
    *       another reference for this page, but for page-cache and LRU we only need one
    *       reference, but do have two references at this point. So one references has to be
    *       dropped! Also, if add_to_page_cache_lru() fails, we need to drop all references to
    *       this page, in order to allow to re-use it for other purposes. However,
    *       add_to_page_cache_lru() can only fail for the call of add_to_page_cache() and at this
    *       point we still only have a single reference.
    *       So in the end we do need to call here page_cache_release(page) - if everything
    *       was fine to drop the 2nd reference, if something failed to drop our single reference.
    *       Code for kernels < 2.6.30 is different and only has to drop the reference if an
    *       error occurred (as pagevec_add() does not get a 2nd reference).
    */
   page_cache_release(page);

   if (addRes)
   {
      LOG_DEBUG_FORMATTED(log, 5, __func__, "add_to_page_cache failed");
      return fhgfs_false;
   }
#endif
   flush_dcache_page(page);
   SetPageUptodate(page);
   unlock_page(page); // locked by add_to_page_cache() and add_to_page_cache_lru()

   return fhgfs_true;
}


/**
 * Read list of pages
 *
 * NOTE: we MUST remove pages we successfully processed from struct list_head* pageList, as
 *       pages still in this list later on will be released from the page cache and due the
 *       mapping we added to those pages there will come up 'bad-pages'
 *
 * @return 0 on success, negative linux error code otherwise
 */
int FhgfsOps_readpagesVec(struct file* file, struct address_space* mapping,
   struct list_head* pageList, unsigned numPages)
{
   struct inode* inode = mapping->host;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   FsFileInfo* fileInfo = __FhgfsOps_getFileInfo(file);
   RemotingIOInfo ioInfo;
   struct page** pageArray;
   FhgfsPageVec pageVec;

   int retVal;
   int numContPages;

   unsigned maxPageArraySize = MIN(numPages, FHGFS_MAX_PAGE_LIST_SIZE);

   struct dentry *dentry = file->f_dentry;
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);

   size_t expectedReadSize;
   ssize_t readRes;
   ssize_t remainingReadRes;

   int pgIdx;
   loff_t offset;

   // only required for 2.6.30, but for code sanity we always initialize it
   struct pagevec lru_pvec;
   pagevec_init(&lru_pvec, 0);

   FhgfsOpsHelper_logOpDebug(app, dentry, __func__, "(num_pages: %u)", numPages);

   pageArray = vmalloc(sizeof(*pageArray) * maxPageArraySize);
   if (!pageArray)
      return -ENOMEM;

   pageVec.pages = vmalloc(sizeof (*pageVec.pages) * maxPageArraySize);
   if (!pageVec.pages)
   {
      retVal = -ENOMEM;
      goto outFreePageArray;
   }

   FsFileInfo_getIOInfo(fileInfo, fhgfsInode, &ioInfo);

   numContPages = _FhgfsOpsPages_pageListToVector(log, pageList, maxPageArraySize, pageArray);
   if (numContPages < 0)
   {
      retVal = numContPages;
      goto outFreePageVec;
   }

   // iterate over all continous pages and add those to our read vector
   for (pgIdx = 0; pgIdx < numContPages; pgIdx++)
   {
      struct page* page = pageArray[pgIdx];
      FhgfsPage* fhgfsPage = &pageVec.pages[pgIdx];

      fhgfsPage->data = kmap(page);
      fhgfsPage->length = PAGE_SIZE;
   }

   pageVec.size = numContPages;
   pageVec.offset = 0; // offset of the very first page of the vector
   offset = pageArray[0]->index << PAGE_CACHE_SHIFT; // offset within the file
   expectedReadSize = numContPages << PAGE_CACHE_SHIFT;

   readRes = FhgfsOpsRemoting_RWfileVec(&pageVec, offset, &ioInfo, FHGFS_RWTYPE_READ);
   if (unlikely(readRes < 0))
   {
      retVal = FhgfsOpsErr_toSysErr(-readRes);
      LOG_DEBUG_FORMATTED(log, 1, __func__, "error: %s", FhgfsOpsErr_toErrString(-readRes) );
      goto outUnMap;
   }

   LOG_DEBUG_FORMATTED(log, 5, __func__, "readRes: %d", readRes );

   /* readRes is the overall size we have read in. From that value we will subtract the size of
    * handled pages (need to add those to the lru-page-cache) and introduce remainingReadRes for
    * that. */
   remainingReadRes = readRes;
   pgIdx = 0;
   while (pgIdx < numContPages && remainingReadRes > 0)
   {
      struct page* page = pageArray[pgIdx];

      /* Remove the page from the list (struct list_head* pageList), to prevent read_pages()
       * from calling release_page() on pages we handled.
       * We can only do that here, once we know the read succeeded */
      list_del(&page->lru);

      if (remainingReadRes < (int)PAGE_SIZE)
      {
         // we dont't have data for a complete page left and need to zero the last bytes
         int zeroOffset = remainingReadRes;

         // zero_user_segment() would be optimal, but not available in older kernels
         // zero_user_segment(page, zeroOffset, FHGFS_PAGE_SIZE);
         // BUT we can use our kmapped Fhgfs_Page directly for zeroing
         FhgfsPage* fhgfsPage = &pageVec.pages[pgIdx];

         memset(fhgfsPage->data + zeroOffset, 0, PAGE_SIZE - remainingReadRes);
      }

      _FhgfsOpsPages_addToLRU(log, mapping, page, &lru_pvec);

      remainingReadRes -= PAGE_SIZE;
      pgIdx++;
   }

   if (unlikely(remainingReadRes > 0) )
   {
      Logger_logTopErrFormatted(log, 2, __func__,
         "Bug: More data read than expected (remainingReadRes (%llu) > 0, expected read size: %llu",
         (long long unsigned) remainingReadRes, (long long unsigned) expectedReadSize);
   }
   else
   if ( pgIdx < numContPages)
   {   // less pages read than expected
      _FhgfsOpsPages_handleShortRead(app, inode, pageArray, &pageVec, pgIdx, &ioInfo, mapping,
         &lru_pvec);
   }

   #if ((LINUX_VERSION_CODE < KERNEL_VERSION(2,6,30) ) )
      pagevec_lru_add_file(&lru_pvec); // not __pagevec_lru_add_file() here, as pvec might be empty
   #endif


   //printk_fhgfs_debug(KERN_INFO, "%s: pageList %p readRes: %d expectedReadSize %d\n",
   //   __func__, pageList, readRes, expectedReadSize);

  FhgfsOpsPages_incInodeFileSizeOnPagedRead(inode, offset, readRes);

   retVal = 0;

outUnMap:
   for (pgIdx = 0; pgIdx < numContPages; pgIdx++)
   {
      struct page* page = pageArray[pgIdx];
      kunmap(page);
   }

outFreePageVec:
   vfree(pageVec.pages);
outFreePageArray:
   vfree(pageArray);

   return retVal;
};
