#include <common/storage/StorageDefinitions.h>
#include <filesystem/FsDirInfo.h>
#include <common/toolkit/vector/StrCpyVec.h>
#include <common/toolkit/vector/IntCpyVec.h>


struct FsDirInfo
{
   FsObjectInfo fsObjectInfo;

   StrCpyVec dirContents;
   IntCpyVec dirContentsTypes; // DirEntryType elements matching dirContents vector
   int64_t serverOffset; // offset for the next incremental list request to the server
   uint64_t incrementalOffset; // seek to n-th elem offset; only used when serverOffset is -1;
   size_t currentContentsPos; // current local pos in dirContents (>=0 && <dirContents_len)
   fhgfs_bool endOfDir; // fhgfs_true if server reached end of dir entries during last query
};


void FsDirInfo_init(FsDirInfo* this, App* app)
{
   FsObjectInfo_init( (FsObjectInfo*)this, app, FsObjectType_DIRECTORY);

   StrCpyVec_init(&this->dirContents);
   IntCpyVec_init(&this->dirContentsTypes);
   this->serverOffset = 0;
   this->incrementalOffset = 0;
   this->currentContentsPos = 0;
   this->endOfDir = fhgfs_false;

   // assign virtual functions
   ( (FsObjectInfo*)this)->uninit = FsDirInfo_uninit;
}

struct FsDirInfo* FsDirInfo_construct(App* app)
{
   struct FsDirInfo* this =
      (FsDirInfo*)os_kmalloc(sizeof(*this) );

   FsDirInfo_init(this, app);

   return this;
}

void FsDirInfo_uninit(FsObjectInfo* this)
{
   FsDirInfo* thisCast = (FsDirInfo*)this;

   IntCpyVec_uninit(&thisCast->dirContentsTypes);
   StrCpyVec_uninit(&thisCast->dirContents);

   FsObjectInfo_uninit( (FsObjectInfo*)this);
}

void FsDirInfo_destruct(FsObjectInfo* this)
{
   FsDirInfo_uninit( (FsObjectInfo*)this);

   os_kfree(this);
}


loff_t FsDirInfo_getServerOffset(FsDirInfo* this)
{
   return this->serverOffset;
}

void FsDirInfo_setServerOffset(FsDirInfo* this, int64_t serverOffset)
{
   this->serverOffset = serverOffset;
}

loff_t FsDirInfo_getIncrementalOffset(FsDirInfo* this)
{
   return this->incrementalOffset;
}

void FsDirInfo_setIncrementalOffset(FsDirInfo* this, uint64_t incrementalOffset)
{
   this->incrementalOffset = incrementalOffset;
}

size_t FsDirInfo_getCurrentContentsPos(FsDirInfo* this)
{
   return this->currentContentsPos;
}

void FsDirInfo_setCurrentContentsPos(FsDirInfo* this, size_t currentContentsPos)
{
   this->currentContentsPos = currentContentsPos;
}

StrCpyVec* FsDirInfo_getDirContents(FsDirInfo* this)
{
   return &this->dirContents;
}

/**
 * @return vector of DirEntryType elements, matching dirContents vector
 */
IntCpyVec* FsDirInfo_getDirContentsTypes(FsDirInfo* this)
{
   return &this->dirContentsTypes;
}

void FsDirInfo_setEndOfDir(FsDirInfo* this, fhgfs_bool endOfDir)
{
   this->endOfDir = endOfDir;
}

fhgfs_bool FsDirInfo_getEndOfDir(FsDirInfo* this)
{
   return this->endOfDir;
}

