#include <app/App.h>
#include <common/net/message/session/rw/ReadLocalFileV2Msg.h>
#include <common/net/message/session/rw/WriteLocalFileMsg.h>
#include <common/net/message/session/rw/WriteLocalFileRespMsg.h>
#include <common/toolkit/MessagingTk.h>
#include <common/toolkit/VersionTk.h>
#include <common/storage/StorageErrors.h>
#include <common/threading/Thread.h>

#define FhgfsOpsCommKitVec
#include "FhgfsOpsCommKitVec.h"
#include <common/FhgfsTypes.h>


/**
 * @param pollEvents the events to poll for (e.g. FHGFS_POLLIN or FHGFS_POLLOUT)
 */
static void __FhgfsOpsCommKitVec_addStateSockToPollSocks(
   RWfileStatesHelper* statesHelper, FileStateVec* curState, short pollEvents)
{
   (statesHelper->pollSocks)[statesHelper->numPollSocks] =
      __FhgfsOpsCommKitCommon_assignPollSocket(curState->sock, pollEvents);

   curState->contPollSocksIndex = statesHelper->numPollSocks;
   statesHelper->numPollSocks++;
}

/**
 * Prepare a read request
 *
 * @return fhgfs_false on error (means 'break' for the surrounding switch statement)
 */
fhgfs_bool __FhgfsOpsCommKitVec_readfileStagePREPARE(RWfileStatesHelper* statesHelper,
   FileStateVec* curState)
{
   TargetMapper* targetMapper = App_getTargetMapper(statesHelper->app);

   FhgfsOpsErr resolveErr;
   NodeConnPool* connPool;
   ReadLocalFileV2Msg readMsg;
   fhgfs_bool supportsSessionCheck;
   fhgfs_bool allowWaitForConn = !statesHelper->numAcquiredConns; // don't wait if we got at least
      // one conn already (this is important to avoid a deadlock between racing commkit processes)

   curState->sock = NULL;
   curState->nodeResult = -FhgfsOpsErr_COMMUNICATION;

   // get the target-node reference
   curState->node = NodeStoreEx_referenceNodeByTargetID(statesHelper->nodes,
      curState->targetID, targetMapper, &resolveErr);
   if(unlikely(!curState->node) )
   { // unable to resolve targetID
      curState->nodeResult = -resolveErr;
      curState->type.read.stage = ReadfileStage_CLEANUP;
      return fhgfs_false;
   }

   supportsSessionCheck = VersionTk_checkRequiredRelease(VERSIONTK_REQUIRED_VERSION_2012_10_R2,
      Node_getFhgfsVersion(curState->node));

   connPool = Node_getConnPool(curState->node);

   // connect
   curState->sock = NodeConnPool_acquireStreamSocketEx(connPool, allowWaitForConn);
   if(!curState->sock)
   { // no conn available => error or didn't want to wait
      if(likely(!allowWaitForConn) )
      { // just didn't want to wait => keep stage and try again later
         NodeStoreEx_releaseNode(statesHelper->nodes, &curState->node);

         statesHelper->numUnconnectable++;
      }
      else
      { // connection error
         if(!statesHelper->connFailedLogged)
         { // no conn error logged yet
            if(Thread_isSignalPending() )
               Logger_logFormatted(statesHelper->log, Log_DEBUG, statesHelper->logContext,
                  "Connect to server canceled by pending signal: %s",
                  Node_getNodeIDWithTypeStr(curState->node) );
            else
               Logger_logFormatted(statesHelper->log, Log_WARNING, statesHelper->logContext,
                  "Unable to connect to server: %s",
                  Node_getNodeIDWithTypeStr(curState->node) );
         }

         statesHelper->connFailedLogged = fhgfs_true;
         curState->type.read.stage = ReadfileStage_CLEANUP;
      }

      // in either case return fhgfs_false to avoid falling through to send stage
      return fhgfs_false;
   }

#if 0
   LOG_DEBUG_FORMATTED(statesHelper->log, Log_SPAM, __func__,
      "Acquire read state: %d node: %s numAcquiredConns %d",
      curState->stateIdx, Node_getNodeIDWithTypeStr(curState->node),
      statesHelper->numAcquiredConns);
#endif

   statesHelper->numAcquiredConns++;


   // prepare message
   ReadLocalFileV2Msg_initFromSession(&readMsg, statesHelper->localNodeID,
      curState->fileHandleID, curState->targetID, curState->accessFlags,
      curState->offset, curState->dataSize);

   if (curState->firstWriteDoneForTarget && supportsSessionCheck)
      NetMessage_addMsgHeaderFeatureFlag( (NetMessage*)&readMsg,
         WRITELOCALFILEMSG_FLAG_SESSION_CHECK);

   ReadLocalFileV2Msg_setDisableIO(&readMsg, App_getNetBenchModeEnabled(statesHelper->app) );

   NetMessage_serialize( (NetMessage*)&readMsg, curState->msgBuf, FHGFS_COMMKIT_MSGBUF_SIZE);

   curState->hdrLen = NetMessage_getMsgLength( (NetMessage*)&readMsg);

   ReadLocalFileV2Msg_uninit( (NetMessage*)&readMsg);

   curState->nodeResult = 0; // ready to read, so set this variable to 0
   curState->type.read.stage = ReadfileStage_SEND;

   return fhgfs_true;
}



/**
 * Tell the server what which part of the file it shall send
 */
void __FhgfsOpsCommKitVec_readfileStageSEND(RWfileStatesHelper* statesHelper,
   FileStateVec* curState)
{
   ssize_t sendRes;

   sendRes = curState->sock->send(
      curState->sock, curState->msgBuf, curState->hdrLen, 0);

   if(unlikely(sendRes <= 0) )
   {
      Logger_logFormatted(statesHelper->log, Log_WARNING, statesHelper->logContext,
         "Failed to send message to %s: %s", Node_getNodeIDWithTypeStr(curState->node),
         Socket_getPeername(curState->sock) );

      curState->type.read.stage = ReadfileStage_SOCKETEXCEPTION;
      return;
   }

   curState->type.read.stage = ReadfileStage_RECVHEADER;

   // prepare polling for the next stage
   __FhgfsOpsCommKitVec_addStateSockToPollSocks(statesHelper, curState, FHGFS_POLLIN);
}

/**
 * Receive how many data the server is going to send next, this is also called after each
 * *complete* read request
 */
void __FhgfsOpsCommKitVec_readfileStageRECVHEADER(RWfileStatesHelper* statesHelper,
   FileStateVec* curState)
{
   ssize_t recvRes;
   char dataLenBuf[sizeof(int64_t)]; // length info in fhgfs network byte order
   int64_t lengthInfo; // length info in fhgfs host byte order
   unsigned lengthInfoBufLen; // deserialization buf length for lengthInfo (ignored)

   // check for incoming data
   if(!__FhgfsOpsCommKitVec_readFileIsDataAvailable(statesHelper, curState) )
      return;

   recvRes = Socket_recvExactT(
      curState->sock, &dataLenBuf, sizeof(int64_t), 0, CONN_LONG_TIMEOUT);

   if(unlikely(recvRes <= 0) )
   { // error
      Logger_logFormatted(statesHelper->log, Log_WARNING, statesHelper->logContext,
         "Failed to receive the length info response from %s: %s",
         Node_getNodeIDWithTypeStr(curState->node), Socket_getPeername(curState->sock) );

      curState->type.read.stage = ReadfileStage_SOCKETEXCEPTION;
      return;
   }

   // got the length info response
   Serialization_deserializeInt64(dataLenBuf, sizeof(int64_t), &lengthInfo, &lengthInfoBufLen);

    // LOG_DEBUG_FORMATTED(statesHelper->log, 5, statesHelper->logContext,
    //  "Received lengthInfo from %s: %lld", Node_getID(curState->node), (long long)lengthInfo);

   if(lengthInfo <= 0)
   { // end of file data transmission

      if(unlikely(lengthInfo < 0) )
         curState->nodeResult = lengthInfo; // error occurred

      curState->type.read.stage = ReadfileStage_CLEANUP;
      return;
   }

   // positive result => node is going to send some file data
   curState->type.read.sendDataLength = lengthInfo;
   curState->type.read.receivedLength = 0;
   curState->type.read.stage = ReadfileStage_RECVDATA;
}



/**
 * Read the data from the server
 */
void __FhgfsOpsCommKitVec_readfileStageRECVDATA(RWfileStatesHelper* statesHelper,
   FileStateVec* curState)
{
   ssize_t recvRes;
   int pgIdx;
   fhgfs_bool needNewRound = fhgfs_false;

   // check for incoming data
   if(!__FhgfsOpsCommKitVec_readFileIsDataAvailable(statesHelper, curState) )
      return;

   if (unlikely(curState->contLastPage >= curState->pageVec.size) )
   {
      Logger_logErrFormatted(statesHelper->log, statesHelper->logContext,
         "BUG: lastPage (%d) >= maxPages (%d)", curState->contLastPage, curState->pageVec.size);

      curState->type.read.stage = ReadfileStage_SOCKETEXCEPTION;
      return;
   }

   // walk over all pages of this state
   pgIdx = curState->contLastPage;  // initially (first net transfer round) contLastPage is 0
   do {
      FhgfsPageVec* pageVec = &curState->pageVec;
      FhgfsPage *page = &pageVec->pages[pgIdx];
      void* pageDataPtr;
      int pgOffset = 0;
      Socket* sock = curState->sock;

      size_t missingPgLen;   // length of a page
      size_t requestLength;

      if (curState->contPageOffset)
      {
         pgOffset = curState->contPageOffset;
         curState->contPageOffset = 0; // reset for the next page
      }

#if 0
      LOG_DEBUG_FORMATTED(statesHelper->log, Log_SPAM, __func__,
         "stateIdx: %d; size: %lld; offset: %lld; "
         "vecSize: %d Vec-Idx: %d pgOffset %d receivedData: %ld",curState->stateIdx,
         (long long)curState->dataSize,
         (long long)curState->offset, curState->pageVec.size, pgIdx, pgOffset,
         curState->type.read.receivedLength);
#endif

      missingPgLen = page->length - pgOffset;
      pageDataPtr = page->data + pgOffset;

      requestLength = MIN(missingPgLen,
         curState->type.read.sendDataLength - curState->type.read.receivedLength);

      // receive available dataPart
      recvRes = curState->sock->recvT(sock, pageDataPtr, requestLength, 0, CONN_LONG_TIMEOUT);

      // LOG_DEBUG_FORMATTED(statesHelper->log, 5, __func__, "pgIdx: %d; requested: %lld; "
      //    "received: %lld", pgIdx, (long long)requestLength, (long long)recvRes);

      if(unlikely(recvRes <= 0) )
      { // something went wrong
         if(recvRes == -EFAULT)
         { // bad buffer address given
            Logger_logFormatted(statesHelper->log, Log_DEBUG, statesHelper->logContext,
               "Bad buffer address");

            curState->nodeResult = -FhgfsOpsErr_ADDRESSFAULT;
            curState->type.read.stage = ReadfileStage_SOCKETINVALIDATE;
            return;
         }
         else
         if(recvRes < 0)
         { // error
            Logger_logErrFormatted(statesHelper->log, statesHelper->logContext,
               "Communication error in RECVDATA stage. Node: %s (SysErr: %d)",
               Node_getNodeIDWithTypeStr(curState->node), (long long)recvRes);
         }
         else
         if(!recvRes)
         { // timeout
            Logger_logErrFormatted(statesHelper->log, statesHelper->logContext,
               "Communication timeout in RECVDATA stage. Node: %s",
               Node_getNodeIDWithTypeStr(curState->node) );
         }

         Logger_logFormatted(statesHelper->log, Log_SPAM, statesHelper->logContext,
            "Request details: missingLength of %s: %lld", Node_getNodeIDWithTypeStr(curState->node),
            (long long)missingPgLen);

         curState->type.read.stage = ReadfileStage_SOCKETEXCEPTION;
         return;
      }

      curState->type.read.receivedLength += recvRes;

      if (recvRes < (ssize_t) requestLength)
      {
         /* save already submitted data, so that we can continue there in the next round with the
          * current page */

         // LOG_DEBUG_FORMATTED(statesHelper->log, 5, __func__, "Less data received than requested "
         //   "(%ld vs. %ld) pgIdx: %d", recvRes, requestLength, pgIdx);
         curState->contLastPage = pgIdx;
         curState->contPageOffset = pgOffset + recvRes;
         needNewRound = fhgfs_true;
         break;
      }
      else
      if (requestLength < missingPgLen)
      {
         /* This should only happen if the server has less data available than requested or if
          * server side "tuneFileReadSize" is not a multiple of
          * the client side page size or if there is a bug somewhere here in the code.
          * In general tuneFileReadSize != n * PAGE_SIZE  it is not well tested code and should
          * be avoided */

         // also continue with the same page, but needNewRound=false
         curState->contLastPage = pgIdx;
         curState->contPageOffset = pgOffset + recvRes;
         break;
      }
      else
      { // all ok, continue with the next page
         pgIdx++;
         curState->contLastPage = pgIdx;
      }

   } while (pgIdx < curState->pageVec.size &&
      curState->type.read.receivedLength < curState->type.read.sendDataLength);

   if (unlikely(pgIdx == curState->pageVec.size &&
      (curState->type.read.sendDataLength - curState->type.read.receivedLength  > 0) ) )
   {
      // server wants to send more data than we have available pages
      curState->nodeResult = -FhgfsOpsErr_INTERNAL;
      curState->type.read.stage = ReadfileStage_SOCKETINVALIDATE;

      Logger_logErrFormatted(statesHelper->log, statesHelper->logContext,
         "Bug: Server wants to send more data than we requested: %ld",
         curState->type.read.sendDataLength - curState->type.read.receivedLength);
   }

   if(!needNewRound)
   {
      // all data for this state and round received, add it up
      curState->nodeResult += curState->type.read.receivedLength;

      // all of the data has been received => receive the next lengthInfo in the header stage
      curState->type.read.stage = ReadfileStage_RECVHEADER;
   }

   // prepare polling for the next round (no matter if we received all of the data or not)
   __FhgfsOpsCommKitVec_addStateSockToPollSocks(statesHelper, curState, FHGFS_POLLIN);
}


/**
 * Checks for immediately available incoming data.
 *
 * @return fhgfs_true if incoming data is available for this state
 */
static fhgfs_bool __FhgfsOpsCommKitVec_readFileIsDataAvailable(RWfileStatesHelper* statesHelper,
   FileStateVec* curState)
{
   // check for a poll() timeout or error (all states have to be cancelled in that case)
   if(unlikely(statesHelper->pollTimedOut) )
   {
      //Logger_logFormatted(statesHelper->log, 2, statesHelper->logContext,
      //   "Canceling due to communication error: %s (%s)",
      //   curState->nodeID, Socket_getPeername(curState->sock) );

      curState->nodeResult = -FhgfsOpsErr_COMMUNICATION;
      curState->type.read.stage = ReadfileStage_SOCKETINVALIDATE;
      return fhgfs_false;
   }


   // check for immediately available data
   if(!(statesHelper->pollSocks)[curState->contPollSocksIndex].revents)
   { // no data available yet => wait for the next round
      __FhgfsOpsCommKitVec_addStateSockToPollSocks(statesHelper, curState, FHGFS_POLLIN);

      return fhgfs_false;
   }

   // incoming data is available

   return fhgfs_true;
}

/**
 * Socket exception happened
 */
void __FhgfsOpsCommKitVec_readfileStageSOCKETEXCEPTION(RWfileStatesHelper* statesHelper,
   FileStateVec* curState)
{
   const char* logCommInterruptedTxt = "Communication interrupted by pending signal";
   const char* logCommErrTxt = "Communication error (SocketException)";

   if(Thread_isSignalPending() )
   { // interrupted by signal
      Logger_logFormatted(statesHelper->log, Log_NOTICE, statesHelper->logContext,
         "%s. Node: %s", logCommInterruptedTxt, Node_getNodeIDWithTypeStr(curState->node) );
   }
   else
   { // "normal" connection error
      Logger_logErrFormatted(statesHelper->log, statesHelper->logContext,
         "%s. Node: %s", logCommErrTxt, Node_getNodeIDWithTypeStr(curState->node) );

      Logger_logFormatted(statesHelper->log, Log_DEBUG, statesHelper->logContext,
         "Sent request: node: %s; fileHandleID: %s; offset: %lld; size: %lld",
         Node_getNodeIDWithTypeStr(curState->node), curState->fileHandleID,
         (long long)curState->offset, (long long)curState->dataSize);
   }

   curState->nodeResult = -FhgfsOpsErr_COMMUNICATION;

   curState->type.read.stage = ReadfileStage_SOCKETINVALIDATE;
}

/**
 * Invalidate our socket
 */
void __FhgfsOpsCommKitVec_readfileStageSOCKETINVALIDATE(RWfileStatesHelper* statesHelper,
   FileStateVec* curState)
{
   NodeConnPool_invalidateStreamSocket(Node_getConnPool(curState->node), curState->sock);
   curState->sock = NULL;
   statesHelper->numAcquiredConns--;

   curState->type.read.stage = ReadfileStage_CLEANUP;
}


/**
 * Clean up the current state after an error or too few data.
 */
void __FhgfsOpsCommKitVec_readfileStageCLEANUP(RWfileStatesHelper* statesHelper,
   FileStateVec* curState)
{
   // actual clean-up

   if (curState->sock)
   {
#if 0
      LOG_DEBUG_FORMATTED(statesHelper->log, Log_SPAM, __func__,
         "Release read state: %d node: %s numAcquiredConns %d",
         curState->stateIdx, Node_getNodeIDWithTypeStr(curState->node),
         statesHelper->numAcquiredConns);
#endif

      NodeConnPool_releaseStreamSocket(Node_getConnPool(curState->node), curState->sock);
      curState->sock = NULL;
      statesHelper->numAcquiredConns--;
   }

   if(likely(curState->node) )
      NodeStoreEx_releaseNode(statesHelper->nodes, &curState->node);

   // prepare next stage

   if(unlikely(curState->nodeResult == -FhgfsOpsErr_COMMUNICATION) )
   { // comm error occurred => check whether we can do a retry

      if(Thread_isSignalPending() )
      { // interrupted by signal => no retry
         curState->nodeResult = -FhgfsOpsErr_INTERRUPTED;
      }
      else
      if( App_getConnRetriesEnabled(statesHelper->app) &&
          ( (!statesHelper->maxNumRetries) ||
            (statesHelper->currentRetryNum < statesHelper->maxNumRetries) ) )
      {
         // we have retries left
         __FhgfsOpsCommKitVec_readfilePrepareRetry(statesHelper, curState);

         return;
      }
   }

   // success or no retries left => done

   statesHelper->numDone++;
   curState->type.read.stage = ReadfileStage_DONE;
}

/**
 * Reset a state for a retry and tell put it into wait state
 */
static void __FhgfsOpsCommKitVec_readfilePrepareRetry(RWfileStatesHelper* statesHelper,
   FileStateVec* curState)
{
   statesHelper->numRetryWaiters++;

   curState->type.read.stage = ReadfileStage_RETRYWAIT;

   // reset for another full chunk transfer
   curState->contLastPage   = 0;
   curState->contPageOffset = 0;
   curState->nodeResult     = -EREMOTEIO;
}

/**
 * We have waited a bit after an error, now set required states into another read state
 */
static void __FhgfsOpsCommKitVec_readfileStartRetry(RWfileStatesHelper* statesHelper,
   FileStateVec* states, size_t numStates)
{
   size_t i;

   // set the new stage for each retry waiter and reset number of waiters

   for(i = 0; i < numStates; i++)
   {
      FileStateVec* curState = &states[i];

      if(curState->type.read.stage == ReadfileStage_RETRYWAIT)
         curState->type.read.stage =  ReadfileStage_PREPARE;
   }

   statesHelper->numRetryWaiters = 0;

   statesHelper->pollTimedOut = fhgfs_false;


   // wait before we actually start the retry

   MessagingTk_waitBeforeRetry(statesHelper->currentRetryNum);

   statesHelper->currentRetryNum++;
}




/**
 * @return fhgfs_false on error (means 'break' for the surrounding switch statement)
 */
fhgfs_bool __FhgfsOpsCommKitVec_writefileStagePREPARE(RWfileStatesHelper* statesHelper,
   FileStateVec* curState)
{
   Config* cfg = App_getConfig(statesHelper->app);

   TargetMapper* targetMapper = App_getTargetMapper(statesHelper->app);

   FhgfsOpsErr resolveErr;
   NodeConnPool* connPool;
   WriteLocalFileMsg writeMsg;
   fhgfs_bool supportsSessionCheck;
   fhgfs_bool supportsQuota;
   fhgfs_bool allowWaitForConn = !statesHelper->numAcquiredConns; // don't wait if we got at least
      // one conn already (this is important to avoid a deadlock between racing commkit processes)

   curState->sock = NULL;
   curState->nodeResult = -FhgfsOpsErr_COMMUNICATION;

   // get the target-node reference
   curState->node = NodeStoreEx_referenceNodeByTargetID(statesHelper->nodes,
      curState->targetID, targetMapper, &resolveErr);
   if(unlikely(!curState->node) )
   { // unable to resolve targetID
      curState->nodeResult = -resolveErr;
      curState->type.write.stage = WritefileStage_CLEANUP;
      return fhgfs_false;
   }

   supportsSessionCheck = VersionTk_checkRequiredRelease(VERSIONTK_REQUIRED_VERSION_2012_10_R2,
      Node_getFhgfsVersion(curState->node));

   supportsQuota = (Config_getQuotaEnabled(cfg) && VersionTk_checkRequiredRelease(
      VERSIONTK_REQUIRED_VERSION_QUOTA_SUPPORT, Node_getFhgfsVersion(curState->node) ) );

   connPool = Node_getConnPool(curState->node);

   // connect
   curState->sock = NodeConnPool_acquireStreamSocketEx(connPool, allowWaitForConn);
   if(!curState->sock)
   { // no conn available => error or didn't want to wait
      if(likely(!allowWaitForConn) )
      { // just didn't want to wait => keep stage and try again later
         NodeStoreEx_releaseNode(statesHelper->nodes, &curState->node);

         statesHelper->numUnconnectable++;
      }
      else
      { // connection error
         if(!statesHelper->connFailedLogged)
         { // no conn error logged yet
            if(Thread_isSignalPending() )
               Logger_logFormatted(statesHelper->log, Log_DEBUG, statesHelper->logContext,
                  "Connect to server canceled by pending signal: %s",
                  Node_getNodeIDWithTypeStr(curState->node) );
            else
               Logger_logFormatted(statesHelper->log, Log_WARNING, statesHelper->logContext,
                  "Unable to connect to server: %s",
                  Node_getNodeIDWithTypeStr(curState->node) );
         }

         statesHelper->connFailedLogged = fhgfs_true;
         curState->type.write.stage = WritefileStage_CLEANUP;
      }

      // in either case return fhgfs_false to avoid falling through to send stage
      return fhgfs_false;
   }

#if 0
   LOG_DEBUG_FORMATTED(statesHelper->log, Log_SPAM, __func__,
      "Acquire write state: %d node: %s numAcquiredConns %d",
      curState->stateIdx, Node_getNodeIDWithTypeStr(curState->node),
      statesHelper->numAcquiredConns);
#endif

   statesHelper->numAcquiredConns++;


   // prepare message
   WriteLocalFileMsg_initFromSession(&writeMsg, statesHelper->localNodeID,
      curState->fileHandleID, curState->targetID, curState->accessFlags,
      curState->offset, curState->dataSize);

   if (supportsQuota)
      WriteLocalFileMsg_setUserdataForQuota(&writeMsg, curState->type.write.userID,
         curState->type.write.groupID);

   if (curState->firstWriteDoneForTarget && supportsSessionCheck)
      NetMessage_addMsgHeaderFeatureFlag( (NetMessage*)&writeMsg,
         WRITELOCALFILEMSG_FLAG_SESSION_CHECK);

   WriteLocalFileMsg_setMirrorToTargetID(&writeMsg, curState->mirrorTargetID);
   WriteLocalFileMsg_setDisableIO(&writeMsg, App_getNetBenchModeEnabled(statesHelper->app) );

   NetMessage_serialize( (NetMessage*)&writeMsg, curState->msgBuf, FHGFS_COMMKIT_MSGBUF_SIZE);

   curState->hdrLen = NetMessage_getMsgLength( (NetMessage*)&writeMsg);

   WriteLocalFileMsg_uninit( (NetMessage*)&writeMsg);

   curState->type.write.stage = WritefileStage_SENDHEADER;

   return fhgfs_true;
}

/**
 * Tell the server how many data we are going to send
 */
void __FhgfsOpsCommKitVec_writefileStageSENDHEADER(RWfileStatesHelper* statesHelper,
   FileStateVec* curState)
{
   ssize_t sendRes;

   sendRes = curState->sock->send(
      curState->sock, curState->msgBuf, curState->hdrLen, 0);

   if(unlikely(sendRes <= 0) )
   {
      Logger_logFormatted(statesHelper->log, Log_WARNING, statesHelper->logContext,
         "Failed to send message to %s: %s", Node_getNodeIDWithTypeStr(curState->node),
         Socket_getPeername(curState->sock) );

      curState->type.write.stage = WritefileStage_SOCKETEXCEPTION;
      return;
   }

   curState->type.write.stage = WritefileStage_SENDDATA;

   // prepare polling for the next stage
   __FhgfsOpsCommKitVec_addStateSockToPollSocks(statesHelper, curState, FHGFS_POLLOUT);
}

/**
 * Send the data (of our current state)
 *
 * @param curState  - current state to send data for
 */
void __FhgfsOpsCommKitVec_writefileStageSENDDATA(RWfileStatesHelper* statesHelper,
   FileStateVec* curState)
{
   FhgfsOpsErr sendDataPartRes;
   int pgIdx;
   fhgfs_bool needNewRound = fhgfs_false;

   // check for a poll() timeout or error
   if(unlikely(statesHelper->pollTimedOut) )
   {
      //Logger_logFormatted(statesHelper->log, 2, statesHelper->logContext,
      //   "Canceling due to communication error: %s (%s)",
      //   curState->nodeID, Socket_getPeername(curState->sock) );

      curState->nodeResult = -FhgfsOpsErr_COMMUNICATION;
      curState->type.write.stage = WritefileStage_SOCKETINVALIDATE;
      return;
   }

   // check if we can send data non-blocking
   if(!(statesHelper->pollSocks)[curState->contPollSocksIndex].revents)
   { // not ready for sending yet => wait for the next round
      __FhgfsOpsCommKitVec_addStateSockToPollSocks(
         statesHelper, curState, FHGFS_POLLOUT);

      return;
   }

   if (unlikely(curState->contLastPage >= curState->pageVec.size) )
   {
      Logger_logErrFormatted(statesHelper->log, statesHelper->logContext,
         "BUG: %s: lastPage >= maxPages", curState->contLastPage, curState->pageVec.size);

      curState->type.write.stage = WritefileStage_SOCKETEXCEPTION;
      return;
   }

   // walk over all pages of this state
   pgIdx = curState->contLastPage;  // initially (first net transfer round) contLastPage is 0

#if 0
   LOG_DEBUG_FORMATTED(statesHelper->log, 5, __func__, "stateIdx: %d; size: %lld; vecOffset: %lld "
      "vecSize: %d, pgIdx: %d",
      curState->stateIdx, (long long)curState->dataSize, (long long)curState->offset,
      curState->pageVec.size, pgIdx);
#endif

   do {
      FhgfsPageVec* pageVec = &curState->pageVec;
      FhgfsPage *page = &pageVec->pages[pgIdx];
      void* data;
      int offset = 0;

      size_t dataLength;
      size_t transmitted = 0; /* how many data have been really transmitted, NOTE: has to be
                               * zero initialized, as WriteLocalFileMsg_sendDataPartInstant()
                               * presently uses it as in/out value, but we already that pointer
                               * magic ourself above */
      if (curState->contPageOffset)
      {
         offset = curState->contPageOffset;
         curState->contPageOffset = 0; // reset for the next page
      }

      dataLength = page->length - offset;
      data = page->data + offset;

      // send dataPart non-blocking
      sendDataPartRes = WriteLocalFileMsg_sendDataPartInstant(statesHelper->app, data,
         dataLength, &transmitted, curState->sock);

      // LOG_DEBUG_FORMATTED(statesHelper->log, 5, __func__,
      //    "state %d pgIdx: %d; size: %lld; transmitted: %lld PgLen: %d",
      //    curState->stateIdx, pgIdx, (long long)dataLength, (long long)transmitted, page->length);

      if (unlikely(sendDataPartRes != FhgfsOpsErr_SUCCESS) )
      {
         if(sendDataPartRes == FhgfsOpsErr_ADDRESSFAULT)
         { // bad buffer address given
            Logger_logFormatted(statesHelper->log, Log_DEBUG, statesHelper->logContext,
               "Bad buffer address");

            curState->nodeResult = -FhgfsOpsErr_ADDRESSFAULT;
            curState->type.write.stage = WritefileStage_SOCKETINVALIDATE;
            return;
         }

         Logger_logErrFormatted(statesHelper->log, statesHelper->logContext,
            "Communication error in SENDDATA stage. "
            "Node: %s; Request details: %lld/%lld pages",
            Node_getNodeIDWithTypeStr(curState->node), curState->contLastPage,
            curState->pageVec.size);

         curState->type.write.stage = WritefileStage_SOCKETEXCEPTION;
         return;
      }

      if (transmitted < dataLength)
      {
         // save already submitted data, so that we can continue there in the next round
         curState->contLastPage = pgIdx;
         curState->contPageOffset = transmitted;
         needNewRound = fhgfs_true;
         break;
      }

      pgIdx++;
   } while (pgIdx < curState->pageVec.size);


   if(needNewRound)
   {  // there is still data to be sent => prepare pollout for the next round
      __FhgfsOpsCommKitVec_addStateSockToPollSocks(statesHelper, curState, FHGFS_POLLOUT);
   }
   else
   {  // all of the data has been sent => proceed to the next stage
      curState->type.write.stage = WritefileStage_RECV;
      __FhgfsOpsCommKitVec_addStateSockToPollSocks(statesHelper, curState, FHGFS_POLLIN);
   }
}


/**
 * Ask the server side for errors
 */
void __FhgfsOpsCommKitVec_writefileStageRECV(RWfileStatesHelper* statesHelper,
   FileStateVec* curState)
{
   ssize_t respRes;

   // check for incoming data
   if(!__FhgfsOpsCommKitVec_writeFileIsDataAvailable(statesHelper, curState) )
      return;

   // incoming data available

   respRes = MessagingTk_recvMsgBuf(statesHelper->app, curState->sock,
      curState->msgBuf, FHGFS_COMMKIT_MSGBUF_SIZE);

   if(unlikely(respRes <= 0) )
   { // error
      Logger_logFormatted(statesHelper->log, Log_WARNING, statesHelper->logContext,
         "Failed to receive response from: %s (%s)", Node_getNodeIDWithTypeStr(curState->node),
         Socket_getPeername(curState->sock) );

      curState->type.write.stage = WritefileStage_SOCKETEXCEPTION;
      return;
   }
   else
   { // got response => deserialize it
      fhgfs_bool deserRes;

      WriteLocalFileRespMsg_init(&curState->type.write.respMsg);

      deserRes = NetMessageFactory_deserializeFromBuf(curState->msgBuf, respRes,
         (NetMessage*)&curState->type.write.respMsg, NETMSGTYPE_WriteLocalFileResp);

      if(unlikely(!deserRes) )
      { // response invalid
         Logger_logFormatted(statesHelper->log, Log_WARNING, statesHelper->logContext,
            "Received invalid response from %s. Expected type: %d. Disconnecting: %s",
            Node_getNodeIDWithTypeStr(curState->node), NETMSGTYPE_WriteLocalFileResp,
            Socket_getPeername(curState->sock) );

         curState->type.write.stage = WritefileStage_SOCKETINVALIDATE;
      }
      else
      { // correct response
         int64_t writeRespValue = WriteLocalFileRespMsg_getValue(&curState->type.write.respMsg);

         LOG_DEBUG_FORMATTED(statesHelper->log, 5, __func__, "writeRespValue: %lld",
            (long long) writeRespValue);

         curState->nodeResult = writeRespValue;

         curState->type.write.stage = WritefileStage_CLEANUP;
      }

      WriteLocalFileRespMsg_uninit( (NetMessage*)&curState->type.write.respMsg);
   }

}

/**
 * Socket exception happend
 */
void __FhgfsOpsCommKitVec_writefileStageSOCKETEXCEPTION(RWfileStatesHelper* statesHelper,
   FileStateVec* curState)
{
   const char* logCommInterruptedTxt = "Communication interrupted by pending signal";
   const char* logCommErrTxt = "Communication error (SocketException)";

   if(Thread_isSignalPending() )
   { // interrupted by signal
      Logger_logFormatted(statesHelper->log, Log_NOTICE, statesHelper->logContext,
         "%s. Node: %s", logCommInterruptedTxt, Node_getNodeIDWithTypeStr(curState->node) );
   }
   else
   { // "normal" connection error
      Logger_logErrFormatted(statesHelper->log, statesHelper->logContext,
         "%s. Node: %s", logCommErrTxt, Node_getNodeIDWithTypeStr(curState->node) );

      Logger_logFormatted(statesHelper->log, Log_DEBUG, statesHelper->logContext,
         "Sent request: node: %s; fileHandleID: %s; offset: %lld; size: %lld",
         Node_getNodeIDWithTypeStr(curState->node), curState->fileHandleID,
         (long long)curState->offset, (long long)curState->dataSize);
   }

   curState->nodeResult = -FhgfsOpsErr_COMMUNICATION;

   curState->type.write.stage = WritefileStage_SOCKETINVALIDATE;
}

/**
 * Invalidate the socket
 */
void __FhgfsOpsCommKitVec_writefileStageSOCKETINVALIDATE(RWfileStatesHelper* statesHelper,
   FileStateVec* curState)
{
   NodeConnPool_invalidateStreamSocket(Node_getConnPool(curState->node), curState->sock);
   curState->sock = NULL;
   statesHelper->numAcquiredConns--;

   curState->type.write.stage = WritefileStage_CLEANUP;
}

/**
 * Clean up a connection, if there was an error
 */
void __FhgfsOpsCommKitVec_writefileStageCLEANUP(RWfileStatesHelper* statesHelper,
   FileStateVec* curState)
{
   if (curState->sock)
   {
#if 0
      LOG_DEBUG_FORMATTED(statesHelper->log, 5, __func__,
         "Release write state: %d node: %s numAcquiredConns %d",
         curState->stateIdx, curState->node->id, statesHelper->numAcquiredConns);
#endif

      NodeConnPool_releaseStreamSocket(Node_getConnPool(curState->node), curState->sock);
      curState->sock = NULL;
      statesHelper->numAcquiredConns--;
   }

   if(likely(curState->node) )
      NodeStoreEx_releaseNode(statesHelper->nodes, &curState->node);

   // prepare next stage

   if(unlikely(curState->nodeResult == -FhgfsOpsErr_COMMUNICATION) )
   { // comm error occurred => check whether we can do a retry

      if(Thread_isSignalPending() )
      { // interrupted by signal => no retry
         curState->nodeResult = -FhgfsOpsErr_INTERRUPTED;
      }
      else
      if( App_getConnRetriesEnabled(statesHelper->app) &&
          ( (!statesHelper->maxNumRetries) ||
            (statesHelper->currentRetryNum < statesHelper->maxNumRetries) ) )
      {
         // we have retries left
         __FhgfsOpsCommKitVec_writefilePrepareRetry(statesHelper, curState);
         return;
      }
   }

   // success or no retries left => done
   statesHelper->numDone++;
   curState->type.write.stage = WritefileStage_DONE;
}

/**
 * Checks for immediately available incoming data.
 *
 * @return fhgfs_true if incoming data is available for this state
 */
fhgfs_bool __FhgfsOpsCommKitVec_writeFileIsDataAvailable(RWfileStatesHelper* statesHelper,
   FileStateVec* curState)
{
   // check for a poll() timeout or error (all states have to be cancelled in that case)

   if(unlikely(statesHelper->pollTimedOut) )
   {
      //Logger_logFormatted(statesHelper->log, 2, statesHelper->logContext,
      //   "Canceling due to communication error: %s (%s)",
      //   curState->nodeID, Socket_getPeername(curState->sock) );

      curState->nodeResult = -FhgfsOpsErr_COMMUNICATION;
      curState->type.write.stage = WritefileStage_SOCKETINVALIDATE;
      return fhgfs_false;
   }


   // check for immediately available data

   if(!(statesHelper->pollSocks)[curState->contPollSocksIndex].revents)
   { // no data available yet => wait for the next round
      __FhgfsOpsCommKitVec_addStateSockToPollSocks(statesHelper, curState, FHGFS_POLLIN);

      return fhgfs_false;
   }

   // incoming data is available

   return fhgfs_true;
}

/**
 * Prepare to completely resend this chunk
 */
void __FhgfsOpsCommKitVec_writefilePrepareRetry(RWfileStatesHelper* statesHelper,
   FileStateVec* curState)
{
   statesHelper->numRetryWaiters++;

   curState->contLastPage = 0;
   curState->contPageOffset = 0;
   curState->nodeResult = -EREMOTEIO;
   curState->type.write.stage = WritefileStage_RETRYWAIT;
}

/**
 * Put failed-retry-states into prepare mode, to resend their chunk
 */
void __FhgfsOpsCommKitVec_writefileStartRetry(RWfileStatesHelper* statesHelper,
   FileStateVec* states, size_t numStates)
{
   size_t i;

   // set the new stage for each retry waiter and reset number of waiters

   for(i = 0; i < numStates; i++)
   {
      FileStateVec* curState = &states[i];

      if(curState->type.write.stage == WritefileStage_RETRYWAIT)
         curState->type.write.stage = WritefileStage_PREPARE;
   }

   statesHelper->numRetryWaiters = 0;

   statesHelper->pollTimedOut = fhgfs_false;


   // wait before we actually start the retry

   MessagingTk_waitBeforeRetry(statesHelper->currentRetryNum);

   statesHelper->currentRetryNum++;
}


/**
 * Note: This method cannot fail (no return value), but the performed communication etc. can fail,
 * of course. Hence, read the results from the states.
 *
 * @param pollSocks array length must match numStates
 */
void FhgfsOpsCommKitVec_readfileCommunicate(App* app, FileStateVec* states, size_t numStates,
   PollSocketEx* pollSocks)
{
   Config* cfg = App_getConfig(app);
   NodeStoreEx* nodes = App_getStorageNodes(app);
   Node* localNode = App_getLocalNode(app);

   RWfileStatesHelper statesHelper =
   {
      .app = app,
      .log = App_getLogger(app),
      .logContext = "readfile (vec communication)",

      .nodes = nodes,
      .localNode = localNode,
      .localNodeID = Node_getID(localNode),

      .numRetryWaiters = 0, // couter for states that encountered a comm error
      .numDone = 0, // counter for finished states
      .numAcquiredConns = 0, // counter for currently acquired conns (to wait only for first conn)

      .pollTimedOut = fhgfs_false,
      .pollTimeoutLogged = fhgfs_false,
      .connFailedLogged = fhgfs_false,

      .pollSocks = pollSocks,

      .currentRetryNum = 0,
      .maxNumRetries = Config_getConnNumCommRetries(cfg),
   };


   size_t state;

   // LOG_DEBUG_FORMATED(statesHelper.log, 5, __func__, "states: %d", numStates);

   // loop until all states increased the done-counter
   do
   {
      statesHelper.numUnconnectable = 0; // will be increased by states the didn't get a conn
      statesHelper.numPollSocks = 0; // will be increased by the states that need to poll

      // let each state do something (before we call poll)
      for(state = 0; state < numStates; state++)
      {
         FileStateVec* curState = &states[state];

         // LOG_DEBUG_FORMATTED(statesHelper.log, 5, __func__, "state: %d stage: %s",
         //   state, readFileStateEnumToStr(curState->type.read.stage));

         switch(curState->type.read.stage)
         {
            case ReadfileStage_PREPARE:
            {
               if(!__FhgfsOpsCommKitVec_readfileStagePREPARE(&statesHelper, curState) )
                  break;

               // LOG_DEBUG_FORMATTED(statesHelper.log, 5, __func__, "state: %d stage: %s",
               //   state, readFileStateEnumToStr(curState->type.read.stage));
            } /* fall through */

            case ReadfileStage_SEND:
            {
               __FhgfsOpsCommKitVec_readfileStageSEND(&statesHelper, curState);
            } break;

            case ReadfileStage_RECVHEADER:
            {
               __FhgfsOpsCommKitVec_readfileStageRECVHEADER(&statesHelper, curState);

               // check if we reached end of transmission
               if(curState->type.read.stage != ReadfileStage_RECVDATA)
               {
                  // do not wait for poll to clean up, but do it as soon as possible, so here
                  if (curState->type.read.stage == ReadfileStage_CLEANUP)
                  {
                     // LOG_DEBUG_FORMATTED(statesHelper.log, 5, __func__, "state: %d stage: %s",
                     //   state, readFileStateEnumToStr(curState->type.read.stage));

                     __FhgfsOpsCommKitVec_readfileStageCLEANUP(&statesHelper, curState);
                  }
                  break;

                  // LOG_DEBUG_FORMATTED(statesHelper.log, 5, __func__, "state: %d stage: %s",
                  //   state, readFileStateEnumToStr(curState->type.read.stage));
               }
            } /* fall through */

            case ReadfileStage_RECVDATA:
            {
               __FhgfsOpsCommKitVec_readfileStageRECVDATA(&statesHelper, curState);
            } break;

            case ReadfileStage_SOCKETEXCEPTION:
            {
               __FhgfsOpsCommKitVec_readfileStageSOCKETEXCEPTION(&statesHelper, curState);
            } /* fall through */

            case ReadfileStage_SOCKETINVALIDATE:
            {
               __FhgfsOpsCommKitVec_readfileStageSOCKETINVALIDATE(&statesHelper, curState);
            } /* fall through */

            case ReadfileStage_CLEANUP:
            {
               __FhgfsOpsCommKitVec_readfileStageCLEANUP(&statesHelper, curState);
            } break;

            case ReadfileStage_RETRYWAIT:
            {
               // nothing to be done here
            } break;

            case ReadfileStage_DONE:
            {
               // nothing to be done here
            } break;

            default:
            { // this should actually never be called
               // nothing to be done here
            } break;

         } // end of switch

      } // end of numStates loop


      if(statesHelper.numPollSocks)
         __FhgfsOpsCommKitCommon_pollStateSocks(&statesHelper, numStates);
      else
      if(unlikely(
         statesHelper.numRetryWaiters &&
         ( (statesHelper.numDone + statesHelper.numRetryWaiters) == numStates) ) )
      { // we have some retry waiters and the rest is done
         // note: this can only happen if we don't have any pollsocks, hence the "else"

         __FhgfsOpsCommKitVec_readfileStartRetry(&statesHelper, states, numStates);
      }

   } while(statesHelper.numDone != numStates);

}


/**
 * Note: This method cannot fail (no return value), but the performed communication etc. can fail,
 * of course. Hence, read the results from the states.
 *
 * @param pollSocks array length must match numStates
 */
void FhgfsOpsCommKitVec_writefileCommunicate(App* app, FileStateVec* states, size_t numStates,
   PollSocketEx* pollSocks)
{
   Config* cfg = App_getConfig(app);
   NodeStoreEx* nodes = App_getStorageNodes(app);
   Node* localNode = App_getLocalNode(app);

   RWfileStatesHelper statesHelper =
   {
      .app = app,
      .log = App_getLogger(app),
      .logContext = "writefile (vec communication)",

      .nodes = nodes,
      .localNode = localNode,
      .localNodeID = Node_getID(localNode),

      .numRetryWaiters = 0, // couter for states that encountered a comm error
      .numDone = 0, // counter for finished states
      .numAcquiredConns = 0, // counter for currently acquired conns (to wait only for first conn)

      .pollTimedOut = fhgfs_false,
      .pollTimeoutLogged = fhgfs_false,
      .connFailedLogged = fhgfs_false,

      .pollSocks = pollSocks,

      .currentRetryNum = 0,
      .maxNumRetries = Config_getConnNumCommRetries(cfg),
   };

   size_t state;

   LOG_DEBUG_FORMATTED(statesHelper.log, 5, __func__, "states: %d", numStates);

   // loop until all states increased the done-counter
   do
   {
      statesHelper.numUnconnectable = 0; // will be increased by states the didn't get a conn
      statesHelper.numPollSocks = 0; // will be increased by the states that need to poll

      // let each state do something (before we call poll)
      for(state = 0; state < numStates; state++)
      {
         FileStateVec* curState = &states[state];

         // LOG_DEBUG_FORMATTED(statesHelper.log, 5, __func__, "state: %d stage: %s",
         //   state, writeFileStateEnumToStr(curState->type.write.stage));

         switch(curState->type.write.stage)
         {
            case WritefileStage_PREPARE:
            {
               if(unlikely(!__FhgfsOpsCommKitVec_writefileStagePREPARE(
                  &statesHelper, curState) ) )
                  break;

               // LOG_DEBUG_FORMATTED(statesHelper.log, 5, __func__, "state: %d stage: %s",
               //   state, writeFileStateEnumToStr(curState->type.write.stage));
            } /* fall through */

            case WritefileStage_SENDHEADER:
            {
               __FhgfsOpsCommKitVec_writefileStageSENDHEADER(&statesHelper, curState);
            } break;

            case WritefileStage_SENDDATA:
            {
               __FhgfsOpsCommKitVec_writefileStageSENDDATA(&statesHelper, curState);
            } break;

            case WritefileStage_RECV:
            {
               __FhgfsOpsCommKitVec_writefileStageRECV(&statesHelper, curState);

               // do not wait for poll to clean up, but do it as soon as possible, so here
               if (curState->type.write.stage == WritefileStage_CLEANUP)
               {
                  // LOG_DEBUG_FORMATTED(statesHelper.log, 5, __func__, "state: %d stage: %s",
                  //    state, writeFileStateEnumToStr(curState->type.write.stage));

                  __FhgfsOpsCommKitVec_writefileStageCLEANUP(&statesHelper, curState);
               }
            } break;

            case WritefileStage_SOCKETEXCEPTION:
            {
               __FhgfsOpsCommKitVec_writefileStageSOCKETEXCEPTION(&statesHelper, curState);
               // LOG_DEBUG_FORMATTED(statesHelper.log, 5, __func__, "state: %d stage: %s",
               //   state, writeFileStateEnumToStr(curState->type.write.stage));
            } /* fall through */

            case WritefileStage_SOCKETINVALIDATE:
            {
               __FhgfsOpsCommKitVec_writefileStageSOCKETINVALIDATE(&statesHelper, curState);
               // LOG_DEBUG_FORMATTED(statesHelper.log, 5, __func__, "state: %d stage: %s",
               //   state, writeFileStateEnumToStr(curState->type.write.stage));
            } /* fall through */

            case WritefileStage_CLEANUP:
            {
               __FhgfsOpsCommKitVec_writefileStageCLEANUP(&statesHelper, curState);
            } break;

            case WritefileStage_RETRYWAIT:
            {
               // nothing to be done here
            } break;

            case WritefileStage_DONE:
            {
               // nothing to be done here
            } break;

            default:
            { // this should actually never be called
               // nothing to be done here
            } break;
         } // end of switch

      } // end of numStates loop


      if(statesHelper.numPollSocks)
         __FhgfsOpsCommKitCommon_pollStateSocks(&statesHelper, numStates);
      else
      if(unlikely(
         statesHelper.numRetryWaiters &&
         ( (statesHelper.numDone + statesHelper.numRetryWaiters) == numStates) ) )
      { // we have some retry waiters and the rest is done
         // note: this can only happen if we don't have any pollsocks, hence the "else"

         __FhgfsOpsCommKitVec_writefileStartRetry(&statesHelper, states, numStates);
      }

   } while(statesHelper.numDone != numStates);

}
