#include <app/App.h>
#include <app/log/Logger.h>
#include <common/storage/Metadata.h>
#include <filesystem/FhgfsOpsHelper.h>
#include <app/config/Config.h>
#include <common/net/message/storage/creating/MkDirMsg.h>
#include <common/net/message/storage/creating/MkDirRespMsg.h>
#include <common/net/message/storage/creating/MkFileMsg.h>
#include <common/net/message/storage/creating/MkFileRespMsg.h>
#include <common/net/message/storage/creating/RmDirMsg.h>
#include <common/net/message/storage/creating/RmDirRespMsg.h>
#include <common/net/message/storage/creating/HardlinkMsg.h>
#include <common/net/message/storage/creating/HardlinkRespMsg.h>
#include <common/net/message/storage/creating/UnlinkFileMsg.h>
#include <common/net/message/storage/creating/UnlinkFileRespMsg.h>
#include <common/net/message/storage/listing/ListDirFromOffsetMsg.h>
#include <common/net/message/storage/listing/ListDirFromOffsetRespMsg.h>
#include <common/net/message/storage/moving/RenameMsg.h>
#include <common/net/message/storage/moving/RenameRespMsg.h>
#include <common/net/message/storage/attribs/RefreshEntryInfoMsg.h>
#include <common/net/message/storage/attribs/RefreshEntryInfoRespMsg.h>
#include <common/net/message/storage/attribs/SetAttrMsg.h>
#include <common/net/message/storage/attribs/SetAttrRespMsg.h>
#include <common/net/message/storage/attribs/StatMsg.h>
#include <common/net/message/storage/attribs/StatRespMsg.h>
#include <common/net/message/storage/TruncFileMsg.h>
#include <common/net/message/storage/TruncFileRespMsg.h>
#include <common/net/message/session/locking/FLockEntryMsg.h>
#include <common/net/message/session/locking/FLockEntryRespMsg.h>
#include <common/net/message/session/locking/FLockRangeMsg.h>
#include <common/net/message/session/locking/FLockRangeRespMsg.h>
#include <common/net/message/session/opening/OpenFileMsg.h>
#include <common/net/message/session/opening/OpenFileRespMsg.h>
#include <common/net/message/session/opening/CloseFileMsg.h>
#include <common/net/message/session/opening/CloseFileRespMsg.h>
#include <common/storage/Path.h>
#include <common/storage/StorageDefinitions.h>
#include <common/storage/StorageErrors.h>
#include <common/toolkit/MessagingTk.h>
#include <common/toolkit/SynchronizedCounter.h>
#include <components/worker/FSyncLocalFileWork.h>
#include <components/worker/ReadLocalFileWorkV2.h>
#include <components/worker/StatStoragePathWork.h>
#include <components/worker/WriteLocalFileWorkV2.h>
#include <nodes/NodeStoreEx.h>
#include <toolkit/NoAllocBufferStore.h>
#include <common/net/message/NetMessage.h>
#include <common/nodes/Node.h>
#include <common/storage/StorageErrors.h>
#include <common/toolkit/MathTk.h>
#include <common/toolkit/MetadataTk.h>
#include <toolkit/SignalTk.h>
#include <common/toolkit/BufferStore.h>
#include <common/components/worker/RetryableWork.h>
#include <common/components/worker/WorkQueue.h>
#include <common/FhgfsTypes.h>
#include <common/toolkit/ackstore/AcknowledgmentStore.h>
#include <common/toolkit/VersionTk.h>
#include <toolkit/FhgfsPage.h>
#include <net/filesystem/FhgfsOpsRemoting.h>
#include <common/Common.h>
#include "FhgfsOpsCommKit.h"
#include "FhgfsOpsCommKitVec.h"

static inline const char* _FhgfsOpsRemoting_RWTypeToString(enum Fhgfs_RWType);

static fhgfs_bool __FhgfsOpsRemoting_writefileVerify(App* app, WritefileState* states, int numWorks,
   ssize_t supposedWritten, ssize_t* outWritten, BitStore* firstWriteDone, size_t firstTargetIndex,
   size_t numStripeNodes);
static fhgfs_bool __FhgfsOpsRemoting_RWfileVerifyVec(App* app, FileStateVec* states,
   int numWorks, ssize_t supposedRW, FhgfsPageVec *pageVec, Fhgfs_RWType rwType,
   BitStore* firstWriteDone, size_t firstTargetIndex, size_t numStripeNodes,
   ssize_t* outDoneRW);
static ssize_t __FhgfsOpsRemoting_writefileStandard(const char __user *buf, size_t size,
   loff_t offset, RemotingIOInfo* ioInfo);
static ssize_t __FhgfsOpsRemoting_writefileClientMirrored(const char __user *buf, size_t size,
   loff_t offset, RemotingIOInfo* ioInfo);
static inline int64_t __FhgfsOpsRemoting_getNodeLocalOffset(int64_t pos, unsigned chunkSize,
   size_t numNodes, size_t stripeNodeIndex);


struct Fhgfs_RWTypeStrEntry
{
      enum Fhgfs_RWType type;
      const char* typeStr;
};

struct Fhgfs_RWTypeStrEntry const __Fhgfs_RWTypeList[] =
{
   {FHGFS_RWTYPE_READ,  "read vec"},
   {FHGFS_RWTYPE_WRITE, "write vec"},
   {0, "NULL"},
};

#define __FHGFSOPS_REMOTING_RW_SIZE \
   ( (sizeof(__Fhgfs_RWTypeList) ) / (sizeof(struct Fhgfs_RWTypeStrEntry) ) - 1)
   /* -1 because last elem is NULL */



const char* _FhgfsOpsRemoting_RWTypeToString(enum Fhgfs_RWType enumType)
{
   size_t type = (size_t)enumType;

   if (likely(type < __FHGFSOPS_REMOTING_RW_SIZE) )
      return __Fhgfs_RWTypeList[type].typeStr;


   printk_fhgfs(KERN_WARNING, "Unknown rwType given to %s(): %d; (dumping stack...)\n",
      __func__, (int) type);

   dump_stack();

   return "Unknown error code";

}

/**
 * Note: Clears FsDirInfo names/types, then sets server offset, adds names/types, sets contents
 * offset and endOfDir in case of RPC success.
 * Note: Also retrieves the optional entry types.
 *
 * @param dirInfo used as input (offsets) and output (contents etc.) parameter
 */
FhgfsOpsErr FhgfsOpsRemoting_listdirFromOffset(const EntryInfo* entryInfo, FsDirInfo* dirInfo,
   unsigned maxOutNames)
{
   FsObjectInfo* fsObjectInfo = (FsObjectInfo*)dirInfo;
   App* app = FsObjectInfo_getApp(fsObjectInfo);
   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);

   Logger* log = App_getLogger(app);
   const char* logContext = "Remoting (list dir from offset)";

   fhgfs_sigset_t oldSignalSet;
   int64_t serverOffset = FsDirInfo_getServerOffset(dirInfo);
   uint64_t incrementalOffset = FsDirInfo_getIncrementalOffset(dirInfo);
   ListDirFromOffsetMsg requestMsg;
   char* respBuf;
   NetMessage* respMsg;
   FhgfsOpsErr requestRes;
   ListDirFromOffsetRespMsg* listDirResp;
   FhgfsOpsErr retVal;

   NodeStoreEx* nodes = App_getMetaNodes(app);
   Node* node;

   node = NodeStoreEx_referenceNode(nodes, entryInfo->ownerNodeID);
   if(unlikely(!node) )
   {
      Logger_logErrFormatted(log, logContext, "Unknown metadata node: %hu", entryInfo->ownerNodeID);
      return FhgfsOpsErr_UNKNOWNNODE;
   }

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   // prepare request msg
   ListDirFromOffsetMsg_initFromEntryInfo(&requestMsg, entryInfo, serverOffset, incrementalOffset,
      maxOutNames);

   // connect & communicate
   requestRes = MessagingTk_requestResponseRetry(app,
      node, (NetMessage*)&requestMsg, NETMSGTYPE_ListDirFromOffsetResp, &respBuf, &respMsg);

   NodeStoreEx_releaseNode(nodes, &node);

   if(unlikely(requestRes != FhgfsOpsErr_SUCCESS) )
   { // clean-up
      retVal = requestRes;
      goto cleanup_request;
   }

   // handle result
   listDirResp = (ListDirFromOffsetRespMsg*)respMsg;
   retVal = (FhgfsOpsErr)ListDirFromOffsetRespMsg_getResult(listDirResp);

   if(likely(retVal == FhgfsOpsErr_SUCCESS) )
   {
      IntCpyVec* dirContentsTypes = FsDirInfo_getDirContentsTypes(dirInfo);
      StrCpyVec* dirContents = FsDirInfo_getDirContents(dirInfo);
      fhgfs_bool endOfDirReached;

      FsDirInfo_setCurrentContentsPos(dirInfo, 0);
      IntCpyVec_clear(dirContentsTypes);
      ListDirFromOffsetRespMsg_parseEntryTypes(listDirResp, dirContentsTypes);
      StrCpyVec_clear(dirContents);
      ListDirFromOffsetRespMsg_parseNames(listDirResp, dirContents);

      FsDirInfo_setServerOffset(dirInfo, ListDirFromOffsetRespMsg_getNewServerOffset(listDirResp) );

      endOfDirReached = (StrCpyVec_length(dirContents) < maxOutNames);
      FsDirInfo_setEndOfDir(dirInfo, endOfDirReached);
   }
   else
   {
      int logLevel = (retVal == FhgfsOpsErr_PATHNOTEXISTS) ? Log_DEBUG : Log_NOTICE;

      Logger_logFormatted(log, logLevel, logContext, "ListDirResp error code: %s",
         FhgfsOpsErr_toErrString(retVal) );
   }

   // clean-up
   NetMessage_virtualDestruct(respMsg);
   NoAllocBufferStore_addBuf(bufStore, respBuf);

cleanup_request:
   ListDirFromOffsetMsg_uninit( (NetMessage*)&requestMsg);

   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return retVal;
}

/**
 * Resolve path to entry owner and stat the entry.
 * Note: This function should *only* be called to stat the root path, as it is the only dir
 * without a parent, for other entries use _statDirect() instead, as the owner info is already
 * available.
 */
FhgfsOpsErr FhgfsOpsRemoting_statRoot(App* app, fhgfs_stat* outFhgfsStat)
{
   fhgfs_sigset_t oldSignalSet;
   FhgfsOpsErr retVal;

   Node* node;
   const char* logContext = "stat meta root node";
   Logger* log = App_getLogger(app);

   NodeStoreEx* nodes = App_getMetaNodes(app);

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   node = NodeStoreEx_referenceRootNode(nodes);
   if(likely(node) )
   {
      uint16_t ownerNodeID      = Node_getNumID(node);
      const char* parentEntryID = "";
      const char* entryID       = META_ROOTDIR_ID_STR;
      const char* fileName      = "";
      DirEntryType entryType    = DirEntryType_DIRECTORY;
      int flags                 = 0;

      EntryInfo entryInfo;
      EntryInfo_init(&entryInfo, ownerNodeID, parentEntryID, entryID, fileName, entryType, flags);

      retVal = FhgfsOpsRemoting_statDirect(app, &entryInfo, outFhgfsStat);

      // clean-up
      NodeStoreEx_releaseNode(nodes, &node);
   }
   else
   {
      Logger_logErr(log, logContext, "Unable to proceed without a working root metadata node");
      retVal = FhgfsOpsErr_UNKNOWNNODE;
   }

   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return retVal;
}

/**
 * Stat directly from entryInfo.
 */
FhgfsOpsErr FhgfsOpsRemoting_statDirect(App* app, const EntryInfo* entryInfo,
   fhgfs_stat* outFhgfsStat)
{
   Logger* log = App_getLogger(app);
   const char* logContext = "Remoting (stat)";
   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);

   fhgfs_sigset_t oldSignalSet;
   StatMsg requestMsg;
   char* respBuf;
   NetMessage* respMsg;
   FhgfsOpsErr requestRes;
   StatRespMsg* statResp;
   FhgfsOpsErr retVal;

   NodeStoreEx* nodes = App_getMetaNodes(app);
   Node* node;

   node = NodeStoreEx_referenceNode(nodes, entryInfo->ownerNodeID);
   if(unlikely(!node) )
   {
      Logger_logErrFormatted(log, logContext, "Unknown metadata node: %hu", entryInfo->ownerNodeID);
      return FhgfsOpsErr_UNKNOWNNODE;
   }

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   // prepare request msg
   StatMsg_initFromEntryInfo(&requestMsg, entryInfo );

   // connect & communicate
   requestRes = MessagingTk_requestResponseRetry(app,
      node, (NetMessage*)&requestMsg, NETMSGTYPE_StatResp, &respBuf, &respMsg);

   NodeStoreEx_releaseNode(nodes, &node);

   if(unlikely(requestRes != FhgfsOpsErr_SUCCESS) )
   { // clean-up
      retVal = requestRes;
      goto cleanup_request;
   }

   // handle result
   statResp = (StatRespMsg*)respMsg;
   retVal = (FhgfsOpsErr)StatRespMsg_getResult(statResp);

   if(retVal == FhgfsOpsErr_SUCCESS)
   {
      StatData* statData = StatRespMsg_getStatData(statResp);
      StatData_getOsStat(statData, outFhgfsStat);
   }
   else
   {
      LOG_DEBUG_FORMATTED(log, Log_DEBUG, logContext, "StatResp error code: %s",
         FhgfsOpsErr_toErrString(retVal) );
      IGNORE_UNUSED_VARIABLE(log);
      IGNORE_UNUSED_VARIABLE(logContext);
   }


   // clean-up
   NetMessage_virtualDestruct(respMsg);
   NoAllocBufferStore_addBuf(bufStore, respBuf);

cleanup_request:
   StatMsg_uninit( (NetMessage*) &requestMsg);

   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return retVal;
}

/**
 * Modify file or dir attributes.
 */
FhgfsOpsErr FhgfsOpsRemoting_setAttr(App* app, const EntryInfo* entryInfo,
   SettableFileAttribs* fhgfsAttr, int validAttribs)
{
   Logger* log = App_getLogger(app);
   Config* cfg = App_getConfig(app);
   const char* logContext = "Remoting (set attr)";
   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);

   fhgfs_sigset_t oldSignalSet;
   SetAttrMsg requestMsg;
   char* respBuf;
   NetMessage* respMsg;
   FhgfsOpsErr requestRes;
   SetAttrRespMsg* setAttrResp;
   FhgfsOpsErr retVal;

   NodeStoreEx* nodes = App_getMetaNodes(app);
   Node* node;

   node = NodeStoreEx_referenceNode(nodes, entryInfo->ownerNodeID);
   if(unlikely(!node) )
   {
      Logger_logErrFormatted(log, logContext, "Unknown metadata node: %hu", entryInfo->ownerNodeID);
      return FhgfsOpsErr_UNKNOWNNODE;
   }

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   // logging
   if(validAttribs & SETATTR_CHANGE_MODE)
      { LOG_DEBUG(log, Log_DEBUG, logContext, "Changing mode"); }

   if(validAttribs & SETATTR_CHANGE_USERID)
      { LOG_DEBUG(log, Log_DEBUG, logContext, "Changing userID"); }

   if(validAttribs & SETATTR_CHANGE_GROUPID)
      { LOG_DEBUG(log, Log_DEBUG, logContext, "Changing groupID"); }

   if(validAttribs & SETATTR_CHANGE_MODIFICATIONTIME)
      { LOG_DEBUG(log, Log_DEBUG, logContext, "Changing modificationTime"); }

   if(validAttribs & SETATTR_CHANGE_LASTACCESSTIME)
      { LOG_DEBUG(log, Log_DEBUG, logContext, "Changing lastAccessTime"); }

   // prepare request msg
   SetAttrMsg_initFromEntryInfo(&requestMsg, entryInfo, validAttribs, fhgfsAttr);

   if (validAttribs & (SETATTR_CHANGE_USERID | SETATTR_CHANGE_GROUPID))
   {
      fhgfs_bool supportsQuota = (Config_getQuotaEnabled(cfg) && VersionTk_checkRequiredRelease(
         VERSIONTK_REQUIRED_VERSION_QUOTA_SUPPORT, Node_getFhgfsVersion(node) ) );

      if(supportsQuota)
         NetMessage_addMsgHeaderFeatureFlag((NetMessage*)&requestMsg, SETATTRMSG_FLAG_USE_QUOTA);
   }

   // connect & communicate
   requestRes = MessagingTk_requestResponseRetry(app,
      node, (NetMessage*)&requestMsg, NETMSGTYPE_SetAttrResp, &respBuf, &respMsg);

   NodeStoreEx_releaseNode(nodes, &node);

   if(unlikely(requestRes != FhgfsOpsErr_SUCCESS) )
   { // clean-up
      retVal = requestRes;
      goto cleanup_request;
   }

   // handle result
   setAttrResp = (SetAttrRespMsg*)respMsg;
   retVal = (FhgfsOpsErr)SetAttrRespMsg_getValue(setAttrResp);

   if(unlikely(retVal != FhgfsOpsErr_SUCCESS) )
   { // error on server
      int logLevel = (retVal == FhgfsOpsErr_PATHNOTEXISTS) ? Log_DEBUG : Log_NOTICE;

      Logger_logFormatted(log, logLevel, logContext, "SetAttrResp error code: %s",
         FhgfsOpsErr_toErrString(retVal) );
   }


   // clean-up
   NetMessage_virtualDestruct(respMsg);
   NoAllocBufferStore_addBuf(bufStore, respBuf);

cleanup_request:
   SetAttrMsg_uninit( (NetMessage*)&requestMsg);

   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return retVal;
}

/**
 * @param outEntryInfo attribs set only in case of success (and must then be kfreed by the
 * caller)
 */
FhgfsOpsErr FhgfsOpsRemoting_mkdir(App* app, const EntryInfo* parentInfo,
   struct CreateInfo* createInfo, EntryInfo* outEntryInfo)
{
   Logger* log = App_getLogger(app);
   const char* logContext = "Remoting (mkdir)";
   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);

   fhgfs_sigset_t oldSignalSet;
   MkDirMsg requestMsg;
   char* respBuf;
   NetMessage* respMsg;
   FhgfsOpsErr requestRes;
   MkDirRespMsg* mkResp;
   FhgfsOpsErr retVal;

   NodeStoreEx* nodes = App_getMetaNodes(app);
   Node* node;

   node = NodeStoreEx_referenceNode(nodes, parentInfo->ownerNodeID);
   if(unlikely(!node) )
   {
      Logger_logErrFormatted(log, logContext, "Unknown metadata node: %hu",
         parentInfo->ownerNodeID);
      return FhgfsOpsErr_UNKNOWNNODE;
   }

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   // prepare request msg
   MkDirMsg_initFromEntryInfo(&requestMsg, parentInfo, createInfo);

   // connect & communicate
   requestRes = MessagingTk_requestResponseRetry(app,
      node, (NetMessage*)&requestMsg, NETMSGTYPE_MkDirResp, &respBuf, &respMsg);

   NodeStoreEx_releaseNode(nodes, &node);

   if(unlikely(requestRes != FhgfsOpsErr_SUCCESS) )
   { // clean-up
      retVal = requestRes;
      goto cleanup_request;
   }

   // handle result
   mkResp = (MkDirRespMsg*)respMsg;
   retVal = (FhgfsOpsErr)MkDirRespMsg_getResult(mkResp);

   if(retVal == FhgfsOpsErr_SUCCESS)
   { // success
      EntryInfo_dup(MkDirRespMsg_getEntryInfo(mkResp), outEntryInfo );
   }
   else
   {
      int logLevel = Log_NOTICE;

      if(retVal == FhgfsOpsErr_EXISTS)
         logLevel = Log_DEBUG; // don't bother user with non-error messages

      Logger_logFormatted(log, logLevel, logContext,
         "MkDirResp ownerID: %d parentID: %s name: %s error code: %s ",
         parentInfo->ownerNodeID, parentInfo->entryID, createInfo->entryName,
         FhgfsOpsErr_toErrString(retVal) );
   }

   // clean-up
   NetMessage_virtualDestruct(respMsg);
   NoAllocBufferStore_addBuf(bufStore, respBuf);

cleanup_request:
   MkDirMsg_uninit( (NetMessage*)&requestMsg);

   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return retVal;
}

FhgfsOpsErr FhgfsOpsRemoting_rmdir(App* app, const EntryInfo* parentInfo, const char* entryName)
{
   Logger* log = App_getLogger(app);
   const char* logContext = "Remoting (rmdir)";
   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);

   fhgfs_sigset_t oldSignalSet;
   RmDirMsg requestMsg;
   char* respBuf;
   NetMessage* respMsg;
   FhgfsOpsErr requestRes;
   RmDirRespMsg* rmResp;
   FhgfsOpsErr retVal;

   NodeStoreEx* nodes = App_getMetaNodes(app);
   Node* node;

   node = NodeStoreEx_referenceNode(nodes, parentInfo->ownerNodeID);
   if(unlikely(!node) )
   {
      Logger_logErrFormatted(log, logContext, "Unknown metadata node: %hu",
         parentInfo->ownerNodeID);
      return FhgfsOpsErr_UNKNOWNNODE;
   }

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   // prepare request msg
   RmDirMsg_initFromEntryInfo(&requestMsg, parentInfo, entryName);

   // connect & communicate
   requestRes = MessagingTk_requestResponseRetry(app,
      node, (NetMessage*)&requestMsg, NETMSGTYPE_RmDirResp, &respBuf, &respMsg);

   NodeStoreEx_releaseNode(nodes, &node);

   if(unlikely(requestRes != FhgfsOpsErr_SUCCESS) )
   { // clean-up
      retVal = requestRes;
      goto cleanup_request;
   }

   // handle result
   rmResp = (RmDirRespMsg*)respMsg;
   retVal = (FhgfsOpsErr)RmDirRespMsg_getValue(rmResp);

   if(retVal != FhgfsOpsErr_SUCCESS)
   {
      int logLevel = Log_NOTICE;

      if( (retVal == FhgfsOpsErr_NOTEMPTY) || (retVal == FhgfsOpsErr_PATHNOTEXISTS) ||
          (retVal == FhgfsOpsErr_INUSE) )
         logLevel = Log_DEBUG; // don't bother user with non-error messages

      Logger_logFormatted(log, logLevel, logContext, "RmDirResp error code: %s",
         FhgfsOpsErr_toErrString(retVal) );
   }

   // clean-up
   NetMessage_virtualDestruct(respMsg);
   NoAllocBufferStore_addBuf(bufStore, respBuf);

cleanup_request:
   RmDirMsg_uninit( (NetMessage*)&requestMsg);

   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return retVal;
}

/**
 * @param outEntryInfo attribs set only in case of success (and must then be kfreed by the
 * caller)
 */
FhgfsOpsErr FhgfsOpsRemoting_mkfile(App* app, const EntryInfo* parentInfo,
   struct CreateInfo* createInfo, EntryInfo* outEntryInfo)
{
   Logger* log = App_getLogger(app);
   const char* logContext = "Remoting (mkfile)";
   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);

   fhgfs_sigset_t oldSignalSet;
   MkFileMsg requestMsg;
   char* respBuf;
   NetMessage* respMsg;
   FhgfsOpsErr requestRes;
   MkFileRespMsg* mkResp;
   FhgfsOpsErr retVal;

   NodeStoreEx* nodes = App_getMetaNodes(app);
   Node* node;

   node = NodeStoreEx_referenceNode(nodes, parentInfo->ownerNodeID);
   if(unlikely(!node) )
   {
      Logger_logErrFormatted(log, logContext, "Unknown metadata node: %hu",
         parentInfo->ownerNodeID);
      return FhgfsOpsErr_UNKNOWNNODE;
   }

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   // prepare request msg
   MkFileMsg_initFromEntryInfo(&requestMsg, parentInfo, createInfo);

   // connect & communicate
   requestRes = MessagingTk_requestResponseRetry(app,
      node, (NetMessage*)&requestMsg, NETMSGTYPE_MkFileResp, &respBuf, &respMsg);

   NodeStoreEx_releaseNode(nodes, &node);

   if(unlikely(requestRes != FhgfsOpsErr_SUCCESS) )
   { // clean-up
      retVal = requestRes;
      goto cleanup_request;
   }

   // handle result
   mkResp = (MkFileRespMsg*)respMsg;
   retVal = (FhgfsOpsErr)MkFileRespMsg_getResult(mkResp);

   if(retVal == FhgfsOpsErr_SUCCESS)
   { // success
      const EntryInfo* entryInfo = MkFileRespMsg_getEntryInfo(mkResp);

      EntryInfo_dup(entryInfo, outEntryInfo);
   }
   else
   {
      int logLevel = Log_NOTICE;

      if(retVal == FhgfsOpsErr_EXISTS)
         logLevel = Log_DEBUG; // don't bother user with non-error messages

      Logger_logFormatted(log, logLevel, logContext,
         "MkFileResp: ownerID: %d parentID: %s name: %s error code: %s",
         parentInfo->ownerNodeID, parentInfo->entryID, createInfo->entryName,
         FhgfsOpsErr_toErrString(retVal) );

   }

   // clean-up
   NetMessage_virtualDestruct(respMsg);
   NoAllocBufferStore_addBuf(bufStore, respBuf);

cleanup_request:
   MkFileMsg_uninit( (NetMessage*)&requestMsg);

   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return retVal;
}

FhgfsOpsErr FhgfsOpsRemoting_unlinkfile(App* app, const EntryInfo* parentInfo,
   const char* entryName)
{
   Logger* log = App_getLogger(app);
   const char* logContext = "Remoting (unlink file)";
   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);

   fhgfs_sigset_t oldSignalSet;
   UnlinkFileMsg requestMsg;
   char* respBuf;
   NetMessage* respMsg;
   FhgfsOpsErr requestRes;
   UnlinkFileRespMsg* unlinkResp;
   FhgfsOpsErr retVal;

   NodeStoreEx* nodes = App_getMetaNodes(app);
   Node* node;

   node = NodeStoreEx_referenceNode(nodes, parentInfo->ownerNodeID);
   if(unlikely(!node) )
   {
      Logger_logErrFormatted(log, logContext, "Unknown metadata node: %hu",
         parentInfo->ownerNodeID);
      return FhgfsOpsErr_UNKNOWNNODE;
   }

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   // prepare request msg
   UnlinkFileMsg_initFromEntryInfo(&requestMsg, parentInfo, entryName);

   // connect & communicate
   requestRes = MessagingTk_requestResponseRetry(app,
      node, (NetMessage*)&requestMsg, NETMSGTYPE_UnlinkFileResp, &respBuf, &respMsg);

   NodeStoreEx_releaseNode(nodes, &node);

   if(unlikely(requestRes != FhgfsOpsErr_SUCCESS) )
   { // clean-up
      retVal = requestRes;
      goto cleanup_request;
   }

   // handle result
   unlinkResp = (UnlinkFileRespMsg*)respMsg;
   retVal = (FhgfsOpsErr)UnlinkFileRespMsg_getValue(unlinkResp);

   if(retVal != FhgfsOpsErr_SUCCESS)
   {
      int logLevel = Log_NOTICE;

      if(retVal == FhgfsOpsErr_PATHNOTEXISTS)
         logLevel = Log_DEBUG; // don't bother user with non-error messages

      Logger_logFormatted(log, logLevel, logContext, "UnlinkFileResp error code: %s",
         FhgfsOpsErr_toErrString(retVal) );
   }

   // clean-up
   NetMessage_virtualDestruct(respMsg);
   NoAllocBufferStore_addBuf(bufStore, respBuf);

cleanup_request:
   UnlinkFileMsg_uninit( (NetMessage*)&requestMsg);

   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return retVal;
}

/**
 * @param ioInfo in and out arg; in case of success: fileHandleID will be set, pattern will be set
 * if it is NULL; values have to be freed by the caller
 */
FhgfsOpsErr FhgfsOpsRemoting_openfile(const EntryInfo* entryInfo, RemotingIOInfo* ioInfo)
{
   App* app = ioInfo->app;
   Logger* log = App_getLogger(app);
   Config* cfg = App_getConfig(app);
   const char* logContext = "Remoting (open file)";
   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);

   fhgfs_sigset_t oldSignalSet;
   char* localNodeID;
   OpenFileMsg requestMsg;
   char* respBuf;
   NetMessage* respMsg;
   FhgfsOpsErr requestRes;
   OpenFileRespMsg* openResp;
   FhgfsOpsErr retVal;

   fhgfs_bool supportsQuota;

   NodeStoreEx* nodes = App_getMetaNodes(app);
   Node* node;

   node = NodeStoreEx_referenceNode(nodes, entryInfo->ownerNodeID);
   if(unlikely(!node) )
   {
      Logger_logErrFormatted(log, logContext, "Unknown metadata node: %hu", entryInfo->ownerNodeID);
      return FhgfsOpsErr_UNKNOWNNODE;
   }

   supportsQuota = (Config_getQuotaEnabled(cfg) && VersionTk_checkRequiredRelease(
      VERSIONTK_REQUIRED_VERSION_QUOTA_SUPPORT, Node_getFhgfsVersion(node) ) );

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   // log open flags
   LOG_DEBUG_FORMATTED(log, Log_DEBUG, logContext, "access: %s%s%s; extra flags: %s|%s",
      ioInfo->accessFlags & OPENFILE_ACCESS_READWRITE ? "rw" : "",
      ioInfo->accessFlags & OPENFILE_ACCESS_WRITE     ? "w" : "",
      ioInfo->accessFlags & OPENFILE_ACCESS_READ      ? "r" : "",
      ioInfo->accessFlags & OPENFILE_ACCESS_APPEND    ? "append" : "",
      ioInfo->accessFlags & OPENFILE_ACCESS_TRUNC     ? "trunc" : "");

   // prepare request msg
   localNodeID = Node_getID(App_getLocalNode(app) );

   OpenFileMsg_initFromSession(&requestMsg, localNodeID, entryInfo, ioInfo->accessFlags);

   if (supportsQuota)
      NetMessage_addMsgHeaderFeatureFlag((NetMessage*)&requestMsg, OPENFILEMSG_FLAG_USE_QUOTA);

   // connect & communicate
   requestRes = MessagingTk_requestResponseRetry(app,
      node, (NetMessage*)&requestMsg, NETMSGTYPE_OpenFileResp, &respBuf, &respMsg);

   NodeStoreEx_releaseNode(nodes, &node);

   if(unlikely(requestRes != FhgfsOpsErr_SUCCESS) )
   { // clean-up
      retVal = requestRes;
      goto cleanup_request;
   }

   // handle result
   openResp = (OpenFileRespMsg*)respMsg;
   retVal = (FhgfsOpsErr)OpenFileRespMsg_getResult(openResp);

   if(likely(retVal == FhgfsOpsErr_SUCCESS) )
   { // success => store file details
      ioInfo->fileHandleID = StringTk_strDup(OpenFileRespMsg_getFileHandleID(openResp) );

      LOG_DEBUG_FORMATTED(log, Log_DEBUG, logContext, "FileHandleID: %s",
         OpenFileRespMsg_getFileHandleID(openResp) );

      if(!ioInfo->pattern)
      { // inode doesn't have pattern yet => create it from response
         StripePattern* pattern = OpenFileRespMsg_createPattern(openResp);

         // check stripe pattern validity
         if(unlikely(StripePattern_getPatternType(pattern) == STRIPEPATTERN_Invalid) )
         { // invalid pattern
            Logger_logErrFormatted(log, logContext,
               "Received invalid stripe pattern from metadata node: %hu; FileHandleID: %s",
               entryInfo->ownerNodeID, OpenFileRespMsg_getFileHandleID(openResp) );

            StripePattern_virtualDestruct(pattern);

            // nevertheless, the file is open now on mds => close it
            FhgfsOpsHelper_closefileWithAsyncRetry(entryInfo, ioInfo);

            kfree(ioInfo->fileHandleID);

            retVal = FhgfsOpsErr_INTERNAL;
         }
         else
            ioInfo->pattern = pattern; // received a valid pattern
      }
   }
   else
   { // error on server
      int logLevel = (retVal == FhgfsOpsErr_PATHNOTEXISTS) ? Log_DEBUG : Log_NOTICE;

      Logger_logFormatted(log, logLevel, logContext, "OpenFileResp error code: %s",
         FhgfsOpsErr_toErrString(retVal) );
   }

   // clean-up
   NetMessage_virtualDestruct(respMsg);
   NoAllocBufferStore_addBuf(bufStore, respBuf);

cleanup_request:
   OpenFileMsg_uninit( (NetMessage*)&requestMsg);

   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return retVal;
}

FhgfsOpsErr FhgfsOpsRemoting_closefile(const EntryInfo* entryInfo, RemotingIOInfo* ioInfo)
{
   return FhgfsOpsRemoting_closefileEx(entryInfo, ioInfo, fhgfs_true);
}

/**
 * Note: You probably want to call FhgfsOpsRemoting_closefile() instead of this method.
 *
 * @param allowRetries usually fhgfs_true, only callers like InternodeSyncer might use fhgfs_false
 * here to make sure they don't block on communication retries too long.
 */
FhgfsOpsErr FhgfsOpsRemoting_closefileEx(const EntryInfo* entryInfo, RemotingIOInfo* ioInfo,
   fhgfs_bool allowRetries)
{
   App* app = ioInfo->app;
   Config* cfg = App_getConfig(app);
   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);

   Logger* log = App_getLogger(app);
   const char* logContext = "Remoting (close file)";

   char* localNodeID;
   int maxUsedTargetIndex = AtomicInt_read(ioInfo->maxUsedTargetIndex);
   fhgfs_sigset_t oldSignalSet;
   CloseFileMsg requestMsg;
   char* respBuf;
   NetMessage* respMsg;
   FhgfsOpsErr requestRes;
   CloseFileRespMsg* closeResp;
   FhgfsOpsErr retVal;

   NodeStoreEx* nodes = App_getMetaNodes(app);
   Node* node = NodeStoreEx_referenceNode(nodes, entryInfo->ownerNodeID);
   if(!node)
   {
      Logger_logErrFormatted(log, logContext, "Unable to proceed because of unknown node: %hu",
         entryInfo->ownerNodeID);
      return FhgfsOpsErr_UNKNOWNNODE;
   }

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   // prepare request msg
   localNodeID = Node_getID(App_getLocalNode(app) );

   CloseFileMsg_initFromSession(&requestMsg, localNodeID, ioInfo->fileHandleID, entryInfo,
      maxUsedTargetIndex);

   if(Config_getTuneEarlyCloseResponse(cfg) ) // (note: linux doesn't return release() res to apps)
      NetMessage_addMsgHeaderFeatureFlag( (NetMessage*)&requestMsg, CLOSEFILEMSG_FLAG_EARLYRESULT);

   // connect & communicate
   if(allowRetries)
   { // normal case
      requestRes = MessagingTk_requestResponseRetry(app,
         node, (NetMessage*)&requestMsg, NETMSGTYPE_CloseFileResp, &respBuf, &respMsg);
   }
   else
   { // caller doesn't want retries => just use the requestResponse tool without retries...
      requestRes = MessagingTk_requestResponse(app,
         node, (NetMessage*)&requestMsg, NETMSGTYPE_CloseFileResp, &respBuf, &respMsg);
   }

   NodeStoreEx_releaseNode(nodes, &node);

   if(unlikely(requestRes != FhgfsOpsErr_SUCCESS) )
   { // clean-up
      retVal = requestRes;
      goto cleanup_request;
   }

   // handle result
   closeResp = (CloseFileRespMsg*)respMsg;
   retVal = (FhgfsOpsErr)CloseFileRespMsg_getValue(closeResp);

   if(likely(retVal == FhgfsOpsErr_SUCCESS) )
   { // success
      LOG_DEBUG_FORMATTED(log, Log_DEBUG, logContext, "FileHandleID: %s", ioInfo->fileHandleID);
   }
   else
   { // error
      Logger_logFormatted(log, Log_NOTICE, logContext, "CloseFileResp error: %s (FileHandleID: %s)",
         FhgfsOpsErr_toErrString(retVal), ioInfo->fileHandleID);
   }

   // clean-up
   NetMessage_virtualDestruct(respMsg);
   NoAllocBufferStore_addBuf(bufStore, respBuf);

cleanup_request:
   CloseFileMsg_uninit( (NetMessage*)&requestMsg);
   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return retVal;
}

FhgfsOpsErr FhgfsOpsRemoting_flockEntry(const EntryInfo* entryInfo, RemotingIOInfo* ioInfo,
   int64_t clientFD, int ownerPID, int lockTypeFlags)
{
   return FhgfsOpsRemoting_flockEntryEx(entryInfo, ioInfo, clientFD, ownerPID, lockTypeFlags,
      fhgfs_true);
}

/**
 * Note: You probably want to call FhgfsOpsRemoting_flockEntry() instead of this method.
 *
 * @param allowRetries usually fhgfs_true, only callers like InternodeSyncer might use fhgfs_false
 * here to make sure they don't block on communication retries too long.
 */
FhgfsOpsErr FhgfsOpsRemoting_flockEntryEx(const EntryInfo* entryInfo, RemotingIOInfo* ioInfo,
   int64_t clientFD, int ownerPID, int lockTypeFlags, fhgfs_bool allowRetries)
{
   /* note: this code is very similar to _flockRange(), so if you change something here, you
      probably want to change it there, too */

   const int resendIntervalMS = 5000;

   App* app = ioInfo->app;
   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);
   AcknowledgmentStore* ackStore = App_getAckStore(app);
   Node* localNode = App_getLocalNode(app);
   char* localNodeID = Node_getID(localNode);
   AtomicInt* atomicCounter = App_getLockAckAtomicCounter(app);

   Logger* log = App_getLogger(app);
   const char* logContext = "Remoting (flock entry)";

   char* lockAckID;
   WaitAck lockWaitAck;
   WaitAckMap waitAcks;
   WaitAckMap receivedAcks;
   WaitAckNotification ackNotifier;

   fhgfs_sigset_t oldSignalSet;
   FLockEntryMsg requestMsg;
   char* respBuf;
   NetMessage* respMsg;
   FhgfsOpsErr requestRes;
   FLockEntryRespMsg* flockResp;
   FhgfsOpsErr retVal;

   NodeStoreEx* nodes = App_getMetaNodes(app);
   Node* node = NodeStoreEx_referenceNode(nodes, entryInfo->ownerNodeID);
   if(unlikely(!node) )
   {
      Logger_logErrFormatted(log, logContext, "Unable to proceed because of unknown node: %hu",
         entryInfo->ownerNodeID);
      return FhgfsOpsErr_UNKNOWNNODE;
   }

   // register lock ack wait...

   // (note: lockAckID must be _globally_ unique)
   lockAckID = StringTk_kasprintf("%X-%s-elck", AtomicInt_incAndRead(atomicCounter), localNodeID);

   WaitAck_init(&lockWaitAck, lockAckID, NULL);
   WaitAckMap_init(&waitAcks);
   WaitAckMap_init(&receivedAcks);
   WaitAckNotification_init(&ackNotifier);

   WaitAckMap_insert(&waitAcks, lockAckID, &lockWaitAck);

   AcknowledgmentStore_registerWaitAcks(ackStore, &waitAcks, &receivedAcks, &ackNotifier);


   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   // prepare request msg
   localNodeID = Node_getID(App_getLocalNode(app) );

   FLockEntryMsg_initFromSession(&requestMsg, localNodeID, ioInfo->fileHandleID, entryInfo,
      clientFD, ownerPID, lockTypeFlags, lockAckID);

   for( ; ; )
   {
      fhgfs_bool continueRequestLoop = fhgfs_false;

      // connect & communicate
      if(allowRetries)
      { // normal case
         requestRes = MessagingTk_requestResponseRetry(app,
            node, (NetMessage*)&requestMsg, NETMSGTYPE_FLockEntryResp, &respBuf, &respMsg);
      }
      else
      { // caller doesn't want retries => just use the requestResponse tool without retries...
         requestRes = MessagingTk_requestResponse(app,
            node, (NetMessage*)&requestMsg, NETMSGTYPE_FLockEntryResp, &respBuf, &respMsg);
      }

      if(unlikely(requestRes != FhgfsOpsErr_SUCCESS) )
      { // clean-up
         retVal = requestRes;
         break;
      }

      // store result
      flockResp = (FLockEntryRespMsg*)respMsg;
      retVal = (FhgfsOpsErr)FLockEntryRespMsg_getResult(flockResp);

      // clean-up response
      // (note: cleanup is done here to not block the buffer while we're waiting for delayed ack)
      NetMessage_virtualDestruct(respMsg);
      NoAllocBufferStore_addBuf(bufStore, respBuf);

      // handle result

      if( (retVal == FhgfsOpsErr_WOULDBLOCK) && !(lockTypeFlags & ENTRYLOCKTYPE_NOWAIT) )
      {
         // not immediately granted and caller wants to wait for the lock

         fhgfs_bool waitRes = AcknowledgmentStore_waitForAckCompletion(
            ackStore, &waitAcks, &ackNotifier, resendIntervalMS);

         if(waitRes)
         { // received lock grant
            LOG_DEBUG_FORMATTED(log, Log_DEBUG, logContext, "received async lock grant. "
               "(handleID: %s)", ioInfo->fileHandleID);

            retVal = FhgfsOpsErr_SUCCESS;
         }
         else
         if(!waitRes && !Thread_isSignalPending() )
         { // no grant received yet => resend request
            LOG_DEBUG_FORMATTED(log, Log_DEBUG, logContext, "no grant yet. re-requesting lock... "
               "(handleID: %s)", ioInfo->fileHandleID);

            continueRequestLoop = fhgfs_true;
         }
         else
         { // signal pending => cancel waiting
            LOG_DEBUG_FORMATTED(log, Log_DEBUG, logContext, "canceling wait due to pending signal "
               "(handleID: %s)", ioInfo->fileHandleID);

            retVal = FhgfsOpsErr_INTERRUPTED;
         }
      }

      if(!continueRequestLoop)
      {
         if(retVal == FhgfsOpsErr_SUCCESS)
         { // success
            LOG_DEBUG_FORMATTED(log, Log_DEBUG, logContext, "received grant (handleID: %s)",
               ioInfo->fileHandleID);
         }
         else
         { // error
            LOG_DEBUG_FORMATTED(log, Log_DEBUG, logContext,
               "FLockEntryResp error: %s (FileHandleID: %s)",
               FhgfsOpsErr_toErrString(retVal), ioInfo->fileHandleID);
         }

         break;
      }

   } // end of for-loop

   /* note: if we were interrupted after sending a lock request and user allowed waiting, we have
      one last chance now to get our ack after unregistering. (this check is also important to avoid
      race conditions because we still send acks to lock grants as long as we're registered. */

   AcknowledgmentStore_unregisterWaitAcks(ackStore, &waitAcks);

   /* now that we're unregistered, we can safely check whether we received the grant after the
      interrupt and before we unregistered ourselves (see corresponding note above)  */

   if( (retVal == FhgfsOpsErr_INTERRUPTED) && WaitAckMap_length(&receivedAcks) )
   { // wow, we received the grant just in time
      retVal = FhgfsOpsErr_SUCCESS;
   }

   WaitAckNotification_uninit(&ackNotifier);
   WaitAckMap_uninit(&receivedAcks);
   WaitAckMap_uninit(&waitAcks);
   WaitAck_uninit(&lockWaitAck);
   os_kfree(lockAckID);


   FLockEntryMsg_uninit( (NetMessage*)&requestMsg);
   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s
   NodeStoreEx_releaseNode(nodes, &node);

   return retVal;
}

FhgfsOpsErr FhgfsOpsRemoting_flockRange(const EntryInfo* entryInfo,
   RemotingIOInfo* ioInfo, int ownerPID, int lockTypeFlags, uint64_t start, uint64_t end)
{
   return FhgfsOpsRemoting_flockRangeEx(entryInfo, ioInfo, ownerPID, lockTypeFlags, start, end,
      fhgfs_true);
}

/**
 * Note: You probably want to call FhgfsOpsRemoting_flockRange() instead of this method.
 *
 * @param allowRetries usually fhgfs_true, only callers like InternodeSyncer might use fhgfs_false
 * here to make sure they don't block on communication retries too long.
 */
FhgfsOpsErr FhgfsOpsRemoting_flockRangeEx(const EntryInfo* entryInfo,
   RemotingIOInfo* ioInfo, int ownerPID, int lockTypeFlags, uint64_t start, uint64_t end,
   fhgfs_bool allowRetries)
{
   /* note: this code is very similar to _flockEntry(), so if you change something here, you
      probably want to change it there, too */

   const int resendIntervalMS = 5000;

   App* app = ioInfo->app;
   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);
   AcknowledgmentStore* ackStore = App_getAckStore(app);
   Node* localNode = App_getLocalNode(app);
   char* localNodeID = Node_getID(localNode);
   AtomicInt* atomicCounter = App_getLockAckAtomicCounter(app);

   Logger* log = App_getLogger(app);
   const char* logContext = "Remoting (flock range)";

   char* lockAckID;
   WaitAck lockWaitAck;
   WaitAckMap waitAcks;
   WaitAckMap receivedAcks;
   WaitAckNotification ackNotifier;

   fhgfs_sigset_t oldSignalSet;
   FLockRangeMsg requestMsg;
   char* respBuf;
   NetMessage* respMsg;
   FhgfsOpsErr requestRes;
   FLockRangeRespMsg* flockResp;
   FhgfsOpsErr retVal;

   NodeStoreEx* nodes = App_getMetaNodes(app);
   Node* node = NodeStoreEx_referenceNode(nodes, entryInfo->ownerNodeID);
   if(unlikely(!node) )
   {
      Logger_logErrFormatted(log, logContext, "Unable to proceed because of unknown node: %hu",
         entryInfo->ownerNodeID);
      return FhgfsOpsErr_UNKNOWNNODE;
   }

   // register lock ack wait...

   // (note: lockAckID must be _globally_ unique)
   lockAckID = StringTk_kasprintf("%X-%s-rlck", AtomicInt_incAndRead(atomicCounter), localNodeID);

   WaitAck_init(&lockWaitAck, lockAckID, NULL);
   WaitAckMap_init(&waitAcks);
   WaitAckMap_init(&receivedAcks);
   WaitAckNotification_init(&ackNotifier);

   WaitAckMap_insert(&waitAcks, lockAckID, &lockWaitAck);

   AcknowledgmentStore_registerWaitAcks(ackStore, &waitAcks, &receivedAcks, &ackNotifier);


   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   // prepare request msg
   localNodeID = Node_getID(App_getLocalNode(app) );

   FLockRangeMsg_initFromSession(&requestMsg, localNodeID, ioInfo->fileHandleID, entryInfo,
      ownerPID, lockTypeFlags, start, end, lockAckID);

   for( ; ; )
   {
      fhgfs_bool continueRequestLoop = fhgfs_false;

      // connect & communicate
      if(allowRetries)
      { // normal case
         requestRes = MessagingTk_requestResponseRetry(app,
            node, (NetMessage*)&requestMsg, NETMSGTYPE_FLockRangeResp, &respBuf, &respMsg);
      }
      else
      { // caller doesn't want retries => just use the requestResponse tool without retries...
         requestRes = MessagingTk_requestResponse(app,
            node, (NetMessage*)&requestMsg, NETMSGTYPE_FLockRangeResp, &respBuf, &respMsg);
      }

      if(unlikely(requestRes != FhgfsOpsErr_SUCCESS) )
      { // clean-up
         retVal = requestRes;
         break;
      }

      // store result
      flockResp = (FLockRangeRespMsg*)respMsg;
      retVal = (FhgfsOpsErr)FLockRangeRespMsg_getResult(flockResp);

      // cleanup response
      // (note: cleanup is done here to not block the buffer while we're waiting for delayed ack)
      NetMessage_virtualDestruct(respMsg);
      NoAllocBufferStore_addBuf(bufStore, respBuf);

      // handle result

      if( (retVal == FhgfsOpsErr_WOULDBLOCK) && !(lockTypeFlags & ENTRYLOCKTYPE_NOWAIT) )
      {
         // not immediately granted and caller wants to wait for the lock

         fhgfs_bool waitRes = AcknowledgmentStore_waitForAckCompletion(
            ackStore, &waitAcks, &ackNotifier, resendIntervalMS);

         if(waitRes)
         { // received lock grant
            LOG_DEBUG_FORMATTED(log, Log_DEBUG, logContext, "received async lock grant. "
               "(handleID: %s)", ioInfo->fileHandleID);

            retVal = FhgfsOpsErr_SUCCESS;
         }
         else
         if(!waitRes && !Thread_isSignalPending() )
         { // no grant received yet => resend request
            LOG_DEBUG_FORMATTED(log, Log_DEBUG, logContext, "no grant yet. re-requesting lock... "
               "(handleID: %s)", ioInfo->fileHandleID);

            continueRequestLoop = fhgfs_true;
         }
         else
         { // signal pending => cancel waiting
            LOG_DEBUG_FORMATTED(log, Log_DEBUG, logContext, "canceling wait due to pending signal "
               "(handleID: %s)", ioInfo->fileHandleID);

            retVal = FhgfsOpsErr_INTERRUPTED;
         }
      }

      if(!continueRequestLoop)
      {
         if(retVal == FhgfsOpsErr_SUCCESS)
         { // success
            LOG_DEBUG_FORMATTED(log, Log_DEBUG, logContext, "received grant (handleID: %s)",
               ioInfo->fileHandleID);
         }
         else
         { // error
            LOG_DEBUG_FORMATTED(log, Log_DEBUG, logContext,
               "FLockEntryResp error: %s (FileHandleID: %s)",
               FhgfsOpsErr_toErrString(retVal), ioInfo->fileHandleID);
         }

         break;
      }

   } // end of for-loop

   /* note: if we were interrupted after sending a lock request and user allowed waiting, we have
      one last chance now to get our ack after unregistering. (this check is also important to avoid
      race conditions because we still send acks to lock grants as long as we're registered. */

   AcknowledgmentStore_unregisterWaitAcks(ackStore, &waitAcks);

   /* now that we're unregistered, we can safely check whether we received the grant after the
      interrupt and before we unregistered ourselves (see corresponding note above)  */

   if( (retVal == FhgfsOpsErr_INTERRUPTED) && WaitAckMap_length(&receivedAcks) )
   { // wow, we received the grant just in time
      retVal = FhgfsOpsErr_SUCCESS;
   }

   WaitAckNotification_uninit(&ackNotifier);
   WaitAckMap_uninit(&receivedAcks);
   WaitAckMap_uninit(&waitAcks);
   WaitAck_uninit(&lockWaitAck);
   os_kfree(lockAckID);


   FLockRangeMsg_uninit( (NetMessage*)&requestMsg);
   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s
   NodeStoreEx_releaseNode(nodes, &node);

   return retVal;
}

/**
 * Check if there have been errors during _writefile()
 *
 * @supposedWritten  - number of bytes already written, including the current step, if an error
 *                     came up, the number of unwritten bytes of this step has to be substracted
 * @outWritten       - number of really written results or error code
 * @firstWriteDone   - bit mask of the targets, the bit of the target is true if a chunk was
 *                     successful written to the target
 * @firstTargetIndex - the index of the first stripe target
 * @numStripeNodes   - count of stripe targets
 * @return           - true if all was fine, false if there was an error
 */
static fhgfs_bool __FhgfsOpsRemoting_writefileVerify(App* app, WritefileState* states, int numWorks,
   const ssize_t supposedWritten, ssize_t* outWritten, BitStore* firstWriteDone,
   size_t firstTargetIndex, size_t numStripeNodes)
{
   int i;

   size_t currentTargetIndex = 0;

   for(i = 0; i < numWorks; i++)
   {
      /* We abort on the first error that comes up to return the number of successfull
       * supposedWritten bytes till this error. With a stripe count > 0 there might
       * (and probably will) be data successfully supposedWritten to other targets, but which
       * will not be returned to the user */
      if(unlikely(states[i].expectedNodeResult != states[i].nodeResult) )
      { // something was other than we expected
         Logger* log = App_getLogger(app);
         const char* logContext = "Remoting (write file)";

         if(states[i].nodeResult >= 0)
         { // node wrote only a part of the data (probably disk full => makes no sense to go on)
            ssize_t numNodeBytesUnwritten = 0;
            ssize_t j;

            Logger_logFormatted(log, Log_NOTICE, logContext,
               "Problem storage targetID: %hu; "
               "Msg: Node wrote only %lld bytes (expected %lld bytes); FileHandle: %s",
               states[i].targetID, states[i].nodeResult, states[i].expectedNodeResult,
               states[i].fileHandleID);

            // add up remaining expected results
            /* note: this will most likely NOT give us the correct result (because remaining nodes
               have supposedWritten some data) but there's no way to return the "correct" result */
            numNodeBytesUnwritten += states[i].expectedNodeResult - states[i].nodeResult;

            /* add data from all other workers here, so that the user really only gets the result
               of successfully written bytes here */
            for(j=i+1; j < numWorks; j++)
               numNodeBytesUnwritten += states[j].expectedNodeResult;

            *outWritten = supposedWritten - numNodeBytesUnwritten;

            // update BitStore with first write done
            currentTargetIndex = (firstTargetIndex + i) % numStripeNodes;
            BitStore_setBit(firstWriteDone, currentTargetIndex, fhgfs_true);

            return fhgfs_false;
         }
         else
         { // error occurred
            FhgfsOpsErr nodeError = -(states[i].nodeResult);

            if(nodeError == FhgfsOpsErr_INTERRUPTED) // this is normal on ctrl+c (=> no logErr() )
               Logger_logFormatted(log, Log_DEBUG, logContext,
                  "Storage targetID: %hu; Msg: %s; FileHandle: %s",
                  states[i].targetID, FhgfsOpsErr_toErrString(nodeError), states[i].fileHandleID);
            else
               Logger_logErrFormatted(log, logContext,
                  "Error storage targetID: %hu; Msg: %s; FileHandle: %s",
                  states[i].targetID, FhgfsOpsErr_toErrString(nodeError), states[i].fileHandleID);

            *outWritten = states[i].nodeResult;
            return fhgfs_false;
         }

      }

      // update BitStore with first write done
      currentTargetIndex = (firstTargetIndex + i) % numStripeNodes;
      BitStore_setBit(firstWriteDone, currentTargetIndex, fhgfs_true);
   } // end of for-loop (result verification)


   *outWritten = supposedWritten;
   return fhgfs_true;
}

/**
 * Check if there have been errors during page vector transfer to the server
 *
 * @supposedRW       max size the states should contain
 * @remainingPages   number of pages already written, including the current step, if an error
 *                   came up, the number of unwritten pages/bytes of this step has to be
 *                   substracted
 * @firstWriteDone   bit mask of the targets, the bit of the target is true if a chunk was
 *                   successful written to the target
 * @firstTargetIndex the index of the first stripe target
 * @numStripeNodes   count of stripe targets
 * @outRW            number of really read/written results or error code
 *
 * @return           true if all was fine, false if there was an error
 */
static fhgfs_bool __FhgfsOpsRemoting_RWfileVerifyVec(App* app, FileStateVec* states,
   int numWorks, ssize_t supposedRW, FhgfsPageVec *pageVec, Fhgfs_RWType rwType,
   BitStore* firstWriteDone, size_t firstTargetIndex, size_t numStripeNodes, ssize_t* outDoneRW)
{
   Logger* log = App_getLogger(app);
   const char* logContext = "Remoting verify";
   const char* rwTypeStr  = _FhgfsOpsRemoting_RWTypeToString(rwType);

   int i;
   ssize_t didRW = 0;

   size_t currentTargetIndex = 0;

   for(i = 0; i < numWorks; i++)
   {
#if 0
      LOG_DEBUG_FORMATTED(log, Log_SPAM, logContext, "%s: state-result: %d", rwTypeStr,
         states[i].nodeResult);
#endif

      /* We abort on the first error that comes up to return the number of successfull
       * supposedWritten bytes till this error. With a stripe count > 0 there might
       * (and probably will) be data successfully supposedWritten to other targets, but which
       * will not be returned to the user */
      if(unlikely(states[i].dataSize != states[i].nodeResult) )
      { // something was other than we expected

         if (states[i].nodeResult < 0)
         { // error occurred
            FhgfsOpsErr nodeError = -(states[i].nodeResult);

            Logger_logErrFormatted(log, logContext,
               "Error storage targetID: %hu; Msg: %s; FileHandle: %s",
               states[i].targetID, FhgfsOpsErr_toErrString(nodeError), states[i].fileHandleID);

            *outDoneRW = states[i].nodeResult;
            return fhgfs_false;
         }

         if (rwType == FHGFS_RWTYPE_WRITE)
         { // node wrote only a part of the data (probably disk full => makes no sense to go on)
            ssize_t numNodeBytesUnhandled;
            ssize_t j;

            Logger_logFormatted(log, Log_NOTICE, logContext,
               "%s, Problem storage targetID: %hu; "
               "Msg: Node handled only %lld bytes (expected %lld bytes); FileHandle: %s",
               rwTypeStr, states[i].targetID, states[i].nodeResult, states[i].dataSize,
               states[i].fileHandleID);

            // add up remaining expected results
            // note: this will most likely NOT give us the correct result (because the remaining
            //    nodes have supposedWritten some data) but it should be good enough for now
            numNodeBytesUnhandled = states[i].dataSize - states[i].nodeResult;

            // add data from all other workers here, so that the user really only gets the result
            // of successfully written bytes here
            for(j=i+1; j < numWorks; j++)
               numNodeBytesUnhandled += states[j].dataSize;

            *outDoneRW = supposedRW - numNodeBytesUnhandled;

            // update BitStore with first write done
            currentTargetIndex = (firstTargetIndex + i) % numStripeNodes;
            BitStore_setBit(firstWriteDone, currentTargetIndex, fhgfs_true);

            return fhgfs_false;
         }
         else
         if (rwType == FHGFS_RWTYPE_READ)
         {
            // we have an end of file here (but some data might have been read)
            *outDoneRW = didRW + states[i].nodeResult;

#if 0
            LOG_DEBUG_FORMATTED(log, Log_SPAM, logContext, "%s: didRW %d, nodeResult: %lld"
               "outRW: %lld", rwTypeStr, didRW, (long long) states[i].nodeResult,
               (long long)*outDoneRW);
#endif

            return fhgfs_false;
         }
      }

      didRW += states[i].nodeResult;

      // update BitStore with first write done
      currentTargetIndex = (firstTargetIndex + i) % numStripeNodes;
      BitStore_setBit(firstWriteDone, currentTargetIndex, fhgfs_true);
   } // end of for-loop (result verification)

   *outDoneRW = didRW;
   return fhgfs_true;
}

/**
 * Wrapper for different single-threaded parallel file write methods.
 *
 * @return number of bytes written or negative fhgfs error code
 */
ssize_t FhgfsOpsRemoting_writefile(const char __user *buf, size_t size, loff_t offset,
   RemotingIOInfo* ioInfo)
{
   App* app = ioInfo->app;
   Config* cfg = App_getConfig(app);
   StripePattern* pattern = ioInfo->pattern;
   UInt16Vec* mirrorTargetIDs = pattern->getMirrorTargetIDs(pattern); // NULL if not raid10

   if(!mirrorTargetIDs || (size > Config_getTuneMaxClientMirrorSize(cfg) ) )
      return __FhgfsOpsRemoting_writefileStandard(buf, size, offset, ioInfo);
   else
      return __FhgfsOpsRemoting_writefileClientMirrored(buf, size, offset, ioInfo);
}

/**
 * Single-threaded parallel file write.
 * Works for mirrored and unmirrored files. In case of a mirrored file, the mirror data will be
 * forwarded by the servers.
 *
 * @return number of bytes written or negative fhgfs error code
 */
ssize_t __FhgfsOpsRemoting_writefileStandard(const char __user *buf, size_t size, loff_t offset,
   RemotingIOInfo* ioInfo)
{
   App* app = ioInfo->app;
   Config* cfg = App_getConfig(app);

   fhgfs_bool verifyRes;
   ssize_t verifyValue;
   ssize_t retVal = size;
   size_t toBeWritten = size;
   loff_t currentOffset = offset;

   fhgfs_sigset_t oldSignalSet;

   StripePattern* pattern = ioInfo->pattern;
   unsigned chunkSize = StripePattern_getChunkSize(pattern);
   UInt16Vec* targetIDs = pattern->getStripeTargetIDs(pattern);
   UInt16Vec* mirrorTargetIDs = pattern->getMirrorTargetIDs(pattern); // NULL if not raid10
   size_t numStripeTargets = UInt16Vec_length(targetIDs);
   const char* fileHandleID = ioInfo->fileHandleID;
   unsigned accessFlags = ioInfo->accessFlags;
   int maxUsedTargetIndex = AtomicInt_read(ioInfo->maxUsedTargetIndex);

   NoAllocBufferStore* bufStore = App_getRWStatesStore(app);
   size_t maxNumRWNodes = Config_getTuneMaxReadWriteNodesNum(cfg);
   size_t maxNumWorks = MIN(maxNumRWNodes, numStripeTargets);

   char* storeBuf = NoAllocBufferStore_waitForBuf(bufStore);

   // partition the storeBuf
   WritefileState* states = (WritefileState*)storeBuf; // starts at storeBuf
   PollSocketEx* pollSocks = (PollSocketEx*)(&states[maxNumRWNodes] ); // starts after states
   char* msgBufs = (char*)(&pollSocks[maxNumRWNodes] ); // starts after pollSocks


   __FhgfsOpsRemoting_logDebugIOCall(__func__, size, offset, ioInfo);

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   while(toBeWritten)
   {
      size_t currentTargetIndex = pattern->getStripeTargetIndex(pattern, currentOffset);
      size_t firstTargetIndex = currentTargetIndex;
      size_t numWorks = 0;

      /* stripeset-loop: loop over one stripe set (using dynamically determined stripe target
         indices). */
      while(toBeWritten && (numWorks < maxNumWorks) )
      {
         size_t currentChunkSize =
            StripePattern_getChunkEnd(pattern, currentOffset) - currentOffset + 1;
         size_t currentWriteSize = MIN(currentChunkSize, toBeWritten);
         loff_t currentNodeLocalOffset = __FhgfsOpsRemoting_getNodeLocalOffset(
            currentOffset, chunkSize, numStripeTargets, currentTargetIndex);

         maxUsedTargetIndex = MAX(maxUsedTargetIndex, (int)currentTargetIndex);

         // prepare the state information
         states[numWorks] = FhgfsOpsCommKit_assignWritefileState(
            &buf[size-toBeWritten], currentNodeLocalOffset, currentWriteSize,
            fileHandleID, accessFlags, UInt16Vec_at(targetIDs, currentTargetIndex),
            &msgBufs[numWorks * FHGFS_COMMKIT_MSGBUF_SIZE],
            BitStore_getBit(ioInfo->firstWriteDone, currentTargetIndex));

         FhgfsOpsCommKit_setUserdataForQuota(&states[numWorks], ioInfo->userID, ioInfo->groupID);

         if(mirrorTargetIDs)
            FhgfsOpsCommKit_setWritefileStateMirror(
               UInt16Vec_at(mirrorTargetIDs, currentTargetIndex), &states[numWorks] );

         App_incNumRemoteWrites(app);

         // prepare for next loop
         currentOffset += currentWriteSize;
         toBeWritten -= currentWriteSize;
         numWorks++;
         currentTargetIndex = pattern->getStripeTargetIndex(pattern, currentOffset);
      }

      // communicate with the nodes
      FhgfsOpsCommKit_writefileV2bCommunicate(app, states, numWorks, pollSocks);

      // verify results
      verifyRes = __FhgfsOpsRemoting_writefileVerify(app, states, numWorks, size - toBeWritten,
         &verifyValue, ioInfo->firstWriteDone, firstTargetIndex, numStripeTargets);

      if(unlikely(!verifyRes) )
      {
         retVal = verifyValue;
         break;
      }
   } // end of while-loop (write out data)

   NoAllocBufferStore_addBuf(bufStore, storeBuf);
   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   AtomicInt_max(ioInfo->maxUsedTargetIndex, maxUsedTargetIndex);

   return retVal;
}

/**
 * Single-threaded parallel file write for mirrored files.
 * Mirroring will be done by client.
 *
 * Note: This method is not intended to be called for unmirrored files.
 *
 * @return number of bytes written or negative fhgfs error code
 */
ssize_t __FhgfsOpsRemoting_writefileClientMirrored(const char __user *buf, size_t size,
   loff_t offset, RemotingIOInfo* ioInfo)
{
   const char* logContext = __func__;

   App* app = ioInfo->app;
   Config* cfg = App_getConfig(app);
   Logger* log = App_getLogger(app);

   fhgfs_bool verifyRes;
   ssize_t verifyValue;
   ssize_t retVal = size;
   size_t toBeWritten = size;
   loff_t currentOffset = offset;

   fhgfs_sigset_t oldSignalSet;

   StripePattern* pattern = ioInfo->pattern;
   unsigned chunkSize = StripePattern_getChunkSize(pattern);
   UInt16Vec* targetIDs = pattern->getStripeTargetIDs(pattern);
   UInt16Vec* mirrorTargetIDs = pattern->getMirrorTargetIDs(pattern); // NULL if not raid10
   size_t numStripeTargets = UInt16Vec_length(targetIDs);
   const char* fileHandleID = ioInfo->fileHandleID;
   unsigned accessFlags = ioInfo->accessFlags;
   int maxUsedTargetIndex = AtomicInt_read(ioInfo->maxUsedTargetIndex);

   NoAllocBufferStore* bufStore = App_getRWStatesStore(app);
   size_t maxNumRWNodes = Config_getTuneMaxReadWriteNodesNum(cfg);
   size_t maxNumWorks = MIN(maxNumRWNodes, numStripeTargets*2); // ("*2" for mirror work)

   char* storeBuf = NoAllocBufferStore_waitForBuf(bufStore);

   // partition the storeBuf
   WritefileState* states = (WritefileState*)storeBuf; // starts at storeBuf
   PollSocketEx* pollSocks = (PollSocketEx*)(&states[maxNumRWNodes] ); // starts after states
   char* msgBufs = (char*)(&pollSocks[maxNumRWNodes] ); // starts after pollSocks

   // make sure we have enough work buffers to mirror data (=> at least 2)
   if(unlikely(maxNumWorks < 2) )
   {
      Logger_logErr(log, logContext, "Invalid config: tuneMaxReadWriteNodesNum must be >= 2");
      return -FhgfsOpsErr_INTERNAL;
   }

   __FhgfsOpsRemoting_logDebugIOCall(__func__, size, offset, ioInfo);

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   while(toBeWritten)
   {
      size_t currentTargetIndex = pattern->getStripeTargetIndex(pattern, currentOffset);
      size_t firstTargetIndex = currentTargetIndex;
      size_t numWorks = 0;

      /* stripeset-loop: loop over one stripe set (using dynamically determined stripe target
         indices). */
      // ("maxNumWorks-1" for mirror, because we're generating two work packages per loop)
      while(toBeWritten && (numWorks < (maxNumWorks-1) ) )
      {
         size_t currentChunkSize =
            StripePattern_getChunkEnd(pattern, currentOffset) - currentOffset + 1;
         size_t currentWriteSize = MIN(currentChunkSize, toBeWritten);
         loff_t currentNodeLocalOffset = __FhgfsOpsRemoting_getNodeLocalOffset(
            currentOffset, chunkSize, numStripeTargets, currentTargetIndex);

         size_t numWorksMirror = numWorks + 1; // array index for mirror work

         maxUsedTargetIndex = MAX(maxUsedTargetIndex, (int)currentTargetIndex);

         // prepare the state information
         states[numWorks] = FhgfsOpsCommKit_assignWritefileState(
            &buf[size-toBeWritten], currentNodeLocalOffset, currentWriteSize,
            fileHandleID, accessFlags, UInt16Vec_at(targetIDs, currentTargetIndex),
            &msgBufs[numWorks * FHGFS_COMMKIT_MSGBUF_SIZE],
            BitStore_getBit(ioInfo->firstWriteDone, currentTargetIndex) );

         FhgfsOpsCommKit_setUserdataForQuota(&states[numWorks], ioInfo->userID, ioInfo->groupID);

         // prepare the mirror state information
         states[numWorksMirror] = FhgfsOpsCommKit_assignWritefileState(
            &buf[size-toBeWritten], currentNodeLocalOffset, currentWriteSize,
            fileHandleID, accessFlags, UInt16Vec_at(mirrorTargetIDs, currentTargetIndex),
            &msgBufs[numWorksMirror * FHGFS_COMMKIT_MSGBUF_SIZE],
            BitStore_getBit(ioInfo->firstWriteDone, currentTargetIndex) );

         FhgfsOpsCommKit_setUserdataForQuota(&states[numWorksMirror], ioInfo->userID,
            ioInfo->groupID);

         FhgfsOpsCommKit_setWritefileStateMirroredFrom(
            UInt16Vec_at(targetIDs, currentTargetIndex), &states[numWorksMirror] );

         App_incNumRemoteWrites(app);

         // prepare for next loop
         currentOffset += currentWriteSize;
         toBeWritten -= currentWriteSize;
         numWorks += 2;
         currentTargetIndex = pattern->getStripeTargetIndex(pattern, currentOffset);
      }

      // communicate with the nodes
      FhgfsOpsCommKit_writefileV2bCommunicate(app, states, numWorks, pollSocks);

      // verify results
      verifyRes = __FhgfsOpsRemoting_writefileVerify(app, states, numWorks, size - toBeWritten,
         &verifyValue, ioInfo->firstWriteDone, firstTargetIndex, numStripeTargets);
      if(unlikely(!verifyRes) )
      {
         retVal = verifyValue;
         break;
      }
   } // end of while-loop (write out data)

   NoAllocBufferStore_addBuf(bufStore, storeBuf);
   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   AtomicInt_max(ioInfo->maxUsedTargetIndex, maxUsedTargetIndex);

   return retVal;
}


/**
 * Write (send) a vector (array) of pages to the server
 *
 * @return number of bytes written or negative fhgfs error code
 */
ssize_t FhgfsOpsRemoting_RWfileVec(FhgfsPageVec *pageVec, loff_t offset, RemotingIOInfo* ioInfo,
   Fhgfs_RWType rwType)
{
   App* app = ioInfo->app;

   Logger* log = App_getLogger(app);
   const char* logContext = "Remoting read/write";
   const char* rwTypeStr  = _FhgfsOpsRemoting_RWTypeToString(rwType);

   Config* cfg = App_getConfig(app);

   fhgfs_bool isVerifyOk = fhgfs_false;

   loff_t curOffset = offset;

   fhgfs_sigset_t oldSignalSet;

   StripePattern* pattern = ioInfo->pattern;
   unsigned chunkSize = StripePattern_getChunkSize(pattern);
   unsigned remainingPages = pageVec->size;
   unsigned pageIdx = 0; // page index

   UInt16Vec* targetIDs = pattern->getStripeTargetIDs(pattern);
   UInt16Vec* mirrorTargetIDs = pattern->getMirrorTargetIDs(pattern); // NULL if not raid10
   size_t numStripeNodes = UInt16Vec_length(targetIDs);
   const char* fileHandleID = ioInfo->fileHandleID;
   unsigned accessFlags = ioInfo->accessFlags;
   int maxUsedTargetIndex = AtomicInt_read(ioInfo->maxUsedTargetIndex);

   NoAllocBufferStore* bufStore = App_getRWStatesStore(app);
   size_t maxNumRWNodes = Config_getTuneMaxReadWriteNodesNum(cfg);
   size_t maxNumWorks = MIN(maxNumRWNodes, numStripeNodes);

   char* storeBuf = NoAllocBufferStore_waitForBuf(bufStore);

   /* Calculate the data supposedSize here. All pages except for the very last page should have
    * a data length of FHGFS_PAGE_SIZE. The last page might have less than that. */
   FhgfsPage* lastPage = &pageVec->pages[pageVec->size - 1];
   ssize_t supposedSize = ( (pageVec->size - 1) * PAGE_SIZE) + lastPage->length;
   ssize_t handledSize = 0;
   FileStateVec* states;
   PollSocketEx* pollSocks;
   char* msgBufs;

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

#ifdef FHGFS_DEBUG
   /* in debug mode we use individual allocations instead of a single big buffer, as we can detect
      array out of bounds for those (if the proper kernel options are set) */
   states = os_kmalloc(sizeof(*states) * maxNumWorks);
   if (!states)
   {
      handledSize = -ENOMEM;
      goto outNoAlloc;
   }

   pollSocks = os_kmalloc(sizeof (*pollSocks) * maxNumWorks);
   if (!pollSocks)
   {
      handledSize = -ENOMEM;
      goto outPollSocks;
   }

   msgBufs = storeBuf; // TODO: would be better to use malloc also for this one
#else
   // in normal mode we use a single pre-alloc'ed buffer and split it up in pieces for performance
   states = (FileStateVec*)storeBuf; // starts at storeBuf
   pollSocks = (PollSocketEx*)(&states[maxNumRWNodes] ); // starts after states
   msgBufs = (char*)(&pollSocks[maxNumRWNodes] ); // starts after pollSocks
#endif // FHGFS_DEBUG

   __FhgfsOpsRemoting_logDebugIOCall(__func__, supposedSize, offset, ioInfo);

   while(remainingPages)
   {
      size_t currentTargetIndex = pattern->getStripeTargetIndex(pattern, curOffset);
      size_t firstTargetIndex = currentTargetIndex;
      size_t numWorks = 0;
      ssize_t verifyValue;

      // stripeset-loop: loop over one stripe set (using dynamically determined stripe node
      // indices).
      while(remainingPages && (numWorks < maxNumWorks) )
      {
         // how many bytes from offset to chunk end
         size_t chunkSizeLeft = StripePattern_getChunkEnd(pattern, curOffset) - curOffset + 1;
         loff_t curNodeLocalOffset = __FhgfsOpsRemoting_getNodeLocalOffset(
            curOffset, chunkSize, numStripeNodes, currentTargetIndex);
         int curNumPages = MIN(chunkSizeLeft / PAGE_SIZE, remainingPages);

         maxUsedTargetIndex = MAX(maxUsedTargetIndex, (int)currentTargetIndex);

         // prepare the state information
         states[numWorks] = FhgfsOpsCommKitVec_assignRWfileState(
            pageVec, pageIdx, curNumPages, curNodeLocalOffset, fileHandleID, accessFlags,
            UInt16Vec_at(targetIDs, currentTargetIndex),
            &msgBufs[numWorks * FHGFS_COMMKIT_MSGBUF_SIZE], numWorks,
            BitStore_getBit(ioInfo->firstWriteDone, currentTargetIndex) );

         if (rwType == FHGFS_RWTYPE_WRITE)
         {
            states[numWorks].type.write.stage = RW_FILE_STAGE_PREPARE;

            FhgfsOpsCommKitVec_setUserdataForQuota(&states[numWorks], ioInfo->userID,
               ioInfo->groupID);

            if(mirrorTargetIDs)
               FhgfsOpsCommKitVec_setRWfileStateMirror(
                  UInt16Vec_at(mirrorTargetIDs, currentTargetIndex), &states[numWorks] );
         }
         else
            states[numWorks].type.read.stage = RW_FILE_STAGE_PREPARE;

         App_incNumRemoteWrites(app);

         // prepare for next loop
         curOffset += chunkSizeLeft;
         pageIdx += curNumPages;
         remainingPages -= curNumPages;
         numWorks++;
         currentTargetIndex = pattern->getStripeTargetIndex(pattern, curOffset);
      }

      // communicate with the nodes
      if (rwType == FHGFS_RWTYPE_WRITE)
         FhgfsOpsCommKitVec_writefileCommunicate(app, states, numWorks, pollSocks);
      else
         FhgfsOpsCommKitVec_readfileCommunicate(app, states, numWorks, pollSocks);

      // verify results
      isVerifyOk = __FhgfsOpsRemoting_RWfileVerifyVec(app, states, numWorks,
         supposedSize - handledSize, pageVec,rwType, ioInfo->firstWriteDone, firstTargetIndex,
         numStripeNodes, &verifyValue);

      handledSize += verifyValue;

      LOG_DEBUG_FORMATTED(log, Log_SPAM, logContext, "%s: sum-result %lld", rwTypeStr, handledSize);
      IGNORE_UNUSED_VARIABLE(log);
      IGNORE_UNUSED_VARIABLE(logContext);
      IGNORE_UNUSED_VARIABLE(rwTypeStr);


      if(!isVerifyOk)
         break;
   } // end of while-loop (write out data)

#ifdef FHGFS_DEBUG

   os_kfree(states);
outPollSocks:
   os_kfree(pollSocks);
outNoAlloc:

#endif // FHGFS_DEBUG

   NoAllocBufferStore_addBuf(bufStore, storeBuf);
   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   AtomicInt_max(ioInfo->maxUsedTargetIndex, maxUsedTargetIndex);

   return handledSize;
}



/**
 * @return number of bytes read or negative fhgfs error code
 */
ssize_t FhgfsOpsRemoting_readfile(char __user *buf, size_t size, loff_t offset,
   RemotingIOInfo* ioInfo)
{
   App* app = ioInfo->app;
   Config* cfg = App_getConfig(app);

   ssize_t retVal = size;
   size_t toBeRead = size;
   loff_t currentOffset = offset;

   fhgfs_sigset_t oldSignalSet;

   StripePattern* pattern = ioInfo->pattern;
   unsigned chunkSize = StripePattern_getChunkSize(pattern);
   UInt16Vec* targetIDs = pattern->getStripeTargetIDs(pattern);
   size_t numStripeNodes = UInt16Vec_length(targetIDs);
   const char* fileHandleID = ioInfo->fileHandleID;
   unsigned accessFlags = ioInfo->accessFlags;
   int maxUsedTargetIndex = AtomicInt_read(ioInfo->maxUsedTargetIndex);

   NoAllocBufferStore* bufStore = App_getRWStatesStore(app);
   size_t maxNumRWNodes = Config_getTuneMaxReadWriteNodesNum(cfg);
   size_t maxNumWorks = MIN(maxNumRWNodes, numStripeNodes);

   char* storeBuf = NoAllocBufferStore_waitForBuf(bufStore);

   // partition the storeBuf
   ReadfileState* states = (ReadfileState*)storeBuf; // starts at storeBuf
   PollSocketEx* pollSocks = (PollSocketEx*)(&states[maxNumRWNodes] ); // starts after states
   char* msgBufs = (char*)(&pollSocks[maxNumRWNodes] ); // starts after pollSocks

   ssize_t usableReadSize = 0; // the amount of usable data that was received from the nodes


   __FhgfsOpsRemoting_logDebugIOCall(__func__, size, offset, ioInfo);

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   while(toBeRead)
   {
      size_t currentTargetIndex = pattern->getStripeTargetIndex(pattern, currentOffset);
      size_t numWorks = 0; // number of work items that we added to the queue
      size_t i;

      // stripeset-loop: loop over one stripe set (using dynamically determined stripe node
      // indices).
      while(toBeRead && (numWorks < maxNumWorks) )
      {
         size_t currentChunkSize =
            StripePattern_getChunkEnd(pattern, currentOffset) - currentOffset + 1;
         size_t currentReadSize = MIN(currentChunkSize, toBeRead);
         loff_t currentNodeLocalOffset = __FhgfsOpsRemoting_getNodeLocalOffset(
            currentOffset, chunkSize, numStripeNodes, currentTargetIndex);

         maxUsedTargetIndex = MAX(maxUsedTargetIndex, (int)currentTargetIndex);

         // prepare the state information
         states[numWorks] = FhgfsOpsCommKit_assignReadfileState(
            &buf[size-toBeRead], currentNodeLocalOffset, currentReadSize,
            fileHandleID, accessFlags, UInt16Vec_at(targetIDs, currentTargetIndex),
            &msgBufs[numWorks * FHGFS_COMMKIT_MSGBUF_SIZE],
            BitStore_getBit(ioInfo->firstWriteDone, currentTargetIndex));

         App_incNumRemoteReads(app);

         // prepare for next loop
         currentOffset += currentReadSize;
         toBeRead -= currentReadSize;
         numWorks++;
         currentTargetIndex = (currentTargetIndex + 1) % numStripeNodes;
      }

      // communicate with the nodes
      FhgfsOpsCommKit_readfileV2bCommunicate(app, states, numWorks, pollSocks);

      // verify results
      for(i=0; i < numWorks; i++)
      {
         if(states[i].nodeResult < states[i].expectedNodeResult)
         { // result not as expected => check cause: end-of-file or error condition
            if(states[i].nodeResult >= 0)
            { // we have an end of file here (but some data might have been read)
               retVal = usableReadSize + states[i].nodeResult;
            }
            else
            { // error occurred
               FhgfsOpsErr nodeError = -(states[i].nodeResult);

               Logger* log = App_getLogger(app);
               const char* logContext = "Remoting (read file)";

               if(nodeError == FhgfsOpsErr_INTERRUPTED) // normal on ctrl+c (=> no logErr() )
                  Logger_logFormatted(log, Log_DEBUG, logContext,
                     "Storage targetID: %hu; Msg: %s; FileHandle: %s",
                     states[i].targetID, FhgfsOpsErr_toErrString(nodeError), fileHandleID);
               else
                  Logger_logErrFormatted(log, logContext,
                     "Error storage targetID: %hu; Msg: %s; FileHandle: %s",
                     states[i].targetID, FhgfsOpsErr_toErrString(nodeError), fileHandleID);

               retVal = states[i].nodeResult;
            }

            goto clean_up;
         }

         // at this point, the node succeeded
         // => nodeResult is exactly what we expected and usableReadSize is exactly what has been
         //    read so far

         usableReadSize += states[i].nodeResult;
      }

   }

clean_up:
   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s
   NoAllocBufferStore_addBuf(bufStore, storeBuf);

   AtomicInt_max(ioInfo->maxUsedTargetIndex, maxUsedTargetIndex);

   return retVal;
}


FhgfsOpsErr FhgfsOpsRemoting_rename(App* app, const char* oldName, unsigned oldLen,
   DirEntryType entryType, EntryInfo* fromDirInfo, const char* newName, unsigned newLen,
   EntryInfo* toDirInfo)
{
   FhgfsOpsErr retVal = FhgfsOpsErr_INTERNAL;
   Logger* log = App_getLogger(app);

   const char* logContext = "Remoting (rename)";
   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);

   RenameMsg requestMsg;
   char* respBuf;
   NetMessage* respMsg;
   FhgfsOpsErr requestRes;
   RenameRespMsg* renameResp;

   NodeStoreEx* nodes = App_getMetaNodes(app);
   uint16_t fromNodeID = fromDirInfo->ownerNodeID;
   Node* fromNode;

   fhgfs_sigset_t oldSignalSet;

   fromNode = NodeStoreEx_referenceNode(nodes, fromNodeID);
   if(unlikely(!fromNode) )
   {
      Logger_logErrFormatted(log, logContext, "Unknown metadata node: %hu", fromNodeID);
      return FhgfsOpsErr_UNKNOWNNODE;
   }

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   // prepare request msg

   RenameMsg_initFromEntryInfo(&requestMsg, oldName, oldLen, entryType, fromDirInfo, newName,
      newLen, toDirInfo);

   // connect & communicate
   requestRes = MessagingTk_requestResponseRetry(app,
      fromNode, (NetMessage*)&requestMsg, NETMSGTYPE_RenameResp, &respBuf, &respMsg);

   NodeStoreEx_releaseNode(nodes, &fromNode);

   if(unlikely(requestRes != FhgfsOpsErr_SUCCESS) )
   { // clean-up
      retVal = requestRes;
      goto cleanup_request;
   }

   // handle result
   renameResp = (RenameRespMsg*)respMsg;
   retVal = (FhgfsOpsErr)RenameRespMsg_getValue(renameResp);

   // clean-up
   NetMessage_virtualDestruct(respMsg);
   NoAllocBufferStore_addBuf(bufStore, respBuf);

cleanup_request:
   RenameMsg_uninit( (NetMessage*)&requestMsg);

   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return retVal;
}

FhgfsOpsErr FhgfsOpsRemoting_truncfile(App* app, const EntryInfo* entryInfo, loff_t size)
{
   Logger* log = App_getLogger(app);
   Config* cfg = App_getConfig(app);
   const char* logContext = "Remoting (trunc file)";
   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);

   fhgfs_sigset_t oldSignalSet;
   TruncFileMsg requestMsg;
   char* respBuf;
   NetMessage* respMsg;
   FhgfsOpsErr requestRes;
   TruncFileRespMsg* truncResp;
   FhgfsOpsErr retVal;

   NodeStoreEx* nodes = App_getMetaNodes(app);
   Node* node;

   fhgfs_bool supportsQuota;

   node = NodeStoreEx_referenceNode(nodes, entryInfo->ownerNodeID);
   if(unlikely(!node) )
   {
      Logger_logErrFormatted(log, logContext, "Unknown metadata node: %hu", entryInfo->ownerNodeID);
      return FhgfsOpsErr_UNKNOWNNODE;
   }

   supportsQuota = (Config_getQuotaEnabled(cfg) && VersionTk_checkRequiredRelease(
      VERSIONTK_REQUIRED_VERSION_QUOTA_SUPPORT, Node_getFhgfsVersion(node) ) );

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   // prepare request msg
   TruncFileMsg_initFromEntryInfo(&requestMsg, size, entryInfo);

   if (supportsQuota)
      NetMessage_addMsgHeaderFeatureFlag((NetMessage*)&requestMsg, TRUNCFILEMSG_FLAG_USE_QUOTA);

   // connect & communicate
   requestRes = MessagingTk_requestResponseRetry(app,
      node, (NetMessage*)&requestMsg, NETMSGTYPE_TruncFileResp, &respBuf, &respMsg);

   NodeStoreEx_releaseNode(nodes, &node);

   if(unlikely(requestRes != FhgfsOpsErr_SUCCESS) )
   { // clean-up
      retVal = requestRes;
      goto cleanup_request;
   }

   // handle result
   truncResp = (TruncFileRespMsg*)respMsg;
   retVal = (FhgfsOpsErr)TruncFileRespMsg_getValue(truncResp);

   if(unlikely(retVal != FhgfsOpsErr_SUCCESS) )
   { // error on server
      int logLevel = (retVal == FhgfsOpsErr_PATHNOTEXISTS) ? Log_DEBUG : Log_NOTICE;

      Logger_logFormatted(log, logLevel, logContext, "TruncFileResp error code: %s",
         FhgfsOpsErr_toErrString(retVal) );
   }

   // clean-up
   NetMessage_virtualDestruct(respMsg);
   NoAllocBufferStore_addBuf(bufStore, respBuf);

cleanup_request:
   TruncFileMsg_uninit( (NetMessage*)&requestMsg);

   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return retVal;
}

FhgfsOpsErr FhgfsOpsRemoting_fsyncfile(RemotingIOInfo* ioInfo, fhgfs_bool forceRemoteFlush,
   fhgfs_bool checkSession, fhgfs_bool doSyncOnClose)
{
   const char* logContext = "Remoting (fsync file)";

   App* app = ioInfo->app;
   Logger* log = App_getLogger(app);

   FhgfsOpsErr retVal = FhgfsOpsErr_SUCCESS;
   fhgfs_sigset_t oldSignalSet;

   StripePattern* pattern = ioInfo->pattern;
   UInt16Vec* targetIDs = pattern->getStripeTargetIDs(pattern);
   UInt16Vec* mirrorTargetIDs = pattern->getMirrorTargetIDs(pattern); // NULL if not raid10
   ssize_t numStripeTargets = UInt16Vec_length(targetIDs);
   ssize_t numMirrorTargets = mirrorTargetIDs ? UInt16Vec_length(mirrorTargetIDs) : 0;
   int maxUsedTargetIndex = AtomicInt_read(ioInfo->maxUsedTargetIndex);
   const char* fileHandleID = ioInfo->fileHandleID;
   FhgfsOpsErr* workResults = os_kmalloc( // 1st half for origs, 2nd half for mirrors (if any)
      (numStripeTargets + numMirrorTargets) * sizeof(FhgfsOpsErr) );
   FhgfsOpsErr* mirrorWorkResults = &workResults[numStripeTargets];

   WorkQueue* workQ = App_getWorkQueue(app);

   SynchronizedCounter counter;
   fhgfs_bool isWorkInterrupted = fhgfs_false;

   int currentNodeIndex = 0;
   int numWorks = 0;
   int numMirrorWorks = 0;
   int numWorksTotal; // numWors + numMirrorWorks
   int i;

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   SynchronizedCounter_init(&counter);

   // send fsync msg via workers to each target...

   while( (numWorks <= maxUsedTargetIndex) && (numWorks < numStripeTargets) )
   {
      FSyncLocalFileWork* work = FSyncLocalFileWork_construct(app,
         currentNodeIndex, fileHandleID, targetIDs, workResults, &counter, forceRemoteFlush,
         BitStore_getBit(ioInfo->firstWriteDone, currentNodeIndex), checkSession, doSyncOnClose);
      FSyncLocalFileWork_setMsgUserID(work, currentFsUserID);
      RetryableWork_setInterruptor( (RetryableWork*)work, &isWorkInterrupted);
      WorkQueue_addWork(workQ, (Work*)work);

      // prepare for next loop
      numWorks++;
      currentNodeIndex++;
   }

   // send fsync msg via workers to each mirror target...

   currentNodeIndex = 0;

   while( (numMirrorWorks <= maxUsedTargetIndex) && (numMirrorWorks < numMirrorTargets) )
   {
      FSyncLocalFileWork* work = FSyncLocalFileWork_construct(app,
         currentNodeIndex, fileHandleID, mirrorTargetIDs, mirrorWorkResults, &counter,
         forceRemoteFlush, BitStore_getBit(ioInfo->firstWriteDone, currentNodeIndex),
         checkSession, doSyncOnClose);
      FSyncLocalFileWork_setMsgUserID(work, currentFsUserID);
      FSyncLocalFileWork_setIsMirror(work, fhgfs_true); // this fsync is for a mirror chunk file
      RetryableWork_setInterruptor( (RetryableWork*)work, &isWorkInterrupted);
      WorkQueue_addWork(workQ, (Work*)work);

      // prepare for next loop
      numMirrorWorks++;
      currentNodeIndex++;
   }

   // wait for all work to be done

   numWorksTotal = numWorks + numMirrorWorks;

   if(!SynchronizedCounter_waitForCountInterruptible(&counter, numWorksTotal) )
   { // interrupted by signal => interrupt workers and wait again
      for(i=0; i < numWorksTotal; i++)
      {
         isWorkInterrupted = fhgfs_true;
         SynchronizedCounter_waitForCount(&counter, numWorksTotal);
      }
   }

   // check target results...

   for(i=0; i < numWorks; i++)
   {
      currentNodeIndex = i;

      // check original chunk file result
      if(workResults[currentNodeIndex] != FhgfsOpsErr_SUCCESS)
      { // something went wrong
         Logger_logFormatted(log, Log_WARNING, logContext, "Error storage target: %hu; Msg: %s",
            UInt16Vec_at(targetIDs, currentNodeIndex),
            FhgfsOpsErr_toErrString(workResults[currentNodeIndex]) );

         retVal = workResults[currentNodeIndex];
         goto cleanup;
      }

   }

   // check mirror results...

   for(i=0; i < numMirrorWorks; i++)
   {
      currentNodeIndex = i;

      // check mirror chunk file result
      if(mirrorWorkResults[currentNodeIndex] != FhgfsOpsErr_SUCCESS)
      { // something went wrong
         uint16_t mirrorTargetID =
            mirrorTargetIDs ? UInt16Vec_at(mirrorTargetIDs, currentNodeIndex) : 0;

         Logger_logFormatted(log, Log_WARNING, logContext,
            "Error storage target (mirror): %hu; Msg: %s",
            mirrorTargetID, FhgfsOpsErr_toErrString(mirrorWorkResults[currentNodeIndex]) );

         retVal = mirrorWorkResults[currentNodeIndex];
         goto cleanup;
      }
   }

cleanup:
   // clean-up
   SynchronizedCounter_uninit(&counter);
   os_kfree(workResults);
   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return retVal;
}

/**
 * @return FhgfsOpsErr_INTERNAL if no targets are known
 */
FhgfsOpsErr FhgfsOpsRemoting_statStoragePath(App* app, int64_t* outSizeTotal, int64_t* outSizeFree)
{
   Logger* log = App_getLogger(app);
   const char* logContext = "Remoting (stat storage targets)";

   WorkQueue* workQ = App_getWorkQueue(app);
   TargetMapper* targetMapper = App_getTargetMapper(app);

   FhgfsOpsErr retVal = FhgfsOpsErr_SUCCESS;
   fhgfs_sigset_t oldSignalSet;

   UInt16List targetIDs;
   UInt16ListIter targetsIter;

   size_t numTargets;
   StatStoragePathData* statDataArray;

   SynchronizedCounter counter;
   fhgfs_bool isWorkInterrupted = fhgfs_false;

   unsigned numWorks;
   unsigned i;

   *outSizeTotal = 0;
   *outSizeFree = 0;

   UInt16List_init(&targetIDs);
   SynchronizedCounter_init(&counter);

   TargetMapper_getTargetIDs(targetMapper, &targetIDs);

   numTargets = UInt16List_length(&targetIDs);

   if(unlikely(!numTargets) )
   { // we treat no known storage servers as an error
      Logger_logFormatted(log, Log_CRITICAL, logContext, "No storage targets known.");
      return FhgfsOpsErr_UNKNOWNTARGET;
   }

   statDataArray = os_kmalloc(numTargets * sizeof(StatStoragePathData) );

   UInt16ListIter_init(&targetsIter, &targetIDs);

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   numWorks = 0;
   while(numWorks < numTargets)
   {
      StatStoragePathWork* work;
      StatStoragePathData* currentStatData = &(statDataArray[numWorks]);
      currentStatData->targetID = UInt16ListIter_value(&targetsIter);

      work = StatStoragePathWork_construct(app, currentStatData, &counter);
      RetryableWork_setInterruptor( (RetryableWork*)work, &isWorkInterrupted);
      WorkQueue_addWork(workQ, (Work*)work);

      // prepare for next loop
      numWorks++;
      UInt16ListIter_next(&targetsIter);
   }

   UInt16ListIter_uninit(&targetsIter);

   // wait for all work to be done
   if(!SynchronizedCounter_waitForCountInterruptible(&counter, numWorks) )
   { // interrupted by signal => interrupt workers and wait again
      for(i=0; i < numWorks; i++)
      {
         isWorkInterrupted = fhgfs_true;
         SynchronizedCounter_waitForCount(&counter, numWorks);
      }
   }

   // handle results
   for(i=0; i < numWorks; i++)
   {
      StatStoragePathData* currentStatData = &(statDataArray[i]);

      if(currentStatData->outResult != FhgfsOpsErr_SUCCESS)
      { // something went wrong
         Logger_logFormatted(log, Log_WARNING, logContext,
            "Error target (storage): %hu; Msg: %s",
            currentStatData->targetID, FhgfsOpsErr_toErrString(currentStatData->outResult) );

         retVal = currentStatData->outResult;
         break;
      }

      *outSizeTotal += currentStatData->outSizeTotal;
      *outSizeFree += currentStatData->outSizeFree;
   }

   // clean-up
   os_kfree(statDataArray);
   SynchronizedCounter_uninit(&counter);
   UInt16List_uninit(&targetIDs);
   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return retVal;
}

/**
 * Validate and stat in a single remote request.
 *
 * @param outStatRes result of stat operation, only set if general return value indicates success
 * @param outFhgfsStat only set if outStatRes indicates success
 * @return validation result (note that stat result comes as separate out arg)
 */
FhgfsOpsErr FhgfsOpsRemoting_validateAndStatSubentry(App* app, const EntryInfo* parentEntryInfo,
   const char* entryName, const EntryInfo* entryInfo,
   FhgfsOpsErr* outStatRes, fhgfs_stat* outFhgfsStat)
{
   fhgfs_sigset_t oldSignalSet;
   FhgfsOpsErr retVal;

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   retVal = MetadataTk_validateAndStatSubentry(app, parentEntryInfo, entryName, entryInfo,
      outStatRes, outFhgfsStat);

   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return retVal;
}

/**
 * Lookup and stat in a single remote request.
 */
FhgfsOpsErr FhgfsOpsRemoting_lookupAndStatSubentry(App* app, const EntryInfo* parentEntryInfo,
   const char* entryName, EntryInfo* outEntryInfo,
   FhgfsOpsErr* outStatRes, fhgfs_stat* outFhgfsStat)
{
   fhgfs_sigset_t oldSignalSet;
   FhgfsOpsErr retVal;

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   retVal = MetadataTk_lookupAndStatSubentry(app, parentEntryInfo, entryName, outEntryInfo,
      outStatRes, outFhgfsStat);

   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return retVal;
}

/**
 * Lookup/create and stat in a single remote request.
 */
FhgfsOpsErr FhgfsOpsRemoting_lookupCreateStatFile(App* app,
   const LookupCreateStatInfoIn* inInfo, LookupCreateStatInfoOut* outInfo)
{
   fhgfs_sigset_t oldSignalSet;
   FhgfsOpsErr retVal;

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   retVal = MetadataTk_lookupCreateStatFile(app, inInfo, outInfo);

   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return retVal;
}


#ifdef LOG_DEBUG_MESSAGES
void __FhgfsOpsRemoting_logDebugIOCall(const char* logContext, size_t size, loff_t offset,
   RemotingIOInfo* ioInfo)
{
   App* app = ioInfo->app;
   Logger* log = App_getLogger(app);

   Logger_logFormatted(log, Log_DEBUG, logContext, "size: %lld; offset: %lld; fileHandleID: %s",
      (long long)size, (long long)offset, ioInfo->fileHandleID);
}
#endif // LOG_DEBUG_MESSAGES


/**
 * Compute the chunk-file offset on a storage server from a given user file position.
 *
 * Note: Make sure that the given stripeNodeIndex is really correct for the given pos.
 */
static int64_t __FhgfsOpsRemoting_getNodeLocalOffset(int64_t pos, unsigned chunkSize, size_t numNodes,
   size_t stripeNodeIndex)
{
   /* the code below is an optimization (wrt division and modulo) of the following three lines:
         int64_t posModChunkSize = (pos % chunkSize);
         int64_t stripeSetStart = pos - posModChunkSize - (stripeNodeIndex*chunkSize);
         return ( (stripeSetStart / numNodes) + posModChunkSize); */

   /* note: "& (chunksize-1) only works as "% chunksize" replacement, because chunksize must be
      a power of two */

   int64_t posModChunkSize = pos & (chunkSize-1);
   int64_t stripeSetStart = pos - posModChunkSize - (stripeNodeIndex*chunkSize);

   int64_t stripeSetStartDivNumNodes;

   // note: if numNodes is a power of two, we can do bit shifting instead of division

   if(MathTk_isPowerOfTwo(numNodes) )
   { // quick path => bit shifting
      stripeSetStartDivNumNodes = stripeSetStart >> MathTk_log2Int32(numNodes);
   }
   else
   { // slow path => division
      // note: do_div(n64, base32) assigns the result to n64 and returns the remainder!
      // (we need do_div to enable this division on 32bit archs)

      stripeSetStartDivNumNodes = stripeSetStart; // will be changed by do_div()
      do_div(stripeSetStartDivNumNodes, (unsigned)numNodes);
   }

   return (stripeSetStartDivNumNodes + posModChunkSize);
}

FhgfsOpsErr FhgfsOpsRemoting_hardlink(App* app, const char* fromName, unsigned fromLen,
   EntryInfo* fromInfo, EntryInfo* fromDirInfo, const char* toName, unsigned toLen,
   EntryInfo* toDirInfo)
{
   FhgfsOpsErr retVal = FhgfsOpsErr_INTERNAL;
   Logger* log = App_getLogger(app);

   const char* logContext = "Remoting (hardlink)";
   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);

   HardlinkMsg requestMsg;
   char* respBuf;
   NetMessage* respMsg;
   FhgfsOpsErr requestRes;
   HardlinkRespMsg* hardlinkResp;

   NodeStoreEx* nodes = App_getMetaNodes(app);
   uint16_t fromNodeID = fromDirInfo->ownerNodeID;
   Node* fromNode;
   fhgfs_bool fromNodeSupportsHardlinks;

   fhgfs_sigset_t oldSignalSet;

   fromNode = NodeStoreEx_referenceNode(nodes, fromNodeID);
   if(unlikely(!fromNode) )
   {
      Logger_logErrFormatted(log, logContext, "Unknown metadata node-id: %hu", fromNodeID);
      return FhgfsOpsErr_UNKNOWNNODE;
   }

   fromNodeSupportsHardlinks = VersionTk_checkRequiredRelease(VERSIONTK_REQUIRED_VERSION_HARDLINK,
      Node_getFhgfsVersion(fromNode) );

   if (unlikely(!fromNodeSupportsHardlinks) )
   {
      NodeStoreEx_releaseNode(nodes, &fromNode);
      return FhgfsOpserr_PERM;
   }

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   // prepare request msg

   HardlinkMsg_initFromEntryInfo(&requestMsg, fromName, fromLen, fromInfo, fromDirInfo, toName,
      toLen, toDirInfo);

   // connect & communicate
   requestRes = MessagingTk_requestResponseRetry(app,
      fromNode, (NetMessage*)&requestMsg, NETMSGTYPE_HardlinkResp, &respBuf, &respMsg);

   NodeStoreEx_releaseNode(nodes, &fromNode);

   if(unlikely(requestRes != FhgfsOpsErr_SUCCESS) )
   { // clean-up
      retVal = requestRes;
      goto cleanup_request;
   }

   // handle result
   hardlinkResp = (HardlinkRespMsg*)respMsg;
   retVal = (FhgfsOpsErr)HardlinkRespMsg_getValue(hardlinkResp);

   if(retVal == FhgfsOpsErr_SUCCESS)
   {
      // success (nothing to be done here)
   }
   else
   {
      int logLevel = Log_NOTICE;

      if( (retVal == FhgfsOpsErr_PATHNOTEXISTS) || (retVal == FhgfsOpsErr_INUSE) )
         logLevel = Log_DEBUG; // don't bother user with non-error messages

      Logger_logFormatted(log, logLevel, logContext, "HardlinkResp error code: %s",
         FhgfsOpsErr_toErrString(retVal) );
   }

   // clean-up
   NetMessage_virtualDestruct(respMsg);
   NoAllocBufferStore_addBuf(bufStore, respBuf);

cleanup_request:
   HardlinkMsg_uninit( (NetMessage*)&requestMsg);

   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return retVal;
}


/**
 * Refresh the entry (meta-data update)
 */
FhgfsOpsErr FhgfsOpsRemoting_refreshEntry(App* app, const EntryInfo* entryInfo)
{
   Logger* log = App_getLogger(app);
   const char* logContext = "Remoting (refresh Entry)";
   NoAllocBufferStore* bufStore = App_getMsgBufStore(app);

   fhgfs_sigset_t oldSignalSet;
   RefreshEntryInfoMsg requestMsg;
   char* respBuf;
   NetMessage* respMsg;
   FhgfsOpsErr requestRes;
   RefreshEntryInfoRespMsg* refreshResp;
   FhgfsOpsErr retVal;

   NodeStoreEx* nodes = App_getMetaNodes(app);
   Node* node;

   node = NodeStoreEx_referenceNode(nodes, entryInfo->ownerNodeID);
   if(unlikely(!node) )
   {
      Logger_logErrFormatted(log, logContext, "Unknown metadata node: %hu", entryInfo->ownerNodeID);
      return FhgfsOpsErr_UNKNOWNNODE;
   }

   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   // prepare request msg
   RefreshEntryInfoMsg_initFromEntryInfo(&requestMsg, entryInfo );

   // connect & communicate
   requestRes = MessagingTk_requestResponseRetry(app,
      node, (NetMessage*)&requestMsg, NETMSGTYPE_RefreshEntryInfoResp, &respBuf, &respMsg);

   NodeStoreEx_releaseNode(nodes, &node);

   if(unlikely(requestRes != FhgfsOpsErr_SUCCESS) )
   { // clean-up
      retVal = requestRes;
      goto cleanup_request;
   }

   // handle result
   refreshResp = (RefreshEntryInfoRespMsg*)respMsg;
   retVal = (FhgfsOpsErr)RefreshEntryInfoRespMsg_getResult(refreshResp);

   if(retVal != FhgfsOpsErr_SUCCESS)
   {
      LOG_DEBUG_FORMATTED(log, Log_DEBUG, logContext, "StatResp error code: %s",
         FhgfsOpsErr_toErrString(retVal) );
      IGNORE_UNUSED_VARIABLE(log);
      IGNORE_UNUSED_VARIABLE(logContext);
   }


   // clean-up
   NetMessage_virtualDestruct(respMsg);
   NoAllocBufferStore_addBuf(bufStore, respBuf);

cleanup_request:
   RefreshEntryInfoMsg_uninit( (NetMessage*) &requestMsg);

   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return retVal;
}


