/*
 * Compatibility functions for older Linux versions
 */

#include <linux/mm.h> // for old sles10 kernels, which forgot to include it in backing-dev.h
#include <linux/backing-dev.h>
#include <linux/pagemap.h>

#include <app/App.h>
#include <app/log/Logger.h>
#include <common/Common.h>
#include <filesystem/FhgfsOpsSuper.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,32) && LINUX_VERSION_CODE < KERNEL_VERSION(2,6,34)
   static atomic_long_t fhgfs_bdiSeq = ATOMIC_LONG_INIT(0);
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,30)
   /**
    * memdup_user - duplicate memory region from user space
    *
    * @src: source address in user space
    * @len: number of bytes to copy
    *
    * Returns an ERR_PTR() on failure.
    */
   void *memdup_user(const void __user *src, size_t len)
   {
      void *p;

      /*
       * Always use GFP_KERNEL, since copy_from_user() can sleep and
       * cause pagefault, which makes it pointless to use GFP_NOFS
       * or GFP_ATOMIC.
       */
      p = kmalloc(len, GFP_KERNEL);
      if (!p)
         return ERR_PTR(-ENOMEM);

      if (copy_from_user(p, src, len)) {
         kfree(p);
         return ERR_PTR(-EFAULT);
      }

      return p;
   }
#endif // memdup_user, LINUX_VERSION_CODE < KERNEL_VERSION(2,6,30)



#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17)
   /*
    * strndup_user - duplicate an existing string from user space
    * @s: The string to duplicate
    * @n: Maximum number of bytes to copy, including the trailing NUL.
    */
   char *strndup_user(const char __user *s, long n)
   {
      char *p;
      long length;

      length = strnlen_user(s, n);

      if (!length)
         return ERR_PTR(-EFAULT);

      if (length > n)
         return ERR_PTR(-EINVAL);

      p = memdup_user(s, length);

      if (IS_ERR(p))
         return p;

      p[length - 1] = '\0';

      return p;
   }
#endif // strndup_user, LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17)

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,32) && \
     LINUX_VERSION_CODE < KERNEL_VERSION(2,6,34)
   /*
    * For use from filesystems to quickly init and register a bdi associated
    * with dirty writeback
    */
   int bdi_setup_and_register(struct backing_dev_info *bdi, char *name,
               unsigned int cap)
   {
      char tmp[32];
      int err;

      bdi->name = name;
      bdi->capabilities = cap;
      err = bdi_init(bdi);
      if (err)
         return err;

      sprintf(tmp, "%.28s%s", name, "-%d");
      err = bdi_register(bdi, NULL, tmp, atomic_long_inc_return(&fhgfs_bdiSeq));
      if (err) {
         bdi_destroy(bdi);
         return err;
      }

      return 0;
   }
#endif



/* NOTE: We can't do a feature detection for find_get_pages_tag(), as
 *       this function is in all headers of all supported kernel versions.
 *       However, it is only _exported_ since 2.6.22 and also only
 *       exported in RHEL >=5.10. */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
   /**
    * find_get_pages_tag - find and return pages that match @tag
    * @mapping:   the address_space to search
    * @index:  the starting page index
    * @tag: the tag index
    * @nr_pages:  the maximum number of pages
    * @pages:  where the resulting pages are placed
    *
    * Like find_get_pages, except we only return pages which are tagged with
    * @tag.   We update @index to index the next page for the traversal.
    */
   unsigned find_get_pages_tag(struct address_space *mapping, pgoff_t *index,
            int tag, unsigned int nr_pages, struct page **pages)
   {
      unsigned int i;
      unsigned int ret;

      read_lock_irq(&mapping->tree_lock);
      ret = radix_tree_gang_lookup_tag(&mapping->page_tree,
               (void **)pages, *index, nr_pages, tag);
      for (i = 0; i < ret; i++)
         page_cache_get(pages[i]);
      if (ret)
         *index = pages[ret - 1]->index + 1;
      read_unlock_irq(&mapping->tree_lock);
      return ret;
   }
#endif // find_get_pages_tag() for <2.6.22


#if LINUX_VERSION_CODE < KERNEL_VERSION(3,3,0)

/**
 * This is the former d_alloc_root with an additional iput on error.
 */
struct dentry *d_make_root(struct inode *root_inode)
{
   struct dentry* allocRes = d_alloc_root(root_inode);
   if(!allocRes)
      iput(root_inode);

   return allocRes;
}
#endif // LINUX_VERSION_CODE (d_make_root)


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)

char *kvasprintf(gfp_t gfp, const char *fmt, va_list ap)
{
   unsigned int len;
   char *p;
   va_list aq;

   va_copy(aq, ap);
   len = vsnprintf(NULL, 0, fmt, aq);
   va_end(aq);

   p = kmalloc(len+1, gfp);
   if (!p)
      return NULL;

   vsnprintf(p, len+1, fmt, ap);

   return p;
}

char *kasprintf(gfp_t gfp, const char *fmt, ...)
{
   va_list ap;
   char *p;

   va_start(ap, fmt);
   p = kvasprintf(gfp, fmt, ap);
   va_end(ap);

   return p;
}

#endif // LINUX_VERSION_CODE kvasprintf / kasprintf

#ifndef KERNEL_HAS_D_MATERIALISE_UNIQUE
/**
 * d_materialise_unique() is available since 2.6.19, but was backported to RHEL5 kernels.
 * It is required to add dir-dentries. We try to emulate it here for older kernels, but that is
 * a hard task, as we don't have access to all required dcache functions.
 */
struct dentry* d_materialise_unique(struct dentry *dentry, struct inode *inode)
{
   struct dentry* aliasDentry = d_find_alias(inode);
   struct dentry* returnDentry;

   if (aliasDentry && !(aliasDentry->d_flags & DCACHE_DISCONNECTED) )
      d_invalidate(aliasDentry);

   dput(aliasDentry);

   aliasDentry = d_add_unique(dentry, inode);
   if (aliasDentry)
   {
      App* app = FhgfsOps_getApp(dentry->d_sb);
      Logger* log = App_getLogger(app);
      const char* logContext = "Add dir entry";

      // returnDentry = aliasDentry; // NFS did that up to 2.6.18, but definitely not entirely
                                     //  correct and with unknown consequences

      // For now we take the simplest approach (-EBUSY and log message)

      dput(aliasDentry);
      returnDentry = ERR_PTR(-EBUSY);

      Logger_logFormatted(log, Log_ERR, logContext, "Failed to add unique dentry. "
         "Please notifiy FhGFS developers or update your kernel to >2.6.19 or 2.6.18-el5");
   }
   else
      returnDentry = dentry;

   return returnDentry;
}
#endif // KERNEL_HAS_D_MATERIALISE_UNIQUE
