/*
 * Compatibility functions for older Linux versions
 */

#ifndef OSCOMPAT_H_
#define OSCOMPAT_H_

#include <common/Common.h>
#include <linux/fs.h>
#include <linux/namei.h>
#include <asm/kmap_types.h> // for 2.6.16
#include <linux/compat.h>

#ifdef KERNEL_HAS_TASK_IO_ACCOUNTING
   #include <linux/task_io_accounting_ops.h>
#endif


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
   extern char *kvasprintf(gfp_t gfp, const char *fmt, va_list ap);
   extern char *kasprintf(gfp_t gfp, const char *fmt, ...);
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17)
   extern char *strndup_user(const char __user *s, long n);
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,30)
   extern void *memdup_user(const void __user *src, size_t len);
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,3,0)
   extern struct dentry *d_make_root(struct inode *root_inode);
#endif

#ifndef KERNEL_HAS_TASK_IO_ACCOUNTING
   static inline void task_io_account_read(size_t bytes);
   static inline void task_io_account_write(size_t bytes);
#endif

static inline int os_generic_permission(struct inode *inode, int mask);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,34)
   extern int bdi_setup_and_register(struct backing_dev_info *bdi, char *name, unsigned int cap);
#endif

/**
 * generic_permission() compatibility function
 *
 * NOTE: Only kernels > 2.6.32 do have inode->i_op->check_acl, but as we do not
 *       support it anyway for now, we do not need a complete kernel version check for it.
 *       Also, in order to skip useless pointer references we just pass NULL here.
 */
int os_generic_permission(struct inode *inode, int mask)
{
   #if LINUX_VERSION_CODE >= KERNEL_VERSION(3,1,0)
      return generic_permission(inode, mask);
   #elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,38)
      return generic_permission(inode, mask, 0, NULL);
   #else
      return generic_permission(inode, mask, NULL);
   #endif
}


#ifndef KERNEL_HAS_TASK_IO_ACCOUNTING

/**
 * No-op if kernel version does not support this.
 */
void task_io_account_read(size_t bytes)
{
}

/**
 * No-op if kernel version does not support this.
 */
void task_io_account_write(size_t bytes)
{
}

#endif // KERNEL_HAS_TASK_IO_ACCOUNTING


/**
 * warning: this is the atomic version, so callers are not allowed to sleep until kunmap_atomic
 * is called.
 *
 * note: this always passes KM_USER0 to kmap_atomic.
 */
static inline void *os_kmap_atomic(struct page *page)
{
   #if LINUX_VERSION_CODE >= KERNEL_VERSION(3,4,0)
      return kmap_atomic(page);
   #else
      return kmap_atomic(page, KM_USER0);
   #endif
}

static inline void os_kunmap_atomic(void *addr)
{
   #if LINUX_VERSION_CODE >= KERNEL_VERSION(3,4,0)
      kunmap_atomic(addr);
   #else
      kunmap_atomic(addr, KM_USER0);
   #endif
}

#ifndef KERNEL_HAS_D_MATERIALISE_UNIQUE
extern struct dentry* d_materialise_unique(struct dentry *dentry, struct inode *inode);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,32)
/**
 * Taken from ext3 dir.c. is_compat_task() does work for all kernels, although it was already there.
 * So we are conservativ and only allow it for recent kernels.
 */
static inline int is_32bit_api(void)
{
#ifdef CONFIG_COMPAT
   return is_compat_task();
#else
   return (BITS_PER_LONG == 32);
#endif
}
#else
static inline int is_32bit_api(void)
{
   return (BITS_PER_LONG == 32);
}
#endif // LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,32)


#endif /* OSCOMPAT_H_ */
